/**
 * This class contains global attributes/methods for the jRelix system.
 * 
 */
import java.io.*;
import java.util.*;

class Global  implements TypeConstants, Constants
{
	//debugging Yi Zheng
	private static boolean DEBUG=false;
  //global Environment/Tables
  public static Environment environment;
  public static CompTable comptable;
  public static ExprTable exprtable;

  // the JRelixInputStream for the parser
  public static JRelixInputStream jin;
  // the status of the current parser
  public static int status;
  // the root of the current JJTREE
  public static SimpleNode root;
  // the interpret time of the current JJTREE
  public static long interpretTime;

//following default values added by Yi Zheng, March 18, 2002
  public static boolean debug = false;    //debug toggle, true for ON
  public static boolean batch = false;    //batch toggle, true for ON
  public static boolean expert = false;   //expert toggle, true for ON
  public static boolean time = false;     //time toggle, true for ON
  public static boolean trace = false;    //trace toggle, true for ON
  public static boolean ssd = false;
  public static boolean ssr = false;

  public static final InputStream instream = System.in;
  public static final PrintStream outstream = System.out;
  public static PrintStream traceoutstream = null;

  private static final String RD_NAME = ".rd";
  private static final int BUF_SIZE = 16 * 1024;

 // static { init();}         -- Modified by Yi Zheng March 18, 2002
 static {
 	try { init(); }
 	catch (Exception e)
 	{ processThrowable(e);}
 }


/**
 * This method implements the action needed after the parsed node is interpreted.
 */
public static void afterInterpret() {
	if (time && (status == Constants.OP_STATEMENT || status == Constants.OP_COMMAND)) {
		println("Interpret time " + ((float)interpretTime)/1000);
	}
	if (status != Constants.OP_IGNORE) {
		Global.print(">"); Global.flush();
	}
}

/**
 * This method implements the action needed after the parsed node is interpreted.
 * only used by Interpreter.executeJRelixCommand
 */
public static void afterSRDInterpret() {
	if (time && (status == Constants.OP_STATEMENT || status == Constants.OP_COMMAND)) {
		println("Interpret time " + ((float)interpretTime)/1000);
	}
	if (status != Constants.OP_IGNORE) {
//		Global.print(">"); 
		Global.flush();
	}
}

/**
 * This method implements the action needed after the user input text
 * is interpreted.
 * From the root type, one could do some processing of the buffered user
 * input text stored in the JRelixInputStream jin. For example, if it is 
 * a computation, one may create a file to store the code for the computation.
 */
public static void afterParse(Environment env)
throws InterpretError
{
	tracePrintln(jin.getText().toString());
	saveCompCode(env);
	if (debug && root != null) {
	//	root.dump("");
	Utility.dump(root, "debug ==>"); //Yi Zheng June 23, 2002
	}
}
/**
 * This method implements the action needed before the parsed node
 * is interpreted.
 */
public static void beforeInterpret() {
}
/**
 * This method implements the action needed before the user input is parsed.
 * The default implementation simply clear the buffered user input text.
 */
public static void beforeParse() {
	if (jin != null) {
		jin.resetText();
	}
}
/**
 * IMPORTANT: all program that want to exit the java VM must call
 * this exit(int) to do housekeeping job.
 * Need to save computation table ???
 * @param status int
 */
public static void exit(int status) {
  environment.domtable.dump(); 
  environment.reltable.dump(); 
  exprtable.dump(); 
  Surrogate.save();
	//added by Yi Zheng March 27, 2002
	Utility.cleanUp("__#");  //remove all temp files used in computation
	Utility.cleanUp(Interpreter.temp_name_prefix);
  System.exit(status); 
}
/**
 * This method was created by a SmartGuide.
 * @param doms Domain[]
 * @param domname java.lang.String
 * @param index int
 */
private static void fillDomArray(Domain[] doms,	String domname, int index, Environment env) 
throws InterpretError //added by Yi Zheng March 18, 2002
{
  Domain doment;
  
  if( (doment = env.lookupDom(domname, true)) != null) {
   doms[index] = doment;
   } else {
	 // This is possible when the domain information in .dom file 
	 // corrupted.
//	 print("loadRDTable: no such domain -" + domname); --commented out by Yi Zheng March 18, 2002
//	 println(". It seems that .dom file is corrupted."); -- commented out by Yi Zheng March 18, 2002

	 // REVISIT: Can we exit here?
//	 exit(1); -- commented out by Yi Zheng, March 18, 2002
	throw new InterpretError("error: Global: fillDomArray(): no such domain -" + domname +
	 ". It seems that .dom file is corrupted.");
	
   }
}
/**
 * This method was created by a SmartGuide.
 */
public static void flush() {
   outstream.flush();
   if (trace && traceoutstream != null) traceoutstream.flush();
}
/**
 * This method was created by a SmartGuide.
 */
private static void init() 
throws InterpretError //added by Yi Zheng March 18, 2002
{
	try {
	environment = new Environment();
	comptable = new CompTable();
	exprtable = new ExprTable(environment, comptable);
	} catch (InterpretError i)
	{}

	jin = new JRelixInputStream(instream);

	loadRDTable(environment);
}
/**
 * This method load the Relation/Domain (RD) tables from the files on disk.
 * @param env Environment
 */
private static void loadRDTable(Environment env) 
throws InterpretError //added by Yi Zheng March 18, 2002
{
	 Domain doment;
	 Domain[] domarray;
	 String relName;
	 String domName; 
	 int pos;
	 Relation rel = new Relation();
	 String rb;
	 Integer tint;
	 
	 RelTable reltable = env.reltable;
	 DomTable domtable = env.domtable;

	 try{
	   BlockInputStream rdis = new BlockInputStream(RD_NAME, BUF_SIZE); 
	   
	   // load data to hash table.
	   while(true){
	  // REVISIT: will need to handle bad record exceptions..
	  try{
		/* modified by Weizhong
		relName = rdis.readString();
		domName = rdis.readString();
		pos = rdis.readInt();
		*/

/*
		rb = rel.strRead(rdis, STRING);
		relName = rb;
		rb = rel.strRead(rdis, STRING);
		domName = rb;
		rb = rel.strRead(rdis, INTEGER);
		tint = new Integer(rb);
		pos = tint.intValue();
		*/
		
		rb = rel.strRead(rdis, STRING);
		relName = rdis.readString(rb);
		rb = rel.strRead(rdis, STRING);
		domName = rdis.readString(rb);
		rb = rel.strRead(rdis, INTEGER);
		tint = new Integer(rb);
		pos = tint.intValue();

		// REVISIT: when insert IDlist, position is not taken 
		// care of (currently) since it's a vector that we're
		// dealing with and we can only read data from disk
		// _sequentially_.  ..need to fix this later.
		
		// It does't matter, since we have pos to indicate the position
		// of the domain within a relation -- Yi Zheng March 18, 2002

			// Note: we don't care if the name is started with a '.'
		// or not. The domains with IDLIST type will be treated
		// exactly the same as a relation. There is one point of
		// confusing, i.e. IDLIST-typed domains will be confued
		// with system relations -- their names all start with '.'.
		doment = env.lookupDom(domName, true);
		if(doment != null){
		  reltable.insertIDlist(relName, doment, pos);
		}else{
		//	  outstream.println("Error: loadRDTable() no such domain-"+domName); --Yi Zheng commented out March 18, 2002
	  // REVISIT: what to do next?
	  
	  
	  //Yi Zheng wrote March 18, 2002:
	  // Since we should have prevented this exception by disallowing the deletion
	  // of a domain while it's still in use, why don't we just throw and exception to 
	  //indicate the fact that the .rd table is corrupt?
	  
	  throw new InterpretError("error: Global: loadRDTable() no such domain-"+domName);
			}
		 }catch(IOException e){
		rdis.close();   // EOF
		break;
	  }
	   }
	 }catch(FileNotFoundException e){
   // Can't find .rd file, must do initialization or error.
	   // insert .rel idlist.
	   domarray = new Domain[5];
   fillDomArray(domarray, ".rel_name", 0, env);
   fillDomArray(domarray, ".tuples", 1, env);
   fillDomArray(domarray, ".attributes", 2, env);
   fillDomArray(domarray, ".rvc", 3, env);
   fillDomArray(domarray, ".sort", 4, env);
	   reltable.insertIDlist(".rel", domarray);
//Yi Zheng: Added populateInitData() March 21, 2002
	populateInitData(".rel", env);
	   // insert .dom idlist.
   domarray = new Domain[4]; //Yi Zheng April 29, 2002
   fillDomArray(domarray, ".dom_name", 0, env);
   fillDomArray(domarray, ".type", 1, env);
   fillDomArray(domarray, ".count", 2, env);
   fillDomArray(domarray, ".isState", 3, env); //April 29, 2002
	   reltable.insertIDlist(".dom", domarray);
//Yi Zheng: Added populateInitData() March 21, 2002
	populateInitData(".dom", env);
	   // insert .rd idlist.
   domarray = new Domain[3]; //April 29, 2002
   fillDomArray(domarray, ".rel_name", 0, env);
   fillDomArray(domarray, ".dom_name", 1, env);
   fillDomArray(domarray, ".position", 2, env);
	   reltable.insertIDlist(".rd", domarray);
//Yi Zheng: Added populateInitData() March 21, 2002
	populateInitData(".rd", env);
	 }catch(Exception e){
 //  outstream.println("Error: loadRDTable() IOException");  (modified by Yi Zheng, March 18, 2002)
 
	println("Error: loadRDTable() Exception");
	
	 }
}

/**
 * This method populates .rel, .dom, .rd system relations at startup
 * and stores them on to disk
 * @param relName java.lang.String the name of the relation
 * @param env Environment
 * @author Yi Zheng March 20, 2002
 */
private static void populateInitData(String relName, Environment env)
throws InterpretError
{
	Relation rel = env.lookupRel(relName, true);
	if (rel == null)
		throw new InterpretError("relation " + relName + " not found");
	if(relName.compareTo(".rel") == 0)
	{
		Object[] d = new Object[5];
		String[] d1 = new String[3];
		d1[0] = ".rel"; d1[1] = ".dom"; d1[2] = ".rd";
		d[0] = d1;
		int[] d2 = new int[3];
		d2[0] = 3; d2[1] = 21; d2[2] = 12;
		d[1] = d2;
		int[] d3 = new int[3];
		d3[0] = 5; d3[1] = 4; //April 29, 2002
		d3[2] = 3;
		d[2] = d3;
		int[] d4 = new int[3];
		d4[0] = RELATION; d4[1] = RELATION; d4[2] = RELATION;
		d[3] = d4;
		int[] d5 = new int[3];
		d5[0] = 0; d5[1] = 0; d5[2] = 0;
		d[4] = d5;
		rel.data = d;
		
	}
	else if(relName.compareTo(".dom") == 0)
	{
		Object[] d = new Object[4]; //April 29, 2002
		String[] d1 = new String[21];
		d[0] = d1;
		int[] d2 = new int[21];
		d[1] = d2;
		int[] d3 = new int[21];
		d[2] = d3;
		
		//April 29, 2002
		int[] d4 = new int[21];
		for (int k = 0; k < 21; k++)
			d4[k] = 0;
		d[3] = d4;
	
		 d1[0] = ".rel_name"; d2[0] = STRING; d3[0] = 2;
		 d1[1] = ".tuples"; d2[1] = INTEGER; d3[1] = 1;
		 d1[2] = ".attributes"; d2[2] = INTEGER; d3[2] = 1;
		 d1[3] = ".rvc"; d2[3] = INTEGER; d3[3] = 1;
		 d1[4] = ".sort"; d2[4] = INTEGER; d3[4] = 1;
		 d1[5] = ".dom_name"; d2[5] = STRING; d3[5] = 2;
		 d1[6] = ".type"; d2[6] = INTEGER; d3[6] = 1;
		 d1[7] = ".count"; d2[7] = INTEGER; d3[7] = 1;
		 d1[8] = ".position"; d2[8] = INTEGER; d3[8] = 1;
		 d1[9] = ".id"; d2[9] = IDLIST; d3[9] = 1;
		 d1[10] = ".bool"; d2[10] = BOOLEAN; d3[10] = 1;
		 d1[11] = ".short"; d2[11] = SHORT; d3[11] = 1;
		 d1[12] = ".integer"; d2[12] = INTEGER; d3[12] = 1;
		 d1[13] = ".long"; d2[13] = LONG; d3[13] = 1;
		 d1[14] = ".float"; d2[14] = FLOAT; d3[14] = 1;
		 d1[15] = ".double"; d2[15] = DOUBLE; d3[15] = 1;
		 d1[16] = ".string"; d2[16] = STRING; d3[16] = 1;
		 d1[17] = ".text"; d2[17] = TEXT; d3[17] = 1;
		 d1[18] = ".expr"; d2[18] = EXPR; d3[18] = 1;
		 d1[19] = ".stmt"; d2[19] = STMT; d3[19] = 1;
		 d1[20] = ".isState"; d2[20] = INTEGER; d3[20] = 1;
		rel.data = d;
	}else if(relName.compareTo(".rd") == 0)
	{
		Object[] d = new Object[3];
		String[] d1 = new String[12];
		d[0] = d1;
		String[] d2 = new String[12];
		d[1] = d2;
		int[] d3 = new int[12];
		d[2] = d3;
		d1[0] = ".rel"; d2[0] = ".rel_name"; d3[0] = 0;
		d1[1] = ".rel"; d2[1] = ".tuples"; d3[1] = 1;
		d1[2] = ".rel"; d2[2] = ".attributes"; d3[2] = 2;
		d1[3] = ".rel"; d2[3] = ".rvc"; d3[3] = 3;
		d1[4] = ".rel"; d2[4] = ".sort"; d3[4] = 4;
		d1[5] = ".dom"; d2[5] = ".dom_name"; d3[5] = 0;
		d1[6] = ".dom"; d2[6] = ".type"; d3[6] = 1;
		d1[7] = ".dom"; d2[7] = ".count"; d3[7] = 2;
		d1[8] = ".dom"; d2[8] = ".isState"; d3[8] = 3;
		d1[9] = ".rd"; d2[9] = ".rel_name"; d3[9] = 0;
		d1[10] = ".rd"; d2[10] = ".dom_name"; d3[10] = 1;
		d1[11] = ".rd"; d2[11] = ".position"; d3[11] = 2;
		rel.data = d;
		
	}
	else{
		throw new InterpretError("Unexpected initial system relation: " + 
		relName);
	}
	rel.save();
//	if(DEBUG)
//		rel.print();
}
/**
 * This method print a String.
 * @param name java.lang.String
 */
public static void print(String name) {
   outstream.print(name);
   if (trace && traceoutstream != null) traceoutstream.print(name);
}
/**
 * This method print a Throwable.
 * @param e java.lang.Throwable
 */
public static void print(Throwable e) {
  e.printStackTrace(outstream);
  if (trace && traceoutstream != null) e.printStackTrace(traceoutstream);
}
/**
 * This method print the jRelix head.
 */
public static void printHead () {
   println("+-------------------------------------------+");
   println("|            Relix Java version 0.80        |");
   println("| Copyright (c) 1997 -- 2002 Aldat Lab      |");
   println("|         School of Computer Science        |");
   println("|             McGill University             |");
   println("+-------------------------------------------+");
}
/**
 * This method println a String.
 * @param name java.lang.String
 */
public static void println(String name) {
   outstream.println(name);
   if (trace && traceoutstream != null) traceoutstream.println(name);
}
/**
 * This method println a Throwable.
 * @param e java.lang.Throwable
 */
public static void println(Throwable e) {
  e.printStackTrace(outstream);
  if (trace && traceoutstream != null) e.printStackTrace(traceoutstream);
}
/**
 * This method processes a Throwable.
 * @param e java.lang.Throwable
 * modified by Yi Zheng Sept 29, 2002
 *	when debug mode is on, print stack trace; otherwise only prints error message
 */
public static void processThrowable(Throwable e) {

 if(debug) {
  e.printStackTrace(outstream);
  if (trace && traceoutstream != null) e.printStackTrace(traceoutstream);
 }
 else
 {
  outstream.println(e.getMessage());
  if (trace && traceoutstream != null) traceoutstream.println(e.getMessage());
 }
  Global.print(">"); Global.flush();
}
/**
 * Save the user input text for the computation declaration if it is.
 */
public static int saveCompCode(Environment env)
throws InterpretError
{  
int ret = 0;
	if (root != null && root.opcode == Constants.OP_COMPUTATION) {
		SimpleNode node = (SimpleNode) root.jjtGetChild(0);
if(DEBUG)
 Utility.dump(node,"saveCompCode ");
/*
int numchild = node.jjtGetNumChildren();
System.out.println("in saveCompCode, numchild: " + numchild);
System.out.println("in saveCompCode, node.name; node.type; node.opcode: " + node.name + ";" + node.type + ";" + node.opcode);
SimpleNode child;
for ( int i = 0; i < numchild; i++ ) {
  child = (SimpleNode) node.jjtGetChild(i);
  System.out.println("child.name; child.type; child.opcode: "  + child.name + ";" + child.type + ";" + child.opcode);
} */

		PrintWriter pw = null;
		String fname = new String("");
                SimpleNode attrlist;
		try {
			if (node.opcode == OP_EVENTNAME) {
				if(DEBUG)
					System.out.println("beginning of fname...");
				fname = environment.getEventName((SimpleNode)node.jjtGetChild(0));
                                if (node.jjtGetNumChildren() != 1) 
                                   attrlist = (SimpleNode) node.jjtGetChild(1);
                                else
                                   attrlist = null;
                                if (attrlist != null) {
         				int childnum = attrlist.jjtGetNumChildren();
                                        SimpleNode childatt;
                                        if (childnum != 0) {
        					childatt = (SimpleNode)attrlist.jjtGetChild(0);
                                                fname = fname + "[" + childatt.name;
            					for (int i = 1; i < childnum; i++) {
							childatt = (SimpleNode)attrlist.jjtGetChild(i);
     							fname = fname + "," + childatt.name;
                                                }
                                                fname = fname + "]";
                                        }
                                }        
               if(DEBUG)
               System.out.println("end of fname...fname = " + fname);
                        }      
			else 
				fname = ((SimpleNode)node.jjtGetChild(0)).name;

                        Relation temprel = env.lookupRel(fname,true);
                        if (temprel != null) {
                        	if(temprel.rvc == COMPUTATION)
                           		Global.println("Warning: Computation/EventHandler exists, please remove it first!");
                           	else if(temprel.rvc == RELATION)
                           		Global.println("Warning: Relation with the same name exist, please remove it first!");
                        } else {

if(DEBUG)
	System.out.println("in saveCompCode, fname: " + fname);
	                	pw = new PrintWriter(new FileOutputStream(fname + ".comp"), true);
	                	if(DEBUG)
	                	System.out.println("got here? ");
	                	//debugging
	                	String codetext = jin.getText().toString();
	                	if(DEBUG)
	                		System.out.println("In Global: saveCompCode: code is: " + codetext);
	             		pw.println(codetext);
                                ret = 1;
                        }
		} catch (IOException e) {
			try{ if (pw != null) 
				pw.close();
			}
			catch(Exception ee)
			{}
			finally{
				Global.println("IOexception for generating file for computation " + root.name);
			}
		}
	}
        return ret;
}  
/**
 * This method print a String to the OutputStream for the trace.
 * @param name java.lang.String
 */
public static void tracePrint(String name) {
   if (trace && traceoutstream != null) traceoutstream.print(name);
}
/**
 * This method println a String to the OutputStream for the trace.
 * @param name java.lang.String
 */
public static void tracePrintln(String name) {
   if (trace && traceoutstream != null) traceoutstream.println(name);
}
}
