
/* JJT: 0.2.2 */

public class SimpleNode implements Node, java.io.Serializable 
{
  protected Node parent;
  protected java.util.Vector children;
  protected String identifier;
  protected Object info;

  protected int type;
  protected int opcode;
  protected String name;
  protected int bits;      //bits for order or existance of options
  protected java.util.Vector list;
  protected int active;   // active = 1/0, for event handler, on/off weizhong
  


  public SimpleNode(String id) {
	identifier = id;
  }    
  /* Override this method if you want to customize how the node dumps
	 out its children. */

  public void dump(String prefix) {
	System.out.println(toString(prefix) + ":" + type + ":" + opcode +
	":" + name + ":" + bits);
	if (children != null) {
	  for (java.util.Enumeration e = children.elements();
	  e.hasMoreElements();) {
   SimpleNode n = (SimpleNode)e.nextElement();
   n.dump(prefix + " ");
	  }
	}
  }    
  public Object getInfo() { return info; }    
  public Object getObject() { return info; }    
  public boolean isBitSet(int pos) {
	return ((bits >>> pos & 0x01) != 0) ? true : false;
  }    
  public void jjtAddChild(Node n) {
	if (children == null) {
	  children = new java.util.Vector();
	}
	children.addElement(n);
  }    
  public void jjtAddChildAtHead(Node n) {
	if (children == null) {
	  children = new java.util.Vector();
	}
	children.insertElementAt(n,0);
  }    
  public void jjtClose() {
	if (children != null) {
	  children.trimToSize();
	}
  }    
  public static Node jjtCreate(String id) {
	return new SimpleNode(id);
  }    
  public static SimpleNode jjtCreateNode(String id, int type,
	int opcode, String name){
	SimpleNode node = new SimpleNode(id);
	node.name = name;
	node.type = type;
	node.opcode = opcode;

	return node;
  }    
  public SimpleNode jjtDeepDuplicate()
  {  SimpleNode result = new SimpleNode(identifier);
	 result.type = type;
	 result.opcode = opcode;
	 result.name = name;
	 result.bits = bits;
	 result.info = info;
	 //REVISIT: result.list = ?? Do we use this field??
	 if (children != null)
	 {  for(int i=0; i<children.size(); i++)
		{  SimpleNode ch = (SimpleNode) children.elementAt(i);
		   SimpleNode chDup = ch.jjtDeepDuplicate();
		   result.jjtAddChild(chDup);
		   chDup.jjtSetParent(result);
		}
	 }
	 return result;
  }    
  /**
   * Duplicate a new node.
   */
  public SimpleNode jjtDuplicate(){
	 SimpleNode node = new SimpleNode(identifier);

	 node.parent = parent;
	 if(children != null)
		node.children = (java.util.Vector)children.clone();
	 node.info = info;
	 node.type = type;
	 node.opcode = opcode;
	 if(name == null)
   node.name = "";
	 else
		node.name = new String(name);
//System.out.println("setting node.bits to " + bits);
	 node.bits = bits;
	 if(list != null)
		node.list = (java.util.Vector)list.clone();

	 return node;
  }    
  public Node jjtGetChild(int i) {
	return (Node)children.elementAt(i);
  }    
  public int jjtGetNumChildren() {
	return (children == null) ? 0 : children.size();
  }    
  public Node jjtGetParent() { 
	return parent; 
  }    
  public void jjtOpen() {
  }    
  // remove and return child indexed by i 
  // if no such child, return null
  public Node jjtRemoveChild(int i) {
	if (children == null || children.size() <= i) {
	  return null;
	} else {
	  Node n = (Node) children.elementAt(i);
	  children.removeElementAt(i);
	  return n; 
	}
  }    
  /**
   * Replace _this_ node with node 'n'.
   */
  public void jjtReplace(SimpleNode n){
	 parent = n.parent;
	 identifier = new String(n.identifier);
	 if(n.children != null)
		children = (java.util.Vector)n.children.clone();
	 info = n.info;
	 type = n.type;
	 opcode = n.opcode;
	 if(n.name == null)
   name = "";
	 else
		name = new String(n.name);
	 bits = n.bits;
	 if(n.list != null)
		list = (java.util.Vector)n.list.clone();
  }    
  /**
   * replace a child with the one indexed as i in this node.
   * if no such child, do nothing.
   */
  public void jjtReplaceChild(int i, Node node) {
	if (children == null || children.size() <= i) {
	   // do nothing but return.
	}else{
	  children.setElementAt(node, i);
	}
  }    
  public void jjtSetParent(Node n) { 
	parent = n; 
  }    
  public void set(int t, int op) {
	type = t; opcode = op;
  }    
  public void set(int t, int op, String s) {
	type = t; opcode = op; name = s;
  }    
  public void setBit(int pos) {
	bits |= 0x01 << pos;
  }    
  /* These two methods provide a very simple mechanism for attaching
	 arbitrary data to the node. */

  public void setInfo(Object i) { info = i; }    
  public void setObject(java.lang.Object o) {
	info = o;
  }    
  /* You can override these two methods in subclasses of SimpleNode to
	 customize the way the node appears when the tree is dumped.  If
	 your output uses more than one line you should override
	 toString(String), otherwise overriding toString() is probably all
	 you need to do. */

  public String toString() { return identifier; }    
  public String toString(String prefix) { return prefix + toString(); }    
}
