// Created by Weizhong Sun
// to keep the triggers for every update level

public class TriggerNode implements Node, java.io.Serializable 
{
  protected Node parent;
  protected java.util.Vector children;
  protected String relname;		// the updated relation name
  public int opcode;			// the operate code on that relation

  protected CTrigger trigger;	// if childnum == 0; fill in trigger which 
				// consists of three pieces 

  protected Environment env;

  // for initializing
  public TriggerNode(String rlname, Environment env, Node p) {
	relname = rlname;
	trigger = new CTrigger(rlname);
	this.env = env;
	parent = p;
  }    

  public TriggerNode(String rlname, Environment env) {
	relname = rlname;
	trigger = new CTrigger(rlname);
	this.env = env;
	parent = null;	// the root
  }    

  public TriggerNode(Environment env) {
	relname = new String("");
	trigger = new CTrigger(env);
	this.env = env;
	parent = null;	// the root
  }    

  public TriggerNode(Environment env, TriggerNode p) {
	relname = new String("");
	trigger = new CTrigger(env);
	this.env = env;
	parent = p;	
  }

  // for duplicating, inside used.
  private TriggerNode(String rlname, Environment env, CTrigger trigger) {
	relname = rlname;
	this.trigger = trigger;
	this.env = env;
  }    

  /* Override this method if you want to customize how the node dumps
	 out its children. */

/*
  public void dump(String prefix) {
//	System.out.println(toString(prefix) + ":" + type + ":" + opcode +
//	":" + name + ":" + bits);
	if (children != null) {
	  for (java.util.Enumeration e = children.elements();
	  e.hasMoreElements();) {
   TriggerNode n = (TriggerNode)e.nextElement();
   n.dump(prefix + " ");
	  }
	}
  }    
*/

  public void jjtAddChild(Node n) {
	if (children == null) {
	  children = new java.util.Vector();
	}
	children.addElement(n);
  }    

  public void jjtAddChildAtHead(Node n) {
	if (children == null) {
	  children = new java.util.Vector();
	}
	children.insertElementAt(n,0);
  }    

  public void jjtClose() {
	if (children != null) {
	  children.trimToSize();
	}
  }    

/*
  public static Node jjtCreate(String id) {
	return new TriggerNode(id);
  }    
*/

  /*
   * duplicate a tree started with a certain node 
   */
  public TriggerNode jjtDeepDuplicate()
  {  TriggerNode result = new TriggerNode(relname, env, trigger);
	 if (children != null)
	 {  for(int i=0; i<children.size(); i++)
		{  TriggerNode ch = (TriggerNode) children.elementAt(i);
		   TriggerNode chDup = ch.jjtDeepDuplicate();
		   result.jjtAddChild(chDup);
		   chDup.jjtSetParent(result);
		}
	 }
	 return result;
  }    

  /**
   * Duplicate a new node.
   */
  public TriggerNode jjtDuplicate(){
	 TriggerNode node = new TriggerNode(relname, env, trigger);

	 node.parent = parent;
	 if(children != null)
		node.children = (java.util.Vector)children.clone();

	 return node;
  }    

  public Node jjtGetChild(int i) {
	return (Node)children.elementAt(i);
  }    

  public int jjtGetNumChildren() {
	return (children == null) ? 0 : children.size();
  }    

  public Node jjtGetParent() { 
	return parent; 
  }    

  public void jjtOpen() {
  }    

  // remove and return child indexed by i 
  // if no such child, return null
  public Node jjtRemoveChild(int i) {
	if (children == null || children.size() <= i) {
	  return null;
	} else {
	  Node n = (Node) children.elementAt(i);
	  children.removeElementAt(i);
	  return n; 
	}
  }    

  public void jjtSetParent(Node n) { 
	parent = n; 
  }    

  public String getName() { return relname; }    

  public void setName(String rname) { relname = rname; }

  public TriggerNode getNode(String rname, int opc) {
	TriggerNode result = null;
	if (relname.compareTo(rname) == 0 && opcode == opc)
		return this;

	 if (children != null)
	 {  for(int i=0; i<children.size(); i++)
		{  TriggerNode ch = (TriggerNode) children.elementAt(i);
		   if (ch.getName().compareTo(rname) == 0 && ch.opcode == opc)
			return ch;
		   else 
			result = ch.getNode(rname, opc);
		}
	 }
	 return result;
  }
}	/* class TriggerNode */

/* commented out by Yi Zheng
class CTrigger
{
	Relation oldp;	// the parts of relation will be updated
	Relation newp;	// the updated parts of relation
	Relation restp;	// the parts of relation won't be updated
	String	relname;	// the name of relation to be updated

	CTrigger(Environment env) {
		oldp = new Relation(env);
		newp = new Relation(env);
		restp = new Relation(env);
		relname = new String("");
	}
	CTrigger(String rlname) {
		oldp = null;
		newp = null;
		restp = null;
		relname = rlname;
	}
}
*/
