import java.util.*;
/**
 * Class Actualizer: for virtual domain actualization.
  * NOTE: CODE FOR Object Data Changes IS NOT USED:
 * Object Data Changes by JRelCaNu: 
 *  private boolean actWithNull = true; // Handles Dc and Dk values explicitely.
 *  
 * Object Functionality Changes by JRelCaNu: 
 * a. OP_CAST is a OP_BIOPERATOR different from others in terms of type checking: 
 *   1. 2 children nodes are explicitely checked for cast compatibility.
 *   2. The cast from type (of the 0th child node) is stored in the 
 *      left.bits of the OP_TYPE node(1st child node).
 * b. to allow project and select of these in nested relational expressions;
 * c. Allows Dk and Dc as literals, for example, in a SELECT.
 *    This is okay because we know what the actualized
 *    type curdom.type is supposed to be.
 *    Also disallow implicite casting from BOOLEAN to IDLIST.
 *    while continue to allow casting from IDLIST to BOOLEAN.
 * d. NOTE: CODE FOR THIS SUB SECTION IS NOT USED:
 *    Changed from actXYZCell() to actXYZCellWithNull() for casting
 *    with Dc and Dk handling.
 *    Set private boolean actWithNull = true to used Dc and Dk handling.
 * e. Use 'CastToSometype.value()' methods instead of Java '(SomeType)' cast.
 * f. NOTE: CODE FOR THIS SUB SECTION IS NOT USED:
 *    info for a node is a Byte object for BOOLEAN type for actXYZCellWithNull.
 *    This facilitates the Dc and Dk handling for BOOLEAN.

 */
//public class Actualizer implements Constants, TypeConstants{
public class Actualizer extends Actualizer76
{
  //debugging
  private static boolean DEBUG = false;
  private static boolean debug_oct_5 = false;
  private static boolean debug_oct_52 = false;
  private static boolean debug_nov_24 = false;

  //   public static RelTable reltable;    // reference to reltable.
  //   public static DomTable domtable;    // reference to domtable.
  
  public static final int ACC_RESET = 100; // from sungsoo
  public static final int ACC_BYPASS= 200; // from sungsoo
  public static final int ACC_ACCUM = 300; // from sungsoo

  private Environment env;
  
  // START Since jRelixNS
  Hashtable relhtbl = null;

  private NREnvironment NRenv;
  // END Since jRelixNS
  
  private static final long startTime = System.currentTimeMillis(); 
  private Relation srcrel;      // source relation.
  private Relation destrel;     // destination relation.
  private Domain[] destdoms;    // destination domains.
  private Vector virdoms;       // vector for virtual domains (Domain object).
  private Vector actdoms;       // vector for actual domains (Domain obj).
  private Vector alldoms;       // vector for all domains involved.
  private Hashtable vvtree;
  private int fdomcounter = 0;  // temp domain counter for RED & EQUIV.
  
  // Transient varibles.
  private int currow;           // the row# currently passing.
  private Vector curpath;       // Vector for detecting recursive definition.
  private Hashtable vertdom;    // domains containing red or equiv operations.
  private int vertlevel = 0;    // current let of vertical operations.
  private int maxvertlevel = 0;
  
  private boolean nt = false; // new technology for nested domain actualization.
  public static Interpreter interpreter; // nt will use interpreter routines.
  private Vector nestdoms;      // vector for nested domains.
  private Hashtable nesthashs;  // each nested domain has a hashtable.
  private Vector tempnames;
  
   private Vector dom_tempnames; //from Andrey

private Domain[] origDoms = null;	//for bug fix May 7, 2002 Yi Zheng; holds the orignal domains passed to ctor
private Hashtable origTreeTable = null;	//for bug fix May 7, 2002 Yi Zheng; holds the original tree of a virtual dom		
 
    //START Since JRelCaNu d.
   //private boolean actWithNull = false; 
   //End Since JRelCaNu d.

   /* Modified by Andrey Rozenberg:
	 Wrapper or shell around real actualizer*/
/* -------------------IMPORTANT NOTICE----------------------
 * The way Andrey's wrapper is works fine for virtual domais that are themselves defined on TRANSPOSE; for example,
 *	let a be [atr, val] transpose A, B;
 * But it will fail in the case of domain aliasing: let b be a;
 *	This is because when b is dealt with in actualizer_old, its tree will be replaced with a's tree; but it's only
 *	in method actulizing() that b's new tree is interpreted and acted upon; if OP_TRANSPOSE occurrs at this stage,
 *	actualizer won't know how to cope!
 * So don't do domain aliasing in this way: U <- [b] in V; //without actualizing a first
 *	Do it this way if you must: U <- [b] in ([a] in V);
 */

Actualizer(Domain[] doms, Relation srcrel, Environment env)
			  throws InterpretError{
	//Bug fix Yi Zheng May 7, 2002
	//Since the code in this function modifies the tree of a virtual dom directly,
	//the definition of the virtual domain will change if nothing is done to prevent such
	//modification; the consequence of such modification is that a virtual domain
	//can only be actualized once;
	//this bug fix stores the original trees in a hashtable and restore them at the end of actualize();
	origDoms = doms;
	for(int i = 0; i < doms.length; i++)
	{
		if (doms[i].tree != null)
		{
			if (origTreeTable == null)
				origTreeTable = new Hashtable(); //first time use
			SimpleNode origTree = ((SimpleNode) doms[i].tree).jjtDeepDuplicate();
			origTreeTable.put(doms[i].name, origTree);
		}
	}
	
	Domain[] domsList,domsArray = null;	
	String tempname;
		
	dom_tempnames = new Vector();
	tempnames = new Vector(); //from actualizer_old(...)

	
	for(int i=0;i<doms.length;i++)
	{
		if(doms[i].tree != null) // in case of a virtual domain
		{
			SimpleNode nd;
			if((nd=IsInTree(doms[i].tree,OP_TRANSPOSE))!=null)
			{			
			
				//Get list of domains to transpose
				SimpleNode tnode = (SimpleNode) nd.jjtGetChild(0);
				domsList =  this.interpreter.ExpressionListToDomains(tnode,new Hashtable(), env); 
				
				//Project the relation to transpose
				String filename = this.interpreter.nextTempName();
				tempnames.addElement(filename);  
				Relation r = srcrel.duplicateLoadAll();
				Relation rr = r.project(domsList, filename);
				r=rr.duplicateLoadAll();
			
				//Get list of domains to project ('Attr' and/or 'val')
				SimpleNode proj = (SimpleNode) nd.jjtGetParent().jjtGetChild(0);
				if(proj.type!=OP_PROJECTOR)
					throw new InterpretError("No projection list for operator transpose");
				SimpleNode nodel = (SimpleNode) proj.jjtGetChild(0);
				domsList = this.interpreter.ExpressionListToDomains(nodel,new Hashtable(),env); 
		
				
				// create a nested domain
				String tempRel = this.interpreter.nextTempName();
				Domain newDom = new Domain(tempRel,IDLIST,domsList,env);
				env.put(newDom);
				dom_tempnames.addElement(newDom.name); // for clean_up

				//replace the "transpose" expression with the nested domain identifier
				//in the tree feild of the virtal domain
				SimpleNode tnd = SimpleNode.jjtCreateNode("Identifier",OP_IDENTIFIER,OP_IDENTIFIER,
																		newDom.name);
				nd = (SimpleNode) nd.jjtGetParent();
				tnd.parent=(SimpleNode)nd.jjtGetParent();
				nd.jjtReplace(tnd);
				nd.children=null;
				
     	
				
				//Make a dummy relation which includes the "transpose" attribute
				String relname = this.interpreter.nextTempName();
				tempnames.addElement(relname);  
				Domain[] domsL = new Domain[r.domains.length+1];
				int p;
				for(p=0;p<r.domains.length;p++)
					domsL[p]=r.domains[p];
				domsL[p] = newDom;
				Relation new_srcrel = new Relation(relname,domsL,0/*r.numtuples*/,domsL.length,0,env);
				Object[] dArray = new_srcrel.createDataArray(new_srcrel.domains,r.numtuples);
				for(p=0;p<r.domains.length;p++)
					dArray[p]=r.data[p];
		
				
				//Build the "transpose" nested relation
				Relation tRel = env.lookupRel('.'+tempRel,true);
				tempnames.addElement(tRel.name);  
				Object[] dataArray = tRel.createDataArray(tRel.domains,r.numattrs*r.numtuples);
				int offset=r.domains.length;
				for(int j=0;j<r.numtuples;j++) // for each tuple in r
				{
					long s= Surrogate.next();
					((long[]) dArray[new_srcrel.domains.length-1])[j]=s;
					for(int l=0;l<offset;l++)  // for each attribute in r
						for(int k=0;k<tRel.domains.length;k++)        // for each attribute in tRel
						{
							if(k==0)
								((long[]) dataArray[k])[j*offset+l]=s;
							else 
							{	if(tRel.domains[k].type==ATTRIBUTE) 
									((String[]) dataArray[k])[j*offset+l]=r.domains[l].name;
										
								else if(tRel.domains[k].type==UNIVERSAL)
								{
										String str = new String();
										Object b = new Object();
										b = r.data[l]; 
										
										Object a = new Object();
										a= java.lang.reflect.Array.get(b,j);
										str = a.toString();
																	
									//	String str =String.valueOf(((Object[]) r.data[l])[j]);
										str = Utility.type2String(r.domains[l].type)+';'+str;
										((String[]) dataArray[k])[j*offset+l]=str;
								}
							}
						}
				}
					
				tRel.append(dataArray,r.numattrs*r.numtuples);
				tRel.load();
				new_srcrel.append(dArray,r.numtuples);
				new_srcrel.load();
				// restore original relation
				srcrel = srcrel.mujoin(OP_IJOIN,null,new_srcrel,null,new_srcrel.name);
				
		//		srcrel = new_srcrel;
			}	
								
			
			
			
			else 
			{
			/*SimpleNode*/ nd = (SimpleNode) doms[i].tree.jjtGetChild(0);
		//	System.out.println("doms[i].tree: ");
		//	Utility.dump(doms[i].tree, " tree: ");
		//	System.out.println("doms[i].tree's first child:");
		//	Utility.dump(nd, "  ");
			if(nd.children!=null)
			{
				if(nd.type==OP_QUOTE)
				{
					nd = (SimpleNode)nd.jjtRemoveChild(0);
					SimpleNode ident = SimpleNode.jjtCreateNode("Literal",OP_LITERAL,OP_ATTRIBUTE,null);
					ident.info=(String) nd.name;
					doms[i].tree.jjtRemoveChild(0);		
					doms[i].tree.jjtAddChild(ident);
					ident.jjtSetParent(doms[i].tree);
				}
				else if(nd.jjtGetNumChildren()>1) //what is this used for?
				{ int oldtype = nd.type;
					SimpleNode oldNode = nd; //Yi Zheng June 24, 2002 	
					SimpleNode nd2 = null;
					//if added July 8, 2002 by Yi Zheng
					if(oldNode.opcode != OP_REDFUNCALL)
					{
						nd = (SimpleNode) nd.jjtGetChild(1);
						nd2 = (oldNode.jjtGetNumChildren() > 2)? (SimpleNode)oldNode.jjtGetChild(2) : null; //Yi Zheng June 24, 2002, for the case of OP_PAR, where both orderlist and bylist are present
					}
					else if(oldNode.jjtGetNumChildren() > 2)
					{
						nd = (SimpleNode) nd.jjtGetChild(2);
						nd2 = (oldNode.jjtGetNumChildren() > 3)? (SimpleNode)oldNode.jjtGetChild(3) : null; //Yi Zheng June 24, 2002, for the case of OP_PAR, where both orderlist and bylist are present
					}
				//	if(nd.type==OP_EXPRLIST)	//changed by Yi Zheng June 23, 2002
					if((nd != null) && (nd.type == OP_EXPRLIST) && (oldtype == OP_EQUIV || oldtype == OP_FUN || oldtype == OP_PAR))	
					{
						Hashtable htable =  new Hashtable();
						Domain domArray[]=this.interpreter.ExpressionListToDomains(nd,htable,env);
						if(domArray.length > 0) //if added by Yi Zheng June 24, 2002
						{
							int NumOfChildren=nd.jjtGetNumChildren();
							for(int j=NumOfChildren-1;j>-1;j--)
								nd.jjtRemoveChild(j);
		
							for(int j=0;j<domArray.length;j++)
							{
								SimpleNode ident = SimpleNode.jjtCreateNode("Identifier",OP_IDENTIFIER,OP_IDENTIFIER,
																			domArray[j].name);
								nd.jjtAddChild(ident);
								ident.jjtSetParent(nd);
							}
						}
					}

					if((nd2 != null) && (nd.type == OP_EXPRLIST) && (oldtype == OP_PAR))	
					{
						Hashtable htable =  new Hashtable();
						Domain domArray[]=this.interpreter.ExpressionListToDomains(nd2,htable,env);
						if(domArray.length > 0) //if added by Yi Zheng June 24, 2002
						{
							int NumOfChildren=nd2.jjtGetNumChildren();
							for(int j=NumOfChildren-1;j>-1;j--)
								nd2.jjtRemoveChild(j);
		
							for(int j=0;j<domArray.length;j++)
							{
								SimpleNode ident = SimpleNode.jjtCreateNode("Identifier",OP_IDENTIFIER,OP_IDENTIFIER,
																			domArray[j].name);
								nd2.jjtAddChild(ident);
								ident.jjtSetParent(nd2);
							}
						}
					}
					
				}
			}
			}
		}
	}	
	Actualizer_old(doms, srcrel, env);
	
}
 
  /**
   * Constructor initializing object variables and checks to see 
   * if the domains exist and are actualizable
   * for the source relation srcrel.
   *
   * JRelixNS modified/enhanced this method:
   *   In jRelixNS, nullary-Relation/IDLIST and boolean are now compatible;
   *   so relax the checks to allow this mix.
   *
   * JRelCaNu modified/enhanced this method:
   * - Allows Dk and Dc as literals.
   * - disallow implicite casting from BOOLEAN to IDLIST added in JRelixNS.
   */
//comments from Andy May 12, 2002
//  Actualizer(Domain[] doms, Relation srcrel, Environment env) //modified to merge with Andrey's code
    private void Actualizer_old(Domain[] doms, Relation srcrel, Environment env)
       throws InterpretError
  {
    
    this.srcrel = srcrel;
    this.destdoms = doms;
    this.env = env;

  //  this.relhtbl = new Hashtable(); // from andi //Andy May 12, 2002
    
    virdoms = new Vector();       // vector for virtual domains (Domain obj).
    alldoms = new Vector();       // vector for all domains involved.
    actdoms = new Vector();       // REVISIT: vector for actual domains.
    curpath = new Vector();      // for detecting recursive definition.
    vertdom = new Hashtable();   // for checking if we need second pass.
    vvtree = new Hashtable();	//maps virtual domain to its tree
    
    if(nt){
      nestdoms = new Vector();
      nesthashs = new Hashtable();
    }
    if (tempnames == null)  //if added by Yi Zheng May 25, 2002, to conform to Andrey's code
    	tempnames = new Vector();
    
    // Fill the alldoms vector with domains in source relation.
    for(int i = 0; i < srcrel.domains.length; i++){
      alldoms.addElement(srcrel.domains[i]);
      //System.out.println("added to alldoms: srcrel.domains["+ i+ "] name = "+ srcrel.domains[i].name);
   //   System.out.println(" i "+ i+ " type "+ srcrel.domains[i].type);
    }
   // System.out.println("get org rel domains" );
 
    // Check to see if the domains exist and are actualizable.
    // This is a big loop going through the virtual domains and in
    // charge of testing their validation, type compatibility, 
    // mutual recursive definitions detection and finally
    // responsible for building a optimized tree for each domain in
    // order to prepare for the tuple-level actualization.
    for(int i = 0; i < destdoms.length; i++){
      Domain curdom = destdoms[i];
      if(debug_oct_52)
      System.out.println(" curdom "+ i+ " name "+ curdom.name);
   //   System.out.println(" curdom "+ i+ " type "+ curdom.type);
  //    if(i != 0){
//	System.out.println(" pre dom "+ (i-1)+ " name "+ destdoms[i-1].name);
//	System.out.println(" pre dom "+ (i-1)+ " type "+ destdoms[i-1].type);
   //   }
      //System.out.println("Actualizer: curdoms initialized");
      // Check each domain in doms[]...
      if(srcrel.isDomainIn(curdom)){
	// This is an actual domain in this relation's domain list.
	// REVISIT: do we need to maintain this 'actdoms'?
	
	
	actdoms.addElement(curdom);
      }
      else if(curdom.tree == null){
	// This is an actual domain which is not declared in this 
	// relation's actual domain list, so we can't do it.
	throw new InterpretError("domain '"+curdom.name+
				 "' is not in the domain list of relation '"+srcrel.name+"'.");
      }
      else{
	// This must be a virtual domain, and we want to investigate.
	// Note: 'curdom' might not be existing in domtable, so we
	// want to analyze the domain's tree field directly.
	SimpleNode vnode = 
	  ((SimpleNode)curdom.tree.jjtGetChild(0)).jjtDuplicate();
	
	curpath.addElement(curdom);
	
	// Have to test OP_RED & OP_EQUIV since they are special.
	if((vnode.type == OP_RED || vnode.type == OP_EQUIV || vnode.type == OP_FUN || vnode.type == OP_PAR)
	  && (curdom.type != IDLIST || !nt) ) // changed according to sungsoo's code. hongyu Feb0402
	  {
	    if(!vertdom.containsKey(curdom) || vertlevel > ((Integer)vertdom.get(curdom)).intValue()){
	      vertdom.put(curdom, new Integer(vertlevel));
	    }
	    
	    // Need to duplicate the expression list node.
	    if(vnode.type == OP_EQUIV || vnode.type == OP_FUN || vnode.type == OP_PAR)
	      // changed according to sungsoo's code. hongyu Feb0402
	      {
			// Get the child of expression list.
			//if added by Yi Zheng July 8, 2002
			if(vnode.opcode != OP_REDFUNCALL)
			{
				SimpleNode child = (SimpleNode)vnode.jjtGetChild(1);
				vnode.jjtReplaceChild(1, child.jjtDuplicate());
			}
			else
			{
				SimpleNode child = (SimpleNode)vnode.jjtGetChild(2);
				vnode.jjtReplaceChild(2, child.jjtDuplicate());
			}
			
	      }
	    
	    if(maxvertlevel < ++vertlevel)
	      maxvertlevel = vertlevel;
	  }
	
	try{
	  // Try to build the whole tree. ..will cause exceptions
	  // in case there are problems.
	  //System.out.println("Actualizer: building tree");
	  int domtype = buildTree(vnode);
	  //System.out.println("Actualizer: building tree done");
	  if(debug_oct_52)
	  	Utility.dump(vnode,"after buildTree for vnode:");
	  // Virtual and actual type mismatch. ..should never happen.
	                             // START Since jRelixNS
                           // In jRelixSJ, nullary-Relation/IDLIST and
                           // boolean are now compatible.
    if (Relation.isSigmaJoinImplemented != true) {
//	  if (env.isSigmaJoinImplemented != true) {  //Andy May 12, 2002
	    if(domtype != curdom.type){
	      throw new InterpretError("virtual type is different"+
				       " from the actualized type.");
	    }
	  } 
	  else {
//	    if(domtype != curdom.type &&
//	       !((domtype == BOOLEAN && curdom.type == IDLIST) ||
//		 (domtype == IDLIST && curdom.type == BOOLEAN ) ))

               //START Since JRelCaNu c.
               // Allows Dk and Dc as literals, for example, in a SELECT.
			   // This is okay because we know what the actualized
			   // type curdom.type is supposed to be.
			   // Also disallow implicite casting from BOOLEAN to IDLIST.
			   if(domtype != curdom.type &&
			   (!(domtype == DK)) && 
			   (!(domtype == DC)) &&
			   (!(//Not allowed => (domtype == BOOLEAN && curdom.type == IDLIST) ||
			     (domtype == IDLIST && curdom.type == BOOLEAN ) )))
                           //End Since JRelCaNu c.

	      {
		throw new InterpretError("virtual type" + Utility.type2String(curdom.type) + " is different"+
					 " from the actualized type " + Utility.type2String(domtype));
	      }
	  }
	}// END Since jRelixNS
	catch(InterpretError e){
	  vnode.dump("InterpretError Dump ==>  ");
	  throw new InterpretError("domain '"+doms[i].name+"' is"+
				   " unactualizable: "+e.toString().substring(16));
	}
	
	curpath.removeElement(curdom);
	//vnode.dump("==>  ");
	
	if((vnode.type == OP_RED || vnode.type == OP_EQUIV || vnode.type == OP_FUN || vnode.type == OP_PAR)
	   && (curdom.type != IDLIST || !nt) ) // changed according to sungsoo's code. hongyu Feb0402
	  vertlevel--;
	else if(curdom.type == IDLIST && nt){
//	  System.out.println(" IDLIST && nt ");
	  vnode.info = curdom;
	  
	  // Each IDLIST domain has an hashtable(newdom, newrel).
	  // e.g.                                    ijoin
	  //  let x be S ijoin T ijoin U;            / x \
	    //                                      ijoin   U
	  //                                      / 0 \
	    // Hence, contents in domrelhash:      S     T
	  //                  Dom      Rel
	  //             1.    x       .x
	  //             2.    0       .0
	  Hashtable domrelhash = new Hashtable();
	  
	  // Relation rel = reltable.getRel("."+curdom.name);
	  // REVISIT: Zhongxia, I used true but check it please
	  Relation rel = env.lookupRel("."+curdom.name, true);
	  
	  if(rel == null)
	    // should never happen..
	    throw new InterpretError("relation '."+curdom.name+
				     "' not found in reltable.");
	  
	  // Any intermediate nested domain e.g. '0' in above example
	  // will be put in this 'domrelhash' hashtable by 
	  // processIDListDom() method.
	  domrelhash.put(curdom, rel);
	  processIDListDom(vnode, domrelhash, true);
	  
	  nesthashs.put(curdom, domrelhash);
	  nestdoms.addElement(curdom);
	}
	else 
	  virdoms.addElement(curdom);
	
	if (debug_oct_52)
		Utility.dump(vnode, "before vvtree.put(" + curdom.name + "):");
		
	vvtree.put(curdom, vnode);
	
	if (debug_oct_52)
	{
		SimpleNode vnode2 = (SimpleNode) vvtree.get(curdom);
		Utility.dump(vnode2, "lookup :");	
	}
	alldoms.addElement(curdom);
      }
    }
	if(debug_oct_52)
    System.out.println("end of declaring actualizer" ); 
  }
  
   /**
	*  Actualize a boolean-typed cell., wrapper by Andrey Rozenberg
	*/
  private boolean actBoolCell(SimpleNode node) throws InterpretError
  {
	  
//	  if(node.type ==OP_BIOPERATOR && node.identifier=="Compare") //cannot compare Strings like that, modified by Yi Zheng May 27, 2002
	if(node.type ==OP_BIOPERATOR && node.identifier.compareTo("Compare") == 0)	

	  {
		  SimpleNode left = (SimpleNode)node.jjtGetChild(0);
		  SimpleNode right = (SimpleNode)node.jjtGetChild(1);
		 // Domain d = env.lookupDom(left.name,true);
		//Bug: the follow statement will cause NullPointerException
		//in this case:R <- [ITEM] quant (# = 2) DEPT in  SUPPLY; where # is not a domain
		//in addition, it is equally likely that the domain appears on the RHS
		//Fixes provided by Yi Zheng May 27, 2002 as follows
	  
		//bug fix by Yi Zheng May 27, 2002
		Domain dleft = env.lookupDom(left.name, true);
		Domain dright = env.lookupDom(right.name, true);

		if(!((dleft == null) && (dright == null))) //either one of the domains exists
		{
		  if(dleft != null && dleft.type==UNIVERSAL)
			{
			//DEBUGGING
			//System.out.println("left is UNIVERSAL");
			    SimpleNode tnode=null;
				//System.out.println("currow = " + currow);
				String str = ((String[])destrel.data[left.bits])[currow];
				 //System.out.println("str = " + str);
				String type = str.substring(0,str.indexOf(';'));
				String value = str.substring(str.indexOf(";")+1);
				if(type.compareTo(Utility.type2String(right.opcode-OP_BOOLEAN))!=0)
					return false;
				
				
				tnode = node.jjtDeepDuplicate();
			//added by Yi Zheng May 27, 2002
				tnode.jjtSetParent(node.jjtGetParent());
			
			SimpleNode tleft = (SimpleNode)tnode.jjtGetChild(0);
		
			//added by Yi Zheng May 27, 2002
				int bitVal = right.opcode - OP_BOOLEAN;
				tnode.bits = bitVal; //this will be needed by actBoolCell_old				
	
			//	switch(right.opcode-OP_BOOLEAN) //modified by Yi Zheng May 27, 2002
				switch(bitVal)
				{
					case STRING:	tleft.info= value;
									break;
					case NUMERIC:	//added by Yiyi June 5'02
						tleft.info = value;
						break;

					case INTEGER:	tleft.info= Integer.valueOf(value);
									break;
					case FLOAT:		tleft.info=Float.valueOf(value);
									break;
					case SHORT:		tleft.info=Short.valueOf(value);
									break;
					case LONG:		tleft.info=Long.valueOf(value);
									break;
					case DOUBLE:	tleft.info=Double.valueOf(value);
									break;
					case BOOLEAN:	//tleft.info=new Boolean(value.compareTo("1")==0);
									//tleft.info=new Boolean((String)value=="1");
									tleft.info=Byte.valueOf(value);
									break;
					default:
							throw new InterpretError("actBoolCell: to be implemented.");
				}
				tleft.identifier = "Literal"; //June 28, 2002
				tleft.type = OP_LITERAL;
				tleft.opcode=right.opcode;
				tleft.name=null;
				tleft.bits=0;
				node = tnode;	//commented out by Yi Zheng May 27, 2002, reverted June 28, 2002
			//added by Yi Zheng May 27, 2002 

			//DEBUGGING
			//System.out.println("tleft is: ");
			//Utility.dump(tleft, " -- ");
		//	node.jjtReplace(tnode);
		
			//Utility.dump(node, "after being replaced by tnode, node is: ");
			//System.out.println("tleft.info is " + ((SimpleNode)node.jjtGetChild(0)).info.toString());
		  } 		

		  //apply the same logic to the RHS
		if(dright != null && dright.type==UNIVERSAL)
			{
			//DEBUGGING
			//System.out.println("right is UNIVERSAL");
			    SimpleNode tnode=null;
				String str = ((String[])destrel.data[right.bits])[currow];
				String type = str.substring(0,str.indexOf(';'));
				String value = str.substring(str.indexOf(";")+1);
				if(type.compareTo(Utility.type2String(left.opcode-OP_BOOLEAN))!=0)
					return false;
				
				
				tnode = node.jjtDeepDuplicate();

				//added by Yi Zheng May 27, 2002
				tnode.jjtSetParent(node.jjtGetParent());

				SimpleNode tright = (SimpleNode)tnode.jjtGetChild(1);
				
				//added by Yi Zheng May 27, 2002
				int bitVal = left.opcode-OP_BOOLEAN;
				tnode.bits = bitVal;	//needed by actBoolCell_old

			//	switch(left.opcode-OP_BOOLEAN) //modified by Yi Zheng May 27, 2002
				switch(bitVal)
				{
					case STRING:	tright.info= value;
									break;
					case NUMERIC://added by Yiyi June 5'02
						tright.info = value;
						break;
					case INTEGER:	tright.info= Integer.valueOf(value);
									break;
					case FLOAT:		tright.info=Float.valueOf(value);
									break;
					case SHORT:		tright.info=Short.valueOf(value);
									break;
					case LONG:		tright.info=Long.valueOf(value);
									break;
					case DOUBLE:	tright.info=Double.valueOf(value);
									break;
					case BOOLEAN:	//tright.info=new Boolean(value.compareTo("1")==0);
									//tright.info=new Boolean((String)value=="1");
									tright.info=Byte.valueOf(value);
									break;
					default:
							throw new InterpretError("actBoolCell: to be implemented.");
				}
				tright.identifier = "Identifier";
				tright.type = OP_LITERAL;
				tright.opcode=left.opcode;
				tright.name=null; //will this cause problem? REVISIT, Yi Zheng May 27, 2002
				tright.bits=0;
				//DEBUGGING
				//System.out.println("tright is: ");
				//Utility.dump(tright, " -- ");
	//			node.jjtReplace(tnode);
				node = tnode;	 
		  }	
		}

		//end of bug fix May 27, 2002

		/* the original code by Andrey
		  if(env.lookupDom(left.name,true).type==UNIVERSAL)
			{
			    SimpleNode tnode=null;
				String str = ((String[])destrel.data[left.bits])[currow];
				String type = str.substring(0,str.indexOf(';'));
				String value = str.substring(str.indexOf(";")+1);
				if(type.compareTo(Utility.type2String(right.opcode-OP_BOOLEAN))!=0)
					return false;
				
				
				tnode = node.jjtDeepDuplicate();
				SimpleNode tleft = (SimpleNode)tnode.jjtGetChild(0);
				switch(right.opcode-OP_BOOLEAN)
				{
					case STRING:	tleft.info= value;
									break;
					case INTEGER:	tleft.info= Integer.valueOf(value);
									break;
					case FLOAT:		tleft.info=Float.valueOf(value);
									break;
					case SHORT:		tleft.info=Short.valueOf(value);
									break;
					case LONG:		tleft.info=Long.valueOf(value);
									break;
					case DOUBLE:	tleft.info=Double.valueOf(value);
									break;
					case BOOLEAN:	//tleft.info=new Boolean(value.compareTo("1")==0);
									//tleft.info=new Boolean((String)value=="1");
									tleft.info=Byte.valueOf(value);
									break;
					default:
							throw new InterpretError("actBoolCell: to be implemented.");
				}
				tleft.type = OP_LITERAL;
				tleft.opcode=right.opcode;
				tleft.name=null;
				tleft.bits=0;
				node = tnode;	 
		  } */
	  }
	  
	  return actBoolCell_old(node);
  }
  
  
  /**
   *  Actualize a boolean-typed cell.
           *  
        *  JRelixNS modified/enhanced this method:
        *   In jRelixNS, if node type is not scalar boolean cell,
        *   check and see if it is a relation cell instead of throwing an
	*   InterpretError. This
        *   because nullary-Relation/IDLIST and boolean are now compatible.

   *///Andy May 12, 2002
 //    private boolean actBoolCell(SimpleNode node) throws InterpretError //modified May 25, 2002 to conform to Andrey's code
  private boolean actBoolCell_old(SimpleNode node) throws InterpretError
  {
//    System.out.println("Cell . Nodename " +node.name );
//    System.out.println("Cell . type " +node.type );
//    System.out.println("Cell . opcode " +node.opcode );
//    System.out.println("Cell . opbits " +node.bits );
    switch(node.type){
    	case OP_FUNCTION:
    	   	throw new InterpretError("actBoolCell: to be implemented.");

    case OP_BIOPERATOR:
      SimpleNode left = (SimpleNode)node.jjtGetChild(0);
      SimpleNode right = (SimpleNode)node.jjtGetChild(1);
      
//      System.out.println("left type " +left.type + " opcode " + left.opcode );
//      System.out.println("right type " +right.type + " opcode " + right.opcode );
      
      switch(node.opcode){
      case OP_OR:
	return(actBoolCell(left) || actBoolCell(right));
      case OP_AND:
	return(actBoolCell(left) && actBoolCell(right));
      case OP_EQ:
	//System.out.println("in OP_EQ, node.bits:" + node.bits);
	switch(node.bits){
	  
	  //case INTEGER:
	  //System.out.println("in case INTEGER");
	  //return actIntCell(left) == actIntCell(right);
	  //case STRING:
	  //System.out.println("in case STRING");
	  //return (actStrCell(left)).compareTo(actStrCell(right)) == 0;
	  
	  //Revised by hongyu Sept,01
	  case BOOLEAN: ///case added by Yi Zheng May 25, 2002
	  	return actBoolCell(left) == actBoolCell(right);
	case SHORT:
	  //throw new InterpretError("actBoolCell: to be implemented..");
		return (short)actIntCell(left) == (short)actIntCell(right);
	case LONG:
	  return actLongCell(left) == actLongCell(right);
	case FLOAT:  
	case DOUBLE:{
//	  System.out.println("DOUBLE -------------------------------");
//	  System.out.println("actDoubleCell(left) = " + actDoubleCell(left));
//	  System.out.println("actDoubleCell(right) = "+ actDoubleCell(right));
	  return actDoubleCell(left) == actDoubleCell(right);
	}
	case NUMERIC://added by Yiyi June 4'02
		return (actNumCell(left).compareTo(actNumCell(right)) == 0);

	
	case INTEGER:{
//	  System.out.println(" INTEGER----------------------------");
	  return actIntCell(left) == actIntCell(right);
	}
	case TEXT: // Hongyu added
	case STMT: // Hongyu added
	case EXPR: // Hongyu added
	case IDLIST:// ??Hongyu added 
	 case UNIVERSAL:  //from Andrey
//				   return (actUniCell(left))==(actUniCell(right));
	   case ATTRIBUTE: //from Andrey
	case STRING:
	  //System.out.println("in case STRING");
	//System.out.println("left is " + actStrCell(left));
   	//System.out.println("right is " + actStrCell(right));
	  return (actStrCell(left)).compareTo(actStrCell(right)) == 0;
                      //START Since JRelCaNu a.
       default:
      throw new InterpretError("actBoolCell: OP_BIO to be implemented..");

	}
	
//	break;
	                        //END Since JRelCaNu a.

      case OP_NEQ:
	switch(node.bits){
	  
	  //case INTEGER:
	  //return actIntCell(left) != actIntCell(right);
	  //case STRING:
	  //return (actStrCell(left)).compareTo(actStrCell(right)) != 0;
	  
	  //Revised by hongyu Sept,01
	  case BOOLEAN: //case added by Yi Zheng Sept. 29, 2002
	  	return actBoolCell(left) != actBoolCell(right);
	case SHORT:
	 // throw new InterpretError("actBoolCell: to be implemented..");
		return (short)actIntCell(left) != (short)actIntCell(right);
	case LONG:
	  return actLongCell(left) != actLongCell(right);
	case FLOAT:  
	case DOUBLE:
	  return actDoubleCell(left) != actDoubleCell(right);
	case NUMERIC://case added by Yiyi June4'02
	  			return (actNumCell(left).compareTo(actNumCell(right)) != 0);
	  
	  
	case INTEGER:
	  return actIntCell(left) != actIntCell(right);
	case TEXT: // Hongyu added
	case STMT: // Hongyu added
	case EXPR: // Hongyu added
	case IDLIST:// ??Hongyu added
	case STRING:
	  return (actStrCell(left)).compareTo(actStrCell(right)) != 0;
	                          //START Since JRelCaNu a.
		           default:
			      throw new InterpretError("actBoolCell: OP_BIO to be implemented..");
	}
	//break;
	                        //END Since JRelCaNu a.

      case OP_GT:
	switch(node.bits){
	  
	  //case INTEGER:
	  //return actIntCell(left) > actIntCell(right);
	  //case STRING:
	  //return (actStrCell(left)).compareTo(actStrCell(right)) > 0;
	  
	  //Revised by hongyu Sept,01
	case SHORT:
		return (short)actIntCell(left) > (short)actIntCell(right);
	//  throw new InterpretError("actBoolCell: to be implemented..");
	case LONG:
	  return actLongCell(left) > actLongCell(right);
	case FLOAT:  
	case DOUBLE:
	  return actDoubleCell(left) > actDoubleCell(right);
	  //case NUMERIC: 
	 case NUMERIC://case added by Yiyi June 4'02
	      	return (actNumCell(left).compareTo(actNumCell(right)) > 0);
	  
	case INTEGER:
	  return actIntCell(left) > actIntCell(right);
	case TEXT:// Hongyu added
	case STMT:// Hongyu added
	case EXPR:// Hongyu added
	case IDLIST:// ??Hongyu added
	case STRING:
	  return (actStrCell(left)).compareTo(actStrCell(right)) > 0;
	                          //START Since JRelCaNu a.
		           default:
			      throw new InterpretError("actBoolCell: OP_BIO to be implemented..");

	}
//	break;
                        //END Since JRelCaNu a.

      case OP_LT:
	switch(node.bits){ 
	  
	  //case INTEGER:
	  //return actIntCell(left) < actIntCell(right);
	  //case STRING:
	  //return (actStrCell(left)).compareTo(actStrCell(right)) < 0;
	  
	  //Revised by hongyu Sept,01
	case SHORT:
	//  throw new InterpretError("actBoolCell: to be implemented..");
		return (short)actIntCell(left) < (short)actIntCell(right);
	case LONG:
	  return actLongCell(left) < actLongCell(right);
	case FLOAT:  
	case DOUBLE:
	  return actDoubleCell(left) < actDoubleCell(right);
	  //case NUMERIC:
	case NUMERIC://case added by Yiyi June 4'02
	      	return (actNumCell(left).compareTo(actNumCell(right)) < 0);
 
	case INTEGER:
	  return actIntCell(left) < actIntCell(right);
	case TEXT:// Hongyu added
	case STMT:// Hongyu added
	case EXPR:// Hongyu added
	case IDLIST://?? Hongyu added
	case STRING:
	  return (actStrCell(left)).compareTo(actStrCell(right)) < 0;
	                          //START Since JRelCaNu a.
		           default:
			      throw new InterpretError("actBoolCell: OP_BIO to be implemented..");

	}
//	break;
                        //END Since JRelCaNu a.

      case OP_GE:
	switch(node.bits){
	  
	  //case INTEGER:
	  //return actIntCell(left) >= actIntCell(right);
	  //case STRING:
	  //return (actStrCell(left)).compareTo(actStrCell(right)) >= 0;
	  
	  //Revised by hongyu Sept,01
	case SHORT:
		return (short)actIntCell(left) >= (short)actIntCell(right);
	//  throw new InterpretError("actBoolCell: to be implemented..");
	case LONG:
	  return actLongCell(left) >= actLongCell(right);
	case FLOAT:  
	case DOUBLE:
	  return actDoubleCell(left) >= actDoubleCell(right);
	  //case NUMERIC: 
	case NUMERIC://case added by Yiyi June 4'02
	      	return (actNumCell(left).compareTo(actNumCell(right)) >= 0);
 
	case INTEGER:
	  return actIntCell(left) >= actIntCell(right);
	case TEXT:// Hongyu added
	case STMT:// Hongyu added
	case EXPR:// Hongyu added
	case IDLIST:// ??Hongyu added
	case STRING:
	  return (actStrCell(left)).compareTo(actStrCell(right)) >= 0;
	                          //START Since JRelCaNu a.
		           default:
			      throw new InterpretError("actBoolCell: OP_BIO to be implemented for " +
			      Utility.type2String(node.bits));

	}
//	break;
                        //END Since JRelCaNu a.

      case OP_LE:
	switch(node.bits){
	  
	  //case INTEGER:
	  //return actIntCell(left) <= actIntCell(right);
	  //case STRING:
	  //return (actStrCell(left)).compareTo(actStrCell(right)) <= 0;
	  
	  //Revised by hongyu Sept,01
	case SHORT:
		return (short)actIntCell(left) <= (short)actIntCell(right);
	 // throw new InterpretError("actBoolCell: to be implemented..");
	case LONG:
	  return actLongCell(left) <= actLongCell(right);
	case FLOAT:  
	case DOUBLE:
	  return actDoubleCell(left) <= actDoubleCell(right);
	  //case NUMERIC: 
	case NUMERIC:
	      	return actNumCell(left).compareTo(actNumCell(right)) <= 0;
	  
	case INTEGER:
	  return actIntCell(left) <= actIntCell(right);
	case TEXT:// Hongyu added
	case STMT:// Hongyu added
	case EXPR:// Hongyu added
	case IDLIST:// ??Hongyu added
	case STRING:
	  return (actStrCell(left)).compareTo(actStrCell(right)) <= 0;
	                          //START Since JRelCaNu a.
		           default:
			      throw new InterpretError("actBoolCell: OP_BIO to be implemented..");

	}
//	break;
                        //END Since JRelCaNu a.
          //START Since JRelCaNu a.
            case OP_CAST:
                    // Cast from type had been stored in left.bits of 
                    // the the left OP_TYPE node in buildTree and
                    // traverseType; so, we can use it in the switch.
                    switch(left.bits){
                    case BOOLEAN:
                       return actBoolCell(right);
                    case STRING:
                       return CastToBool.value(CastToBoolByte.value(actStrCell(right)));
                    case SHORT:
                       return CastToBool.value(CastToBoolByte.value((short)CastToShort.value(actIntCell(right))));
                    case INTEGER:
                       return CastToBool.value(CastToBoolByte.value(actIntCell(right)));
                    case LONG:
                       return CastToBool.value(CastToBoolByte.value(actLongCell(right)));
                    case DOUBLE:
                       return CastToBool.value(CastToBoolByte.value(actDoubleCell(right)));
                    case FLOAT:
                       return CastToBool.value(CastToBoolByte.value(CastToFloat.value(actDoubleCell(right))));
                    case IDLIST:
                       return(actBoolRelCell(right,null,null,null,null,null));
//YYY
                    case NUMERIC:
                       return CastToBool.value(CastToBoolByte.value(actNumCell(right)));
                  default:
                       Utility.dump(node,"switched on node.bits"+node.bits+" in actBoolCellWithNull->");
                       throw new InterpretError("actBoolCellWithNull: to be implemented.");
                    }
          //END Since JRelCaNu a.
      default:
//	throw new InterpretError("actBoolCell: OP_BIO to be implemented.."); //From Andy Sept 5
      }
      //return false; // This is very bad and will cause problems. hongyu Sept 01
      throw new InterpretError("actBoolCell: OP_BIO to be implemented..");
    case OP_UOPERATOR:
      // node.opcode must be OP_UNOT.
      return(!actBoolCell((SimpleNode)node.jjtGetChild(0)));
    case OP_IDENTIFIER:{
      // this node must of boolean (byte) type.
      //destrel.print();
      // BEGIN jRelixNS BUGFIX 
// Check if an OP_IDENTIFIER is defined as a BOOLEAN or IDLIST 
// in the original relation and handle accordingly.
// Specifically, have to call actBoolRelCell for IDLIST.
                 if (Relation.isSigmaJoinImplemented == true) {
		 switch(destrel.domains[node.bits].type){
			case BOOLEAN:
		 // this node must of boolean (byte) type.
		return ((byte[])destrel.data[node.bits])[currow] == (byte)1;
			case IDLIST:
		 // an IDLIST type node is also acceptible.
                return(actBoolRelCell(node,null,null,null,null,null));
		        default:
			throw new InterpretError("actBoolCell: OP_BIO to be implemented..");
		 }
		 }else {
		 // this node must of boolean (byte) type.
	return ((byte[])destrel.data[node.bits])[currow] == (byte)1;
		}
// END jRelixNS BUGFIX 

     // return ((byte[])destrel.data[node.bits])[currow] == (byte)1;
    }
    case OP_LITERAL:
          //START Since JRelCaNu a.
             if(node.info instanceof Byte)
	    return(((Byte)node.info).byteValue() == BOOL_TRUE);
             else 
             {
                Utility.dump(node,"actBoolCell OP_LITERAL Node --->");
                throw new InterpretError("actBoolCellWithNull: Not a boolean literal.");
	 }

      //REVISIT: node.opcode must be boolean type.
      //return(((Byte)node.info).intValue() == 1);
                //END Since JRelCaNu a.

    case OP_RED:
    case OP_EQUIV:// add songsoo? hongyu asked this question Feb 2002
    case OP_FUN:
    case OP_PAR:
     if(node.opcode != OP_REDFUNCALL)
     {
      boolean actValue= actBoolCell((SimpleNode)node.jjtGetChild(0));
      
      switch(node.opcode){
      case OP_OR:
	if(node.info == null)  node.info = new Boolean(false);
	node.info = new Boolean(((Boolean)node.info).booleanValue() ||
				actValue);
	break;
      case OP_AND:
	if(node.info == null)  node.info = new Boolean(true);
	node.info = new Boolean(((Boolean)node.info).booleanValue() &&
				actValue);
	break;
		//addition by Yi Zheng July 5, 2002
		case OP_NOP:
		if (node.info == null)
			node.info = new Boolean(actValue);
		else // a nop b --> randomly pick a or b
		{
			boolean rb = new Random().nextBoolean();
			if (rb) 
				node.info = new Boolean(actValue);
		}
		break;
		//end of addition July 5, 2002
      default:
	throw new InterpretError("actBoolCell: Red/Equiv to be implemented.");
      }
      return actValue;
    }
    else //added by Yi Zheng July 9, 2002
    {
    	boolean actValue = actBoolCell((SimpleNode)node.jjtGetChild(1));
    	RedFunCallAccumNode(node, BOOLEAN, new Boolean(actValue));
    	return actValue;
    }
    case OP_IFTHENELSE:
      if(actBoolCell((SimpleNode)node.jjtGetChild(0)))
	return actBoolCell((SimpleNode)node.jjtGetChild(1));
      else
	return actBoolCell((SimpleNode)node.jjtGetChild(2));
	        //added by Yi Zheng July 2, 2002
    // to implement level lifting
    case OP_PROJECT:
    case OP_TSELECT: {
    	Vector dnames = new Vector();
    	Vector rname = new Vector();
    	Vector rindx = new Vector();
    	relhtbl = new Hashtable();
    	String tname0 = interpreter.nextTempName();
    	String tname1 = interpreter.nextTempName();
    	
    	SimpleNode projector = (SimpleNode) node.jjtGetChild(0);
    	if(projector.isBitSet(1)) {
	SimpleNode exprlist = (SimpleNode) projector.jjtGetChild(0);
    	if(exprlist.jjtGetNumChildren() != 1)
    		throw new InterpretError("actBoolCell: to be implemented.");
		SimpleNode rednode = (SimpleNode) exprlist.jjtGetChild(0);
		if((rednode.name == null) || (rednode.name.compareTo("") == 0))
    		throw new InterpretError("actBoolCell: to be implemented.");
		Domain redDom = env.lookupDom(rednode.name, true);
		if((redDom == null) || (redDom.isRedTemp == 0))
    		throw new InterpretError("actBoolCell: to be implemented.");
    	Relation tempRel = actRelCell(node,  dnames, rname,
			      rindx, relhtbl, tname0, tname1);
		if(tempRel.domains.length != 1)
			throw new InterpretError("actBoolCell: to be implemented");
		
		if(tempRel.domains[0].type != BOOLEAN)
			throw new InterpretError("actBoolCell: type mismatch");
		tempRel.load(true);
		byte retVal = ((byte[]) tempRel.data[0])[0];
		tempRel.unload();
		return (retVal == 1);
	}
    }
    //end of Yi's addition July 2, 2002

	//added by Yi Zheng July 5, 2002
	//to implement: let A be a nop b;
	case OP_JOINOPERATOR: {
		if((node.opcode == OP_NOP) && (node.jjtGetNumChildren() == 2))
		{
			SimpleNode leftb = (SimpleNode) node.jjtGetChild(0);
			SimpleNode rightb = (SimpleNode) node.jjtGetChild(1);
			boolean Ran = new Random().nextBoolean();
			if(Ran)
				return actBoolCell(leftb);
			else
				return actBoolCell(rightb);
		}
		//deliberately allow to fall through to the default case
	}
	//end of addition July 5, 2002
    default:
    //  return(actBoolRelCell(node,null,null,null,null,null));
      //throw new InterpretError("actBoolCell: to be implemented..");
                      // START Since jRelixNS
                 if (Relation.isSigmaJoinImplemented == true)
                   return(actBoolRelCell(node,null,null,null,null,null));
                 else //if (Relation.isSigmaJoinImplemented != true)
		   throw new InterpretError("actBoolCell: to be implemented..");
                 // END Since jRelixNS //Andy May 12, 2002
    }
  }     
  /**
   *  Actualize a double-typed cell.
   */
  private double actDoubleCell(SimpleNode node) throws InterpretError
  {
//    System.out.println("Double . Nodename " +node.name );
//    System.out.println("Double . type " +node.type );
//    System.out.println("Double . opcode " +node.opcode );
    
    switch(node.type){
    	//OP_FUNCTION case added by Yi Zheng May 6, 2002
    	case OP_FUNCTION:
    		SimpleNode argument = (SimpleNode) node.jjtGetChild(0);
			
			double ret;
			switch(node.opcode)
			{
				case OP_ABS://assume argument is of type double
				{
					double arg = actDoubleCell(argument);
					ret = (double) Math.abs(arg);
					return ret;
				}
				case OP_SQRT://assume argument is of type double
				{
					double arg = actDoubleCell(argument);
					ret = (double) Math.sqrt(arg);
					return ret;		
				}			
				case OP_SIN://assume argument is of type double
				{
					double arg = actDoubleCell(argument);
					ret = (double) Math.sin(arg);
					return ret;
				}
				case OP_ASIN://assume argument is of type double
				{
					double arg = actDoubleCell(argument);
					ret = (double) Math.asin(arg);
					return ret;	
				}
				case OP_COS://assume argument is of type double
				{
					double arg = actDoubleCell(argument);
					ret = (double) Math.cos(arg);
					return ret;	
				}
				case OP_ACOS://assume argument is of type double
				{
					double arg = actDoubleCell(argument);
					ret = (double) Math.acos(arg);
					return ret;	
				}
				case OP_TAN://assume argument is of type double
				{
					double arg = actDoubleCell(argument);
					ret = (double) Math.tan(arg);
					return ret;	
				}
				case OP_ATAN://assume argument is of type double
				{
					double arg = actDoubleCell(argument);
					ret = (double) Math.atan(arg);
					return ret;	
				}		
				case OP_LOG://assume argument is of type double
				{
					double arg = actDoubleCell(argument);
					ret = (double) Math.log(arg);
					return ret;	
				}		
				case OP_LOG10://assume argument is of type double
				{
					double arg = actDoubleCell(argument);
					double denom = (double) Math.log(10);
					ret = (double) Math.log(arg) / denom;
					return ret;
				}
				case OP_ROUND://assume argument is of type double
				{//round() returns integer, not double
					throw new InterpretError("Bad use of round()");
				}
				case OP_CEIL://assume argument is of type double
				{
					throw new InterpretError("Bad use of ceil()");
				}
				case OP_FLOOR://assume argument is of type double
				{
					throw new InterpretError("Bad use of floor()");
				}
				case OP_CHR:
					throw new InterpretError("Bad use of chr()");					
				case OP_ORD:
				{
					String s = actStrCell(argument);
					char startC = '\u0000';
					char arg;
					if(s == null)
						throw new InterpretError("Empty string in ord()");
					else
						ret = (double) ((char) s.charAt(0) - startC);
					return ret;
						
				}
				case OP_ISKNOWN:
				case OP_SINH:
				case OP_COSH:
				case OP_TANH:
					throw new InterpretError("Not implemented");
				default:
					throw new InterpretError("Unknown function operator");
			}	

    		
    		
    case OP_IDENTIFIER:
      switch(destrel.domains[node.bits].type){
      case DOUBLE:
	return ((double[])destrel.data[node.bits])[currow];
      case FLOAT:
	return (double) ((float[])destrel.data[node.bits])[currow];
      case INTEGER:
	return (double) ((int[])destrel.data[node.bits])[currow];
      case SHORT:
	return (double) ((short[])destrel.data[node.bits])[currow];
      case LONG:
	return (double) ((long[])destrel.data[node.bits])[currow];
	    case NUMERIC://case added by Yiyi June 4'02
		return (((number[])destrel.data[node.bits])[currow]).numToDouble();

      default:
	throw new InterpretError("actDoubleCell: to be implemented.");
      }
    case OP_BIOPERATOR:
      SimpleNode left = (SimpleNode)node.jjtGetChild(0);
      SimpleNode right = (SimpleNode)node.jjtGetChild(1);
      
      switch(node.opcode){
	
	//case OP_PLUS:
	//return (actDoubleCell(left) + actDoubleCell(right));
	
	// Revised by hongyu. 
      case OP_PLUS:{
	switch(node.bits){
	case SHORT:
	 return (double)((short)(actIntCell(left)) + (short)(actIntCell(right)));
	 // throw new InterpretError("actDoubleCell: to be implemented.");// Hongyu Need to be implement
	case LONG:
	  return (double)(actLongCell(left) + actLongCell(right));
	case FLOAT:  
	case DOUBLE:{
//	  System.out.println("DOUBLE -------------------------------");
	  return actDoubleCell(left) + actDoubleCell(right);
	}
	//case NUMERIC: 
	case INTEGER:{
//	  System.out.println(" INTEGER----------------------------");
	  return (double)(actIntCell(left) + actIntCell(right));
	}
		case NUMERIC://case added by Yiyi June 4'02
		    return (actNumCell(left).add(actNumCell(right))).numToDouble();

	case TEXT: // Hongyu added
	case STMT: // Hongyu added
	case EXPR: // Hongyu added
	case IDLIST:// ??Hongyu added 
	case STRING:
	  //System.out.println("in case STRING");
	default:
	  throw new InterpretError("actDoubleCell: plus to be implemented.. .");
	}
      }
      
      //case OP_MINUS:
      //return(actDoubleCell(left) - actDoubleCell(right));
      
      // Revised by hongyu.
      case OP_MINUS:{
	switch(node.bits){
	case SHORT:
		 return (double)((short)(actIntCell(left)) - (short)(actIntCell(right)));

	 // throw new InterpretError("actDoubleCell: to be implemented."); //Hongyu Need to be implement
	case LONG:
	  return (double)(actLongCell(left) - actLongCell(right));
	case FLOAT:  
	case DOUBLE:{
//	  System.out.println("DOUBLE -------------------------------");
	  return actDoubleCell(left) - actDoubleCell(right);
	}
	//case NUMERIC: 
	case INTEGER:{
//	  System.out.println(" INTEGER----------------------------");
	  return (double)(actIntCell(left) - actIntCell(right));
	}
	case NUMERIC://case added by YIYI 
		    return (actNumCell(left).subtract(actNumCell(right))).numToDouble();

	case TEXT: // Hongyu added
	case STMT: // Hongyu added
	case EXPR: // Hongyu added
	case IDLIST://?? Hongyu added 
	case STRING:
	  //System.out.println("in case STRING");
	default:
	  throw new InterpretError("actDoubleCell: minus to be implemented.. .");
	}
      }   
      
      //case OP_MUL:
      //return(actDoubleCell(left) * actDoubleCell(right));
      
      // Revised by hongyu.
      case OP_MUL:{
	switch(node.bits){
	case SHORT:
		 return (double)((short)(actIntCell(left)) * (short)(actIntCell(right)));

//	  throw new InterpretError("actDoubleCell: to be implemented.");//Hongyu Need to be implement
	case LONG:
	  return (double)(actLongCell(left) * actLongCell(right));
	case FLOAT:  
	case DOUBLE:{
//	  System.out.println("DOUBLE -------------------------------");
	  return actDoubleCell(left) * actDoubleCell(right);
	}
	//case NUMERIC: 
	case INTEGER:{
//	  System.out.println(" INTEGER----------------------------");
	  return (double)(actIntCell(left) * actIntCell(right));
	}
		case NUMERIC://case added by Yiyi June 4'02
		    return (actNumCell(left).multiply(actNumCell(right))).numToDouble();

	case TEXT: // Hongyu added
	case STMT: // Hongyu added
	case EXPR: // Hongyu added
	case IDLIST:// ??Hongyu added 
	case STRING:
	  //System.out.println("in case STRING");
	default:
	  throw new InterpretError("actDoubleCell: mul to be implemented.. .");
	}
      }  
      
      //case OP_DIV:
      //return(actDoubleCell(left) / actDoubleCell(right));
      
      // Revised by hongyu.
      case OP_DIV:{
	switch(node.bits){
	case SHORT:{
		      short temp = (short)actIntCell(right);
		      if(temp != 0 )
		      return (double)((short)actIntCell(left)) /(short)actIntCell(right);
		      else 
		      throw new InterpretError("actDoubleCell: divided by zero");

	 // throw new InterpretError("actDoubleCell:  div to be implemented.. . ");
	}
	case LONG:{
	  long temp = actLongCell(right);
	  if(temp != 0L )
	    return (double)(actLongCell(left) / temp);
	  else 
	    throw new InterpretError("actDoubleCell: divided by zero");
	}
	case FLOAT:
	case DOUBLE:{
//	  System.out.println("DOUBLE -------------------------------");
	  double temp = actDoubleCell(right);
	  if(temp != 0.0)
	    return actDoubleCell(left) / temp;
	  else 
	    throw new InterpretError("actDoubleCell: divided by zero");
	}
	//case NUMERIC: 	   
	 case NUMERIC://added by Yiyi June 4'02
		return (actNumCell(left).divide(actNumCell(right))).numToDouble();

	case INTEGER:{
//	  System.out.println(" INTEGER----------------------------");
	  int temp = actIntCell(right);
	  if(temp !=0)
	    return (double)(actIntCell(left) / temp);
	  else 
	    throw new InterpretError("actDoubleCell: divided by zero");
	}
	case TEXT: // Hongyu added
	case STMT: // Hongyu added
	case EXPR: // Hongyu added
	case IDLIST://?? Hongyu added 
	case STRING:
	  //System.out.println("in case STRING");
	default:
	  throw new InterpretError("actDoubleCell: div to be implemented.. .");
	}
      }
      
      // added by hongyu. Sept 2001
      case OP_MOD:{
	switch(node.bits){
	case SHORT:{
		      short temp = (short)actIntCell(right);
		      if(temp != 0)
		      return (double)((int)(actIntCell(left)) % (int)(actIntCell(right)));
		      else 
		      throw new InterpretError("actDoubleCell: mod by zero");
		      
	//  throw new InterpretError("actDoubleCell:  mod to be implemented.. . ");
	}
        case LONG:{
	  long temp =actLongCell(right);
	  if(temp != 0L)
	    return (double)(actLongCell(left) % temp);
	  else 
	    throw new InterpretError("actDoubleCell: mod by zero");
	}
	case FLOAT:  
	  throw new InterpretError("actDoubleCell: mod by value of float type");
	case DOUBLE:
	  throw new InterpretError("actDoubleCell: mod by value of doubel type");
	  //case NUMERIC: 
	  case NUMERIC:{ //added by Yiyi
			number tmp = actNumCell(right);
			if(!number.iszero(tmp))
				return (number.mod(actNumCell(left),tmp)).numToDouble();
			else
				throw new InterpretError("actDoubleCell: mod by 0");
		}

	case INTEGER:{
//	  System.out.println(" INTEGER----------------------------");
	  int temp = actIntCell(right);
	  if(temp !=0)
	    return (double)(actIntCell(left) % temp);
	  else 
	    throw new InterpretError("actDoubleCell: mod by zero");
	}
	case TEXT: // Hongyu added
	case STMT: // Hongyu added
	case EXPR: // Hongyu added
	case IDLIST:// ??Hongyu added 
	case STRING:
	  //System.out.println("in case STRING");
	default:
	  throw new InterpretError("actDoubleCell: mod to be implemented.. .");
	}
      }
      	    case OP_MIN:{//yiyi added
				switch(node.bits){
				case FLOAT:
				case DOUBLE:
					return Math.min(actDoubleCell(left),actDoubleCell(right));
				case INTEGER:
					return (double)(Math.min(actIntCell(left),actIntCell(right)));
				case LONG:
					return (double)(Math.min(actLongCell(left),actLongCell(right)));
				case SHORT:
					throw new InterpretError("actDoubleCell: min to be implemented");
				case NUMERIC:
					return number.min(actNumCell(left),actNumCell(right)).numToDouble();
				}
			}
	    case OP_MAX:{//yiyi added
				switch(node.bits){
				case FLOAT:
				case DOUBLE:
					return Math.max(actDoubleCell(left),actDoubleCell(right));
				case INTEGER:
					return (double)(Math.max(actIntCell(left),actIntCell(right)));
				case LONG:
					return (double)(Math.max(actLongCell(left),actLongCell(right)));
				case SHORT:
					throw new InterpretError("actDoubleCell: max to be implemented");
				case NUMERIC:
					return number.max(actNumCell(left),actNumCell(right)).numToDouble();
				}
			}
	    case OP_POW:{ //Yiyi added
				switch(node.bits){
				case FLOAT:
				case DOUBLE:
					return Math.pow(actDoubleCell(left),actDoubleCell(right));
				case INTEGER:
					return Math.pow((double)(actIntCell(left)),(double)(actIntCell(right)));
				case LONG:
					return Math.pow((double)(actLongCell(left)),(double)(actLongCell(right)));
				case SHORT:
					throw new InterpretError("actDoubleCell: pow to be implemented");
				case NUMERIC:
					return Math.pow(actNumCell(left).numToDouble(), actNumCell(right).numToDouble());
				default:
					throw new InterpretError("actDoubleCell: pow to be implemented.");
				}
			}
      
      
                //START Since JRelCaNu a.
		 case OP_CAST:
                        // Cast from type stored in left.bits of 
			// the the left OP_TYPE node in buildTree and
			// traverseType.
		        switch(left.bits){
		        case BOOLEAN:
			   return CastToDouble.value(CastToBoolByte.value(actBoolCell(right)));
		        case STRING:
			   return CastToDouble.value(actStrCell(right));
		        case INTEGER:
			   return CastToDouble.value(actIntCell(right));
		        case SHORT:
			   return CastToDouble.value((short)CastToShort.value(actIntCell(right)));
		        case LONG:
			   return CastToDouble.value(actLongCell(right));
		        case FLOAT:
		        //bug fix by Yi Zheng Sept 28, 2002
			   //return CastToDouble.value(CastToFloat.value(actDoubleCell(right)));
		       double adv = actDoubleCell(right);
		       if (adv == DOUBLE_DC)
		       	return DOUBLE_DC;
		       else if (adv == DOUBLE_DK)
		       	return DOUBLE_DK;
		       else if (adv == (double)FLOAT_DC)
		       	return DOUBLE_DC;
		       else if (adv == (double)FLOAT_DK)
		       	return DOUBLE_DK;
		       else
		       	return CastToDouble.value(CastToFloat.value(adv));
		       
		       //end of bug fix Sept 28, 2002
		        case DOUBLE:
			   return CastToDouble.value(actDoubleCell(right));
			   
			   //added by Yi Zheng Sept 28, 2002
			  case DC:
			  	return DOUBLE_DC;
			  case DK:
			  	return DOUBLE_DK;
			   
			   //end of addition Sept 28, 2002
//YYY		    
		        case NUMERIC:
			   return CastToDouble.value(actNumCell(right));
		        default:
		             Utility.dump(node,"actDoubleCell OP_CAST Node --->");
			     throw new InterpretError("actDoubleCell: not implemented.");
		        }
          //END Since JRelCaNu a.
/* commented out by Yiyi
      case OP_MIN:{// Hongyu added
      		double lval = actDoubleCell(left);
      		double rval = actDoubleCell(right);
      		if (lval < rval) return lval;
      		return rval;
    	}
      case OP_MAX:{// Hongyu added
      		double lval = actDoubleCell(left);
      		double rval = actDoubleCell(right);
      		if (lval > rval) return lval;
      		return rval;
    	}
      case OP_POW:// Hongyu added
     */
      default:
	throw new InterpretError("actDoubleCell: bio to be implemented.. .");
      }
    case OP_UOPERATOR:
      switch(node.opcode){
      case OP_UPLUS:
	return(actDoubleCell((SimpleNode)node.jjtGetChild(0)));
      case OP_UMINUS:
	return(0.0 - actDoubleCell((SimpleNode)node.jjtGetChild(0)));
      default:
	throw new InterpretError("actDoubleCell: uni to be implemented.");
      }
    case OP_LITERAL:
      if(node.info instanceof Double)
	return ((Double)node.info).doubleValue();
      else if(node.info instanceof Float)
	return (double)((Float)node.info).floatValue();
	    else if(node.info instanceof number)//case added by Yiyi June 5'02
		return ((number)node.info).numToDouble();

      else if(node.info instanceof Integer)
	return (double)((Integer)node.info).intValue();
      else if(node.info instanceof Short)
	return (double)((Short)node.info).shortValue();
      else if(node.info instanceof Long)
	return (double)((Long)node.info).longValue();
      else
	throw new InterpretError("actDoubleCell: to be implemented.");
      
    case OP_RED:
    case OP_EQUIV:
    case OP_FUN://hongyu added Feb 2002
    case OP_PAR://hongyu added Feb 2002
    if(node.opcode != OP_REDFUNCALL)
    {
      double actValue= actDoubleCell((SimpleNode)node.jjtGetChild(0));
      
      switch(node.opcode){
      case OP_PLUS:
	if(node.info == null)  node.info = new Double(0.0);
	node.info = new Double(((Double)node.info).doubleValue()+actValue);
	break;
      case OP_MUL:
	if(node.info == null)  node.info = new Double(1.0);
	node.info = new Double(((Double)node.info).doubleValue()*actValue);
	break;
      case OP_MIN:
	if(node.info == null)  node.info = new Double(Double.MAX_VALUE);
	node.info = new Double(
			       Math.min(((Double)node.info).doubleValue(), actValue));
	break;
      case OP_MAX:
	if(node.info == null)  node.info = new Double(Double.MIN_VALUE);
	node.info = new Double(
			       Math.max(((Double)node.info).doubleValue(), actValue));
	break;
	//addition by Yi Zheng July 5, 2002 to implement the case:
	//let a be red nop of b; where a is randomly selected from b
	case OP_NOP:
		if (node.info == null)
			node.info = new Double(actValue);
		else
		{
			boolean rb = new Random().nextBoolean();
			if(rb)
				node.info = new Double(actValue);
		}
		break;
	default:
		throw new InterpretError("actDoubleCell: to be implemented");
	
	//end of addition July 5, 2002
      }
      return actValue;
    }else //added by Yi Zheng July 9, 2002
    {
        double actValue= actDoubleCell((SimpleNode)node.jjtGetChild(1));
		RedFunCallAccumNode(node, DOUBLE, new Double(actValue));
		return actValue;
    	
    }
    case OP_IFTHENELSE:
      if(actBoolCell((SimpleNode)node.jjtGetChild(0)))
	return actDoubleCell((SimpleNode)node.jjtGetChild(1));
      else
	return actDoubleCell((SimpleNode)node.jjtGetChild(2));
    case OP_QTOPERATOR:{// Added by hongyu, Sept 15,01
      switch(node.bits){
      	    case NUMERIC://case added by Yiyi June 5'02
		return (actNumCell((SimpleNode)node.jjtGetChild(0))).numToDouble();

      case INTEGER:
	return (double)actIntCell((SimpleNode)node.jjtGetChild(0));
      case DOUBLE:
	return actDoubleCell((SimpleNode)node.jjtGetChild(0));
	// End of Hongyu Added. Sept 15,01  
      }
    }
        //added by Yi Zheng July 2, 2002
    // to implement level lifting
    case OP_PROJECT:
    case OP_TSELECT: {
    	Vector dnames = new Vector();
    	Vector rname = new Vector();
    	Vector rindx = new Vector();
    	relhtbl = new Hashtable();
    	String tname0 = interpreter.nextTempName();
    	String tname1 = interpreter.nextTempName();
    	
    	SimpleNode projector = (SimpleNode) node.jjtGetChild(0);
    	if(!projector.isBitSet(1))
    		throw new InterpretError("actDoubleCell: to be implemented.");
    	SimpleNode exprlist = (SimpleNode) projector.jjtGetChild(0);
    	if(exprlist.jjtGetNumChildren() != 1)
    		throw new InterpretError("actDoubleCell: to be implemented.");
		SimpleNode rednode = (SimpleNode) exprlist.jjtGetChild(0);
		if((rednode.name == null) || (rednode.name.compareTo("") == 0))
    		throw new InterpretError("actDoubleCell: to be implemented.");
		Domain redDom = env.lookupDom(rednode.name, true);
		if((redDom == null) || (redDom.isRedTemp == 0))
    		throw new InterpretError("actDoubleCell: to be implemented.");
    	Relation tempRel = actRelCell(node,  dnames, rname,
			      rindx, relhtbl, tname0, tname1);
		if(tempRel.domains.length != 1)
			throw new InterpretError("actDoubleCell: to be implemented");
		
		if((tempRel.domains[0].type != DOUBLE) && (tempRel.domains[0].type != FLOAT))
			throw new InterpretError("actDoubleCell: type mismatch");
		tempRel.load(true);
 		double retVal;
		if(tempRel.domains[0].type == DOUBLE)
		  retVal = ((double[]) tempRel.data[0])[0];
		else
		  retVal = (double)(((float[]) tempRel.data[0])[0]);
		tempRel.unload();
		return retVal;
	}
    //end of Yi's addition July 2, 2002
    
    //added by Yi Zheng July 5, 2002
    //let a be b nop c;
    case OP_JOINOPERATOR:
    	if ((node.opcode == OP_NOP) && (node.jjtGetNumChildren() == 2))
    	{
    		boolean rb = new Random().nextBoolean();
    		if(rb)
    			return actDoubleCell((SimpleNode)node.jjtGetChild(0));
    		else
    			return actDoubleCell((SimpleNode)node.jjtGetChild(1));
    	}
    	//deliberately allow to fall through to the default case
    //end of addition July 5, 2002
    default:
      throw new InterpretError("actDoubleCell: to be implemented.");
    }
  } 
  
  
    /**
   *  Actualize a numeric-typed cell
   *
   * Added by Yiyi, June 2
   */

    private number actNumCell(SimpleNode node) throws InterpretError
    {
	switch (node.type){
	case OP_FUNCTION:
	    SimpleNode argument = (SimpleNode)node.jjtGetChild(0);
	    number arg = actNumCell(argument);
	    number ret;
	    switch(node.opcode){
	    case OP_ABS:{
		ret = number.abs(arg);
		return ret;
	    }
	    case OP_CEIL:{
		ret = number.ceil(arg);
		return ret;
	    }
	    case OP_FLOOR:{
		ret = number.floor(arg);
		return ret;
	    }
	    case OP_ROUND:{
		ret = number.round(arg);
		return ret;
	    }
			case OP_SQRT:{
				ret = number.sqrt(arg);
				return ret;
			}
			case OP_SIN:{
				ret = number.sin(arg);
				return ret;
			}
			case OP_COS:{
				ret = number.cos(arg);
				return ret;
			}
			case OP_TAN:{
				ret = number.tan(arg);
				return ret;
			}
			case OP_ASIN:{
				ret = number.arcsin(arg);
				return ret;
			}
			case OP_ACOS:{
				ret = number.arccos(arg);
				return ret;
			}
			case OP_ATAN:{
				ret = number.arctan(arg);
				return ret;
			}
			case OP_SINH:{
				ret = number.sinh(arg);
				return ret;
			}
			case OP_COSH:{
				ret = number.cosh(arg);
				return ret;
			}
			case OP_TANH:{
				ret = number.tanh(arg);
				return ret;
			}
			case OP_LOG:{
				ret = number.ln(arg);
				return ret;
			}
			case OP_LOG10:{
				ret = number.log(arg);
				return ret;
			}

	    default:
		throw new InterpretError("Not implemented.");
	    }			  
	case OP_IDENTIFIER:
	    switch (destrel.domains[node.bits].type){
	    case NUMERIC:
		return((number[])destrel.data[node.bits])[currow];
	    case INTEGER:
		return number.intToNum(((int[])destrel.data[node.bits])[currow]);
	    case SHORT:
		return number.shortToNum(((short[])destrel.data[node.bits])[currow]);
	    case LONG:
		return number.longToNum(((long[])destrel.data[node.bits])[currow]);
	    case DOUBLE:
		return number.doubleToNum(((double[])destrel.data[node.bits])[currow]);
	    case FLOAT:
		return number.floatToNum(((float[])destrel.data[node.bits])[currow]);
	    default:
		throw new InterpretError("actNumCell: to be implemented.");
	    }
	case OP_BIOPERATOR:
	    SimpleNode left = (SimpleNode)node.jjtGetChild(0);
	    SimpleNode right = (SimpleNode)node.jjtGetChild(1);

	    switch (node.opcode){
	    case OP_PLUS:{
		switch (node.bits){
		case NUMERIC:
			    return actNumCell(left).add(actNumCell(right));
		case INTEGER:
		    return number.intToNum(actIntCell(left)+ actIntCell(right));
		case LONG:
		    return number.longToNum(actLongCell(left)+actLongCell(right));
		case FLOAT:
		case DOUBLE:
		    return number.doubleToNum(actDoubleCell(left) + actDoubleCell(right));
		case SHORT:
		    throw new InterpretError("actNumCell: to be implemented.");
		default:
		    throw new InterpretError("actNumCell: plus to be implemented.");
		}
	    }
	    case OP_MINUS:{
		switch (node.bits){
		case NUMERIC:
		    return actNumCell(left).subtract(actNumCell(right));
		case INTEGER:
		    return number.intToNum(actIntCell(left) - actIntCell(right));
		case LONG:
		    return number.longToNum(actLongCell(left) - actLongCell(right));
		case FLOAT:
		case DOUBLE:
		    return number.doubleToNum(actDoubleCell(left) - actDoubleCell(right));
		case SHORT:
			    throw new InterpretError("actNumCell: to be implemented.");
		default:
			    throw new InterpretError("actNumCell:minus to be implemented.");
		}
	    }
	    case OP_MUL:{
		switch (node.bits){
		case NUMERIC:
		    return actNumCell(left).multiply(actNumCell(right));
		case INTEGER:
		    return number.intToNum(actIntCell(left)*actIntCell(right));
		case LONG:
		    return number.longToNum(actLongCell(left)*actLongCell(right));
		case FLOAT:
		case DOUBLE:
		    return number.doubleToNum(actDoubleCell(left)*actLongCell(right));
		case SHORT:
		    throw new InterpretError("actNumCell: to be implemented.");
		default:
		    throw new InterpretError("actNumCell: multiply to be implemented.");
		}
	    }
	    case OP_DIV:{
		switch (node.bits){
		case NUMERIC:
		    return actNumCell(left).divide(actNumCell(right));
		case INTEGER:{
		    int temp = actIntCell(right);
		    if(temp != 0){
			int k = actIntCell(left) / temp;
			return number.intToNum(k);
		    }
		    else
			throw new InterpretError("actNumCell:divided by zero.");
		}
		case LONG:{
		    long temp = actLongCell(right);
		    if(temp != 0L){
			long k = actLongCell(left)/temp;
			return number.longToNum(k);
		    }
		    else
			throw new InterpretError("actNumCell: divided by zero.");
		}
		case FLOAT:
		case DOUBLE:{
		    double temp = actDoubleCell(right);
		    if(temp != 0.0){
			double k = actDoubleCell(left) / temp;
			return number.doubleToNum(k);
		    }
		    else
			throw new InterpretError("actNumCell: divided by zero.");
		}
		case SHORT:
		    throw new InterpretError("actNumCell: to be implemented.");
		default:
		    throw new InterpretError("actNumCell: divide to be implmented.");
		}
	    }
	    case OP_MIN:{
		switch(node.bits){
		case NUMERIC:
		    return number.min(actNumCell(left),actNumCell(right));
		case INTEGER:
		    return number.intToNum(Math.min(actIntCell(left), actIntCell(right)));
		case LONG:
		    return number.longToNum(Math.min(actLongCell(left), actLongCell(right)));
		case FLOAT:
		case DOUBLE:
		    return number.doubleToNum(Math.min(actDoubleCell(left), actDoubleCell(right)));
		case SHORT:
		    throw new InterpretError("to be implemented.");
		default:
		    throw new InterpretError("min to be implemented.");
		}
	    }
	    case OP_MAX:{
		switch(node.bits){
		case NUMERIC:
		    return number.max(actNumCell(left), actNumCell(right));
		case INTEGER:
		    return number.intToNum(Math.max(actIntCell(left), actIntCell(right)));
		case LONG:
		    return number.longToNum(Math.max(actLongCell(left),actLongCell(right)));
		case FLOAT:
		case DOUBLE:
		    return number.doubleToNum(Math.max(actDoubleCell(left), actDoubleCell(right)));
		case SHORT:
		    throw new InterpretError("to be implmented.");
		default:
		    throw new InterpretError("max to be implemented.");
		}
	    }
		case OP_MOD:{
			switch(node.bits){
			case NUMERIC:{
				number tmp = actNumCell(right);
				if(!number.iszero(tmp))
					return number.mod(actNumCell(left),tmp);
				else
					throw new InterpretError("actNumCell: mod by 0");
			}
			case SHORT:{
		    throw new InterpretError("actNumCell: mod to be implemented.. .");
			}
			case LONG:{
		    long temp = actLongCell(right);
		    if(temp != 0L)
					return number.longToNum(actLongCell(left) % temp);
		    else
					throw new InterpretError("actNumCell: mod by zero");
			}
			case FLOAT:
		    throw new InterpretError("actNumCell: mod by value of float type");
			case DOUBLE:
		    throw new InterpretError("actNumCell: mod by value of double type");
			case INTEGER:{
				int temp = actIntCell(right);
		    if(temp!=0)
					return number.intToNum(actIntCell(left) % temp);
		    else
					throw new InterpretError("actNumCell: divided by zero");
			}
	    default:
				throw new InterpretError("actNumCell: mod to be implemented");
	   }
		}
		case OP_POW:{
			switch(node.bits){
			case NUMERIC:
				return number.pow(actNumCell(left),actNumCell(right));
			case INTEGER:
				return number.pow(number.intToNum(actIntCell(left)), number.intToNum(actIntCell(right)));
			case SHORT:
				throw new InterpretError("actNumCell: pow to be implemented");
			case LONG:
				return number.pow(number.longToNum(actLongCell(left)),number.longToNum(actLongCell(right)));
			case FLOAT:
			case DOUBLE:
				return number.pow(number.doubleToNum(actDoubleCell(left)),number.doubleToNum(actDoubleCell(right)));
			default:
				throw new InterpretError("actNumCell: pow to be implemented.");
			}
		}
		default:
			throw new InterpretError("actNumCell: bio to be implmented.");
		}
	case OP_UOPERATOR:
	    switch (node.opcode){
	    case OP_UPLUS:
		return (actNumCell((SimpleNode)node.jjtGetChild(0)));
	    case OP_UMINUS:{
		number zero = new number("+000 0");
		return (zero.subtract(actNumCell((SimpleNode)node.jjtGetChild(0))));
	    }
	    default:
		throw new InterpretError("actNumCell: uni to be implemented.");
	    }
	case OP_LITERAL:
	    if(node.info instanceof number)
		return (number)node.info;
	    else if(node.info instanceof Integer)
		return number.intToNum(((Integer)node.info).intValue());
	    else if(node.info instanceof Short)
		return number.shortToNum(((Short)node.info).shortValue());
	    else if(node.info instanceof Long)
		return number.longToNum(((Long)node.info).longValue());
	    else if(node.info instanceof Float)
		    return number.floatToNum(((Float)node.info).floatValue());
	    else if(node.info instanceof Double)
		return number.doubleToNum(((Double)node.info).doubleValue());
		else
		    throw new InterpretError("actNumCell: to be implemented.");
	    
	case OP_RED:
	case OP_EQUIV:
	    number actValue = actNumCell((SimpleNode)node.jjtGetChild(0));
	    switch (node.opcode){
	    case OP_PLUS:
		if(node.info == null) node.info = new number("+000+0");
		node.info = ((number)node.info).add(actValue);
		break;
	    case OP_MUL:
		if(node.info == null) node.info = new number("+000+1");
		node.info = ((number)node.info).multiply(actValue);
		break;
		}
	    return actValue;
	    
	case OP_FUN:
	case OP_PAR:
	case OP_IFTHENELSE:
		if(actBoolCell((SimpleNode)node.jjtGetChild(0)))
		    return actNumCell((SimpleNode)node.jjtGetChild(1));
		else
		    return actNumCell((SimpleNode)node.jjtGetChild(2));
		
	case OP_QTOPERATOR:
	    switch (node.bits){
	    case NUMERIC:
		return actNumCell((SimpleNode)node.jjtGetChild(0));
	    case INTEGER:
		return number.intToNum(actIntCell((SimpleNode)node.jjtGetChild(0)));
	    case DOUBLE:
		return number.doubleToNum(actDoubleCell((SimpleNode)node.jjtGetChild(0)));
	    default:
		throw new InterpretError("QT operator to be implmenetd.");
	    }
	default:
	    throw new InterpretError("actNumCell: to be implemented.");
	}
    }
    
 
  /**
   *  Actualize a integer-typed cell.
   */
  private int actIntCell(SimpleNode node) throws InterpretError
  {
//    System.out.println("node.type int "+node.type);
//    System.out.println("node.bits int "+node.bits); 
//    System.out.println("node.opcode int "+node.opcode); 
//    System.out.println("node.id int "+node); 
    switch(node.type){
    		case OP_FUNCTION: //added by Yi Zheng May 6, 2002
		//REVISIT: the type of argument is currently determined by the 
		 //type of argument accepted by corresponding Java methods
			SimpleNode argument = (SimpleNode) node.jjtGetChild(0);
			
			int ret;
			switch(node.opcode)
			{
				case OP_ABS://assume argument is of type int
				{
					int arg = actIntCell(argument);
					ret = (int) Math.abs(arg);
					return ret;
				}
				case OP_SQRT://assume argument is of type double
				{
					double arg = actDoubleCell(argument);
					ret = (int) Math.sqrt(arg);
					return ret;		
				}			
				case OP_SIN://assume argument is of type double
				{
					throw new InterpretError("Bad use of sin()");
				}
				case OP_ASIN://assume argument is of type double
				{
					double arg = actDoubleCell(argument);
					ret = (int) Math.asin(arg);
					return ret;	
				}
				case OP_COS://assume argument is of type double
				{
					throw new InterpretError("Bad use of cos()");
				}
				case OP_ACOS://assume argument is of type double
				{
					double arg = actDoubleCell(argument);
					ret = (int) Math.acos(arg);
					return ret;	
				}
				case OP_TAN://assume argument is of type double
				{
					double arg = actDoubleCell(argument);
					ret = (int) Math.tan(arg);
					return ret;	
				}
				case OP_ATAN://assume argument is of type double
				{
					double arg = actDoubleCell(argument);
					ret = (int) Math.atan(arg);
					return ret;	
				}		
				case OP_LOG://assume argument is of type double
				{
					double arg = actDoubleCell(argument);
					ret = (int) Math.log(arg);
					return ret;	
				}		
				case OP_LOG10://assume argument is of type double
				{
					double arg = actDoubleCell(argument);
					double denom = (double) Math.log(10);
					ret = (int) (Math.log(arg) / denom);
					return ret;
				}
				case OP_ROUND://assume argument is of type float
				{
					float arg = (float)actDoubleCell(argument);
					ret = (int) Math.round(arg);
					return ret;
				}
				case OP_CEIL://assume argument is of type double
				{
					double arg = actDoubleCell(argument);
					ret = (int) Math.ceil(arg);
					return ret;
				}
				case OP_FLOOR://assume argument is of type double
				{
					double arg = actDoubleCell(argument);
					ret = (int) Math.floor(arg);
					return ret;
				}
				case OP_CHR:
					//assumption: only deals with characters 0 - 255
				{
					throw new InterpretError("Bad use of chr()");
				}					
				case OP_ORD:
				{
					String s = actStrCell(argument);
					char startC = '\u0000';
					char arg;
					if(s == null)
						throw new InterpretError("Empty string in ord()");
					else
						ret = (int) ((char) s.charAt(0) - startC);
					return ret;
						
				}
				case OP_ISKNOWN:
				case OP_SINH:
				case OP_COSH:
				case OP_TANH:
					throw new InterpretError("Not implemented");
				default:
					throw new InterpretError("Unknown function operator");
			}
    case OP_IDENTIFIER:
      switch(destrel.domains[node.bits].type){
      case INTEGER:
	int k =  ((int[])destrel.data[node.bits])[currow];
//	System.out.println("currow int "+currow);
//	System.out.println("value int "+k);
	return ((int[])destrel.data[node.bits])[currow];
      case FLOAT:
	return (int) ((float[])destrel.data[node.bits])[currow];
      case SHORT:
	//System.out.println("in actIntCell, case SHORT");
	return (int) ((short[])destrel.data[node.bits])[currow];
      case LONG:
	return (int) ((long[])destrel.data[node.bits])[currow];
      case DOUBLE:{
	int a = (int) ((double[])destrel.data[node.bits])[currow];
//	System.out.println("double to int"+a);
	return (int) ((double[])destrel.data[node.bits])[currow];
      }
      default:
	throw new InterpretError("actIntCell: to be implemented.");
      }
    case OP_BIOPERATOR:
      
      SimpleNode left = (SimpleNode)node.jjtGetChild(0);
      SimpleNode right = (SimpleNode)node.jjtGetChild(1);
//      System.out.println("==============================");
      
      switch(node.opcode){
	//case OP_PLUS:
	//return(actIntCell(left) + actIntCell(right));
	
	//Revised by hongyu. Sept 01
      case OP_PLUS:{
	switch(node.bits){
	case SHORT:
	 return (int)((short)actIntCell(left) + (short)actIntCell(right));
	//  throw new InterpretError("actIntCell: to be implemented.");
	case LONG:
	  return (int)(actLongCell(left) + actLongCell(right));
	case FLOAT:  
	case DOUBLE:
	  return (int)(actDoubleCell(left) + actDoubleCell(right));
	  //case NUMERIC: 
	 		case NUMERIC: //added by YIYI June 4'02
		    return (actNumCell(left).add(actNumCell(right))).numToInt();

	case INTEGER:{
//	  System.out.println(" INTEGER----------------------------");
	  return actIntCell(left) + actIntCell(right);
	}
	case TEXT: // Hongyu added
	case STMT: // Hongyu added
	case EXPR: // Hongyu added
	case IDLIST://?? Hongyu added 
	case STRING:
	  //System.out.println("in case STRING");
	default:
	  throw new InterpretError("actIntCell: plus to be implemented.. .");
	}
      }
      
      //case OP_MINUS:
      //return(actIntCell(left) - actIntCell(right));
      
      //Revised by hongyu. Sept 2001
      case OP_MINUS:{
	switch(node.bits){
	case SHORT:
		 return (int)((short)actIntCell(left) - (short)actIntCell(right));

	//  throw new InterpretError("actIntCell: to be implemented.");
	case LONG:
	  return (int)(actLongCell(left) - actLongCell(right));
	case FLOAT:  
	case DOUBLE:
	  return (int)(actDoubleCell(left) - actDoubleCell(right));
	  //case NUMERIC: 
		case NUMERIC://case added by Yiyi June 4'02
		    return (actNumCell(left).subtract(actNumCell(right))).numToInt();

	case INTEGER:{
//	  System.out.println(" INTEGER----------------------------");
	  return actIntCell(left) - actIntCell(right);
	}
	case TEXT: // Hongyu added
	case STMT: // Hongyu added
	case EXPR: // Hongyu added
	case IDLIST://?? Hongyu added 
	case STRING:
	  //System.out.println("in case STRING");
	default:
	  throw new InterpretError("actIntCell: minus to be implemented.. .");
	}
      }
      
      //case OP_MUL:
      //return(actIntCell(left) * actIntCell(right));
      
      //Revised by hongyu. Sept 2001
      case OP_MUL:{
	switch(node.bits){
	case SHORT:
		 return (int)((short)actIntCell(left) * (short)actIntCell(right));

	//  throw new InterpretError("actIntCell: to be implemented.");
	case LONG:
	  return (int)(actLongCell(left) * actLongCell(right));
	case FLOAT:  
	case DOUBLE:
	  return (int)(actDoubleCell(left) * actDoubleCell(right));
	  //case NUMERIC: 
	  		case NUMERIC: //case added by Yiyi June 4'02
		    return (actNumCell(left).multiply(actNumCell(right))).numToInt();

	case INTEGER:{
//	  System.out.println(" INTEGER----------------------------");
	  return actIntCell(left) * actIntCell(right);
	}
	case TEXT: // Hongyu added
	case STMT: // Hongyu added
	case EXPR: // Hongyu added
	case IDLIST:// ??Hongyu added 
	case STRING:
	  //System.out.println("in case STRING");
	default:
	  throw new InterpretError("actIntCell: mul to be implemented.. .");
	}
      }
      
      //case OP_DIV:
      //return(actIntCell(left) / actIntCell(right));
      
      //Revised by hongyu. Sept 2001
      case OP_DIV:{
	switch(node.bits){
	case SHORT:{
		      short temp = (short)actIntCell(right);
		      if(temp != 0)
		      return (int)((short)actIntCell(left) / (short)actIntCell(right)); //Hongyu Need to be implement
		      else
		      throw new InterpretError("actIntCell: divided by zero");
		      
//	  throw new InterpretError("actIntCell: div to be implemented.. .");
	}
	case LONG:{
	  long temp = actLongCell(right);
	  if(temp !=0L)
	    return (int)(actLongCell(left) / temp);
	  else
	    throw new InterpretError("actIntCell: divided by zero");
	}
	case FLOAT:
	case DOUBLE:{
	  double temp = actDoubleCell(right);
	  if(temp !=0.0)
	    return (int)(actDoubleCell(left) / temp);
	  else
	    throw new InterpretError("actIntCell: divided by zero");
	}
	//case NUMERIC: 
		case NUMERIC://case added by Yiyi June 4'02
		    return (actNumCell(left).divide(actNumCell(right))).numToInt();

	case INTEGER:{
//	  System.out.println(" INTEGER----------------------------");
	  int temp = actIntCell(right);
	  if(temp !=0)
	    return actIntCell(left) / temp;
	  else
	    throw new InterpretError("actIntCell: divided by zero");
	}
	case TEXT: // Hongyu added
	case STMT: // Hongyu added
	case EXPR: // Hongyu added
	case IDLIST://?? Hongyu added 
	case STRING:
	  //System.out.println("in case STRING");
	default:
	  throw new InterpretError("actIntCell: div to be implemented.. .");
	}
      }
      
      //added by hongyu sept 2001
      case OP_MOD:{
	switch(node.bits){
	case SHORT:{
		      short temp = (short)actIntCell(right);
		      if(temp != 0)
		      return (int)((short)actIntCell(left) % (short)actIntCell(right)); //Hongyu Need to be implement
		      else
		      throw new InterpretError("actIntCell: mod by zero");
		      
//	  throw new InterpretError("actIntCell: mod to be implemented.. .");
	}     
	
	case LONG:{
	  long temp = actLongCell(right);
	  if(temp != 0L)
	    return (int)(actLongCell(left) % temp);
	  else
	    throw new InterpretError("actIntCell: mod by zero");
	}
	case FLOAT:
	  throw new InterpretError("actIntCell: mod by value of float type");
	case DOUBLE:
	  throw new InterpretError("actIntCell: mod by value of double type");
	  
	  //case NUMERIC: 
	 		case NUMERIC:{
			number tmp = actNumCell(right);
			if(!number.iszero(tmp))
				return (number.mod(actNumCell(left),tmp)).numToInt();
			else
				throw new InterpretError("actIntCell: mod by zero.");
			}

	case INTEGER:{
//	  System.out.println(" INTEGER----------------------------");
	  int temp = actIntCell(right);
	  if(temp!=0)
	    return actIntCell(left) % temp;
	  else
	    throw new InterpretError("actIntCell: divided by zero");
	}
	case TEXT: // Hongyu added
	case STMT: // Hongyu added
	case EXPR: // Hongyu added
	case IDLIST://?? Hongyu added 
	case STRING:
	  //System.out.println("in case STRING");
	default:
	  throw new InterpretError("actIntCell: mod to be implemented.. .");
	}
      }
                //START Since JRelCaNu a.
		 case OP_CAST:
                // Cast from type stored in left.bits of 
                // the the left OP_TYPE node in buildTree and
                // traverseType.
                switch (left.bits) {
                case BOOLEAN:
                    return CastToInt.value(CastToBoolByte.value(actBoolCell(right)));

                case STRING:
                    return CastToInt.value(actStrCell(right));

                case INTEGER:
                    return CastToInt.value(actIntCell(right));

                case FLOAT:
                    return CastToInt.value(CastToFloat.value(actDoubleCell(right)));

                case SHORT:
                   //bug fix by Yi Zheng Sept 28, 2002
                   // return CastToInt.value((short)CastToShort.value(actIntCell(right)));

					int aiv = actIntCell(right);
					if (aiv == INT_DC)
						return INT_DC;
					else if (aiv == INT_DK)
						return INT_DK;
					else if (aiv == (int) SHORT_DC)
						return INT_DC;
					else if (aiv == (int) SHORT_DK)
						return INT_DK;
					else
						return CastToInt.value((short)CastToShort.value(aiv));
					//end of bug fix Sept 28, 2002
                case LONG:
                    return CastToInt.value(actLongCell(right));

                case DOUBLE:
                    return CastToInt.value(actDoubleCell(right));
//YYY
		case NUMERIC:
		    return CastToInt.value(actNumCell(right));
			    
			    //added by Yi Zheng Sept 28,2002
			    case DC:
			    	return INT_DC;
			    
			    case DK:
			    	return INT_DK;
			    
			    //end of addition Sept 28, 2002

                case IDLIST:

                default:
		    Utility.dump(node,"actIntCell OP_CAST Node --->");
                    throw new InterpretError("actIntCell: not implemented for " + Utility.type2String(left.bits));
                }
          //END Since JRelCaNu a.
         /* commented out by yiyi
      case OP_MIN:{//added by hongyu Sept 2001
      	int lval = actIntCell(left);
      	int rval = actIntCell(right);
      	if (lval < rval) return lval;
      	return rval;
      	}
      case OP_MAX:{//added by hongyu Sept 2001
      	int lval = actIntCell(left);
      	int rval = actIntCell(right);
      	if (lval > rval) return lval;
      	return rval;
      
      	}
      case OP_POW://added by hongyu Sept 2001
     */
	    case OP_MIN:{//yiyi added
				switch(node.bits){
				case INTEGER:
					return Math.min(actIntCell(left),actIntCell(right));
				case LONG:
					return (int)(Math.min(actLongCell(left),actLongCell(right)));
				case SHORT:
					throw new InterpretError("actIntCell: min to be implemented");
				case NUMERIC:
					return number.min(actNumCell(left),actNumCell(right)).numToInt();
				case FLOAT:
				case DOUBLE:
					return (int)(Math.min(actDoubleCell(left),actDoubleCell(right)));
				default:
     throw new InterpretError("actIntCell: min to be implmented.");
				}
			}
			case OP_MAX:{//yiyi added
				switch(node.bits){
				case INTEGER:
					return Math.max(actIntCell(left),actIntCell(right));
				case LONG:
					return (int)(Math.max(actLongCell(left),actLongCell(right)));
				case SHORT:
					throw new InterpretError("actIntCell: max to be implemented");
				case NUMERIC:
					return number.max(actNumCell(left),actNumCell(right)).numToInt();
				case FLOAT:
				case DOUBLE:
					return (int)(Math.max(actDoubleCell(left),actDoubleCell(right)));
				default:
     throw new InterpretError("actIntCell: max to be implmented.");
				}
			}
	    case OP_POW:{//yiyi added
				switch(node.bits){
				case INTEGER:
					return (int)(Math.pow((double)(actIntCell(left)),(double)(actIntCell(right))));
				case LONG:
					return (int)(Math.pow((double)(actLongCell(left)),(double)(actLongCell(right))));
				case SHORT:
					throw new InterpretError("actIntCell: pow to be implemented.");
				case FLOAT:
				case DOUBLE:
					return (int)(Math.pow(actDoubleCell(left),actDoubleCell(right)));
				case NUMERIC:
					return number.pow(actNumCell(left),actNumCell(right)).numToInt();
				default:
					throw new InterpretError("actIntCell: pow to be implmented.");
				}
			}
     
      default:
	throw new InterpretError("actIntCell: bio to be implemented.");
      }
    case OP_UOPERATOR:
      switch(node.opcode){
      case OP_UPLUS:
	return(actIntCell((SimpleNode)node.jjtGetChild(0)));
      case OP_UMINUS:
	return(0-actIntCell((SimpleNode)node.jjtGetChild(0)));
      default:
	throw new InterpretError("actIntCell: uni to be implemented.");
      }
    case OP_LITERAL:
      if(node.info instanceof Integer)
	return ((Integer)node.info).intValue();
      else if(node.info instanceof Long)
	return (int) (((Long)node.info).longValue()); 
      else if(node.info instanceof Short)
	return (int)((Short)node.info).shortValue();	    
	else if(node.info instanceof number)
		return ((number)node.info).numToInt();

      if(node.info instanceof Double)
	return (int) ((Double)node.info).doubleValue();
      else if(node.info instanceof Float)
	return (int)((Float)node.info).floatValue();
	 else if(node.info instanceof Byte) //from Andrey
	return (int)((Byte)node.info).byteValue();//from Andrey
      else
	throw new InterpretError("actDoubleCell: to be implemented.");
    case OP_RED:
    case OP_EQUIV:
    case OP_FUN:// hongyu added. Feb 2002
    case OP_PAR:// hongyu added. Feb 2002
     if(node.opcode != OP_REDFUNCALL) //if added by Yi Zheng July 8, 2002
     {
      int actValue= actIntCell((SimpleNode)node.jjtGetChild(0));
      
      switch(node.opcode){
      case OP_PLUS:
	// if(node.info ==null) System.out.println("node.info is null");
	if(node.info == null)  node.info = new Integer(0);
	//System.out.println("actValue="+actValue);
	//System.out.println("node.info="+((Integer)node.info).intValue());
	node.info = new Integer(((Integer)node.info).intValue()+actValue);
	break;
      case OP_MUL:
	if(node.info == null)  node.info = new Integer(1);
	node.info = new Integer(((Integer)node.info).intValue()*actValue);
	break;
      case OP_MIN:
	if(node.info == null){
	  node.info = new Integer(Integer.MAX_VALUE);
	}
	node.info = new Integer(
				Math.min(((Integer)node.info).intValue(), actValue));
	break;
      case OP_MAX:
	if(node.info == null)  node.info = new Integer(Integer.MIN_VALUE);
	node.info = new Integer(
				Math.max(((Integer)node.info).intValue(), actValue));
	break;
	
	//added by Yi Zheng July 5, 2002
	//let a be red nop of b
	case OP_NOP:
	if(node.info == null)
		node.info = new Integer(actValue);
	else
	{
		boolean rb = new Random().nextBoolean();
		if(rb)
			node.info = new Integer(actValue);
	}
	break;
	
	default:
		throw new InterpretError("actIncCell: to be implemented");
	//end of addition July 5, 2002
      }
      return actValue;
    }
    else //Yi Zheng July 8, 2002
    {
    	int actValue= actIntCell((SimpleNode)node.jjtGetChild(1));
		RedFunCallAccumNode(node, INTEGER, new Integer(actValue));
		return actValue;
    }
    case OP_IFTHENELSE:
      if(actBoolCell((SimpleNode)node.jjtGetChild(0)))
	return actIntCell((SimpleNode)node.jjtGetChild(1));
      else
	return actIntCell((SimpleNode)node.jjtGetChild(2));
      
      //added by hongyu
    case OP_QTOPERATOR:{  // has two children. "." = "# /Total " .Added by hongyu
//      System.out.println(" act QT "+((SimpleNode)(node.jjtGetChild(0))).type);
      switch(node.bits){
      case INTEGER:
	return actIntCell((SimpleNode)node.jjtGetChild(0));
      case DOUBLE:
	return (int)actDoubleCell((SimpleNode)node.jjtGetChild(0));
      }
    }
    //added by Yi Zheng July 2, 2002
    // to implement level lifting
    case OP_PROJECT:
    case OP_TSELECT: {
    	//System.out.println("in actIntCell: node is:");
    	//Utility.dump(node, "-->");
    	Vector dnames = new Vector();
    	Vector rname = new Vector();
    	Vector rindx = new Vector();
    	relhtbl = new Hashtable();
    	String tname0 = interpreter.nextTempName();
    	String tname1 = interpreter.nextTempName();
    	
    	SimpleNode projector = (SimpleNode) node.jjtGetChild(0);
    	if(!projector.isBitSet(1))
    		throw new InterpretError("actIntCell: to be implemented.");
    	SimpleNode exprlist = (SimpleNode) projector.jjtGetChild(0);
    	if(exprlist.jjtGetNumChildren() != 1)
    		throw new InterpretError("actIntCell: to be implemented.");
		SimpleNode rednode = (SimpleNode) exprlist.jjtGetChild(0);
		if((rednode.name == null) || (rednode.name.compareTo("") == 0))
    		throw new InterpretError("actIntCell: to be implemented.");
		Domain redDom = env.lookupDom(rednode.name, true);
		if((redDom == null) || (redDom.isRedTemp == 0))
    		throw new InterpretError("actIntCell: to be implemented.");
	//	redDom.isRedTemp = 0; //avoid to be used by evaluateProject				    		
    	Relation tempRel = actRelCell(node,  dnames, rname,
			      rindx, relhtbl, tname0, tname1);
//		redDom.isRedTemp = 1;
		if(tempRel.domains.length != 1)
			throw new InterpretError("actIntCell: to be implemented");
		
		if(tempRel.domains[0].type != INTEGER)
			throw new InterpretError("actIntCell: type mismatch");
		tempRel.load(true);
		int retVal = ((int[]) tempRel.data[0])[0];
		tempRel.unload();
		return retVal;
	}
    //end of Yi's addition July 2, 2002
    
    //added by Yi Zheng July 5, 2002
    //let a be b nop c;
    case OP_JOINOPERATOR:
    	if((node.opcode == OP_NOP) && (node.jjtGetNumChildren() == 2))
    	{
    		boolean rb = new Random().nextBoolean();
    		if(rb)
    			return actIntCell((SimpleNode)node.jjtGetChild(0));
    		else
    			return actIntCell((SimpleNode)node.jjtGetChild(1));
    	}
    	//deliberately allow to fall through to the default case
    //end of additin July 5
    default:
      throw new InterpretError("actIntCell: to be implemented.");
    }
  } 
  
  
  /**
   * Actualize a COMP  cell
   *
   * @author Yi Zheng June 17, 2002
   */
  private long actCompCell(SimpleNode node) throws InterpretError
  {
  	switch(node.type)
  	{
  		case OP_IDENTIFIER:
  		//System.out.println("in actCompCell, node.name = " + node.name);
  			//first check to see if there is any existing surrogate for this computation
  			Domain d = getDom(node.name);
  			long oldSur = -1 ,sur = -1;
  			if(d != null)
  			{
  				//this is an exising COMP type domain in destrel
  				//so just copy the surrogate over
  				int ind = alldoms.indexOf(d);
  				return ((long[])destrel.data[ind])[currow];
  			}
  			else if((oldSur = Global.comptable.lookup(node.name)) > 0)
  			{
  				//this computation has already been actualized in previous rows
  				return oldSur;
  			}
  			else //finally, check if this is a top-level computation
 			{
 				Relation r = env.lookupRel(node.name, true);
 				if((r != null) && (r.rvc == COMPUTATION))
 				{
 					//generate a new surrogate for the computation
 					sur = Surrogate.next();
 					//add the sur/comp mapping to Global.comptable
 					Global.comptable.addEntry(sur, (Computation)r);
 					Global.comptable.addEntry(node.name, new Long(sur));
 					return sur;
 				}
 				else
 					throw new InterpretError("Compuation " + node.name + " doesn't exist");
 			}
  		default:
  			throw new InterpretError("to be implemented");
  	}
  }     
  /**
   *  Actualize a long-typed cell.
   */
  private long actLongCell(SimpleNode node) throws InterpretError
  {
    
    switch(node.type){
 	    		case OP_FUNCTION: //added by Yi Zheng May 6, 2002
		//REVISIT: the type of argument is currently determined by the 
		 //type of argument accepted by corresponding Java methods
			SimpleNode argument = (SimpleNode) node.jjtGetChild(0);
			
			long ret;
			switch(node.opcode)
			{
				case OP_ABS://assume argument is of type long
				{
					long arg = actLongCell(argument);
					ret = (long) Math.abs(arg);
					return ret;
				}
				case OP_SQRT://assume argument is of type double
				{
					double arg = actDoubleCell(argument);
					ret = (long) Math.sqrt(arg);
					return ret;		
				}			
				case OP_SIN://assume argument is of type double
				{
					throw new InterpretError("Bad use of sin()");
				}
				case OP_ASIN://assume argument is of type double
				{
					double arg = actDoubleCell(argument);
					ret = (long) Math.asin(arg);
					return ret;	
				}
				case OP_COS://assume argument is of type double
				{
					throw new InterpretError("Bad use of cos()");
				}
				case OP_ACOS://assume argument is of type double
				{
					double arg = actDoubleCell(argument);
					ret = (long) Math.acos(arg);
					return ret;	
				}
				case OP_TAN://assume argument is of type double
				{
					double arg = actDoubleCell(argument);
					ret = (long) Math.tan(arg);
					return ret;	
				}
				case OP_ATAN://assume argument is of type double
				{
					double arg = actDoubleCell(argument);
					ret = (long) Math.atan(arg);
					return ret;	
				}		
				case OP_LOG://assume argument is of type double
				{
					double arg = actDoubleCell(argument);
					ret = (long) Math.log(arg);
					return ret;	
				}		
				case OP_LOG10://assume argument is of type double
				{
					double arg = actDoubleCell(argument);
					double denom = (double) Math.log(10);
					ret = (long) (Math.log(arg) / denom);
					return ret;
				}
				case OP_ROUND://assume argument is of type float
				{
					float arg = (float)actDoubleCell(argument);
					ret = (long) Math.round(arg);
					return ret;
				}
				case OP_CEIL://assume argument is of type double
				{
					double arg = actDoubleCell(argument);
					ret = (long) Math.ceil(arg);
					return ret;
				}
				case OP_FLOOR://assume argument is of type double
				{
					double arg = actDoubleCell(argument);
					ret = (long) Math.floor(arg);
					return ret;
				}
				case OP_CHR:
					//assumption: only deals with characters 0 - 255
				{
					throw new InterpretError("Bad use of chr()");
				}					
				case OP_ORD:
				{
					String s = actStrCell(argument);
					char startC = '\u0000';
					char arg;
					if(s == null)
						throw new InterpretError("Empty string in ord()");
					else
						ret = (long) ((char) s.charAt(0) - startC);
					return ret;
						
				}
				case OP_ISKNOWN:
				case OP_SINH:
				case OP_COSH:
				case OP_TANH:
					throw new InterpretError("Not implemented");
				default:
					throw new InterpretError("Unknown function operator");
			}
    case OP_IDENTIFIER:
      switch(destrel.domains[node.bits].type){
      case LONG:
	return ((long[])destrel.data[node.bits])[currow];
      case INTEGER:
	return (long) ((int[])destrel.data[node.bits])[currow];
      case SHORT:
	return (long) ((short[])destrel.data[node.bits])[currow];
      case FLOAT:
	return (long) ((float[])destrel.data[node.bits])[currow];
      case DOUBLE:
	return (long) ((double[])destrel.data[node.bits])[currow];
      default:
	throw new InterpretError("actLongCell: to be implemented.");
      }
    case OP_BIOPERATOR:
      SimpleNode left = (SimpleNode)node.jjtGetChild(0);
      SimpleNode right = (SimpleNode)node.jjtGetChild(1);
      
      switch(node.opcode){
	//case OP_PLUS:
	//return(actLongCell(left) + actLongCell(right));
	
	//Revised by hongyu. Sept 2001
      case OP_PLUS:{
	switch(node.bits){
	case SHORT:
	return (long)((short)actIntCell(left) + (short)actIntCell(right));
	 // throw new InterpretError("actLongCell: to be implemented.");
	case LONG:{
//	  System.out.println(" LONG----------------------------");
	  return actLongCell(left) + actLongCell(right);
	}
	case FLOAT:  
	case DOUBLE:
	  return (long)(actDoubleCell(left) + actDoubleCell(right));
	  //case NUMERIC: 
		case NUMERIC://case added by Yiyi June 4'02
		    return (actNumCell(left).add(actNumCell(right))).numToLong();

	case INTEGER:
	  return (long)(actIntCell(left) + actIntCell(right));
	case TEXT: // Hongyu added
	case STMT: // Hongyu added
	case EXPR: // Hongyu added
	case IDLIST:// ??Hongyu added 
	case STRING:
	  //System.out.println("in case STRING");
	default:
	  throw new InterpretError("actLongCell:  plus to be implemented.. .");
	}
      }
      
      //case OP_MINUS:
      //return(actLongCell(left) - actLongCell(right));
      
      //Revised by hongyu. Sept 2001
      case OP_MINUS:{
	switch(node.bits){
	case SHORT:
		return (long)((short)actIntCell(left) - (short)actIntCell(right));

//	  throw new InterpretError("actLongCell: to be implemented.");
	case LONG:{
//	  System.out.println(" LONG----------------------------");
	  return actLongCell(left) - actLongCell(right);
	}
	case FLOAT: 
	case DOUBLE:
	  return (long)(actDoubleCell(left) - actDoubleCell(right));
	  //case NUMERIC: 
		case NUMERIC:
		    return (actNumCell(left).subtract(actNumCell(right))).numToLong();

	case INTEGER:
	  return (long)(actIntCell(left) -  actIntCell(right));
	case TEXT: // Hongyu added
	case STMT: // Hongyu added
	case EXPR: // Hongyu added
	case IDLIST:// ??Hongyu added 
	case STRING:
	  //System.out.println("in case STRING");
	default:
	  throw new InterpretError("actLongCell:  minus to be implemented.. .");
	}
      }
      case OP_MUL:{
	switch(node.bits){
	case SHORT:
		return (long)((short)actIntCell(left) * (short)actIntCell(right));

//	  throw new InterpretError("actLongCell: to be implemented.");
	case LONG:{
//	  System.out.println(" LONG----------------------------");
	  return actLongCell(left) * actLongCell(right);
	}
	case FLOAT:  
	case DOUBLE:
	  return (long)(actDoubleCell(left) * actDoubleCell(right));
	  //case NUMERIC: 
       case NUMERIC: 
	   return (actNumCell(left).multiply(actNumCell(right))).numToLong();

	case INTEGER:
	  return (long)(actIntCell(left) *  actIntCell(right));
	case TEXT: // Hongyu added
	case STMT: // Hongyu added
	case EXPR: // Hongyu added
	case IDLIST://?? Hongyu added 
	case STRING:
	  //System.out.println("in case STRING");
	default:
	  throw new InterpretError("actLongCell:  mul to be implemented.. .");
	}
      }
      case OP_DIV:{
	switch(node.bits){
	case SHORT:{
		      short temp = (short)actIntCell(right);
		      if(temp!=0)
		      return (long)((short)actIntCell(left) / (short)actIntCell(right)); //Hongyu Need to be implement
		      else
		      throw new InterpretError("actLongCell: divided by zero");
		      
//	  throw new InterpretError("actLongCell:  div to be implemented.. .");
	}
	case LONG:{
//	  System.out.println(" LONG----------------------------");
	  long temp = actLongCell(right);
	  if(temp != 0L)
	    return actLongCell(left) / temp;
	  else
	    throw new InterpretError("actLongCell: divided by zero");
	}
	case FLOAT:
	case DOUBLE:{
	  double temp = actDoubleCell(right);
	  if(temp !=0.0)
	    return (long)(actDoubleCell(left) / temp);
	  else
	    throw new InterpretError("actLongCell: divided by zero");
	}
	//case NUMERIC: 
	case NUMERIC: 
	    return (actNumCell(left).divide(actNumCell(right))).numToLong();

	case INTEGER:{
	  int temp = actIntCell(right);
	  if(temp != 0)
	    return (long)(actIntCell(left) / temp);
	  else
	    throw new InterpretError("actLongCell: divided by zero");
	} 
	case TEXT: // Hongyu added
	case STMT: // Hongyu added
	case EXPR: // Hongyu added
	case IDLIST:// ??Hongyu added 
	case STRING:
	  //System.out.println("in case STRING");
	default:
	  throw new InterpretError("actLongCell:  div to be implemented.. .");
	}
      }
      
      //added by hongyu. Sept 2001
      case OP_MOD:{
	switch(node.bits){
	case SHORT:{
		      short temp = (short)actIntCell(right);
		      if(temp!=0)
		      return (long)((short)actIntCell(left) % (short)actIntCell(right));// Hongyu Need to be implement
		      else
		      throw new InterpretError("actLongCell: mod by zero");
		      
//	  throw new InterpretError("actLongCell:  mod to be implemented.. .");
	}
	
	case LONG:{
//	  System.out.println(" LONG----------------------------");
	  long temp = actLongCell(right);
	  if(temp != 0L)
	    return actLongCell(left) % temp;
	  else
	    throw new InterpretError("actLongCell: mod by zero");
	}
	case FLOAT:
	  throw new InterpretError("actLongCell: mod by value of float type");
	case DOUBLE:
	  throw new InterpretError("actLongCell: mod by value of double type");
	  
	  //case NUMERIC: 
			case NUMERIC: {
			number tmp = actNumCell(right);
			if(!number.iszero(tmp))
				return (number.mod(actNumCell(left),tmp)).numToLong();
			else
				throw new InterpretError("actLongCell: mod by 0");
			}

	case INTEGER:{
	  int temp = actIntCell(right);
	  if(temp !=0)
	    return (long)(actIntCell(left) %temp);
	  else
	    throw new InterpretError("actLongCell: mod by zero");
	} 
	case TEXT: // Hongyu added
	case STMT: // Hongyu added
	case EXPR: // Hongyu added
	case IDLIST:// ??Hongyu added 
	case STRING:
	  //System.out.println("in case STRING");
	default:
	  throw new InterpretError("actLongCell:  mod to be implemented.. .");
	}
      }
                //START Since JRelCaNu a.
		 case OP_CAST:
                // Cast from type stored in left.bits of 
                // the the left OP_TYPE node in buildTree and
                // traverseType.
                switch (left.bits) {
                case BOOLEAN:
                    return CastToLong.value(CastToBoolByte.value(actBoolCell(right)));

                case STRING:
                    return CastToLong.value(actStrCell(right));

                case INTEGER:
                    return CastToLong.value(actIntCell(right));

                case FLOAT:
                    return CastToLong.value(CastToFloat.value(actDoubleCell(right)));

                case SHORT:
                   // return CastToLong.value((short)CastToShort.value(actIntCell(right)));
					//bug fix by Yi Zheng Sept 28, 2002
					int aiv = actIntCell(right);
					if (aiv == INT_DC)
						return LONG_DC;
					else if (aiv == INT_DK)
						return LONG_DK;
					else if (aiv == (int) SHORT_DC)
						return LONG_DC;
					else if (aiv == (int) SHORT_DK)
						return LONG_DK;
					else
						return CastToLong.value((short)CastToShort.value(aiv));
					
					//end of bug fix Sept 28, 2002
                case LONG:
                    return actLongCell(right);

                case DOUBLE:
                    return CastToLong.value(actDoubleCell(right));
                
                
                //added by Yi Zheng Sept 28, 2002
                case DC:
                	return LONG_DC;
                
                case DK:
                	return LONG_DK;
                
                //end of addition Sept 28, 2002
//YYY
		case NUMERIC:
		    return CastToLong.value(actNumCell(right));
                case IDLIST:

                default:
		    Utility.dump(node,"actLongCell OP_CAST Node --->");
                    throw new InterpretError("actLongCell: not implemented.");
                }
          //END Since JRelCaNu a.
        /* commented out by yiyi
      case OP_MIN:{//added by hongyu
      	long lval = actLongCell(left);
      	long rval = actLongCell(right);
      	if (lval < rval) return lval;
      	return rval;
    	}
      case OP_MAX:{//added by hongyu
      	long lval = actLongCell(left);
      	long rval = actLongCell(right);
      	if (lval > rval) return lval;
      	return rval;
    	}
      case OP_POW://added by hongyu
      */
 		case OP_MIN:{//yiyi added
				switch(node.bits){
				case INTEGER:
					return (long)(Math.min(actIntCell(left),actIntCell(right)));
				case LONG:
					return Math.min(actLongCell(left),actLongCell(right));
				case SHORT:
					throw new InterpretError("actLongCell: min to be implemented");
				case NUMERIC:
					return number.min(actNumCell(left),actNumCell(right)).numToLong();
				case FLOAT:
				case DOUBLE:
					return (long)(Math.min(actDoubleCell(left),actDoubleCell(right)));
				default:
     throw new InterpretError("actLongCell: min to be implmented.");
				}
			}
		case OP_MAX:{//yiyi added
				switch(node.bits){
				case INTEGER:
					return (long)(Math.max(actIntCell(left),actIntCell(right)));
				case LONG:
					return Math.max(actLongCell(left),actLongCell(right));
				case SHORT:
					throw new InterpretError("actLongCell: max to be implemented");
				case NUMERIC:
					return number.max(actNumCell(left),actNumCell(right)).numToLong();
				case FLOAT:
				case DOUBLE:
					return (long)(Math.max(actDoubleCell(left),actDoubleCell(right)));
				default:
     throw new InterpretError("actLongCell: max to be implmented.");
				}
			}
	    case OP_POW:{//yiyi added
				switch(node.bits){
				case LONG:
					return (long)(Math.pow((double)(actLongCell(left)),(double)(actLongCell(right))));
				case INTEGER:
					return (long)(Math.pow((double)(actIntCell(left)),(double)(actIntCell(right))));
				case SHORT:
					throw new InterpretError("actLongCell: pow to be implmented");
				case FLOAT:
				case DOUBLE:
					return (long)(Math.pow(actDoubleCell(left),actDoubleCell(right)));
				case NUMERIC:
					return number.pow(actNumCell(left),actNumCell(right)).numToLong();
				default:
					throw new InterpretError("actLongCell: pow to be implemented.");
				}
			}     
      default:
	throw new InterpretError("actLongCell: bio to be implemented.. .");
      }
    case OP_UOPERATOR:
      switch(node.opcode){
      case OP_UPLUS:
	return(actLongCell((SimpleNode)node.jjtGetChild(0)));
      case OP_UMINUS:
	return(0L-actLongCell((SimpleNode)node.jjtGetChild(0)));
      default:
	throw new InterpretError("actLongCell: uni to be implemented.");
      }
    case OP_LITERAL:
      if(node.info instanceof Long)
	return(((Long)node.info).longValue()); 
      else if(node.info instanceof Integer)
	return (long)((Integer)node.info).intValue();
      else if(node.info instanceof Short)
	return (long)((Short)node.info).shortValue();
      if(node.info instanceof Double)
	return (long) ((Double)node.info).doubleValue();
      else if(node.info instanceof Float)
	return (long)((Float)node.info).floatValue();
      else
	throw new InterpretError("actDoubleCell: to be implemented.");
    case OP_RED:
    case OP_EQUIV:
    case OP_FUN:// hognyu added
    case OP_PAR:// hognyu added
    if(node.opcode != OP_REDFUNCALL)
    {
      long actValue= actLongCell((SimpleNode)node.jjtGetChild(0));
      
      switch(node.opcode){
      case OP_PLUS:
	if(node.info == null)  node.info = new Long(0);
	node.info = new Long(((Long)node.info).longValue()+actValue);
	break;
      case OP_MUL:
	if(node.info == null)  node.info = new Long(1);
	node.info = new Long(((Long)node.info).longValue()*actValue);
	break;
      case OP_MIN:
	if(node.info == null)  node.info = new Long(Long.MAX_VALUE);
	node.info = new Long(
			     Math.min(((Long)node.info).longValue(), actValue));
	break;
      case OP_MAX:
	if(node.info == null)  node.info = new Long(Long.MIN_VALUE);
	node.info = new Long(
			     Math.max(((Long)node.info).longValue(), actValue));
	break;
	//added by Yi Zheng July 5, 2002
	//let a be red nop of b
	case OP_NOP:
	if(node.info == null)
		node.info = new Long(actValue);
	else
	{
		boolean rb = new Random().nextBoolean();
		if(rb)
			node.info = new Long(actValue);
	}
	break;
	
	default:
		throw new InterpretError("actLongCell: to be implemented");	
	
	//end of addition July 5
      }
      return actValue;
    }
    else //added by Yi Zheng July 9, 2002
    {
  	     long actValue= actLongCell((SimpleNode)node.jjtGetChild(1));
		RedFunCallAccumNode(node, LONG, new Long(actValue));
		return actValue;
    	
    }
    case OP_IFTHENELSE:
      if(actBoolCell((SimpleNode)node.jjtGetChild(0)))
	return actLongCell((SimpleNode)node.jjtGetChild(1));
      else
	return actLongCell((SimpleNode)node.jjtGetChild(2));
      
    case OP_QTOPERATOR:{// Added by hongyu, Sept 15,01
      switch(node.bits){
      case INTEGER:
	return (long)actIntCell((SimpleNode)node.jjtGetChild(0));
      case DOUBLE:
	return (long)actDoubleCell((SimpleNode)node.jjtGetChild(0));
	// End of Hongyu Added. Sept 15,01  
      }
    } 
            //added by Yi Zheng July 2, 2002
    // to implement level lifting
    case OP_PROJECT:
    case OP_TSELECT: {
    	Vector dnames = new Vector();
    	Vector rname = new Vector();
    	Vector rindx = new Vector();
    	relhtbl = new Hashtable();
    	String tname0 = interpreter.nextTempName();
    	String tname1 = interpreter.nextTempName();
    	
    	SimpleNode projector = (SimpleNode) node.jjtGetChild(0);
    	if(!projector.isBitSet(1))
    		throw new InterpretError("actLongCell: to be implemented.");
    	SimpleNode exprlist = (SimpleNode) projector.jjtGetChild(0);
    	if(exprlist.jjtGetNumChildren() != 1)
    		throw new InterpretError("actLongCell: to be implemented.");
		SimpleNode rednode = (SimpleNode) exprlist.jjtGetChild(0);
		if((rednode.name == null) || (rednode.name.compareTo("") == 0))
    		throw new InterpretError("actLongCell: to be implemented.");
		Domain redDom = env.lookupDom(rednode.name, true);
		if((redDom == null) || (redDom.isRedTemp == 0))
    		throw new InterpretError("actLongCell: to be implemented.");
    	Relation tempRel = actRelCell(node,  dnames, rname,
			      rindx, relhtbl, tname0, tname1);
		if(tempRel.domains.length != 1)
			throw new InterpretError("actLongCell: to be implemented");
		
		if(tempRel.domains[0].type != LONG)
			throw new InterpretError("actLongCell: type mismatch");
		tempRel.load(true);
		long retVal = ((long[]) tempRel.data[0])[0];
		tempRel.unload();
		return retVal;
	}
    //end of Yi's addition July 2, 2002

	//added by Yi Zheng July 5, 2002
	case OP_JOINOPERATOR:
		if((node.opcode == OP_NOP) && (node.jjtGetNumChildren() == 2))
		{
			boolean rb = new Random().nextBoolean();
			if(rb)
				return actLongCell((SimpleNode)node.jjtGetChild(0));
			else
				return actLongCell((SimpleNode)node.jjtGetChild(1));
		}	
	//deliberately allow to fall through to the default case
	//end of addition July 5, 2002
    default:
      throw new InterpretError("actLongCell: to be implemented.");
    }
  }      
  /**
   * Actualize an IDLIST-typed cell.
   */
  private Relation actRelCell(SimpleNode node, Vector dnames, Vector rname,
			      Vector rindx, Hashtable relhtbl, String tname0, String tname1)
       throws InterpretError
  {
    Relation r = null, subr;
    
//    System.out.println(" actRelCell");
//    System.out.println(" node type "+node.type);
 	   
    switch(node.type){
    case OP_RED:
    case OP_EQUIV:
    case OP_FUN:// from sungsoo, hongyu added Feb 2002
    case OP_PAR:// from sungsoo, hongyu added Feb 2002

      if(node.opcode != OP_REDFUNCALL)
      {
	      SimpleNode srcnode = (SimpleNode)node.jjtGetChild(0);
	      r = actRelCell(srcnode, dnames, rname, rindx, relhtbl, tname1, null);
	      
	      if(node.info == null)   {
			node.info = r;
		//((Relation)node.info).print();//added by hongyu
	      }
	      else{
	      	if(node.opcode == OP_NOP) //added by Yi Zheng July 5, 2002
	      	{
	      		boolean rb = new Random().nextBoolean();
	      		if (rb)
	      			node.info = r;
	      		
	      	}else
	      	{
				node.info = ((Relation)node.info).mujoin(node.opcode, null,
									 r, null, tname0);
			//	((Relation)node.info).print();
				((Relation)node.info).load(true);
			}
	      }
    }
    else //added by Yi Zheng July 9, 2002
    {
    	  SimpleNode srcnode = (SimpleNode)node.jjtGetChild(1);
	      r = actRelCell(srcnode, dnames, rname, rindx, relhtbl, tname1, null);
	      RedFunCallAccumNode(node, IDLIST, r);
    }
      
      break;

    default:
      /* for(int i = 0; i < rname.size(); i++){
//	 System.out.println(" rname i "+i);
	 int index = ((Integer)rindx.elementAt(i)).intValue();
	 long surg = ((long[])destrel.data[index])[currow];
	 subr = Relation.getRelation((String)dnames.elementAt(i), surg);
	 subr.name = (String)rname.elementAt(i);
	 
	 relhtbl.put(subr.name, subr);
	 }
	 */
      // Note, Biao has changed the getRelation() method so that we don't
      // need to save above 'subr', and evaluateTLExpression() will not
      // try to load relations from disk if there're data in memory.
      // However, evaluateTLExpression() and Relation.mujoin() methods
      // still saves their result to disk.
//      System.out.println(" here");
      //Yi Zheng June 24, 2002
      
      //added by Yi Zheng June 24, 2002
      destrel.load(true);
      interpreter.destrel = destrel;
      interpreter.currow = currow; //June 25, 2002
      //end of addition June 24, 2002
      if(this.relhtbl != null)
      	r = interpreter.evaluateTLExpression(node, tname0, this.relhtbl, env);
      else
       	r = interpreter.evaluateTLExpression(node, tname0, new Hashtable(), env);
     	
     	interpreter.destrel = null; //June 24, 2002
     	interpreter.currow = -1; //June 25, 2002
	//bug fix Yi Zheng May 7, 2002
	  if((r.data == null) && (r.name.compareTo("") != 0))
	  	r.load();
		r.name = tname0;
		if(r.data != null)
			r.save();
	//end of bug fix May 7, 2002

	//DEBUGGING
//	System.out.println("in actRelCell: r to be returned:");
//	r.print();
	r.load(true);
	
	//added by Yi Zheng July 2, 2002 to handle level lifting
	// let A be [red ujoin of B] in C;
	 if((node.type == OP_PROJECT) || (node.type == OP_TSELECT))
	 {
   // 	System.out.println("in actRelCell: node is:");
   // 	Utility.dump(node, "-->");
    	
    	
    	SimpleNode projector = (SimpleNode) node.jjtGetChild(0);
    	SimpleNode exprlist = null;
    	if(projector.isBitSet(1))
    		exprlist = (SimpleNode) projector.jjtGetChild(0);
    	if((exprlist == null) || (exprlist.jjtGetNumChildren() != 1))
    		break;
		SimpleNode rednode = (SimpleNode) exprlist.jjtGetChild(0);
		if((rednode.name == null) || (rednode.name.compareTo("") == 0))
    		break;
		Domain redDom = env.lookupDom(rednode.name, true);
		if((redDom == null) || (redDom.type != IDLIST) || (redDom.isRedTemp == 0))
    		break;
    	
   // 	System.out.println("relation " + r.name + " has the following domains");
   // 	for(int i = 0; i < r.domains.length; i++)
   // 		System.out.println(r.domains[i].name);	
		if(!r.isDomainIn(redDom, r.domains))
			throw new InterpretError("domain " + redDom.name + " not found in " + r.name);
		
		//now extract the relation represented by redDom
		int index = r.indexOfDomain(redDom);
		if(index == ERROR)
			throw new InterpretError("domain " + redDom.name + " not found in " + r.name);
		long surr = ((long[]) r.data[index])[0];
		Relation dotRel = env.lookupRel("." + redDom.name, true);
		if(dotRel == null)
			throw new InterpretError("." + redDom.name + " not found");
//		dotRel.print();
//		System.out.println("surr = " + surr);
		Relation retVal = dotRel.getRelation(surr);
		retVal.name = tname1;
		if(retVal.data != null)
			retVal.save();
//		System.out.println("retVal is:");
//		retVal.print();
		retVal.load(true);	
		
		return retVal;
	}
    //end of Yi's addition July 2, 2002   
	
      break;
    }
    //relhtbl.clear();
    return r;
  }         
  /**
   *  Actualize a string-typed cell.
   */
  private String actStrCell(SimpleNode node) throws InterpretError
  {
    if(debug_oct_52)
    	Utility.dump(node, "actStrCell: ");
    switch(node.type){
    //case OP_FUNCTION added by Yi Zheng May 6, 2002
    case OP_FUNCTION:
    		SimpleNode argument = (SimpleNode) node.jjtGetChild(0);
	
			String ret;
			switch(node.opcode)
			{
				case OP_CHR:
					//assumption: only deals with characters 0 - 255
				{
					int code = (int)actDoubleCell(argument);
					if(code < 0 || code > 255)
						throw new InterpretError("Invalid argument to chr()");
					else
					{
						char startC = '\u0000';
						char[] newC = new char[1];
						newC[0] = (char)(startC + ((char) code));
						ret = new String(newC);
						return ret;
					}
				}					
				default:
					throw new InterpretError("Unknown function operator");
			}	

    case OP_IDENTIFIER:
      return ((String[])destrel.data[node.bits])[currow];
    case OP_BIOPERATOR:
      SimpleNode left = (SimpleNode)node.jjtGetChild(0);
      SimpleNode right = (SimpleNode)node.jjtGetChild(1);
      
      switch(node.opcode){
      case OP_CAT:
	return actStrCell(left) + actStrCell(right);
	          //START Since JRelCaNu a.
		 case OP_CAST:
                        // Cast from type stored in left.bits of 
			// the the left OP_TYPE node in buildTree and
			// traverseType.
		        switch(left.bits){
		        case BOOLEAN:
			   return CastToStrn.value(CastToBoolByte.value(actBoolCell(right)));
		        case SHORT:
			   // Must cast to short first, otherwise, we may get
			   // a bigger number than possible for short in the
			   // string result.
			   return  CastToStrn.value((short)CastToShort.value(actIntCell(right)));
		        case INTEGER:
			   return  CastToStrn.value(actIntCell(right));
		        case LONG:
			   return  CastToStrn.value(actLongCell(right));
		        case FLOAT:
			   // Must cast to float first, otherwise, we may get
			   // a bigger number than possible for float in the
			   // string result.
			   return CastToStrn.value(CastToFloat.value(actDoubleCell(right)));
		        case DOUBLE:
			   return  CastToStrn.value(actDoubleCell(right));
		        case STRING:
			   return  CastToStrn.value(actStrCell(right));
			   
			   //added by Yi Zheng Sept 28, 2002
			   case DC:
			   	return STRING_DC;
			   case DK:
			    return STRING_DK;
			   
			   //end of addition Sept 28, 2002
//YYY
		        case NUMERIC:
		           return CastToStrn.value(actNumCell(right));
		        default:
			     Utility.dump(node,"actStrCell OP_CAST Node --->");
			     throw new InterpretError("actStrCell: not implemented.");
		        }
          //END Since JRelCaNu a.
      default:
	throw new InterpretError("actStrCell: bio to be implemented.");
      }
    case OP_UOPERATOR:
      switch(node.opcode){
      default:
	throw new InterpretError("actStrCell: uni to be implemented.");
      }
    case OP_LITERAL:
      // REVISIT: node.opcode must be OP_STRING.
	//System.out.println("LITERAL");
      return((String)node.info); 
    case OP_RED:
    case OP_EQUIV: // should have OP_FUN and OP_PAR? hongyu Feb 2002
    //  return ""; //modified by Yi Zheng July 5, 2002
    //added by Yi Zheng July 5, 2002
    case OP_FUN:
    case OP_PAR:
    if(node.opcode != OP_REDFUNCALL)
    {
      String actValue= actStrCell((SimpleNode)node.jjtGetChild(0));
      
      switch(node.opcode){
      case OP_MIN:
	if(node.info == null)  node.info = new String(actValue);
	else
		node.info = ((String)node.info).compareTo(actValue) < 0? (String)node.info: actValue;
	break;
      case OP_MAX:
	if(node.info == null)  node.info = new String(actValue);
	else
		node.info = ((String)node.info).compareTo(actValue) > 0? (String)node.info: actValue;
	break;

		//let a be red nop of b
		case OP_NOP:
		
		if(node.info == null)
			node.info = new String(actValue);
		else
		{
			boolean rb = new Random().nextBoolean();
			if(rb)
				node.info = new String(actValue);
		}
		break;
		
		default:
			break;	
      }
      return actValue;	
	
    //end of Yi's addition July 5
	}
	else //added by Yi Zheng July 8, 2002
	{
		String actValue = actStrCell((SimpleNode)node.jjtGetChild(1));
		RedFunCallAccumNode(node, STRING, actValue);
		return actValue;
	}
    case OP_IFTHENELSE:
   //   return "";
   
   //bug fix Sept 28, 2002
   SimpleNode cond = (SimpleNode) node.jjtGetChild(0);
   if(debug_oct_52)
   {
   	System.out.println("about to call actBolCell with:");
   	Utility.dump(cond, "cond: ");
   }
   	
   boolean bl = actBoolCell(cond);
   if(debug_oct_52)
   	System.out.println("after calling actBoolCell.");
   if (bl) 
   {
   		return actStrCell((SimpleNode)node.jjtGetChild(1));
   	
   }
   else
   {
   		return actStrCell((SimpleNode)node.jjtGetChild(2));
   }
   
   
   //end of bug fix Sept 28, 2002
   
   
            //added by Yi Zheng July 2, 2002
    // to implement level lifting
    case OP_PROJECT:
    case OP_TSELECT: {
    	Vector dnames = new Vector();
    	Vector rname = new Vector();
    	Vector rindx = new Vector();
    	relhtbl = new Hashtable();
    	String tname0 = interpreter.nextTempName();
    	String tname1 = interpreter.nextTempName();
    	
    	SimpleNode projector = (SimpleNode) node.jjtGetChild(0);
    	if(!projector.isBitSet(1))
    		throw new InterpretError("actStrCell: to be implemented.");
    	SimpleNode exprlist = (SimpleNode) projector.jjtGetChild(0);
    	if(exprlist.jjtGetNumChildren() != 1)
    		throw new InterpretError("actStrCell: to be implemented.");
		SimpleNode rednode = (SimpleNode) exprlist.jjtGetChild(0);
		if((rednode.name == null) || (rednode.name.compareTo("") == 0))
    		throw new InterpretError("actStrCell: to be implemented.");
		Domain redDom = env.lookupDom(rednode.name, true);
		if((redDom == null) || (redDom.isRedTemp == 0))
    		throw new InterpretError("actStrCell: to be implemented.");
    	Relation tempRel = actRelCell(node,  dnames, rname,
			      rindx, relhtbl, tname0, tname1);
		if(tempRel.domains.length != 1)
			throw new InterpretError("actStrCell: to be implemented");
		
		if(tempRel.domains[0].type != STRING)
			throw new InterpretError("actStrCell: type mismatch");
		tempRel.load(true);
		String retVal = ((String[]) tempRel.data[0])[0];
		tempRel.unload();
		return retVal;
	}
    //end of Yi's addition July 2, 2002

	//added by Yi Zheng July 5, 2002
	case OP_JOINOPERATOR:
		if((node.opcode == OP_NOP) && (node.jjtGetNumChildren() == 2))
		{
			boolean rb = new Random().nextBoolean();
			if(rb)
				return actStrCell((SimpleNode)node.jjtGetChild(0));
			else
				return actStrCell((SimpleNode)node.jjtGetChild(1));
		}		
		//deliberately to allow fall through to the default case
	//end of addition July 5, 2002
    default:
      throw new InterpretError("actStrCell: to be implemented.");
    }
  } 
  
      /**
	*  Actualize a string-typed cell. -- Andrey Rozenberg
	*/
	//note this method is never used
   private long actUniCell(SimpleNode node) throws InterpretError
   {
     switch(node.type)
	 {
		case OP_IDENTIFIER:
				String str =  ((String[])destrel.data[node.bits])[currow];
				String value = str.substring(str.indexOf(";")+1);
				String type = str.substring(0,str.indexOf(";"));
/*				switch(type)
				{
					case "integer":
					case "string":
					
				}
*/			
	    case OP_LITERAL:
		    	 if(node.info instanceof Long)
					return(((Long)node.info).longValue()); 
				else if(node.info instanceof Integer)
					return (long)((Integer)node.info).intValue();
				else if(node.info instanceof Short)
					return (long)((Short)node.info).shortValue();
				if(node.info instanceof Double)
					return (long) ((Double)node.info).doubleValue();
				else if(node.info instanceof Float)
					return (long)((Float)node.info).floatValue();
				/*else if(node.info instanceof String)
					return  (long)  (node.info);*/
				else
					throw new InterpretError("actUniCell: literal - to be implemented.");
		 
			
		default:
			throw new InterpretError("actUniCell: to be implemented.");
	 }
   }
    
       
  /**
   * Actualize the virtual domains and save the resulting relation to
   * disk relation named 'relname'.
   * Usual calling procedure :
   *          actualizer = new Actualizer(...);
   *          destrel = actualizer.actualize();
   */
  public Relation actualize() throws InterpretError
  {
    Vector vdomsv = new Vector();
    Vector rdomsv = new Vector();
    Domain[] vdoms, rdoms;
    
    // Generate domain array for generating destination relation.
    Domain[] domsarray = new Domain[alldoms.size()];
    alldoms.copyInto(domsarray);
    
    // Create a copy of the source relation and put it to the
    // destination relation.
    destrel = srcrel.duplicateLoadAll();    // we can load now.
    destrel.domains = domsarray;
    destrel.numattrs = domsarray.length;      // set it to actual length.
    
    // REVISIT: not necessary to create those data arrays that
    // already exists in srcrel.domains.
    
    Object[] data =destrel.createDataArray(destrel.domains,destrel.numtuples);
    
   
    for(int i = 0; i < destrel.data.length; i++)
      data[i] = destrel.data[i];
    destrel.data = data;
    
    // We may need additional pass to actualize the domains of RED 
    // and/or EQUIV.
    maxvertlevel--;
    

    // Check for vertical operations firstly. Actualize all those 
    // RED & EQUIV virtual domains before other normal domains.
    while(!vertdom.isEmpty()){
      for(Enumeration e = vertdom.keys(); e.hasMoreElements();){
	Domain d = (Domain)e.nextElement();
	if(debug_oct_52)
		System.out.println("About to actualize vertdom: " + d.name);
	if(((Integer)vertdom.get(d)).intValue() == maxvertlevel){
	  // Find this domain's tree and check if its Equiv.
	  // Note: we need to actualize equiv domains one by one.
	  SimpleNode tree = (SimpleNode)vvtree.get(d);
	  
	  if(tree.type == OP_EQUIV){
	    //this part should be changed to call either nested or normal domain
	    // from sungsoo. hongyu added Feb 2002
	    actualizeEquiv(d);
	  }
	  else if(tree.type == OP_RED){
	    vdomsv.addElement(d);
	    
	  }
	  else if(tree.type == OP_FUN)// from sungsoo. hongyu added Feb 2002
	    {
	      //this part should be changed to call either nested or normal domain
	      actualizeFun(d);
	    }
	  else if(tree.type == OP_PAR)// from sungsoo. hongyu added Feb 2002
	    {
	      actualizeParFun(d);
	    }
	  else{
	  	if(debug_oct_52)
	  		System.out.println(d.name + "added to rdomsv");
	    rdomsv.addElement(d);
	  }
	  vertdom.remove(d);
	}
      }
      maxvertlevel--;
      
      // Vector rdomsv contains the intermediate virtual domains.
      // E.g. let x be equiv+ of A by A+B;   here A+B will be a temp
      // domain which must be actualized prior to the actualization 
      // of x.
      if(rdomsv.size() != 0){
	vdoms = new Domain[rdomsv.size()];
	rdomsv.copyInto(vdoms);
	rdomsv.removeAllElements();   // clear it for next run.
	if(debug_oct_52)
		System.out.println("about to call actualizing on rdomsv");
	actualizing(vdoms, 0);
      }
      
      // Vector vdomsv contains the intermediate domains for vertical
      // operations. E.g. let x be A + equiv+ of A by B;  here 
      // 'equiv+ of A by B' as a whole is a domain which must be 
      // actualized prior to the actualization of 'x'.
      if(vdomsv.size() != 0){
	vdoms = new Domain[vdomsv.size()];
	vdomsv.copyInto(vdoms);
	vdomsv.removeAllElements();   // clear it for next run.
	actualizing(vdoms, 1);
      }
    }
    
    if(virdoms.size() > 0) //if added by Yi Zheng July 9, 2002
    {
	    // Actualize the normal virtual domains (but not nested domains).
	    vdoms = new Domain[virdoms.size()];
	    virdoms.copyInto(vdoms);
	    actualizing(vdoms, 0);
	}
    
    //this is never used !!!! this is all messed up. cant understand previous programmers intention
    // of putting this junky code here.
    // Above comments are from sungsoo. hongyu added
    
    // Actualize the nested domains at last, if we want to use the new
    // technology of whole relation-level actualization (but not tuple-
    // by-tuple approach).
    
    if(nt && nestdoms.size() != 0){
//      System.out.println(" nt && nestdoms.size() != 0 ");
      // We will have to give a temporary name to destrel and save it
      // to disk file, since the mujoin will need it.  ;-(
      // REVISIT: this is too bad... Can we avoid this???
      destrel.name = interpreter.nextTempName();
      destrel.save();
      
      for(int i = 0; i < nestdoms.size(); i++){
	// Get the nested domain.
	Domain dom = (Domain)nestdoms.elementAt(i);
	
	Hashtable htable = (Hashtable)nesthashs.get(dom);
	Hashtable htbl = new Hashtable();
	
	// Put this (operand) relation to htbl for
	// evaluateTLExpression() to look for, since the hashtable
	// will be searched first, and then reltable is consulted.
	htbl.put(destrel.name, destrel);
	
	SimpleNode node = (SimpleNode)vvtree.get(dom);
	actualizeNestedDom(node, htable, htbl);
      }
      
      // Remove the temporary destrel disk file
      interpreter.deleteFile(destrel.name);
      destrel.name = "";
    }
    
    // Added by hongyu.
    if(qtActive){
      if (destrel.numtuples != 0){
	
	Vector tempAll = new Vector();
	for(int i=0;i<alldoms.size();i++)
	  tempAll.addElement(alldoms.elementAt(i)); //Store alldoms
	alldoms = new Vector();
	for(int i=0;i<destdoms.length;i++){
	  alldoms.addElement(destdoms[i]);
	}
	int tempCurrow = currow;
	destrel.sort(destdoms);
	destrel= evalQTSelector(destrel.data,destrel.numtuples);
	alldoms = tempAll;
	currow = tempCurrow;
	
      }
    }
    //  End of Added by Hongyu. July 15,01 
    
	//restore the original trees in virtual domains: Yi Zheng May 7, 2002, bug fix
	 restoreTree();

    // Return the resulting relation.
    return(destrel);
  }   
  
  //restore the original virtual domain trees
  //Bug fix
  //Yi Zheng May 7, 2002
  private void restoreTree()
  {
	for(int i = 0; i < origDoms.length; i++)
	{
		if(origDoms[i].tree != null)
		{
			SimpleNode oldTree = (SimpleNode) origTreeTable.get(origDoms[i].name);
			origDoms[i].tree = oldTree;
		}
	}
  }
   
  /**
   * Actualize the EQUIV-typed virtual domains. Note: we are now 
   * actualizing it one by one. An optimization could be made by 
   * actualize the domains with exactly the same by-list.
           *
        * JRelCaNu modified/enhanced this method:
        * - Optionally handle Nulls and Casts with actualizeEquivWithNull();

   */
  private void actualizeEquiv(Domain dom) throws InterpretError
  {
  	          //START Since JRelCaNu d.
          //if (actWithNull)  {actualizeEquivWithNull(dom); return;};
          //END Since JRelCaNu d.

    int i, oldpos, virtype = dom.type;
    SimpleNode virtree = (SimpleNode)vvtree.get(dom);
    
    // Variables for IDLIST-typed domain and when nt == false;
    long sur =0; // from sungsoo. hongyu added Feb 2002
    String tname0 = null, tname1 = null;
    Hashtable relhtbl = null;
    Vector rname = null;
    Vector rindx = null;
    Vector vdom = null;
    Relation dest = null;
    
    // First we need to sort on the by-list.
    Vector bydoms = new Vector();
    //if added by Yi Zheng July 9, 2002
    SimpleNode exprlist = null;
    if(virtree.opcode != OP_REDFUNCALL)
    	exprlist = (SimpleNode)virtree.jjtGetChild(1);
    else
    	exprlist = (SimpleNode)virtree.jjtGetChild(2);
    	
    for(i = 0; i < exprlist.jjtGetNumChildren(); i++){
      SimpleNode child = (SimpleNode) exprlist.jjtGetChild(i);
      Domain d = getDom(child.name);
      if(d != null)
	bydoms.addElement(d);
      else
	// Should never happen...
	throw new InterpretError("actualizeEquiv: domain not found.");
    }
    Domain[] byarray = new Domain[bydoms.size()];
    bydoms.copyInto(byarray);
    
    if(debug_nov_24)
    {
    	System.out.println("byarray: ");
    	int kk = 0;
    	for(;kk < byarray.length;kk++)
    		System.out.print(" " + byarray[kk].name);
    	System.out.println("");	
    }
    
//    Global.println("age "+age()+": start sort");
    destrel.sort(byarray);
    
    
//    Global.println("age "+age()+": finish sort");
    
    // Since we did sort on the by-list, we must rearrange the 
    // alldoms vector elements correspondingly.
    for(i = byarray.length - 1; i >=0; i--){
      alldoms.removeElement(byarray[i]);
      alldoms.insertElementAt(byarray[i], 0);
    }
    
    // Since alldoms changed, we must reset the sequence number stored
    // in node.bits.
    for(Enumeration e = vvtree.elements(); e.hasMoreElements();){
      resetNodeBits((SimpleNode)e.nextElement());
    }
    
    
    if(virtype == IDLIST && !nt){
      relhtbl = new Hashtable();
      rname = new Vector();
      rindx = new Vector();
      vdom = new Vector();
      tname0 = interpreter.nextTempName();
      tname1 = interpreter.nextTempName();
      
      tempnames.addElement(tname0);  // remember to clear this temp rel.
      tempnames.addElement(tname1);  // remember to clear this temp rel.
      
      // if((dest = reltable.getRel("." + dom.name)) == null)
      // REVISIT: Zhongxia, I used true but check it please
      if((dest = env.lookupRel("." + dom.name, true)) == null)
	
	// Should never happen..
	throw new InterpretError("actualizeequiv: relation .'" +
				 dom.name + "' not found.");
      
      // Fill the vectors.
      preProcIDListDom(virtree, vdom, rname, rindx);
    }
    
    NREnvironment NRenv = env.addNRFrame(destrel);//from sungsoo. hongyu added Feb 2002
    // Remember this domain's position in 'alldom'.
    int pos = alldoms.indexOf(dom);
    //bug fix by Yi Zheng June 23, 2002
    virtree.bits = pos;
    
    //end of bug fix June 23, 2002
    // Get the position of the domains in by-list.
    int bypos[] = new int[byarray.length];
    for(i = 0; i < byarray.length; i++)
      bypos[i] = alldoms.indexOf(byarray[i]);
    
    
    if(debug_nov_24)
    {
    	System.out.println("bypos: ");
    	int kk = 0;
    	for(;kk < bypos.length;kk++)
    		System.out.print(" " + bypos[kk]);
    	System.out.println("");	
    	
    }
    // Get the initial values for sort attributes.
    currow = oldpos = 0;
    
    //YuLing changed the following loop  in March 2001 to 
    //fix the by STRING problem
    Vector vals=new Vector();
    
    //For 1. from sungsoo, hongyu added Feb 2002
    for(i = 0; i < byarray.length; i++){
      
      
      switch(byarray[i].type){
      	case NUMERIC:
	    number numval = ((number[])destrel.data[bypos[i]])[currow];
	    vals.addElement(numval);
	    break;

      case INTEGER:
      case SHORT:
//      case BOOLEAN://modified by Yi Zheng May 7, 2002
	Integer intval=new
	  Integer(((int[])destrel.data[bypos[i]])[currow]);
	vals.addElement (intval);
	break;
	       case BOOLEAN: //modified by Yi Zheng May 7, 2002
		 Byte bval=new
		 Byte (((byte[])destrel.data[bypos[i]])[currow]);
	         vals.addElement (bval);
		 break;
      case LONG:
	Long longval=new
	  Long(((long[])destrel.data[bypos[i]])[currow]);
	vals.addElement (longval);
	break;
      case FLOAT:
	Float floatval=new
	  Float(((float[])destrel.data[bypos[i]])[currow]);
	vals.addElement (floatval);
	break;
      case DOUBLE:
	Double doubleval=new
	  Double(((double[])destrel.data[bypos[i]])[currow]);
	vals.addElement (doubleval);
	break;
	      case UNIVERSAL:  //from Andrey
	   case ATTRIBUTE: //from Andrey
      case TEXT:
      case STRING:
      if(debug_nov_24)
      	System.out.println("String added: " + ((String[])destrel.data[bypos[i]])[currow]);
      case EXPR:
	vals.addElement(((String[])destrel.data[bypos[i]])[currow]);
	break;
      case IDLIST:
	//Sungoo's change
	//No need for checkin the value?
	break;
      default:
	throw new InterpretError("actualizeEquiv(): to be implemented..");
      }
    }
    
    //YuLing's modification ends here
    
    
    // Now actualize the virtual domain tuple by tuple.
    //For 2
    for(; currow < destrel.numtuples; currow++){
      // If the sort values are changed, we must save the values to
      // the destrel on the fly.
      
      NRenv.row = currow;// from sungsoo. hongyu added Feb 2002
      int breakflag=0;
     	  if(debug_nov_24)
	  	System.out.println("currow = " + currow);

      for (i=0;i < byarray.length; i++)
	{
	  //This loop is changed by YuLing in March, 2001
	  //to deal with the by String problem of equiv
	  
	  if(debug_nov_24)
	  	System.out.println("byarray[" + i + "]");
	  	
	  switch(byarray[i].type){
	  	
	 	  case NUMERIC:
	      if(((number)vals.elementAt(i)).compareTo(((number[])destrel.data[bypos[i]])[currow]) != 0){
		  vals.setElementAt(
				    (((number[])destrel.data[bypos[i]])[currow]),i);
		  breakflag = 1;
	      }
	      break;

	  case INTEGER:
	  case SHORT:
	    if(((Integer)(vals.elementAt(i))).intValue() !=
	       ((int[])destrel.data[bypos[i]])[currow]){
	      vals.setElementAt( new Integer(
					     ((int[])destrel.data[bypos[i]])[currow]),i);
	      breakflag=1;
	    }
	    break;
		case BOOLEAN:	//case added by Yi Zheng May 7, 2002
		     if(((Byte)(vals.elementAt(i))).byteValue() !=
		          ((byte[])destrel.data[bypos[i]])[currow]){
		           vals.setElementAt( new Byte(
		          ((byte[])destrel.data[bypos[i]])[currow]),i);
			  breakflag=1;
			  }
		     break;

	  case LONG:
	    if(((Long)(vals.elementAt(i))).longValue() !=
	       ((long[])destrel.data[bypos[i]])[currow]){
	      vals.setElementAt( new Long(
					  ((long[])destrel.data[bypos[i]])[currow]),i);
	      breakflag=1;
	    }
	    break;
	  case DOUBLE:
	    if(((Double)(vals.elementAt(i))).doubleValue() !=
	       ((double[])destrel.data[bypos[i]])[currow]){
	      vals.setElementAt( new Double(
					    ((double[])destrel.data[bypos[i]])[currow]),i);
	      breakflag=1;
	    }
	    break;
	  case FLOAT:
	    if(((Float)(vals.elementAt(i))).floatValue() !=
	       ((float[])destrel.data[bypos[i]])[currow]){
	      vals.setElementAt( new Float(
					   ((float[])destrel.data[bypos[i]])[currow]),i);
	      breakflag=1;
	    }
	    break;
		     case UNIVERSAL:  //from Andrey
			case ATTRIBUTE: //from Andrey
	  case STRING: 
	  	if(debug_nov_24)
	  	{
	  		System.out.println("String!" + ((String[])destrel.data[bypos[i]])[currow]);
			System.out.println("vals: " + (String)vals.elementAt(i));
		}
	  case TEXT:
	  case EXPR:
	    
	    if(((String)vals.elementAt(i)).compareTo(
						     ((String[])destrel.data[bypos[i]])[currow])!=0){
	      vals.setElementAt (
				 (((String[])destrel.data[bypos[i]])[currow]),i);
	      breakflag=1;
	    }
	    if(debug_nov_24)
	    	System.out.println("breakflag = " + breakflag);
	    break;
	    
	  case IDLIST:
	    
	    break;//Sungsoo's change
	  default:
	    throw new InterpretError("actualizeEquiv(): to be implemented..");
	    
	  }//end of switch
	//  if(breakflag==1) break;//commented out by Yi Zheng Nov. 24, 2002 for bug fix
	}//end of for loop
      //YuLing's modification ends here
      
      	
   //   if(i != byarray.length){//modified by Yi Zheng Nov. 24, 2002 for bug fix
   	if(breakflag > 0) {
	// sort value changed, so must write the sub-sum value to
	// the cell.
	//for 3
	      if(debug_nov_24)
      	System.out.println("value changed");

	//sungsoo's code starts here. hongyu added Feb 2002
	if(virtype == IDLIST)
	  {
	    //add surrogated value here
	    sur = Surrogate.next();
	    
	    Relation r = (Relation) virtree.info;
	    r.load(true);  // have to load it. It seems be saved and cleared.
	    // We must write the new surrogate for the new sub-relation.
	    Domain[] darry = new Domain[r.domains.length + 1];
	    
	    //darry[0] = domtable.getDom(".id");
	    //REVISIT: Zhongxia, I used true but check it please
	    darry[0] = env.lookupDom(".id", true);
	    
	    for(int k = 0; k < r.domains.length; k++)
	      darry[k+1] = r.domains[k];
	    
	    Object[] dat = Relation.createDataArray(darry, r.numtuples);
	    for(int k = 0; k < r.domains.length; k++)
	      dat[k+1] = r.data[k];
	    for(int k = 0; k < r.numtuples; k++)
	      ((long[])dat[0])[k] = sur;
	    
	    dest.append(dat, r.numtuples);
	    dest.load();  // have to load it since append to disk.
	    
	  }
	//end of sungsoo's code
	if(debug_nov_24)
		System.out.println("oldpos = " + oldpos + " currow - 1 = " + (currow-1));
	
	for(int j = oldpos; j < currow; j++){
	  switch(virtype){
	 	  case NUMERIC:
	      ((number[])destrel.data[pos])[j] = (number)virtree.info;
	      break;

	  case INTEGER:
	    ((int[])destrel.data[pos])[j] = 
	      ((Integer)virtree.info).intValue();
	    break;
	  case SHORT:
	    ((short[])destrel.data[pos])[j] = 
	      (short)((Integer)virtree.info).intValue();
	    break;
	  case LONG:
	    ((long[])destrel.data[pos])[j] = 
	      ((Long)virtree.info).longValue();
	    break;
	  case FLOAT:
	  if(debug_nov_24)
	  	System.out.println("The value to be stored is: " + (float)((Double)virtree.info).doubleValue());
	  
	    ((float[])destrel.data[pos])[j] = 
	      (float)((Double)virtree.info).doubleValue();
	    break;
	  case DOUBLE:
	    ((double[])destrel.data[pos])[j] = 
	      ((Double)virtree.info).doubleValue();
	    break;
			case UNIVERSAL:  //from Andrey
			case ATTRIBUTE:  //from Andrey
	  case STRING:
	  case STMT:
	  case EXPR:
	  case TEXT:
	    ((String[])destrel.data[pos])[j] = (String)virtree.info;
	    break;
	  case IDLIST:
	    ((long[])destrel.data[pos])[j] =sur;
	    break;
	  default:
	    throw new InterpretError("actualizeEquiv(): to be implemented..");
	  }
	}
	
	// reset the sum field.
	if(debug_nov_24)
		System.out.println("resetting the sum field");
	switch(virtype){
		case NUMERIC:
	    virtree.info = null;
	    break;

	case INTEGER:
	case SHORT:
	  //virtree.info = new Integer(0);
	  virtree.info = null;//YuLing to fix equiv min
	  break;
	case LONG:
	  //virtree.info = new Long(0);
	  virtree.info=null;//YuLing to fix equiv min
	  break;
	case DOUBLE:
	case FLOAT:
	  //virtree.info = new Double(0.0);
	  virtree.info = null; //YuLing to fix equiv min
	  break;
		 case UNIVERSAL:  //from Andrey
	   case ATTRIBUTE: //from Andrey
	case STRING:
	case STMT:
	case EXPR:
	case TEXT:
	  virtree.info = "";
	  break;
	case IDLIST:
	  virtree.info = null;//from sungsoo. hongyu added Feb 2002
	  break;
	default:
	  throw new InterpretError("actualizeEquiv(): to be implemented..");
	}
	oldpos = currow;
      }
      
      // Main loops come here...
      switch(virtype){
         case NUMERIC:
	  actNumCell(virtree);
	  break;

      case INTEGER:
      case SHORT:
	actIntCell(virtree);   // the sum value is stored in the node.
	break;
      case LONG:
	actLongCell(virtree);
	break;
      case FLOAT:
      if(debug_nov_24)
      	System.out.println("actDoubleCell");
      case DOUBLE:
	actDoubleCell(virtree);
	break;
		 case UNIVERSAL:  //from Andrey
	   case ATTRIBUTE: //from Andrey
      case STRING:
      case STMT:
      case EXPR:
      case TEXT:
	actStrCell(virtree);
	break;
      case IDLIST:
	Relation r = actRelCell(virtree, vdom, rname, rindx, relhtbl,
				tname0, tname1);
	break;
      default:
	throw new InterpretError("actualizeEquiv(): to be implemented..");
      }
    }
    
    // Write the sub-sum value for the last group in data[] array.
    //for 4
    //sunsoo's code starts here. hongyu added. Feb 2002
//    System.out.println(" END equiv start 10  here  " );
    if(virtype == IDLIST)
      {
	//add surrogated value here
	sur = Surrogate.next();

	Relation r = (Relation) virtree.info;
	r.load(true);  // have to load it. It seems be saved and cleared.
	// We must write the new surrogate for the new sub-relation.
	Domain[] darry = new Domain[r.domains.length + 1];
	
	//darry[0] = domtable.getDom(".id");
	//REVISIT: Zhongxia, I used true but check it please
	darry[0] = env.lookupDom(".id", true);
	
	for(int k = 0; k < r.domains.length; k++)
	  darry[k+1] = r.domains[k];
	
	Object[] dat = Relation.createDataArray(darry, r.numtuples);
	for(int k = 0; k < r.domains.length; k++)
	  dat[k+1] = r.data[k];
	for(int k = 0; k < r.numtuples; k++)
	  ((long[])dat[0])[k] = sur;
	
	dest.append(dat, r.numtuples);
	dest.load();  // have to load it since append to disk.
	
	
      }
    //End of sungsoo's code. hongyu added Feb 2002
    
    if(debug_nov_24)
    	System.out.println("finally oldpos = " + oldpos);
    	
    for(int j = oldpos; j < currow; j++){
      switch(virtype){
      case NUMERIC:
	  ((number[])destrel.data[pos])[j] = (number)virtree.info;
	  break;

      case INTEGER:
	((int[])destrel.data[pos])[j] = 
	  ((Integer)virtree.info).intValue();
	break;
      case SHORT:
	((short[])destrel.data[pos])[j] = 
	  (short)((Integer)virtree.info).intValue();
	break;
      case LONG:
	((long[])destrel.data[pos])[j] = 
	  ((Long)virtree.info).longValue();
	break;
      case FLOAT:
      if(debug_nov_24)
      	System.out.println("float value: " + (float)((Double)virtree.info).doubleValue());
	((float[])destrel.data[pos])[j] = 
	  (float)((Double)virtree.info).doubleValue();
	break;
      case DOUBLE:
	((double[])destrel.data[pos])[j] = 
	  ((Double)virtree.info).doubleValue();
	break;
			 case UNIVERSAL:  //from Andrey
	   case ATTRIBUTE: //from Andrey
      case STRING:
      case STMT:
      case EXPR:
      case TEXT:
	((String[])destrel.data[pos])[j] = 
						 (String)virtree.info;
	break;
      case IDLIST:
	((long[])destrel.data[pos])[j] =sur;
	break;
      default:
	throw new InterpretError("actualizeEquiv(): to be implemented..");
      }
    } 
  }         
  /** 
   * Actualize a nested domain using new technology (nt == true), i.e.
   * actualize the domain on the top-level relation but not using the
   * tuple-by-tuple approach.
   */
  private void actualizeNestedDom(SimpleNode jnode, Hashtable htable,
				  Hashtable relhtbl) throws InterpretError
  {
    
    
    switch(jnode.type){
    case OP_JOINOPERATOR:
      actualizeNestedJoinDom(jnode, htable, relhtbl);
      break;
    case OP_PROJECT:
    case OP_TSELECT:
    case OP_SELECT:
      actualizeNestedPrjSelDom(jnode, htable, relhtbl);
      break;
    case OP_RED:
      actualizeNestedRedDom(jnode, htable, relhtbl);
      break;
    case OP_EQUIV:// should have OP_FUN, OP_PAR? hognyu Feb 2002
      actualizeNestedEquivDom(jnode, htable, relhtbl);
      break;
    default:
      throw new InterpretError("actualizeNestedDom: to be implemented.");
    }
  }      
  /** 
   * Actualize a nested domain (equivalence) using new technology
   * (nt == true).
   *
   * E.g. R( S(A,  B)   C )                    S(.id,  A,  B) 
   *           1,  2    10    (surrogates)         1,  1,  2
   *           3,  4             1    10           1,  3,  4 
   *         ---------------     --------          2,  2,  3
   *           2,  3    20       2    20           2,  4,  5
   *           4,  5             --------          2,  6,  7
   *           6,  7             3    20           3,  2,  5
   *         ---------------                       3,  3,  4
   *           2,  5    20
   *           3,  4
   * To actualize "let x be equiv ujoin of S by C; tmp<-[S, x] in R;"
   *             tmp( S,   x)              .x( .id,  A,  B)
   *                  1,   4                     4,  1,  2
   *                  2,   5                     4,  3,  4
   *                  3,   5                     5,  2,  3
   *                                             5,  2,  5
   *                                             5,  3,  4
   *                                             5,  4,  5
   *                                             5,  6,  7
   */
  private void actualizeNestedEquivDom(SimpleNode jnode,
				       Hashtable htable, Hashtable relhtbl) 
       throws InterpretError
  {
    
    // First we need to sort on the by-list.
    Vector bydoms = new Vector();
    SimpleNode exprlist = (SimpleNode)jnode.jjtGetChild(1);
    for(int i = 0; i < exprlist.jjtGetNumChildren(); i++){
      SimpleNode child = (SimpleNode) exprlist.jjtGetChild(i);
      
      //         Domain d = domtable.getDom(child.name);
      //REVISIT: Zhongxia, I used true but check it please
      Domain d = env.lookupDom(child.name, true);
      
      if(d != null)
	bydoms.addElement(d);
      else // Should never happen...
	throw new InterpretError("actualizeEquiv: domain not found.");
    }
    Domain[] byarray = new Domain[bydoms.size()];
    bydoms.copyInto(byarray);
    
    // Sort on the by-list...
    destrel.sort(byarray);
    //destrel.print(System.out);
    
    // In order to reuse the actualizeNestedRedDom() method, we must
    // generate the sub-relation which has same by-list value on the
    // fly and pass this sub-rel to the actualizeNestedRedDom() routine.
    
    // Get the position of the domains in by-list.
    int[] pos = new int[byarray.length];
    for(int i = 0; i < byarray.length; i++){
      pos[i] = alldoms.indexOf(byarray[i]);
    }
    
    // Get the initial values for sort attributes.
    int oldpos = 0;
    int[] vals = new int[byarray.length];
    
    // Scan the destrel tuple-by-tuple to find the sub-relation.
    int curow = 0;
    int skipnum = 0;
    /*      while(curow < destrel.numtuples){
	    for(int i = 0; i < byarray.length; i++){
	    switch(byarray[i].type){
	    case INTEGER:
	    vals[i] = ((int[])destrel.data[pos[i]])[curow];
	    break;
	    default:
	    throw new InterpretError("actualizeNestedEquivDom: to be impremented.");
	    }
	    }
	    
	    skipnum = equivSubrel(destrel, vals);
	    curow += skipnum;
	    }
	    */
    throw new InterpretError("out");
  } 
  /** 
   * Actualize a nested domain (join operations) using new technology 
   * (nt == true), i.e. actualize the domain on the top-level relation
   * but not using the buple-by-tuple approach.
   *
   * E.g. R(S(A,B) T(A,C))  R(S, T)     S(.id,A,B)  T(.id,A,C)
   *          2,5    2,7  (surrogates)      1,2,5       2,2,7
   *          5,3    2,9      1  2          1,5,3       2,2,9
   *      ---------------    ------         3,8,9       4,5,8
   *          8,9    5,8      3  4                      4,8,1
   *                 8,1                                4,8,3
   *                 8,3
   *      ---------------     
   * To actualize "let x be S ijoin T;"
   * 1) R1 <- R[S ijion .id]S;          R1(S, id, T, A, B)
   *                                        1, 1, 2, 2, 5
   *                                        1, 1, 2, 5, 3
   *                                        3, 3, 4, 8, 9
   * 2) let .id be T;
   *    R2 <- [.id,A,B] in [T,A,B] in R1;     R2(.id, A, B)
   *                                               2, 2, 5
   *                                               2, 5, 3
   *                                               4, 8, 9
   * 3) R3 <- R2 ijoin T;         R3(.id, A, B, C)
   *                                   2, 2, 5, 7
   *                                   2, 2, 5, 9
   *                                   4, 8, 9, 1
   *                                   4, 8, 9, 3
   *
   * 4) Generate new surrogates for '.id' in R3 and insert the
   *    tuples in new relation.
   */
  private void actualizeNestedJoinDom(SimpleNode jnode, Hashtable htable,
				      Hashtable relhtbl) 
       throws InterpretError
  {
    int lflag = 0;
    int rflag = 0;
    SimpleNode left = null;
    SimpleNode right = null;
    
    left = (SimpleNode)jnode.jjtGetChild(0);
    
    // Handle the case of 'let x be S[A,B:ijion:A,C]T;', i.e.
    // jnode has 4 children including 2 ExpressionList.
    if(jnode.jjtGetNumChildren() == 2)
      right = (SimpleNode)jnode.jjtGetChild(1);
    else  // There must be 4 children.
      right = (SimpleNode)jnode.jjtGetChild(3);
    
    if(left.type == OP_IDENTIFIER && right.type == OP_IDENTIFIER){
      // For the case:     ijoin
      //                   /   \
      //                  S     T
      //
      // Builing the tree of 'R[S:ijoin:.id]S' for procedure 1).
      
      // if(domtable.getDom(left.name) == null){
      //REVISIT: Zhongxia, I used true but check it please
      if(env.lookupDom(left.name, true) == null){
	
	// if(reltable.getRel(left.name) == null)    // Should never happen.
	//REVISIT: Zhongxia, I used true but check it please
	if(env.lookupRel(left.name, true) == null) // Should never happen.
	  
	  throw new InterpretError("join operand '" + left.name
				   + "' not found in system table.");
	else  lflag = 1;   // left is a top-level relation.
      }
      
      // if(domtable.getDom(right.name) == null){
      //REVISIT: Zhongxia, I used true but check it please
      if(env.lookupDom(right.name, true) == null){
	
	
	// if(reltable.getRel(right.name) == null)
	//REVISIT: Zhongxia, I used true but check it please
	if(env.lookupRel(right.name, true) == null)
	  
	  throw new InterpretError("join operand '" + right.name
				   + "' not found in system table.");
	else  rflag = 1;   // left is a top-level relation.
      }
      
      if(lflag == 1 && rflag == 1)
	throw new InterpretError("both join operands are top-level "+
				 "relations, to be implemented.");
      else if(lflag == 1){
	// Must exchange to make sure that left node is always nested
	// domain.
	SimpleNode tmpnode = left;
	left = right;
	right = tmpnode;
	rflag = 1;   // We always check right-flag..  '1' indicates a
	// top-level relation.
      }
      
      SimpleNode node = SimpleNode.jjtCreateNode(jnode.identifier,
						 jnode.type, jnode.opcode, "");
      
      SimpleNode lid = SimpleNode.jjtCreateNode("Identifier", 
						OP_IDENTIFIER, OP_IDENTIFIER, destrel.name);
      SimpleNode rid = SimpleNode.jjtCreateNode("Identifier",
						OP_IDENTIFIER, OP_IDENTIFIER, "."+left.name);
      
      SimpleNode lattr1 = SimpleNode.jjtCreateNode("Identifier",
						   OP_IDENTIFIER, OP_IDENTIFIER, left.name);
      SimpleNode rattr1 = SimpleNode.jjtCreateNode("Identifier",
						   OP_IDENTIFIER, OP_IDENTIFIER, ".id");
      
      SimpleNode lexpr = SimpleNode.jjtCreateNode("ExpressionList",
						  OP_EXPRLIST, OP_EXPRLIST,"");
      SimpleNode rexpr = SimpleNode.jjtCreateNode("ExpressionList",
						  OP_EXPRLIST, OP_EXPRLIST,"");
      lexpr.jjtAddChild(lattr1);
      rexpr.jjtAddChild(rattr1);
      
      // Construct the new JOIN node which will be passed to mujoin 
      // routines in Relation class.
      node.jjtAddChild(lid);
      node.jjtAddChild(lexpr);
      node.jjtAddChild(rexpr);
      node.jjtAddChild(rid);
      //node.dump("=>  ");
      
      // Must check common attributes before doing actual mu-joins.
      /*if(srcrel.domains[i].name.equals("A")){
	Domain dupdom = new Domain(srcrel.domains[i].name,
	srcrel.domains[i].type,
	srcrel.domains[i].numref,
	srcrel.domains[i].tree);
	dupdom.name = "0A";
	alldoms.addElement(dupdom);
	}*/
      //---> here!!!
      
      // Do the joins to get the resulting relation. Got the result of 2).
      String rname = interpreter.nextTempName();
      Relation r = interpreter.evaluateTLExpression(node,
						    rname, this.relhtbl, env);
      //r.print(System.out);
      
      // Have to load the relation from disk, since after calling
      // evaluateTLExpression(), the result is saved to disk and data[]
      // in relation is cleared to null. REVISIT: can we avoid this?
      r.load(true);
      
      // Generate new domain array for the resulting relation.
      Relation leftr;
      if(left.info == null){
	
	// if((leftr = reltable.getRel("."+left.name)) == null)
	//REVISIT: Zhongxia, I used true but check it please
	if((leftr = env.lookupRel("."+left.name, true)) == null)
	  
	  // Should never happen.
	  throw new InterpretError("relation '"+left.name+
				   "' can not be found.");
      }
      else{
	if((leftr = (Relation)htable.get((Domain)left.info)) == null)
	  
	  // if((leftr = reltable.getRel("."+left.name)) == null)
	  //REVISIT: Zhongxia, I used true but check it please
	  if((leftr = env.lookupRel("."+left.name, true)) == null)
	    
	    // Should never happen.
	    throw new InterpretError("relation '"+left.name+
				     "' can not be found.");
      }
      
      // The domains[] of result of procedure 2, i.e. 'R2' will be
      // having same contents of 'leftr'.
      Domain[] doms = leftr.domains;
      
      // Place holder for the new data array.
      Object[] data = Relation.createDataArray(doms, r.numtuples);
      
      if(rflag == 0) // if right relation is not top-level relation.
	// Get the 'T' surrogate values and make them values of '.id'.
	
	// data[0] = r.data[r.indexOfDomain(domtable.getDom(right.name))];
	//REVISIT: Zhongxia, I used true but check it please
	data[0] = r.data[r.indexOfDomain(env.lookupDom(right.name,true))];
      
      else  // if right relation is a top-level relatin.
	// We keep the same '.id' surrogate value.
	
	//            data[0] = r.data[r.indexOfDomain(domtable.getDom(".id"))];
	//REVISIT: Zhongxia, I used true but check it please
	data[0] = r.data[r.indexOfDomain(env.lookupDom(".id", true))];
      
      
      for(int i = 1; i < doms.length; i++)
	data[i] = r.data[r.indexOfDomain(doms[i])];
      
      // Adjust the resulting relation to new features.
      r.data = data;
      r.domains = doms;
      r.numattrs = doms.length;
      
      // Note: we must save this 'r' to disk, otherwise, next mujoin
      // will never load it since we didn't ever save it since its
      // creation.
      r.save();
      
      // Now we got the result relation of procedure 2).
      // Have to save it to record the change... This is for print()
      // only...  
      //r.print(System.out); // note: can't use print here since it will
      // clear data[] in r, hence r can't save()
      //r.load();
      
      // Building the tree of 'R ijoin T' for procedure 3). Since we have
      // a JOIN tree, we can just modify it to fit in new case.
      
      // Change left relaton to 'rname' and right to '.T'.
      ((SimpleNode)node.jjtGetChild(0)).name = rname;
      if(rflag == 0)
	((SimpleNode)node.jjtGetChild(3)).name = "."+right.name;
      else // if right relation is a top-level relation.
	((SimpleNode)node.jjtGetChild(3)).name = right.name;
      
      // Remove the expression lists for join attributes as we
      // don't have them now.
      if(jnode.jjtGetNumChildren() == 2){
	node.jjtRemoveChild(1);
	node.jjtRemoveChild(1);
      }
      else{
	// In case there're join attributes, we need to put the join
	// attributes' nodes into the tree. The followings are the
	// two expression lists.
	SimpleNode eleft = (SimpleNode)jnode.jjtGetChild(1);
	SimpleNode eright = (SimpleNode)jnode.jjtGetChild(2);
	
	// Make new expression lists which include the '.id' nodes.
	SimpleNode expr1 = SimpleNode.jjtCreateNode(eleft.identifier,
						    eleft.type, eleft.opcode, eleft.name);
	SimpleNode expr2 = SimpleNode.jjtCreateNode(eright.identifier,
						    eright.type, eright.opcode, eright.name);
	
	// For top-level relation, we don't need the '.id' stuffs..
	if(rflag == 0){
	  SimpleNode id1 = SimpleNode.jjtCreateNode("Identifier",
						    OP_IDENTIFIER, OP_IDENTIFIER, ".id");
	  SimpleNode id2 = SimpleNode.jjtCreateNode("Identifier",
						    OP_IDENTIFIER, OP_IDENTIFIER, ".id");
	  expr1.jjtAddChild(id1);
	  expr2.jjtAddChild(id2);
	}
	
	for(int i = 0; i < eleft.jjtGetNumChildren(); i++){
	  expr1.jjtAddChild((SimpleNode)eleft.jjtGetChild(i));
	  expr2.jjtAddChild((SimpleNode)eright.jjtGetChild(i));
	}
	node.jjtReplaceChild(1, expr1);
	node.jjtReplaceChild(2, expr2);
      }
      //node.dump("****** ");
      
      // Do procedure 3), i.e. r ijoin T;
      relhtbl.put(rname, r);
      String rname1 = interpreter.nextTempName();
      r = interpreter.evaluateTLExpression(node,rname1,this.relhtbl, env);
      
      // Have to load since after evaluateTLExpression(), data[] is save
      // to disk and then is cleared.
      r.load(true);
      
      // Have to load again after printing..
      //r.print(System.out);
      //r.load();
      
      // Fill the domain's position.
      Domain d = (Domain)jnode.info;
      
      // 'posd' is the pos_index of "x" in "let x be S ijoin T;".
      // 'posr' is the pos_index of "T" in "R3 <- R2 ijoin T;").
      int posd = alldoms.indexOf((Domain)jnode.info);
      int posr;
      if(rflag == 0)
	
	// posr = alldoms.indexOf(domtable.getDom(right.name));
	//REVISIT: Zhongxia, I used true but check it please
	posr = alldoms.indexOf(env.lookupDom(right.name, true));
      
      else
	
	// posr = alldoms.indexOf(domtable.getDom(left.name));
	//REVISIT: Zhongxia, I used true but check it please
	posr = alldoms.indexOf(env.lookupDom(left.name, true));
      
      
      if(posd == ERROR || posr == ERROR)
	// should never happen..
	throw new InterpretError("system error: domain not found..");
      
      long sur, refsur;
      for(currow = 0; currow < destrel.numtuples; currow++){
	sur = Surrogate.next();
	// Change the .id of "x" in destrel..
	((long[])destrel.data[posd])[currow] = sur;
	refsur = ((long[])destrel.data[posr])[currow];
	fillSurrogate(r, refsur, sur);
      }
      
      // The following 3 lines are for print().
      //r.save();
      //r.print(System.out);
      //r.load();
      
      // Here jnode.info will store '.x' in case 'let x be S ijoin T;'.
      
      //    Relation dest = reltable.getRel("."+((Domain)jnode.info).name);
      //REVISIT: Zhongxia, I used true but check it please
      Relation dest = env.lookupRel("."+((Domain)jnode.info).name, true);
      
      if(dest != null){
	if(rflag != 0){
	  // In case that the right relation is a top-level rel,
	  // the result relation of procedure 3 will have its domain
	  // array as follows:  "A, .id, B, C", since '.id' is not a
	  // common join attributes. This will incur a problem for
	  // appending a relation if we do nothing for the sequence.
	  Object[] dat = Relation.createDataArray(dest.domains, r.numtuples);
	  for(int i = 0; i < dest.domains.length; i++){
	    int index = r.indexOfDomain(dest.domains[i]);
	    dat[i] = r.data[index];
	  }
	  r.data = dat;
	  r.domains = dest.domains;
	  //r.save();
	}
	dest.append(r.data, r.numtuples);
	
	// Must use load() here, since the append() routine doesn't
	// append to data[] in memory, but write to disk file. ;-(
	dest.load();
	
	//dest.print(System.out);
	//dest.load();
      }
      else{  // Should never happen.
	throw new InterpretError(
				 "system error: intermediate relation not exist.");
      }
      
      // Have to save changes to 'destrel.data' to disk file in order
      // for next mu-join to use.
      destrel.save();
      
      interpreter.deleteFile(rname);
      interpreter.deleteFile(rname1);
      
    }
    else if(right.type == OP_IDENTIFIER){
      // For the case 'let x be S ijoin T ijoin U;'
      //                   ijoin
      //                   /   \
      //                ijoin   U
      //                /   \
      //               S     T
      
      // Fill the result of left-side-ijoin to rel '.0'.
      actualizeNestedJoinDom(left, htable, relhtbl);
      
      // Prepare for the next top-level ijoin.
      Relation r = (Relation)htable.get((Domain)jnode.info);
      relhtbl.put((((Domain)jnode.info).name).substring(1), r);
      
      // Replace the first 'ijoin' with the intermediate domain e.g. '0'
      // i.e.     ijoin
      //          /   \               ijoin("0")           ijoin
      //       ijoin   U       =>      /   \             /   \
	//       /   \                  S     T          "0"    U
      //      S     T
      String newdomname = ((Domain)left.info).name;
      SimpleNode newnode = SimpleNode.jjtCreateNode("Identifier", 
						    OP_IDENTIFIER, OP_IDENTIFIER, newdomname);
      jnode.jjtReplaceChild(0, newnode);
      
      // Now we can actualize this domain. REVISIT: an InterpretError
      // exception might be thrown here, we need to catch it and 
      // remove the intermediate domain in that case.
      actualizeNestedJoinDom(jnode, htable, relhtbl);
      
    }
    else if(left.type == OP_IDENTIFIER){
      // For the case 'let x be S ijoin T ijoin U;', if any.
      //                   ijoin
      //                   /   \
      //                  S    ijoin
      //                       /   \
      //                      T     U
      
      // Fill the result of left-side-ijoin to rel '.0'.
      actualizeNestedJoinDom(right, htable, relhtbl);
      
      // Prepare for the next top-level ijoin.
      Relation r = (Relation)htable.get((Domain)jnode.info);
      relhtbl.put((((Domain)jnode.info).name).substring(1), r);
      
      // Replace the first 'ijoin' with the intermediate domain e.g. '0'
      // i.e.     ijoin
      //          /   \               ijoin("0")           ijoin
      //         S    ijoin       =>      /   \             /   \
      //              /  \               T     U           S    "0"
      //             T    U
      String newdomname = ((Domain)right.info).name;
      SimpleNode newnode = SimpleNode.jjtCreateNode("Identifier", 
						    OP_IDENTIFIER, OP_IDENTIFIER, newdomname);
      if(jnode.jjtGetNumChildren() == 2)
	jnode.jjtReplaceChild(1, newnode);
      else
	jnode.jjtReplaceChild(3, newnode);
      
      // Now we can actualize this domain. REVISIT: an InterpretError
      // exception might be thrown here, we need to catch it and 
      // remove the intermediate domain in that case.
      actualizeNestedJoinDom(jnode, htable, relhtbl);
      
    }
    else{
      // For the case:     ijoin
      //                   /   \
      //               ijoin   ijoin
      //               /   \   /   \
      //              ..   .. ..   ..
      
      // Fill the result of left/right-side-ijoin to rel '.0/.1'.
      actualizeNestedJoinDom(left, htable, relhtbl);
      actualizeNestedJoinDom(right, htable, relhtbl);
      
      // Prepare for the next top-level ijoin.
      Relation r = (Relation)htable.get((Domain)jnode.info);
      relhtbl.put((((Domain)jnode.info).name).substring(1), r);
      
      // Replace the sub-'ijoin's with the intermediate domain e.g. '0/1'
      String newdomname = ((Domain)left.info).name;
      SimpleNode newnode = SimpleNode.jjtCreateNode("Identifier", 
						    OP_IDENTIFIER, OP_IDENTIFIER, newdomname);
      jnode.jjtReplaceChild(0, newnode);
      
      newdomname = ((Domain)right.info).name;
      newnode = SimpleNode.jjtCreateNode("Identifier", 
					 OP_IDENTIFIER, OP_IDENTIFIER, newdomname);
      if(jnode.jjtGetNumChildren() == 2)
	jnode.jjtReplaceChild(1, newnode);
      else
	jnode.jjtReplaceChild(3, newnode);
      
      // Now we can actualize this domain.
      actualizeNestedJoinDom(jnode, htable, relhtbl);
    }
  }    
  /** 
   * Actualize a nested domain (project operation) using new technology
   * (nt == true).
   *
   * E.g. R(S(A,B) T(A,C))  R(S, T)     S(.id,A,B) 
   *          2,5    2,7  (surrogates)      1,2,5
   *          5,3    2,9      1  2          1,5,3 
   *      ---------------    ------         3,8,9
   *          8,9    5,8      3  4    
   *                 8,1            
   *                 8,3
   *      ---------------     
   * To actualize "let x be [A] in S;"
   * 1) R1 <- R[S ijion .id]S;          R1(S, id, T, A, B)
   *                                        1, 1, 2, 2, 5
   *                                        1, 1, 2, 5, 3
   *                                        3, 3, 4, 8, 9
   * 2) R2 <- [.id,A] in R1;      R2(.id, A)
   *                                   1, 2
   *                                   1, 5
   *                                   3, 8
   *
   * 3) Generate new surrogates for '.id' in R2 and insert the
   *    tuples in new relation.
   */
  private void actualizeNestedPrjSelDom(SimpleNode jnode, Hashtable htable,
					Hashtable relhtbl) 
       throws InterpretError
  {
    
    // Get the source relation node.
    SimpleNode srcnode = (SimpleNode)jnode.jjtGetChild(1);
    
    // This node must be an IDENTIFIER typed node, i.e. we got a true
    // source relation.
    if(srcnode.type != OP_IDENTIFIER)
      // Should never happen.
      throw new InterpretError("actualizeNestedPrjSelDom: " +
			       "source relation is not a true relation.");
    
    //      Relation srcr = reltable.getRel("."+srcnode.name);
    //REVISIT: Zhongxia, I used true but check it please
    Relation srcr = env.lookupRel("."+srcnode.name, true);
    
    if(srcr == null)
      // Should never happen.
      throw new InterpretError("source relation '"+srcnode.name+
			       "' is not found.");
    
    Domain[] doms = null;
    
    // If projection or tselection (they are quite similar).
    if(jnode.type == OP_PROJECT || jnode.type == OP_TSELECT){
      // Get the projection domain list.
      SimpleNode exprlist = (SimpleNode)jnode.jjtGetChild(0).jjtGetChild(0);
      Vector domvec = new Vector();
      
      //         domvec.addElement(domtable.getDom(".id"));
      //REVISIT: Zhongxia, I used true but check it please
      domvec.addElement(env.lookupDom(".id", true));
      
      for(int i = 0; i < exprlist.jjtGetNumChildren(); i++){
	String dname = ((SimpleNode)exprlist.jjtGetChild(i)).name;
	
	//       Domain d = domtable.getDom(dname);
	//REVISIT: Zhongxia, I used true but check it please
	Domain d = env.lookupDom(dname, true);
	
	if(d == null)
	  throw new InterpretError("domain "+dname+" not found.");
	domvec.addElement(d);
      }
      doms = new Domain[domvec.size()];
      domvec.copyInto(doms);
    }else{    // if selection..
      // The resulting domain list must be the source relation's
      // domain list.
      doms = srcr.domains;
    }
    
    // Builing the tree of 'R[S:ijoin:.id]S' for procedure 1).
    SimpleNode node = SimpleNode.jjtCreateNode("Join",
					       OP_JOINOPERATOR, OP_IJOIN, "");
    
    SimpleNode lid = SimpleNode.jjtCreateNode("Identifier", 
					      OP_IDENTIFIER, OP_IDENTIFIER, destrel.name);
    SimpleNode rid = SimpleNode.jjtCreateNode("Identifier",
					      OP_IDENTIFIER, OP_IDENTIFIER, "."+srcnode.name);
    
    SimpleNode lattr1 = SimpleNode.jjtCreateNode("Identifier",
						 OP_IDENTIFIER, OP_IDENTIFIER, srcnode.name);
    SimpleNode rattr1 = SimpleNode.jjtCreateNode("Identifier",
						 OP_IDENTIFIER, OP_IDENTIFIER, ".id");
    
    SimpleNode lexpr = SimpleNode.jjtCreateNode("ExpressionList",
						OP_EXPRLIST, OP_EXPRLIST,"");
    SimpleNode rexpr = SimpleNode.jjtCreateNode("ExpressionList",
						OP_EXPRLIST, OP_EXPRLIST,"");
    lexpr.jjtAddChild(lattr1);
    rexpr.jjtAddChild(rattr1);
    
    // Construct the new JOIN node which will be passed to mujoin 
    // routines in Relation class.                   
    node.jjtAddChild(lid);
    node.jjtAddChild(lexpr);
    node.jjtAddChild(rexpr);
    node.jjtAddChild(rid);
    
    // Do the joins to get the resulting relation. Got the result of 2).
    String rname = interpreter.nextTempName();
    Relation r = interpreter.evaluateTLExpression(node, 
						  rname, this.relhtbl, env);
    
    // Have to load the relation from disk, since after calling
    // evaluateTLExpression(), the result is saved to disk and data[]
    // in relation is cleared to null. REVISIT: can we avoid this?
    //r.print(System.out);
    r.load(true);
    
    // Have to do the operations correspondingly, i.e. project or
    // select.  But for selection, we also do a projection first.
    SimpleNode opnode = null;
    if(jnode.type == OP_PROJECT || jnode.type == OP_TSELECT){
      opnode = SimpleNode.jjtCreateNode("Project", OP_PROJECT,
					OP_WHENIN, "");
      SimpleNode projector = SimpleNode.jjtCreateNode("Projector",
						      OP_PROJECTOR, OP_PROJECTOR, "");
      SimpleNode exprnode = SimpleNode.jjtCreateNode("ExpressionList",
						     OP_EXPRLIST, OP_EXPRLIST, "");
      for(int i = 0; i < doms.length; i++){
	exprnode.jjtAddChild(SimpleNode.jjtCreateNode("Identifier",
						      OP_IDENTIFIER, OP_IDENTIFIER, doms[i].name));
      }
      projector.jjtAddChild(exprnode);
      projector.setBit(1);   // indicate that it contains EXPRLIST. This
      // is used by Biao's projection routine.
      opnode.jjtAddChild(projector);
      opnode.jjtAddChild(SimpleNode.jjtCreateNode("Identifier",
						  OP_IDENTIFIER, OP_IDENTIFIER, rname));
    }else{   // if selection..
      opnode = jnode.jjtDuplicate();
      opnode.jjtRemoveChild(1);
      opnode.jjtAddChild(SimpleNode.jjtCreateNode("Identifier",
						  OP_IDENTIFIER, OP_IDENTIFIER, rname));
    }
    
    // Do procedure 2), i.e. the projection or selection.
    relhtbl.put(rname, r);
    String rname1 = interpreter.nextTempName();
    r = interpreter.evaluateTLExpression(opnode, rname1, this.relhtbl, env);
    
    // Have to load since after evaluateTLExpression(), data[] is save
    // to disk and then is cleared.
    //r.print(System.out);
    r.load(true);
    
    // 'posd' is the pos_index of "x" in "let x be S ijoin T;".
    // 'posr' is the pos_index of "T" in "R3 <- R2 ijoin T;").
    int posd = alldoms.indexOf((Domain)jnode.info);
    
    // int posr = alldoms.indexOf(domtable.getDom(srcnode.name));
    //REVISIT: Zhongxia, I used true but check it please
    int posr = alldoms.indexOf(env.lookupDom(srcnode.name, true));
    
    
    if(posd == ERROR || posr == ERROR)
      // should never happen..
      throw new InterpretError("system error: domain not found..");
    
    long sur, refsur;
    for(currow = 0; currow < destrel.numtuples; currow++){
      sur = Surrogate.next();
      // Change the .id of "x" in destrel..
      ((long[])destrel.data[posd])[currow] = sur;
      refsur = ((long[])destrel.data[posr])[currow];
      fillSurrogate(r, refsur, sur);
    }
    
    // The following 3 lines are for print().
    //r.save();
    //r.print(System.out);
    //r.load();
    
    // If op_type is SELECTION, we must project the results to the
    // domain list of 'S' in case "let x be where A=2 in S;"
    String rname2 = null;
    if(jnode.type == OP_SELECT){
      opnode = SimpleNode.jjtCreateNode("Project", OP_PROJECT,
					OP_WHENIN, "");
      SimpleNode projector = SimpleNode.jjtCreateNode("Projector",
						      OP_PROJECTOR, OP_PROJECTOR, "");
      SimpleNode exprnode = SimpleNode.jjtCreateNode("ExpressionList",
						     OP_EXPRLIST, OP_EXPRLIST, "");
      for(int i = 0; i < doms.length; i++){
	exprnode.jjtAddChild(SimpleNode.jjtCreateNode("Identifier",
						      OP_IDENTIFIER, OP_IDENTIFIER, doms[i].name));
      }
      projector.jjtAddChild(exprnode);
      projector.setBit(1);   // indicate that it contains EXPRLIST. This
      // is used by Biao's projection routine.
      opnode.jjtAddChild(projector);
      opnode.jjtAddChild(SimpleNode.jjtCreateNode("Identifier",
						  OP_IDENTIFIER, OP_IDENTIFIER, rname1));
      
      // Must save 'r' since the surrogates are changed.
      r.save();
      relhtbl.put(rname1, r);
      rname2 = interpreter.nextTempName();
      r = interpreter.evaluateTLExpression(opnode, rname2, this.relhtbl, env);
      
      // Have to load since after evaluateTLExpression(), data[] is save
      // to disk and then is cleared.
      //r.print(System.out);
      r.load(true);
    }
    
    // Here jnode.info will store '.x' in case 'let x be S ijoin T;'.
    
    //    Relation dest = reltable.getRel("."+((Domain)jnode.info).name);
    //REVISIT: Zhongxia, I used true but check it please
    Relation dest = env.lookupRel("."+((Domain)jnode.info).name, true);
    
    if(dest != null){
      dest.append(r.data, r.numtuples);
      
      // Must use load() here, since the append() routine doesn't
      // append to data[] in memory, but write to disk file. ;-(
      
      //dest.print(System.out);
      dest.load();
    }else{
      // Should never happen.
      throw new InterpretError(
			       "system error: intermediate relation not exist.");
    }
    
    // Have to save changes to 'destrel.data' to disk file in order
    // for next mu-join to use.
    destrel.save();
    
    interpreter.deleteFile(rname);
    interpreter.deleteFile(rname1);
    if(jnode.type == OP_SELECT)
      interpreter.deleteFile(rname2);
  }    
  /** 
   * Actualize a nested domain (reduction) using new technology (nt==true).
   *
   * E.g. R(S(A,B) T(A,C))  R(S, T)     S(.id,A,B) 
   *          2,5    2,7  (surrogates)      1,2,5
   *          5,3    2,9      1  2          1,5,3 
   *      ---------------    ------         3,8,9
   *          8,9    5,8      3  4    
   *                 8,1            
   *                 8,3
   *      ---------------     
   * To actualize "let x be red ujoin of S;"
   * 1) R1 <- R[S ijion .id]S;          R1(S, id, T, A, B)
   *                                        1, 1, 2, 2, 5
   *                                        1, 1, 2, 5, 3
   *                                        3, 3, 4, 8, 9
   * 2) R2 <- [A, B] in R1;         R2(A, B)
   *                                   2, 5
   *                                   5, 3
   *                                   8, 9
   *
   * 3) Generate new surrogates for '.id' in R2 and insert the
   *    tuples in new relation, i.e. the result of:
   *         tmp<-[S, x] in R;  -->
   *             tmp( S,   x)              .x( .id,  A,  B)
   *                  1    5                     5   2   5
   *                  3    5                     5   5   3
   *                                             5   8   9
   */
  private void actualizeNestedRedDom(SimpleNode jnode, Hashtable htable,
				     Hashtable relhtbl) 
       throws InterpretError
  {
    
    // Get the source relation node.
    SimpleNode srcnode = (SimpleNode)jnode.jjtGetChild(0);
    
    // This node must be an IDENTIFIER typed node, i.e. we got a true
    // source relation.
    if(srcnode.type != OP_IDENTIFIER)
      // Should never happen.
      throw new InterpretError("actualizeNestedPrjSelDom: " +
			       "source relation is not a true relation.");
    
    //      Relation srcr = reltable.getRel("."+srcnode.name);
    //REVISIT: Zhongxia, I used true but check it please
    Relation srcr = env.lookupRel("."+srcnode.name, true);
    
    if(srcr == null)   // Should never happen...
      throw new InterpretError("source relation '"+srcnode.name+
			       "' is not found.");
    
    // Builing the tree of 'R[S:ijoin:.id]S' for procedure 1).
    SimpleNode node = SimpleNode.jjtCreateNode("Join",
					       OP_JOINOPERATOR, OP_IJOIN, "");
    
    SimpleNode lid = SimpleNode.jjtCreateNode("Identifier", 
					      OP_IDENTIFIER, OP_IDENTIFIER, destrel.name);
    SimpleNode rid = SimpleNode.jjtCreateNode("Identifier",
					      OP_IDENTIFIER, OP_IDENTIFIER, "."+srcnode.name);
    
    SimpleNode lattr1 = SimpleNode.jjtCreateNode("Identifier",
						 OP_IDENTIFIER, OP_IDENTIFIER, srcnode.name);
    SimpleNode rattr1 = SimpleNode.jjtCreateNode("Identifier",
						 OP_IDENTIFIER, OP_IDENTIFIER, ".id");
    
    SimpleNode lexpr = SimpleNode.jjtCreateNode("ExpressionList",
						OP_EXPRLIST, OP_EXPRLIST,"");
    SimpleNode rexpr = SimpleNode.jjtCreateNode("ExpressionList",
						OP_EXPRLIST, OP_EXPRLIST,"");
    lexpr.jjtAddChild(lattr1);
    rexpr.jjtAddChild(rattr1);
    
    // Construct the new JOIN node which will be passed to mujoin 
    // routines in Relation class.                   
    node.jjtAddChild(lid);
    node.jjtAddChild(lexpr);
    node.jjtAddChild(rexpr);
    node.jjtAddChild(rid);
    
    // Do the joins to get the resulting relation. Got the result of 2).
    String rname = interpreter.nextTempName();
    Relation r = interpreter.evaluateTLExpression(node, 
						  rname, this.relhtbl, env);
    
    //r.print(System.out);
    
    // Have to load the relation from disk, since after calling
    // evaluateTLExpression(), the result is saved to disk and data[]
    // in relation is cleared to null. REVISIT: can we avoid this?
    r.load(true);
    
    String rname1 = null;
    String rname2 = null;
    String rname3 = null;
    SimpleNode opnode = null;
    if(jnode.type == OP_RED && jnode.opcode == OP_UJOIN){
      // If it's an UJOIN, we must project out the '.id' first.
      // And the result will be what we want.
      opnode = SimpleNode.jjtCreateNode("Project", OP_PROJECT,
					OP_WHENIN, "");
      SimpleNode projector = SimpleNode.jjtCreateNode("Projector",
						      OP_PROJECTOR, OP_PROJECTOR, "");
      SimpleNode exprnode = SimpleNode.jjtCreateNode("ExpressionList",
						     OP_EXPRLIST, OP_EXPRLIST, "");
      
      for(int i = 0; i < srcr.domains.length; i++){
	if(!srcr.domains[i].name.equals(".id"))
	  exprnode.jjtAddChild(SimpleNode.jjtCreateNode("Identifier",
							OP_IDENTIFIER, OP_IDENTIFIER, srcr.domains[i].name));
      }
      projector.jjtAddChild(exprnode);
      projector.setBit(1);   // indicate that it contains EXPRLIST. This
      // is used by Biao's projection routine.
      opnode.jjtAddChild(projector);
      opnode.jjtAddChild(SimpleNode.jjtCreateNode("Identifier",
						  OP_IDENTIFIER, OP_IDENTIFIER, rname));
      
      relhtbl.put(rname, r);
      rname1 = interpreter.nextTempName();
      r = interpreter.evaluateTLExpression(opnode, rname1, this.relhtbl, env);
      
    }else if(jnode.type == OP_RED && jnode.opcode == OP_IJOIN){
      // If it's an IJOIN, we need to find the sub-relation with
      // the smallest numtuples, and then ijoin it with all the sub-
      // relations until finish or the smallest sub-rel becomes null.
      Relation mset = getMiniSet(r, srcr.domains);
      rname1 = interpreter.nextTempName();
      mset.name = rname1;
      
      // Note: getMiniSet() receives two arguments, i.e. 'r': the big
      // source relation, 'doms': the destination domain array. It will
      // return a relation which has the same tuple# as 'r', but arranged
      // in the domain array 'doms'. It also finds out the '.id' value
      // of the smallest sub-relation set (with same '.id') in 'r', and
      // put the value in the returned relation's 'numsortattrs' field;
      // the #tuples of the smallest sub-relation set is put in the
      // returned relation's 'capacity' field.
      for(int i = mset.numsortattrs;
	  i < mset.numsortattrs+mset.capacity; i++)
	((byte[])mset.data[mset.domains.length-1])[i] = 1;
      
      // Get the common tuples, i.e. the result of ijoin.
      compareTwoRows(mset);
      mset.save();
      //mset.print(System.out);
      //mset.load();
      
      // Now, the common tuples are marked by 'flag==true', so we
      // need to project them out. First we need to *select* out all
      // those tuples whose flags are 'true'.
      opnode = SimpleNode.jjtCreateNode("Selector", OP_SELECT,
					OP_WHEREIN, null);
      opnode.jjtAddChild(SimpleNode.jjtCreateNode("Identifier",
						  OP_IDENTIFIER, OP_IDENTIFIER, ".bool"));
      opnode.jjtAddChild(SimpleNode.jjtCreateNode("Identifier",
						  OP_IDENTIFIER, OP_IDENTIFIER, rname1));
      relhtbl.put(rname1, mset);
      rname2 = interpreter.nextTempName();
      r = interpreter.evaluateTLExpression(opnode, rname2, this.relhtbl, env);
      //r.print(System.out);
      //r.load();
      
      // Second, we need to *project* out those domains that we want.
      opnode = SimpleNode.jjtCreateNode("Project", OP_PROJECT,
					OP_WHENIN, "");
      SimpleNode projector = SimpleNode.jjtCreateNode("Projector",
						      OP_PROJECTOR, OP_PROJECTOR, "");
      SimpleNode exprnode = SimpleNode.jjtCreateNode("ExpressionList",
						     OP_EXPRLIST, OP_EXPRLIST, "");
      for(int i = 0; i < mset.domains.length - 2; i++)
	exprnode.jjtAddChild(SimpleNode.jjtCreateNode("Identifier",
						      OP_IDENTIFIER, OP_IDENTIFIER, mset.domains[i].name));
      projector.jjtAddChild(exprnode);
      projector.setBit(1);   // indicate that it contains EXPRLIST. This
      // is used by Biao's projection routine.
      opnode.jjtAddChild(projector);
      opnode.jjtAddChild(SimpleNode.jjtCreateNode("Identifier",
						  OP_IDENTIFIER, OP_IDENTIFIER, rname2));
      
      relhtbl.put(rname2, r);
      rname3 = interpreter.nextTempName();
      r = interpreter.evaluateTLExpression(opnode, rname3, this.relhtbl, env);
    }else // Should never happen.
      throw new InterpretError("actNestedRed: to be implemented.");
    
    // Now we got the result relation of procedure 2).
    // Have to save it to record the change... This is for print()
    // only...  
    //r.print(System.out); // note: can't use print here since it will
    // clear data[] in r, hence r can't save()
    
    // Have to load since after evaluateTLExpression(), data[] is save
    // to disk and then is cleared.
    r.load(true);
    
    // 'posd' is the pos_index of "x" in "let x be red ujoin of S;".
    int posd = alldoms.indexOf((Domain)jnode.info);
    
    if(posd == ERROR)
      // should never happen..
      throw new InterpretError("system error: domain not found..");
    
    long sur = Surrogate.next();
    for(currow = 0; currow < destrel.numtuples; currow++){
      // Change the .id of "x" in destrel..
      ((long[])destrel.data[posd])[currow] = sur;
    }
    
    // And now change '.id' value of .x.  First we need to add a '.id'
    // field in relation 'r' since we threw away the 'id' when we did
    // the projection before.
    Domain[] doms = new Domain[r.domains.length + 1];
    
    //      doms[0] = domtable.getDom(".id");
    //REVISIT: Zhongxia, I used true but check it please
    doms[0] = env.lookupDom(".id", true);
    
    for(int i = 0; i < r.domains.length; i++)
      doms[i + 1] = r.domains[i];
    
    Object[] data = Relation.createDataArray(doms, r.numtuples);
    // Fill surrogates..
    for(int i = 0; i < r.numtuples; i++) ((long[])data[0])[i] = sur;
    
    // Fill actual data.
    for(int i = 0; i < r.domains.length; i++)
      data[i + 1] = r.data[i];
    
    // Change to recognize the new features.
    r.data = data;
    r.domains = doms;
    r.numattrs = doms.length;
    
    // Must save this 'r' in order for next operation to work.
    r.save();
    
    // Here jnode.info will store '.x' in case 'let x be S ijoin T;'.
    
    //      Relation dest = reltable.getRel("."+((Domain)jnode.info).name);
    //REVISIT: Zhongxia, I used true but check it please
    Relation dest = env.lookupRel("."+((Domain)jnode.info).name, true);
    
    if(dest != null){
      dest.append(r.data, r.numtuples);
      
      // Must use load() here, since the append() routine doesn't
      // append to data[] in memory, but write to disk file. ;-(
      dest.load();
      
      //dest.print(System.out);
      //dest.load();
    }else // Should never happen.
      throw new InterpretError(
			       "system error: intermediate relation not exist.");
    
    // Have to save changes to 'destrel.data' to disk file in order
    // for next mu-join to use.
    destrel.save();
    
    interpreter.deleteFile(rname);
    interpreter.deleteFile(rname1);
    if(jnode.opcode == OP_IJOIN){
      interpreter.deleteFile(rname2);
      interpreter.deleteFile(rname3);
    }
  }      
  /**
   * Actualize the virtual domains (including RED virtual domain).
        *  
        *  JRelixNS modified/enhanced this method:
        *   1) Put the relhtbl in object scope this.relhtbl, if it is null;
        *   as nullary-Relation/IDLIST and boolean are now compatible,
        *   we need to pass the relhtbl for scalars as well.
        *   2) We must set the NRenv.row for all types as scalar expressions
        *   may now contain IDLISTs.
	*   3) For empty nested relation, use Surrogate 0.
	        *
        * JRelCaNu modified/enhanced this method:
        * - Optionally handle Nulls and Casts with actualizingWithNull();

   *///Andy May 12, 2002
  private void actualizing(Domain[] doms, int type) //if type ==1, go for another pass
       throws InterpretError
  {
              //START Since JRelCaNu d.
          //if (actWithNull)  {actualizingWithNull(doms, type); return;};
          //END Since JRelCaNu d.

    // Intermediate arrays are used for speeding up the tuple-level
    //  actualization.
    //DEBUGGING
    //System.out.println(" enter into actualizing");
    //for(int i=0;i<doms.length;i++)
     // System.out.println(" doms["+i+ "] " +doms[i].name+ " type " + Utility.type2String(doms[i].type));
    
    int[] virtype = new int[doms.length];
    SimpleNode[] virtree = new SimpleNode[doms.length];
    int pos[] = new int[doms.length];
    
    // Varialbes for IDLIST domains and when nt == false.
    String[] tname0 = null, tname1 = null;
    int idcount = 0; //counter for IDLIST typed domains
    Hashtable relhtbl = null;
    Vector[] rname = null;
    Vector[] rindx = null;
    Vector[] vdom = null;
    Relation[] dest = null;
    
    // Following loop is only for IDLIST domains and when nt == false.
//    System.out.println("nt:" + nt + " IDLIST " + IDLIST);
    if(!nt){
  //    System.out.println("doms.length:" + doms.length);
      for(int i = 0; i < doms.length; i++) {
//	System.out.println("doms[" + i + "].type:" + doms[i].type);
	if(doms[i].type == IDLIST) idcount++;
      }
      
      relhtbl = new Hashtable();
      // START Since jRelixNS
      // Put the relhtbl in this.relhtbl, if it is null.
//      if (env.isSigmaJoinImplemented == true){ //Andy May 12, 2002
	    if (Relation.isSigmaJoinImplemented == true){
//	System.out.println("env.isSigmaJoinImplemented == true");
	if (this.relhtbl == null) 
	  this.relhtbl = relhtbl;
      }
      // END Since jRelixNS
      rname = new Vector[idcount];
      rindx = new Vector[idcount];
      vdom = new Vector[idcount];
      dest = new Relation[idcount];
      tname0 = new String[idcount];
      tname1 = new String[idcount];
  //    System.out.println(" idcount = "+idcount);
      idcount = 0;    // reset idcount.
    }
    
    for(int i = 0; i < doms.length; i++){
      // Fill the virtual domain's tree array.
      virtree[i] = (SimpleNode)vvtree.get(doms[i]);
      
      if (debug_oct_52)
      {
      	System.out.println("virtree[" + i + "] for dom " + doms[i].name + " is:");
      	Utility.dump(virtree[i], "virtree: ");
    	}
      // Fill the types array of virtual domains. 
      if((virtype[i] = doms[i].type) == IDLIST && !nt){
//	System.out.println("((virtype[i] = doms[i].type) == IDLIST && !nt)");//added by hongyu
	vdom[idcount] = new Vector();
	rname[idcount] = new Vector();
	rindx[idcount] = new Vector();
	tname0[idcount] = interpreter.nextTempName();
	tname1[idcount] = interpreter.nextTempName();
	
	// Add this name to tempnames vector for future clean-up.
	tempnames.addElement(tname0[idcount]);
	tempnames.addElement(tname1[idcount]);
	
	
	preProcIDListDom(virtree[i], vdom[idcount], rname[idcount],
			 rindx[idcount]);
	
	// dest[idcount] = reltable.getRel("."+doms[i].name);
	//REVISIT: Zhongxia, I used true but check it please
	dest[idcount] = env.lookupRel("."+doms[i].name, true);
	if (dest[idcount] == null)
		throw new InterpretError("dest[" + idcount + "] is null, relation ." + doms[i].name + " not found");
	idcount++;
      }
//      System.out.println(" AFTER ((virtype[i] = doms[i].type) == IDLIST && !nt)");//added 
      
 /*     // Fill the types array of virtual domains. 
      if(((virtype[i] = doms[i].type) == BOOLEAN && !nt)
	 && (env.isSigmaJoinImplemented != true )){
//	System.out.println(" here");
	vdom[idcount] = new Vector();
	rname[idcount] = new Vector();
	rindx[idcount] = new Vector();
	tname0[idcount] = interpreter.nextTempName();
	tname1[idcount] = interpreter.nextTempName();
	
	// Add this name to tempnames vector for future clean-up.
	tempnames.addElement(tname0[idcount]);
	tempnames.addElement(tname1[idcount]);
	
	
//	preProcBoolean(virtree[i], vdom[idcount], rname[idcount],
//				 rindx[idcount]); 
	preProcIDListDom(virtree[i], vdom[idcount], rname[idcount],
				 rindx[idcount]);
	
	// dest[idcount] = reltable.getRel("."+doms[i].name);
	//REVISIT: Zhongxia, I used true but check it please
	dest[idcount] = env.lookupRel("."+doms[i].name, true);
	
	idcount++;
      }
  */    //Andy May 12, 2002
      // Fill the virtual domain's position array.
      pos[i] = alldoms.indexOf(doms[i]);
      //System.out.println("pos[" + i + "]:" + pos[i]);
    }
    
    // NREnvironment nrEnv = new NREnvironment(destrel, null, env);
    // relhtbl.put("#env2", nrEnv);
    NREnvironment NRenv = env.addNRFrame(destrel);
   if (Relation.isSigmaJoinImplemented == true) //Andy May 12, 2002
    this.NRenv = NRenv;
    //Now actualize the virtual domain tuple by tuple
    //System.out.println(" Now actualize the virtual domain tuple by tuple");
    for(currow = 0; currow < destrel.numtuples; currow++){
     // System.out.println("currow:" + currow);
      idcount = 0;
      for(int j = 0; j < doms.length; j++){
	//System.out.println("virtype[j]:" + Utility.type2String(virtype[j]));
	// START Since jRelixNS
	// Since it is NOT only the IDLIST domain tree may contain an IDLIST
	// domain with nullary-Relation/IDLIST - boolean compatiblity,
	// (e.g. IF-THEN-ELSE statements that may returning any type,
	// such as numeric or string, may still have an IDLIST
	// in the IF clause.) we must set the NRenv.row for all types. 
     if (Relation.isSigmaJoinImplemented == true)
//	if (env.isSigmaJoinImplemented == true) Andy, May 12, 2002
	  NRenv.row = currow;
	// END Since jRelixNS
	switch(virtype[j]){
	    case NUMERIC:
		((number[])destrel.data[pos[j]])[currow] = actNumCell(virtree[j]);
		break;

	case INTEGER:
	  ((int[])destrel.data[pos[j]])[currow]
	    = actIntCell(virtree[j]);
	  break;
	case BOOLEAN:
	  if(actBoolCell(virtree[j])) 
	    ((byte[])destrel.data[pos[j]])[currow] = 1;
	  else
	    ((byte[])destrel.data[pos[j]])[currow] = 0;
	  break;
	case IDLIST: {
	  // START Since jRelixNS
   if (Relation.isSigmaJoinImplemented != true)
//	  if (env.isSigmaJoinImplemented != true) //Andy May 12, 2002
	    NRenv.row = currow;
	  // END Since jRelixNS
	  
//	  System.out.println(" Start to actRelCell");
	  Relation r = actRelCell(virtree[j], vdom[idcount],
				  rname[idcount], rindx[idcount], relhtbl,
				  tname0[idcount],tname1[idcount]);
//	  r.print();
//	  System.out.println(" End of  actRelCell");
	  
	  if(virtree[j].type != OP_RED && virtree[j].type != OP_EQUIV && virtree[j].type != OP_PAR && virtree[j].type != OP_FUN){
	    r.load(true);   // have to load since its saved and cleared.
//	    System.out.println("r.domains.length " +r.domains.length);
//	    System.out.println("r.domains[0].name "+r.domains[0].name);
//	    System.out.println(" r.isBoolCompatible() "+r.isBoolCompatible());
/*	    // START Since jRelixNS
	    // Added for boolean-IDLIST compatiblity.
	    if (env.isSigmaJoinImplemented == true) {
//	      System.out.println(" env.isSigmaJoinImplemented == true");
	      
	      if (r.isBoolCompatible()) {
//		System.out.println(" r.isBoolCompatible()");
//		System.out.println(" r.boolCompatibleValue()" +r.boolCompatibleValue());
		if (r.boolCompatibleValue()){
//		  System.out.println(" byte[])destrel.data[pos[j]])[currow] = 1");
		  ((byte[])destrel.data[pos[j]])[currow] = 1;
		}
		else{
//		  System.out.println(" j" +j);
//		  System.out.println(" pos[j] " +pos[j]);
//		  System.out.println(" currow " + currow);
		  ((byte[])destrel.data[pos[j]])[currow] = 0;
		}
		if (currow == 1)
		  destrel.domains[pos[j]].type = IDLIST;
	      } else {
*///Andy May 12, 2002		
		// Generate new surrogate in destrel.
//		long sur = Surrogate.next();
		                  // START Since jRelixNS
				  // For empty nested relation, use Surrogate 0.
				  long sur;
                  if (Relation.isSigmaJoinImplemented != true)
				    sur = Surrogate.next();
                   else
                    {
				    if (r.numtuples==0)
				      sur = 0;
                    else
				      sur = Surrogate.next();
                    }
          // END Since jRelixNS //Andy, May 12, 2002

		((long[])destrel.data[pos[j]])[currow] = sur;
		
		// We must write the new surrogate for the new sub-relation.
		// REVISIT: can be well fine-tuned!
		Domain[] darry = new Domain[r.domains.length + 1];
		
		// darry[0] = domtable.getDom(".id");
		//REVISIT: Zhongxia, I used true but check it please
		darry[0] = env.lookupDom(".id", true);
		
		for(int i = 0; i < r.domains.length; i++)
		  darry[i+1] = r.domains[i];
		
		Object[] dat = Relation.createDataArray(darry, r.numtuples);
		for(int i = 0; i < r.domains.length; i++)
		  dat[i+1] = r.data[i];
		for(int i = 0; i < r.numtuples; i++)
		  ((long[])dat[0])[i] = sur;
		  
		  //DEBUGGING
		  if (dest[idcount] == null)
		  	throw new InterpretError("dest[idcount] is null");
		  else if (r == null)
		  	throw new InterpretError("r is null");
//		  System.out.println("about to call append(), dat.length = " + dat.length);
//		  if(dest[idcount].numtuples == 0)
//			System.out.println("relation " + dest[idcount].name + " has zero tuples"); 
		  //added by Yi Zheng June 24, 2002
		  if (dest[idcount].data == null && dest[idcount].numtuples == 0) 
		  	dest[idcount].createDataArray(0);
		  else if(dest[idcount].data == null)
		  	dest[idcount].load();
		  	
	//	  dest[idcount].project(darry);//to make sure domains line up with dat
		  dest[idcount].swapColumns(darry);
		  //end of addition June 24, 2002
		dest[idcount].append(dat, r.numtuples);
		dest[idcount].data = null; //Yi Zheng June 28, 2002
	//	if(dest[idcount].data != null)
	//		dest[idcount].save(); //June 24, 2002
	 //     }
	    } else {
	    // type is OP_RED or OP_EQUIV, Do nothing.
	    // what about OP_FUN and OP_PAR? hognyu Feb 2002
	  }
	  idcount++;
	  break;
	}
	case SHORT:
//	  ((short[])destrel.data[pos[j]])[currow]
//	    = (short) actIntCell(virtree[j]);

//bug fix by Yi Zheng Sept 28, 2002 to allow:
//let x be (short) dc;
	  int aiv = actIntCell(virtree[j]);
	  if (aiv == INT_DC)
	  	((short[])destrel.data[pos[j]])[currow] = SHORT_DC;
	  else if (aiv == INT_DK)
	  	((short[])destrel.data[pos[j]])[currow] = SHORT_DK;
	  else
	  	((short[])destrel.data[pos[j]])[currow] = (short) aiv;
//end of bug fix Sept 28, 2002	 
	 
	  break;
	case LONG:
	  ((long[])destrel.data[pos[j]])[currow]
	    = actLongCell(virtree[j]);
	  break;
	case FLOAT:
	//  ((float[])destrel.data[pos[j]])[currow] 
	//    = (float) actDoubleCell(virtree[j]);
	//bug fix by Yi Zheng Sept 28, 2002
	double adv = actDoubleCell(virtree[j]);
	if (adv == DOUBLE_DC)
		((float[])destrel.data[pos[j]])[currow] = FLOAT_DC;
	else if (adv == DOUBLE_DK)
		((float[])destrel.data[pos[j]])[currow] = FLOAT_DK;
	else
		((float[])destrel.data[pos[j]])[currow] = (float) adv;
	//end of bug fix Sept 28, 2002
	  break;
	case DOUBLE:
	  ((double[])destrel.data[pos[j]])[currow]
	    = actDoubleCell(virtree[j]);
	  break;
	case COMP://Yi Zheng
	  
	   ((long[])destrel.data[pos[j]])[currow]
	    = actCompCell(virtree[j]);
	   
	   break;
	  
	   case UNIVERSAL:  //from Andrey
	   case ATTRIBUTE: //from Andrey
	case STRING:
	case STMT:
	case EXPR:
	case TEXT:
	  ((String[])destrel.data[pos[j]])[currow] = actStrCell(virtree[j]);
	  break;
	default:
	  throw new InterpretError("actualizing(): to be implemented..");
	}
      }
    }
    env.removeNRFrame();
//    System.out.println(" End of current row");
    // Check if there're OP_RED domains. They require special process.
    idcount = 0;
    for(int i = 0; i < doms.length; i++){
    	//System.out.println("in special treatment for OP_RED, i = " + i);
      if(virtype[i] == IDLIST && virtree[i].type == OP_RED){
	// Generate a new surrogate.
		//System.out.println("virtype[i] == IDLIST");
	long sur = Surrogate.next();
	
	// Get the reducted relation and try to append it to
	// nested domain's data field.
	Relation r = (Relation) virtree[i].info;
	r.load(true);  // have to load it. It seems be saved and cleared.
	
	// We must write the new surrogate for the new sub-relation.
	Domain[] darry = new Domain[r.domains.length + 1];
	
	// darry[0] = domtable.getDom(".id");
	//REVISIT: Zhongxia, I used true but check it please
	darry[0] = env.lookupDom(".id", true);
	
	for(int j = 0; j < r.domains.length; j++)
	  darry[j+1] = r.domains[j];
	
	Object[] dat = Relation.createDataArray(darry, r.numtuples);
	for(int j = 0; j < r.domains.length; j++)
	  dat[j+1] = r.data[j];
	for(int j = 0; j < r.numtuples; j++)
	  ((long[])dat[0])[j] = sur;
	
	dest[idcount].append(dat, r.numtuples);
	dest[idcount].load();  // have to load it since append to disk.
	
	// Save the new surrogate into virtree[i].info for next step's
	// use. Note, we don't need the reducted relation stored in
	// virtree[i].info now.
	virtree[i].info = new Long(sur);
      }
      idcount++;
    }
    
    // Must go for another pass for RED operations.
    if(type == 1){
    	//System.out.println("Special treatment for RED when type == 1");
      for(currow = 0; currow < destrel.numtuples; currow++){
      	//System.out.println("currow = " + currow);
	   for(int j = 0; j < doms.length; j++){
	   	//System.out.println("doms[" + j + "] " + doms[j].name);
	     switch(virtype[j]){
	    		case NUMERIC:
		    //((number[])destrel.data[srcrel.numattrs+j])[currow] =
          ((number[])destrel.data[pos[j]])[currow] = 
			(number)virtree[j].info;
		    break;

	     case INTEGER:
	       ((int[])destrel.data[pos[j]])[currow]
		 = ((Integer)virtree[j].info).intValue();
	       break;
	     case BOOLEAN:
	       if(((Boolean)virtree[j].info).booleanValue()) 
		 ((byte[])destrel.data[pos[j]])[currow] = 1;
	       else
		 ((byte[])destrel.data[pos[j]])[currow] = 0;
	       break;
	     case IDLIST:
	       ((long[])destrel.data[pos[j]])[currow] = 
		 ((Long)virtree[j].info).longValue();
	       break;
	     case SHORT:
	 //      ((short[])destrel.data[srcrel.numattrs+j])[currow]
	 ((short[])destrel.data[pos[j]])[currow]
		 = (short)((Integer)virtree[j].info).intValue();
	       break;
	     case LONG:
	 //      ((long[])destrel.data[srcrel.numattrs+j])[currow]
	 ((long[])destrel.data[pos[j]])[currow]
		 = ((Long)virtree[j].info).longValue();
	       break;
	     case FLOAT://REVISIT
	       //System.out.println("in Actualizing 2280, srcrel.numattrs, j, currow, relation name:" + srcrel.numattrs + " " + j + " " + currow + " " + destrel.name);
	 //      ((float[])destrel.data[srcrel.numattrs+j])[currow]
	 ((float[])destrel.data[pos[j]])[currow]
		 = (float)((Double)virtree[j].info).floatValue();
	       break;
	     case DOUBLE://REVISIT
	 //      ((double[])destrel.data[srcrel.numattrs+j])[currow]
	 ((double[])destrel.data[pos[j]])[currow]
		 = ((Double)virtree[j].info).doubleValue();
	       break;
	       		case UNIVERSAL:  //from Andrey
	   case ATTRIBUTE: //from Andrey
	     case STRING:
	     case STMT:
	     case EXPR:
	     case TEXT:
	//     System.out.println("(String)virtree[j].info is" + (String)virtree[j].info);
	//       ((String[])destrel.data[srcrel.numattrs+j])[currow]
	((String[])destrel.data[pos[j]])[currow]
		 = (String)virtree[j].info;
	       break;
	     default:
	       throw new InterpretError("actualize(): to be implemented..");
	     }
	   }
      }
    }
    
    
  }
  
  /* This is for QT evaluation. Added by Hongyu. July 15,01 */
  
  private Relation evalQTSelector(Object[] data, int numtuples)
       throws InterpretError
  { 
//    System.out.println("==== here is evalQTSelector ===");
    int len = destdoms.length,tempType;
    String filename = interpreter.nextTempName();
    boolean resultNoNull=false;// [] quant (#>2) in SUPPLY;
    
    if(projIsSth){
      p = 1;
//      System.out.println(" projIsSth");
    }
    
    if(p == 0){// empty project domain list, return an empty relation which is defined on srcrel domains
      len=srcrel.domains.length;
      return new Relation(filename,srcrel.domains,RELATION,0,len,len,null,env);
    }
    
    else{// do the QTSelection
      Relation resultR = new Relation(filename,destdoms,RELATION,0,len,len,null,env);
      int[] qtCount = new int[p+q+1];
      int current =0,previous = 0;
      
 //     for(int i=0;i<destdoms.length;i++){
//	System.out.println("domsOrder"+ i +" " + destdoms[i].name );
 //     }
      
      intType = new Vector();
      doubleType = new Vector();
      
      for(int k=0;k<q;k++){
	curdom =predList[k];
//	System.out.println("curdom "+k + " " + curdom.name);
	
	tempType = 
	  buildTree((SimpleNode)(predList[k].tree.jjtGetChild(0)));
	
      }
      
 //     System.out.println(" intType "+intType.size());
      SimpleNode parent, child; //Have to check datattype for casting purpose
      if(!intType.isEmpty()){
	for(int i=0; i<intType.size();i++){
	  parent = (SimpleNode)((SimpleNode)(intType.elementAt(i))).jjtGetParent();
	  child = (SimpleNode)parent.jjtGetChild(1); //this is the datatype of value after "#" ( # > 2, 2's type )
	  if((child.bits>INTEGER) && (child.bits <=DOUBLE))// type is long, float and double
	    parent.bits = child.bits; /* force to be child2 type */
	  else parent.bits=INTEGER;
	}
      }
      intType.removeAllElements();
      
      if(!doubleType.isEmpty()){
	for(int i=0; i<doubleType.size();i++)
	  ((SimpleNode)((SimpleNode)(doubleType.elementAt(i))).jjtGetParent()).bits = DOUBLE;/* force to be double type */
      }
      doubleType.removeAllElements();
      
      // To get the total diffrent values in QT QUANT DOMAINS
      Vector[] qtTotalV = new Vector[q];
      for(int i =0;i<q;i++)
	qtTotalV[i] = new Vector();
      for(int i =0;i<numtuples;i++){
	for(int j=0;j<q;j++){
	  Domain d = null;
	  int pos=0;
	  if(projIsSth)
	    pos=j;
	  else pos=p+j;
	  d = destdoms[pos];
//	  System.out.println("domsOrder "+pos + " " + d.name);
	  qtCountValue(i,data,d,pos,qtTotalV[j]);
	}
      }
//      System.out.println("===========================================================");
      
      for(int k=0;k<q;k++){
	int total = qtTotalV[k].size();
//	if(projIsSth)
//	  System.out.println("here is QTTotal " +destdoms[k].name + " "+total);
//	else
//	  System.out.println("here is QTTotal " +destdoms[p+k].name + " "+total);
	Vector qtT = (Vector)(predList[k].tree.list.elementAt(1));
	int kk = qtT.size(); 
	for( int k1=0;k1<kk;k1++)
	  ((SimpleNode)(qtT.elementAt(k1))).info = new Integer(total);
      }
      
      for(int i=0;i<(p+q+1);i++)
	qtCount[i] = 0;
      boolean[] pred = new boolean[p+q+1];
      pred[p+q] = false;
      int brkPos = 0;
      Vector qtC; // qtC for  #, and qtT for . and %  
      
      for(int jj =0;jj<=numtuples;jj++){
	if(jj==0){
	  current = jj;
	  currow = jj;
	  brkPos = p+q-1;
	}
	else if (jj==numtuples){
	  previous = current;
	  currow = current;
	  brkPos = 0;
	}
	else{
	  previous = current;
	  currow = current;
	  current = jj;
	  brkPos = qtFindBrkPos(previous,current,data,destdoms);
	  if(projIsSth)
	    brkPos = brkPos+1;
	}
//	System.out.println("break " +jj+ " " + brkPos);
	
	
	if (brkPos < (p+q-1)){
	  for(int j=(p+q-2);j>=(Math.max(brkPos ,p-1));j--){
//	    System.out.println("count " +qtCount[j+1] );
//	    System.out.println("name " +predList[j+1-p].name);
	    
	    qtC = (Vector)(predList[j+1-p].tree.list.elementAt(0));
	    int kk = qtC.size(); 
	    for( int k1 =0;k1<kk;k1++)
	      ((SimpleNode)(qtC.elementAt(k1))).info = new Integer(qtCount[j+1]);
	    
	    pred[j+1] = actBoolCell((SimpleNode)(predList[j+1-p].tree.jjtGetChild(0)));
//	    System.out.println("j " +j + " " +pred[j+1] );
	    
	    if(pred[j+1]){
//	      System.out.println(" pred[j+1] is true. j " +j + "  p" +p );
	      if(j>=p)
		qtCount[j]++;
	      else{
//		System.out.println("here is copying--------------------------------");
		Object[] datal =destrel.createDataArray(destdoms,1);
		for(int k=0;k<len;k++)
		  System.arraycopy(destrel.data[k],previous,datal[k],0,1);
		resultR.append(datal,1);
		resultNoNull=true;//There is sth in the result relation
	      }
	    }
	    qtCount[j+1]=0;
	  }
	}
	
	if(projIsSth && resultNoNull)//There is sth in the result relation, so no need to do further eval
	  break;
	
	if ((brkPos <= (p+q-1)) ||
	    ((brkPos >(p+q-1)) && (!pred[p+q]))){
	  if(jj<numtuples){
	    if(cond)
	      pred[p+q] = true;
	    else{
	      if(projIsSth)
		pred[p+q]=((( (byte[])data[q])[current])== 1);
	      else pred[p+q]=((( (byte[])data[p+q])[current])== 1);
	      
	    }
	    
//	    System.out.println(" i =" +jj+ " " +pred[p+q]);
	    if(pred[p+q]){
	      qtCount[p+q-1]++;
//	      System.out.println(" info =" +jj + " " +qtCount[p+q-1]);
	      qtC = (Vector)(predList[q-1].tree.list.elementAt(0));
	      int kk = qtC.size(); 
	      for( int k1 =0;k1<kk;k1++)
		((SimpleNode)(qtC.elementAt(k1))).info = new Double((double)(qtCount[p+q-1]));
	    }
	  } 
	}
      }
      return resultR;
    }
  }
  
  
  private void qtCountValue(int current,Object[] data,Domain d,int i,Vector v)
       throws InterpretError
  { 
    switch(d.type){
      case NUMERIC:{
	number num = ((number[])data[i])[current];
	int k = v.size(), kk = -1;
	for(int j=0; j<k; j++){
	    number rnum = (number)(v.elementAt(j));
	    if(num.compareTo(rnum) == 0){
		kk = j;
		break;
	    }
	}
	if(kk == (-1))
	    v.addElement(num);
	break;}


    case INTEGER : {
      int b = ((int[])data[i])[current];
      Integer bm = new Integer(b);
      int k = v.size(),kk = -1;
      for(int j=0;j<k;j++){
	Integer r = (Integer)(v.elementAt(j));
	if(bm.equals(r))
	  kk =j; 
      }
      if(kk == (-1))
	v.addElement(bm);
      break;
    }
    case STRING:
    case TEXT:
    case STMT:
    case EXPR:{
      String bm =((String[])data[i])[current];
      int k = v.size(),kk = -1;;
      for(int j=0;j<k;j++){
	String r = (String)(v.elementAt(j));
	if(bm.compareTo(r) ==0){
	  kk =j; 
	  break;
	}
      }
      if(kk == (-1))
	v.addElement(bm);
 //     System.out.println("VVV "+v.size());
      break;
    }
    case BOOLEAN:{
      byte b =((byte[])data[i])[current];
      Byte bm = new Byte(b);
      int k = v.size(),kk = -1;;
      for(int j=0;j<k;j++){
	Byte r = (Byte)(v.elementAt(j));
	if(bm.equals(r)){
	  kk =j; 
	  break;
	}
      }
      if(kk == (-1))
	v.addElement(bm);
      break;
    }
    case SHORT:{
      short b =((short[])data[i])[current];
      Short bm = new Short(b);
      int k = v.size(),kk = -1;;
      for(int j=0;j<k;j++){
	Short r = (Short)(v.elementAt(j));
	if(bm.equals(r)){
	  kk =j; 
	  break;
	}
      }
      if(kk == (-1))
	v.addElement(bm);
      break;
    } 
    case LONG:
    case IDLIST: 
      {
	long b =((long[])data[i])[current];
	Long bm = new Long(b);
	int k=v.size(),kk = -1;;
	for(int j=0;j<k;j++){
	  Long r = (Long)(v.elementAt(j));
	  if(bm.equals(r)){
	    kk =j; 
	    break;
	  }
	}
	if(kk == (-1))
	  v.addElement(bm);
	break;
      }
    
    default :  
      throw new InterpretError("not implemented: Finding # of different values in QT.");
    }
  }
  
  private int qtFindBrkPos(int previous,int current,Object[] data,
			   Domain[] doms)
       throws InterpretError
  {
    int brkPos = 0,type;
    
    for(int i=0;i<doms.length;i++){
      brkPos = -1;
      
      type =doms[i].type;
      switch(type){
      	case NUMERIC:{
	    number bm = ((number[])data[i])[previous];
	    number bn = ((number[])data[i])[current];
	    if(bm.compareTo(bn) != 0)
		brkPos = i;
	    break;
	}

      case INTEGER : {
	int bm =((int[])data[i])[previous];
	int bn =((int[])data[i])[current];
	if(bm != bn)
	  brkPos = i;
	break;
      }
      case STRING:
      case TEXT:
      case STMT:
      case EXPR:{
	String bm =((String[])data[i])[previous];
//	System.out.println(" bm " + bm);
	String bn =((String[])data[i])[current];
//	System.out.println(" bn " + bn);
	if(bm.compareTo(bn) != 0)
	  brkPos = i;
	break;
      }
      case BOOLEAN:{
	byte bm =((byte[])data[i])[previous];
	byte bn =((byte[])data[i])[current];
	if(bm != bn)
	  brkPos = i;
	break;
      }
      case SHORT:{
	short bm =((short[])data[i])[previous];
	short bn =((short[])data[i])[current];
	if(bm != bn)
	  brkPos = i;
	break;
      }
      case LONG:
      case IDLIST:{
	long bm =((long[])data[i])[previous];
	long bn =((long[])data[i])[current];
	if(bm != bn)
	  brkPos = i;
	break; 
      }
      default : throw 
		  new InterpretError("not implemented: data type");
      }
      if( brkPos != -1)
	break;
    }
    return brkPos;
  }
  
  // End of Added. Hongyu. July 15,01
  
  
  /**
	* Display current age in mili-seconds.
	*/
  public static final long age(){
    return System.currentTimeMillis() - startTime;
  }      
  /**
   * Build/expand the tree for expression list.  It seems only EQUIV
	* operation's by-list contains expression list in domain algebra, so
	* this building is based on equivalent operation. For other cases
	* the method must be modified.
	*/
  private void buildExpressionList(SimpleNode n)
       throws InterpretError
  {
    Domain dd; 
    for(int k = 0; k < n.jjtGetNumChildren(); k++){
      SimpleNode child = (SimpleNode)n.jjtGetChild(k);
      if(child.type == OP_IDENTIFIER){
	
	// Domain d = domtable.(child.name);
	//REVISIT: Zhongxia, I used true but check it please
	Domain d = env.lookupDom(child.name, true);
	
	if(d == null){
	  // We touched a non-declared domain.
	  throw new InterpretError("'"+child.name+
				   "' in the by-list not found.");
	}else if(srcrel.isDomainIn(d)){
	//  n.bits = alldoms.indexOf(d);//Yi Zheng Oct 5, 2002
		child.bits = alldoms.indexOf(d);
	}else if(d.tree == null){
	  // We touched a normal domain which is however not
	  // in the relation's domain list.
	  throw new InterpretError("'"+d.name+"' in the "+
				   "virtual tree can not be actualized in buildExpressionList.");
	}else{
	  // OK, this must be a virtual domain, and we go
	  if(curpath.contains(d))
	    throw new InterpretError("a recursive loop exists.");
	  
	  // Must check the d.tree for recursive definition.
	  curpath.addElement(d);
	  
	  // buildTree(domtable.getTree(d.name).jjtDuplicate());
	  //PATCHANGE:
	  dd = env.lookupDom(d.name, true);
	  SimpleNode newnode = dd.getTree().jjtDuplicate();
//	  buildTree(dd.getTree().jjtDuplicate());
	  buildTree(newnode);
	  
	  curpath.removeElement(d);
	  
	  // SimpleNode newnode = 
	  //         domtable.getTree(d.name).jjtDuplicate();
	  //PATCHANGE:
	  dd = env.lookupDom(d.name, true);
	 // SimpleNode newnode = dd.getTree().jjtDuplicate();
	  
	  if(!vertdom.containsKey(d) || vertlevel >
	     ((Integer)vertdom.get(d)).intValue()){
	     	if(debug_oct_5)
	     		System.out.println(d.name + " added to vertdom");
	    vertdom.put(d, new Integer(vertlevel));
	    if(!alldoms.contains(d)){
	      // for generating doms array in destrel.
	      alldoms.addElement(d); 
	     // n.bits = alldoms.indexOf(d);//Yi Zheng Oct 5, 2002
	     child.bits = alldoms.indexOf(d);
	    }
	    if(maxvertlevel < ++vertlevel)
	      maxvertlevel = vertlevel;
	    
	    // Suppose stop here, this domain will be
	    // actualized in first pass.
	    vvtree.put(d, newnode);
	  }
	}
      }else{   // Must be some expressions. E.g.
	// let x be equiv+ of A by A+B, A-B;
	
	// Have to replace it with an new 'Id' node. Refer to the
	// case of RED & EQUIV.
	
	// We need to take off this node and put it in vertdom.
	SimpleNode newnode = child.jjtDuplicate();
	
	// Will actualize this special domain in firstly.
	Domain newdom = new Domain("."+fdomcounter++,0,0,newnode);
	
	// This must be a new domain so we don't need to check if
	// it's contained in the vertdom hashtable.
	vertdom.put(newdom, new Integer(vertlevel));
	
	if(maxvertlevel < ++vertlevel)
	  maxvertlevel = vertlevel;
	
	// Get newtype since 'newdom' needs it.
	int newtype = buildTree(newnode);
	newdom.type = newtype;    // set the correct type value.
	
	// reset the vertlevel.
	vertlevel--;
	
	alldoms.addElement(newdom);
	vvtree.put(newdom, newnode);
	
	// Current node needs to stop here, i.e. we need to
	// generate a 'domain' node (of 'Identifier' type).
	SimpleNode leaf = new SimpleNode("Identifier");
	leaf.type = leaf.opcode = OP_IDENTIFIER;
	
	// Will search this domain column later.
	leaf.name = newdom.name;  
	leaf.bits = alldoms.indexOf(newdom);
	n.jjtReplaceChild(k, leaf);
      }
    }
  }    
  
  /**
   * method to get the domain in allD that has the same name as d
   * bug fix for buildTree()
   * @author Yi Zheng July 9, 2002
   */
  private Domain Contains(Vector allD, Domain d)
  {
  	Domain retD = null;
  	for(Enumeration e = allD.elements(); e.hasMoreElements();)
  	{
  		Domain next = (Domain)e.nextElement();
  		if(next.name.compareTo(d.name) == 0)
  		{
  			retD = next;
  			//System.out.println("found retD " + retD.name);
  			return retD;
  		}
  	}
  	return retD;
  }
  /**
   * Building the actulization tree after making sure the definition of
   * the virtual domain has no problem.
        *  
        *  JRelixNS modified/enhanced this method:
        *   Allow IDLIST where BOOLEAN is expected.
                *
        * JRelCaNu modified/enhanced this method:
        *  - Handle OP_CAST type check explicitely.  
        *    Also, the OP_CAST cast from type is stored in 
        *    bits of its first child node.
        *  - to allow project and select in nested relational expressions.

   */
  private int buildTree(SimpleNode node)
       throws InterpretError
  {
    int numChildren = node.jjtGetNumChildren();
    
    if(numChildren == 0){
      // Current node has single child. Must be as follows:
      //  e.g.  let a be 23;
      //        let x be y;
      switch(node.type){
      case OP_LITERAL:
	// Node without children e.g. let b be 1;
	return(Utility.typeConvert(node.opcode));
      case OP_IDENTIFIER:{
	// Node without children e.g. let b be a; Or node from
	// the upper level.
	
	//            Domain d = domtable.getDom(node.name);
	//REVISIT: Zhongxia, I used true but check it please
	Domain d = env.lookupDom(node.name, true);
	Domain D = null; //added by Yi Zheng July 9, 2002
	if(d == null){
	  if(node.opcode == OP_JOINOPERATOR || node.opcode == OP_IDENTIFIER){ //OP_IDENTIFIER added by Yi Zheng May 13, 2002
	    // This is for the case 'let x be (S ijoin R);' //should be let x be R (Yi Zheng, May 13, 2002)
	    // where R is a top-level relation.
	    
	    //                  Relation r = reltable.getRel(node.name);
	    //REVISIT: Zhongxia, I used true but check it please
	    Relation r = env.lookupRel(node.name, true);
	    
	    if(r != null){
	   	  //if swtich added by Yi Zheng, May 13, 2002
	   	  if(r.rvc == RELATION)
	   	  {
	   	  	if(DEBUG)
	   	  		System.out.println("virtual domain defined on top level Relation: " + r.name);
	      	return IDLIST;
	      }
	      else if(r.rvc == COMPUTATION)
	      {
	      	if(DEBUG)
	   	  		System.out.println("virtual domain defined on top level Computation: " + r.name);
	      	return COMP;
	      }
	    }
	    else{
	      throw new InterpretError("relation (or domain) '"+node.name+
				       "' in virtual tree not found.");
	    }
	  }
	  else
	    // We touched a non-declared domain.
	    throw new InterpretError("'"+node.name+
				     "' in virtual tree not found.");
	}
//	else if(alldoms.contains(d)){
	else if ((D = Contains(alldoms, d)) != null) {
	  // This is a domain declared in this relation's
	  // domain list, or it's a virtual domain we already
	  // processed. First we save its index in alldoms
	  // vector in this node's bit field for later use.
	  //System.out.println("Setting node.bits at 2481 to " + alldoms.indexOf(d));
//	  node.bits = alldoms.indexOf(d); //Yi Zheng July 9, 2002
	  node.bits = alldoms.indexOf(D);
	  // For safety sake, we duplicate this IDENTIFIER node.
	  // Usually it's not necessary, but for certain case where
	  // the actualizer writes something to a node in the tree,
	  // the real domain tree should not touched.
	  // REVISIT: the following line is only for the tuple-by-
	  // tuple actualization for a reduction virtual domain,
	  // since the actualizer writes temperary sub-relation names
	  // to the IDENTIFIER node.
	  node.jjtReplace(node.jjtDuplicate());
	  
	  // Just return its type.
	  
	  //               return(domtable.getType(node.name));
	  //REVISIT: Check out the method domtable.getType()
	  //         and adjust the action that is taken here.
	  Domain dd = env.lookupDom(node.name, true);
	  if (dd != null)
	    {  return dd.type;
	    }
	   else
	   	throw new InterpretError ("domain: " + node.name + " not found"); //Yi Zheng May 3, 2002
	}else if(d.tree == null){
	  // We touched a normal domain which is however not
	  // in the relation's domain list.
	  throw new InterpretError("'"+d.name+"' in the "+
				   "virtual tree can not be actualized: tree is null.");
	}else{
	  // This is a virtual doamin. e.g. let x be y; and
	  // 'y' is also a virtual domain. We go one level deeper,
	  // i.e. we look into 'y'. But before that, we must insert
	  // the lower level tree into our big tree.
	  // Check if current domain recursively depends on itself.
	  if(curpath.contains(d))
	    throw new InterpretError("a recursive loop exists involving: " + d.name);
	  
	  // Get the lower-level tree.
	  //              SimpleNode newnode = domtable.getTree(d.name).jjtDuplicate();
	  //PATCHANGE:
	  Domain dd = env.lookupDom(d.name, true);
	  SimpleNode newnode = dd.getTree().jjtDuplicate();
	  
      // Have to test OP_RED & OP_EQUIV since they are spcial.
	  // Well, it's kind of dirty, but I don't want to add one
	  // more global variable for 'current domain'.
	  if((newnode.type == OP_RED || newnode.type == OP_EQUIV || newnode.type == OP_FUN || newnode.type == OP_PAR) //FUN/PAR added by Yi Zheng May 13, 2002
	     && (d.type != IDLIST || !nt) ){
	    
	    if(!vertdom.containsKey(d) || vertlevel >
	       ((Integer)vertdom.get(d)).intValue()){
	      vertdom.put(d, new Integer(vertlevel));
	      if(!alldoms.contains(d)){//this is always true if control gets here
		// for generating doms array in destrel.
		alldoms.addElement(d); 
		//System.out.println("Setting node.bits at 2536 to " + alldoms.indexOf(d));
		node.bits = alldoms.indexOf(d);
	      }
	      if(maxvertlevel < ++vertlevel)
		maxvertlevel = vertlevel;
	      
	      // Suppose stop here, this domain will be actualized in
	      // first pass.  But we must call buildTree() first, in
	      // order to duplicate lower-level nodes since
	      // actualizer will write to ID node for tuple-by-
	      // tuple reduction.  I just added the node duplication
	      // for IDENTIFIER node for safety sake.
	      buildTree(newnode);
	      vvtree.put(d, newnode);
	    }
	  }
	  else{
	    // Continue to extend the next node.
	    node.jjtReplace(newnode);
	  }
	  
	  curpath.addElement(d);
	  int type = buildTree(node);    // 'newnode' should be the same
	       curpath.removeElement(d);         // as 'node'.
	       
	       if((newnode.type == OP_RED || newnode.type == OP_EQUIV || newnode.type == OP_FUN || newnode.type == OP_PAR) //last two cases added by Yi Zheng May 13, 2002
		  && (d.type != IDLIST || !nt) )
		 vertlevel--;
	       
	       return(type);
	}
//	throw new InterpretError("buildTree: to be implemented."); Yi Zheng May 3, 2002
      } //end of case OP_IDENTIFIER
      
      case OP_QTOPERATOR:  { // Added by Hongyu.
	SimpleNode qtCount = new SimpleNode(" ");
	switch(node.opcode)
	  {
	  case OP_NUMBER:
	    { 
	      qtCount.type =OP_LITERAL;
	      qtCount.opcode =OP_LITERAL;
	      qtCount.bits =INTEGER;
	      node.bits =INTEGER;
	      ((Vector)(curdom.tree.list.elementAt(0))).addElement(qtCount);
	      node.jjtAddChild(qtCount);
	      
	      if((node.name == null) || (node.name.compareTo("qtPerDot") != 0))
		intType.addElement(node);
	      return INTEGER;
	    }
	  case OP_DOT:
	  case OP_PERCENT:
	    { 
	      qtCount.set(OP_QTOPERATOR,OP_NUMBER);  // change to "#"
	      qtCount.name = "qtPerDot";
	      node.type = OP_BIOPERATOR;         // replace by # / Total
	      node.opcode =OP_DIV;
	      node.bits = DOUBLE;
	      node.jjtAddChild(qtCount);
	      SimpleNode qtTotal = new SimpleNode(" "); // Total;
	      qtTotal.type =OP_LITERAL;
	      qtTotal.opcode =OP_LITERAL;
	      qtTotal.bits = DOUBLE;
	      node.jjtAddChild(qtTotal);
	      ((Vector)(curdom.tree.list.elementAt(1))).addElement(qtTotal);
	      doubleType.addElement((SimpleNode)node.jjtGetParent());
	      int r = buildTree(qtCount);
	      return DOUBLE;
	      
	    }
	  default:
	    throw new InterpretError("QT: to be implemented."); 
	  }
	
      }
      default:
	throw new InterpretError("not implemented.");
      }
    }// End of Children == 0 Added by Hongyu.
    else{     // Current node has children. 
      // Has to go through each child to check the type compatibility.
      int i = 0;                                // type counter.
      int[] typeArray;
      
      switch(node.type){
      	case OP_ARRAY:
      		return IDLIST; //added by Yi Zheng June 21, 2002
      			//case OP_TRANSPOSE added by Yi Zheng May 7, 2002, REVISIT
		case OP_TRANSPOSE:
			return IDLIST;		

      	//OP_FUNCTION case added by Yi Zheng May 6, 2002
      	case OP_FUNCTION:
			 SimpleNode argument = (SimpleNode) node.jjtGetChild(0);
			 int argType = buildTree(argument);
		            if (argType == NUMERIC) return argType;

		     switch (node.opcode)
		     {
				case OP_ABS:
					if (!Utility.isNumeric(argType))
						throw new InterpretError("function argument type mismatch");
					else
						return argType;
				case OP_SQRT:
				case OP_SIN:
				case OP_ASIN:
				case OP_COS:
				case OP_ACOS:
				case OP_TAN:
				case OP_ATAN:		
				case OP_LOG:		
				case OP_LOG10:
					if (!Utility.isNumeric(argType))
						throw new InterpretError("function argument type mismatch");
					else
						return DOUBLE;
				case OP_ROUND:
				case OP_CEIL:
				case OP_FLOOR:
				case OP_CHR:
					if (!Utility.isNumeric(argType))
						throw new InterpretError("function argument type mismatch");
					else
					{
						if (node.opcode != OP_CHR)
							return INTEGER; //elevate to highest level of precision
						else return STRING;
					}
				case OP_ORD:
					if (! Utility.isText(argType))
						throw new InterpretError("function argument type mismatch");
					else
						return 	INTEGER;
						
				
				case OP_ISKNOWN:
				case OP_SINH:
				case OP_COSH:
				case OP_TANH:
					throw new InterpretError("Not implemented");
				default:
					throw new InterpretError("Unknown function operator");
		     }
		
      case OP_IFTHENELSE:{
             if (Relation.isSigmaJoinImplemented != true) { //Andy May 12, 2002
//	if (env.isSigmaJoinImplemented != true) {
	  int type = 0;
	  
	  SimpleNode ifNode = ((SimpleNode)node.jjtGetChild(0)).jjtDuplicate();
	  SimpleNode thenNode = ((SimpleNode)node.jjtGetChild(1)).jjtDuplicate();
	  SimpleNode elseNode = ((SimpleNode)node.jjtGetChild(2)).jjtDuplicate();
	  node.jjtReplaceChild(0, ifNode);
	  node.jjtReplaceChild(1, thenNode);
	  node.jjtReplaceChild(2, elseNode);
	  
	  if(BOOLEAN != buildTree(ifNode)){
	    throw new InterpretError("type of 'if' clause "+
				     "is not boolean.");
	  }else{
	    type = buildTree(thenNode);
	    if(type != buildTree(elseNode)){
	      throw new InterpretError("type mismatch between "+
					    "'then' clause and 'else' clause.");
	    }else{
	      return(type);
	    }
	  }
	}
	else {
	  int type = 0;
	  
	  SimpleNode ifNode = ((SimpleNode)node.jjtGetChild(0)).jjtDuplicate();
	  SimpleNode thenNode = ((SimpleNode)node.jjtGetChild(1)).jjtDuplicate();
	  SimpleNode elseNode = ((SimpleNode)node.jjtGetChild(2)).jjtDuplicate();
	  node.jjtReplaceChild(0, ifNode);
	  node.jjtReplaceChild(1, thenNode);
	  node.jjtReplaceChild(2, elseNode);
	  
	  // START Since jRelixNS
	  // Added for boolean-IDLIST compatiblity.
	  // Allow IDLIST where BOOLEAN is expected.
//	  if (env.isSigmaJoinImplemented != true){
	      if (Relation.isSigmaJoinImplemented != true){
	    if(BOOLEAN != buildTree(ifNode)){
	      throw new InterpretError("type of 'if' clause "+
				       "is not boolean.");
	    }
	    else{
	      type = buildTree(thenNode);
	      if(type != buildTree(elseNode)){
		throw new InterpretError("type mismatch between "+
					 "'then' clause and 'else' clause.");
	      }
	      else{
		return(type);
	      }
	    }
	  }
	  else {
	    type = buildTree(ifNode);
	    if (!Utility.isBoolOrIDListType(type)){
	      throw new InterpretError("buildTree: type of 'if' clause "+
				       "is not boolean nor IDLIST.");
	    }
	    else{
	      type = buildTree(thenNode);
		  int typeElse = buildTree(elseNode); //Yi Zheng May 13, 2002
	      
	   //  if(type != buildTree(elseNode)){ Yi Zheng May 13, 2002
 		     if(type != typeElse){  //Yi Zheng May 13, 2002

	//	if (Utility.isBoolOrIDListType(type)) //Yi Zheng, May 13, 2002
		if (Utility.isBoolOrIDListType(type) && Utility.isBoolOrIDListType(typeElse)) //Yi Zheng, May 13, 2002

		  return(BOOLEAN);
		else
		  throw new InterpretError("type mismatch between "+
					   "'then' clause and 'else' clause.");
	      }
	      else{
	      	
	      	if(debug_oct_52)
	      		Utility.dump(node, "ifthenelse: ");
		return(type);
	      }
	    }
	  }
	  // END Since jRelixNS
	}
    
      }
      case OP_JOINOPERATOR:{
	if(node.jjtGetNumChildren() == 2){
		if(node.opcode != OP_NOP) {
	  SimpleNode left = ((SimpleNode)node.jjtGetChild(0)).jjtDuplicate();
	  SimpleNode right = ((SimpleNode)node.jjtGetChild(1)).jjtDuplicate();
	  
	  // 'left' and 'right' must be OP_IDENTIFIER. Change their opcode
	  // to make it possible for buildTree() to differentiate them from
	  // normal IDENTIFIER, since they need special check, i.e.
	  // to check the existance of top-level relation. e.g.
	  //    let x be S ijoin R;    -> where 'R' is a top-level rel.
	  if(left.type == OP_IDENTIFIER) left.opcode = OP_JOINOPERATOR;
	  if(right.type == OP_IDENTIFIER) right.opcode = OP_JOINOPERATOR;
	  
	  node.jjtReplaceChild(0, left);
	  node.jjtReplaceChild(1, right);
	  
	  if(left.type == OP_RED || left.type == OP_EQUIV)
	    processRedNode(node, left, 0);
	  else if(right.type == OP_RED || right.type == OP_EQUIV)
	    processRedNode(node, right, 1);
	 // else if(buildTree(left) != IDLIST || buildTree(right) != IDLIST)
	 //   throw new InterpretError("join operands type mismatch.");
	// modified by Yi Zheng May 13, 2002 as follows:
	  else
	  {
		int ltype = buildTree(left);
		int rtype = buildTree(right);
		if ((ltype != IDLIST && ltype != COMP) || (rtype != IDLIST && rtype != COMP))
			throw new InterpretError("join operands type mismatch.");
	  }
	 }
	 else //OP_NOP added by Yi Zheng July 5, 2002
	 {
	  SimpleNode left = ((SimpleNode)node.jjtGetChild(0)).jjtDuplicate();
	  SimpleNode right = ((SimpleNode)node.jjtGetChild(1)).jjtDuplicate();
	  int ltype, rtype;
	  node.jjtReplaceChild(0, left);
	  node.jjtReplaceChild(1, right);
	  
	  if(left.type == OP_RED || left.type == OP_EQUIV)
	    ltype = processRedNode(node, left, 0);
	   else
	    ltype = buildTree(left);
	    
	  if(right.type == OP_RED || right.type == OP_EQUIV)
	    rtype = processRedNode(node, right, 1);
	  else
	  	rtype = buildTree(right);
	 // else if(buildTree(left) != IDLIST || buildTree(right) != IDLIST)
	 //   throw new InterpretError("join operands type mismatch.");
	// modified by Yi Zheng May 13, 2002 as follows:
	 	
	 if (ltype != rtype)
			throw new InterpretError("NOP operands type mismatch.");
		return ltype;
	 }
	}
	else if(node.jjtGetNumChildren() == 4){
	  // Handle the join relations.. (the same as above.)
	  SimpleNode left = ((SimpleNode)node.jjtGetChild(0)).jjtDuplicate();
	  SimpleNode right = ((SimpleNode)node.jjtGetChild(3)).jjtDuplicate();
	  
	  if(left.type == OP_IDENTIFIER) left.opcode = OP_JOINOPERATOR;
	  if(right.type == OP_IDENTIFIER) right.opcode = OP_JOINOPERATOR;
	  
	  node.jjtReplaceChild(0, left);
	  node.jjtReplaceChild(3, right);
	  
//	  if(buildTree(left) != IDLIST || buildTree(right) != IDLIST) //modified by Yi Zheng June 21, 2002
	  int leftype = buildTree(left);
	  int rightype = buildTree(right);
	  if(((leftype != IDLIST) && (leftype != COMP) && (rightype != IDLIST) && (rightype != COMP))
	   || ((leftype == COMP) && (rightype == COMP)))
	    throw new InterpretError("join operands type mismatch.");
	  
	  // handle the join attributes.
	  SimpleNode exprl = (SimpleNode)node.jjtGetChild(1);
	  SimpleNode exprr = (SimpleNode)node.jjtGetChild(2);
	  
	  // Get the join relation.
	}
	else{
	  // Should never happen.
	  throw new InterpretError("join tree has wrong number of children.");
	}
	return IDLIST;
      }
      
      case OP_PROJECT:
      case OP_SELECT:
      case OP_TSELECT:
      case OP_QSELECT:
	{
	  
	  // Get the source relation.
	  Relation srcrel = null;
	  // SimpleNode srcnode = ((SimpleNode)node.jjtGetChild(1)).jjtDuplicate(); 
	  // modified by hongyu
	  // First, get the source relation
	  SimpleNode srcnode = null;
	  if((node.type == OP_PROJECT) ||(node.type == OP_SELECT)){
	    srcnode = ((SimpleNode)node.jjtGetChild(1)).jjtDuplicate();
	    node.jjtReplaceChild(1, srcnode);
	  }
	  else if(node.type == OP_QSELECT){
	    if(node.jjtGetNumChildren() == 2){
	      srcnode =((SimpleNode)node.jjtGetChild(1)).jjtDuplicate();
	      //node.jjtReplaceChild(2, srcnode);
	      node = node.jjtDuplicate();
	    }
	    else{
	      srcnode = ((SimpleNode)node.jjtGetChild(2)).jjtDuplicate();
	      node = node.jjtDuplicate();
	    }
	  }
	  else{
	    srcnode = (SimpleNode)node.jjtGetChild(1);
	    node.jjtReplaceChild(1, srcnode);
	  }
	  
	  //node.jjtReplaceChild(1, srcnode);
	  //End of Hongyu modification
	  
	 int treeType = buildTree(srcnode);
	  if((treeType != IDLIST) && (treeType != COMP))
	    // Should never happen.
	    throw new InterpretError("source relation is not a nested domain.");
	  
	  switch(srcnode.type){
	  case OP_IDENTIFIER:{
	    
	    //          srcrel = reltable.getRel("." + srcnode.name);
	    //REVISIT: Zhongxia, I used true but check it please
	    srcrel = env.lookupRel("."+srcnode.name, true);
	    
	    if(srcrel == null){
	      
	      //        srcrel = reltable.getRel(srcnode.name);
	      //REVISIT: Zhongxia, I used true but check it please
	      srcrel = env.lookupRel(srcnode.name, true);
	      
	      if(srcrel == null)
		throw new InterpretError(
					 "source relation '"+srcnode.name+"' not found.");
	    }
	    break;
	  }
	  case OP_SELECT:
	  case OP_PROJECT:
	  case OP_TSELECT:
	  case OP_QSELECT: // Hongyu  
	  case OP_JOINOPERATOR:{
	    // We are doing nothing here, since processIDListDom() will
	    // throw exceptions if anything wrong.
	    break;}
	  //default:
	  //  throw new InterpretError("to be implemented in OP_PRJ.");
			case OP_TRANSPOSE: //Yi Zheng May 7, 2002 REVISIT
				return IDLIST;
	  
	  // AC:
	  case OP_UOPERATOR:{
		if(Relation.isSigmaJoinImplemented == true){ //Andy, May 12, 2002
//	    if(env.isSigmaJoinImplemented == true){
	      SimpleNode unode = (SimpleNode)node.jjtGetChild(0);
	      unode = unode.jjtDuplicate();
	      node.jjtReplaceChild(0, unode);
	      return buildTree(unode);
	    }
	  }
	  //added by Yi Zheng July 5, 2002
	  case OP_ARRAY:
	  	break;
	  	//end of addition July 5
	  default:
	    throw new InterpretError("to be implemented in OP_PRJ.");
	  }
	  // AC: END
	  
	  //added by Yi Zheng July 2, 2002
	  //to allow for level lifting of the following form:
	  //let A be [red + of a] in B; or
	  //let A be [red ujoin of C] in B;
	  if((node.type == OP_PROJECT) || (node.type == OP_TSELECT))
	  {
	  	SimpleNode projector = (SimpleNode) node.jjtGetChild(0);
	  	
	  	SimpleNode exprlist = null;
	  	if((projector.isBitSet(0)) && (projector.isBitSet(1)))
	  		exprlist = (SimpleNode) projector.jjtGetChild(1);
	  	else if ((!projector.isBitSet(0)) && (projector.isBitSet(1)))
	  		exprlist = (SimpleNode) projector.jjtGetChild(0);
	  	
		
	    if((exprlist != null) && (exprlist.jjtGetNumChildren() == 1) 
			&& (((SimpleNode) exprlist.jjtGetChild(0)).type == OP_RED))
		{
			
			SimpleNode firstChild = (SimpleNode) exprlist.jjtGetChild(0);
			
			//create a reddom
			SimpleNode ndCopy = (SimpleNode)firstChild.jjtDeepDuplicate();
			
			//create a complet LET node
			SimpleNode letNode = SimpleNode.jjtCreateNode("Declaration", OP_DECLARATION,
				OP_LET, null);
			
			
			String domname = "reddom" + Surrogate.next();
			SimpleNode tnd = SimpleNode.jjtCreateNode("Identifier",OP_IDENTIFIER,OP_IDENTIFIER,
																	domname);
			
			ndCopy.jjtSetParent(letNode);
			letNode.jjtAddChild(ndCopy);
			
			letNode.bits = 0;

			
			Domain newDom = interpreter.newVirDomain(domname, letNode, env);
	   		//System.out.println("virtual domain " + domname + " is added to environment");
	   		Relation rel_saved = env.lookupRel("."+domname,true);
	   		env.put(newDom);
	   		newDom.isRedTemp = 1;
			if (rel_saved != null)
	     		env.put(rel_saved);


			//replace the red expression with an identifier node
			exprlist.jjtReplaceChild(0,tnd);
			if((!Utility.isPrimitiveType(ndCopy.opcode)) && (ndCopy.opcode != OP_NOP))
				return IDLIST;
			//addition by Yi Zheng July 7, 2002
			else if(ndCopy.opcode == OP_NOP)
			{
				String dName = ((SimpleNode)ndCopy.jjtGetChild(0)).name;
				if(dName != null && dName.compareTo("") != 0)
				{
					Domain rDom = env.lookupDom(dName, true);
					if(rDom != null)
						return rDom.type;
					else
						throw new InterpretError("domain " + dName + " not found");
				}
				else
					throw new InterpretError("buildTree: to be implemented");
				
			}
			//end of addition July 7, 2002
			return buildTree(tnd);			
		}
	  }
	  //end of Yi's addition July 2, 2002
	  
	  if((node.type == OP_PROJECT || node.type == OP_TSELECT)|| //Revised by hongyu
	    (node.type == OP_SELECT || node.type == OP_QSELECT))
	    return IDLIST;
	  else
	    throw new InterpretError("node type is not decalred ");
	}
      
      // end of case OP_TSELECT. hongyu
      // end of hongyu modification
      
      // Following is the default case. Added by Hongyu
      default:
                              //START Since JRelCaNu a.
			if(node.opcode == OP_CAST) {// &&(node.type == OP_BIOPERATOR)
			   //Utility.dump(node,"Sample OP_CAST Node --->");
			   //Sample OP_CAST Node --->Cast:OP_BIOPERATOR:OP_CAST::0
			   //Sample OP_CAST Node ---> Type:OP_TYPE:OP_DOUBLE:null:0
			   //Sample OP_CAST Node ---> Literal:OP_LITERAL:OP_FLOAT:null:0
                           // OP_CAST is a OP_BIOPERATOR different from others in terms
			   // of type checking: 
			   // 1. Here the 2 children nodes are explicitely 
                           //    checked for cast compatibility.
			   // 2. The cast from type (of the 0th child node) is stored in the 
			   //    left.bits of the OP_TYPE node(1st child node).
			   SimpleNode left = ((SimpleNode)node.jjtGetChild(0)).jjtDuplicate();
			   SimpleNode right = ((SimpleNode)node.jjtGetChild(1)).jjtDuplicate();
			   node.jjtReplaceChild(0, left);
			   node.jjtReplaceChild(1, right);
			   // left should always be a OP_TYPE type node
			   int CastToType = Utility.typeConvert(left.opcode);
			   int CastFromType = buildTree(right);
			   left.bits = CastFromType;
			   if (Utility.domsCastCompatible(CastToType,CastFromType)) {
			       return(CastToType);
                           } else{
                                 node.dump("InterpretError buildTree==>");
				 throw new InterpretError("buildTree: cast types not compatible.");
			}
			}
                        //End Since JRelCaNu a.
	if(node.type != OP_EQUIV && node.type != OP_FUN && node.type != OP_PAR)//Form sungsoo. hongyu added. Feb 2002
	{
		//if added by Yi Zheng July 8, 2002
		if(node.opcode != OP_REDFUNCALL)
	  		typeArray = new int[numChildren];   // type array.
	  	else
	  		typeArray = new int[numChildren - 1];
	}
	else if(node.type == OP_PAR)
	{
		//if added by Yi Zheng July 8, 2002
		if(node.opcode != OP_REDFUNCALL)
	  		typeArray = new int[numChildren-2]; // Modified by Sungsoo Kang for OP_PAR,hongyu added. Feb 2002
		else
			typeArray = new int[numChildren - 3];
	}
	else
	{
		//if added by Yi Zheng July 8, 2002
		if(node.opcode != OP_REDFUNCALL)
		  typeArray = new int[numChildren - 1]; // by-list counts no type.
	 	else
	 		typeArray = new int[numChildren - 2];
	}
	// Check each child of the current node.
	if ((debug_oct_52) && (node.opcode == OP_GE))
		Utility.dump(node, "default:");
	for(int j = 0; j < numChildren; j++){
		if((node.opcode == OP_REDFUNCALL) && (j == 0)) //skip OP_REDFUNCALL
			continue;
	  SimpleNode n = (SimpleNode)node.jjtGetChild(j);
	  switch(n.type){
	  case OP_LITERAL:
	    typeArray[i++] = Utility.typeConvert(n.opcode);
	    break;
	  case OP_IDENTIFIER:
	  case OP_QTOPERATOR: 
	    // For OP_QTOPERATOR, " . " and "% "Added by Hongyu. 
	    // OP_FUN and OP_PAR have not been implemented.
	    // Added by Hongyu. Sept 15,01
	  case OP_BIOPERATOR:

	  case OP_UOPERATOR:
	    // BEGIN Since jRelixNS
	    // This is a bug fix to jRelix76?
//	  case OP_UNOT: //Andy May 12, 2002
	    // END Since jRelixNS
	  case OP_JOINOPERATOR:
	                             //START Since JRelCaNu b.
			   // - to allow project and select in 
			   //   nested relational expressions;
			   //   for example, in a SELECT expression.
			   case OP_SELECT:
			   case OP_PROJECT:
			   case OP_TSELECT:
                           //End Since JRelCaNu b.

	  case OP_IFTHENELSE:
	  case OP_FUNCTION: //Yi Zheng May 6, 2002
	  case OP_ARRAY:	//Yi Zheng June 21, 2002
	  case OP_TRANSPOSE:{ //Yi Zheng June 21, 2002
	    SimpleNode newnode = n.jjtDuplicate();
	    node.jjtReplaceChild(j, newnode);
	    typeArray[i++] = buildTree(newnode);
	    break;
	  }
	  case OP_RED:
	  case OP_EQUIV:
	  case OP_FUN:
	  case OP_PAR:// sungsoo's code. Added by hongyu  
	  
	    typeArray[i++] = processRedNode(node, n, j);
	    break;
	  case OP_EXPRLIST:{
	    // REVISIT: many repetition codes here. Can we avoid it?
	    buildExpressionList((SimpleNode)n.jjtDuplicate());
//	    if(debug_oct_52)
//	    	Utility.dump(n,"case OP_EXPRLIST:");
	 //   buildExpressionList(n);
	    break;
	  }
//	  case OP_FUNCTION: //Yi Zheng, May 6, 2002 //moved up June 21, 2002
//	    typeArray[i++] = buildTree(n);
//	    break;
	  default:
	    n.dump("InterpretError buildTree==>");
	    throw new InterpretError("buildTree: to be implemented.");
	  }
	}
	// REVISIT: this is for expression comparison e.g.
	// let x be (a+b-c) = (c*a-b); See actBoolCell().
	// Note: we're using the last type in the typeArray[], and
	// this might cause problems.
	//System.out.println("Setting node.bits at 2743 to " + typeArray[i-1]);
	node.bits = typeArray[i - 1];
	if ((debug_oct_52) && (node.opcode == OP_GE))
		System.out.println("node.bits = " + Utility.type2String(node.bits));
	/* A bug. 
	   The following is added by hongyu for type casting. If a node has two children, 
	   then if node.bits = typeArray[i - 1], the type casting might be wrong. 
	   e.g  node 
	       / div \
	       2.0     3   node will be built up as type INT, but it should be double(long) type
	       according to Java type casring mechanism.
	       */
	switch(node.type){
	case OP_BIOPERATOR: { 
	  switch(node.opcode){
	  case OP_PLUS:
	  case OP_MINUS:
	  case OP_MUL:
	  case OP_DIV:
	  case OP_MOD:
	  case OP_EQ:
	  case OP_NEQ:
	  case OP_GT:
	  case OP_LT:
	  case OP_GE:
	  case OP_LE:
	    // and more ...
	    {
	    	if((debug_oct_52) && (node.opcode == OP_GE))
	    	{
	    		Utility.dump(node,"BIOP node: ");
	    		System.out.println("typeArray[0] = " + Utility.type2String(typeArray[0]));
	    		System.out.println("typeArray[1] = " + Utility.type2String(typeArray[1]));	
	    		
	    	}
	      if(typeArray[0]>typeArray[1])
		node.bits=typeArray[0];
	      break;
	    }
	  default:
	    break;
	  }
	}
	//and more 
	default:
	  break; 
	}
	
	// End of hongyu modification. Feb 2002    
	  
//	System.out.println("buildtree node bits "+node.bits);
	if((debug_oct_52) && (node.opcode == OP_GE))
		Utility.dump(node, "after buildTree: ");
	return Utility.typeCheck(typeArray, node.opcode);
      }
    }
  }  
     /**
    * wrapper added by Andrey Rozenberg
    */
   public void cleanup()
   {
	   for(int i = 0; i < dom_tempnames.size(); i++)		
		  env.domtable.rmDom((String)dom_tempnames.elementAt(i));

	   cleanup_old();
   }
    
  public void cleanup_old() //changed from cleanup according to Andrey's code
  {  if(tempnames.size() != 0)
    {  for(int i = 0; i < tempnames.size(); i++)
      {
	if(nt)
	  //REVISIT: This is taken out temporarily. It MUST be put back in.
	  //  domtable.rmDom((String)tempnames.elementAt(i));
	  return;
	else
	  interpreter.deleteFile((String)tempnames.elementAt(i));
      }
    }
  }      
  /**
   * This method is used with getMiniSet(). It accepts a relation
   * with a column of boolean 'flag'. Only the smallest sub-relation
   * set's 'flag' fields are set to 'true'. 'compareTwoRows()' will
   * compare the samllest set again each of other sets (who has same
   * '.id' values, and reset the 'true' flag field to 'false' if same
   * rwo is not found in any of the other sub-sets.  This is used
   * by the new technology (nt == true) to actualize the following
   * virtual domain:    "let x be red ijoin of S;"
   *  E.g.  (A,   B,   C,  .id   flag )
   *         1    2    3    1    false
   *         2    1    2    1    false
   *         4    2    1    2    true    <--- mini-set
   *         2    3    1    3    false
   *         5    4    2    3    false
   * Control flow is as follows:
   *        loop the mini-set and get one record from it (i)
   *            loop the whole relation (0 -> numtuples) (j)
   *                curid <= id[j]
   *                flag <= false
   *                while same id
   *                     if two tuples have same value
   *                          flag <= true
   *                if falg equals true
   *                    continue next group with same id
   *                else
   *                    curid tuple (from mini-set) is not in current tuple
   *                    so curid tuple shouldn't be in ijoin result
   *                    break (no need to check further)
   */
  private void compareTwoRows(Relation rel)
       throws InterpretError
  {
    long curid = 0L;
    boolean flag = false;
    int pos = rel.numsortattrs;
    int len = rel.capacity;
    int[] types = rel.toTypes(rel.domains);
    int leng = rel.domains.length - 2;
    boolean[] flags = new boolean[leng];
    
    
    //      Domain iddom = domtable.getDom(".id");
    //REVISIT: Zhongxia, I used true but check it please
    Domain iddom = env.lookupDom(".id", true);
    
    int index = rel.indexOfDomain(iddom);
    
    for(int i = pos; i < pos + len; i++){
      if(((byte[])rel.data[rel.domains.length-1])[i] == 1){
	
	int j = 0;
	while(j < rel.numtuples){
	  curid = ((long[])rel.data[index])[j];
	  flag = false;
	  
	  while(j < rel.numtuples && curid == ((long[])rel.data[index])[j]){
	    
	    if(rel.compareTwoRows(rel.data, i, rel.data, j,
				  leng, types, flags, rel.domains, rel.domains) == EQ){
	      flag = true;
	    }
	    j++;
	  }
	  if(!flag){
	    ((byte[])rel.data[rel.domains.length-1])[i] = 0;
	    break;
	  }
	}
      }
    }
  }     
  private int equivSubrel(Relation destrel, int[] vals)
  {
 if(DEBUG)  
  Global.println("equivSubrel");
    return 10;
  }         
  private void fillSurrogate(Relation rel, long ref, long sur)
  {
    
    //      int index = rel.indexOfDomain(domtable.getDom(".id"));
    //REVISIT: Zhongxia, I used true but check it please
    int index = rel.indexOfDomain(env.lookupDom(".id", true));
    
    
    for(int i = 0; i < rel.numtuples; i++){
      if(((long[])rel.data[index])[i] == ref)
	((long[])rel.data[index])[i] = sur;
    }
  }      
  /** 
   * Get those domains in the by-list of equivalent operations.
   * The domain is stored in alldoms vector and might be some
   * intermediate domain generated on the fly when building the tree.
   */
  private Domain getDom(String name)
  {
    for(int i = 0; i < alldoms.size(); i++){
      Domain d = (Domain)alldoms.elementAt(i);
      if(name.compareTo(d.name) == 0){
	return d;
      }
    }
    return null;
  }      
  /**
   * getMiniSet() receives two arguments, i.e. 'r': the bit
   * source relation, 'doms': the destination domain array. It will
   * return a relation which has the same tuple# as 'r', but arranged
   * in the domain array 'doms'. It also finds out the '.id' value
   * of the smallest sub-relation set (with same '.id') in 'r', and
   * put the value in the returned relation's 'numsortattrs' field;
   * the #tuples of the smallest sub-relation set is put in the
   * returned relation's 'capacity' field.
   */
  private Relation getMiniSet(Relation srcr, Domain[] srcdoms)
       throws InterpretError
  {
    int c1, c0, pos, oldi;
    long oid, tid;
    
    // Domain iddom = domtable.getDom(".id");
    //REVISIT: Zhongxia, I used true but check it please
    Domain iddom = env.lookupDom(".id", true);
    
    
    c0 = Integer.MAX_VALUE;
    c1 = 0;
    int index = srcr.indexOfDomain(iddom);
    if(index == ERROR)   // Should never happen.
      throw new InterpretError("no domain '.id' in "+srcr.name);
    
    oid = tid = ((long[])srcr.data[index])[0];
    pos = oldi = 0;
    for(int i = 0; i < srcr.numtuples; i++){
      if(tid != ((long[])srcr.data[index])[i]){
	if(c0 > c1){
	  c0 = c1;      // reset current minimum counter.
	  oid = tid;    // remember the .id value.
	  pos = oldi;
	}
	oldi = i;
	tid = ((long[])srcr.data[index])[i];
	c1 = 1;          // reset the current counter.
      }else
	c1++;
    }
    // Process the last set of tuples.
    if(c0 > c1){
      c0 = c1;      // reset current minimum counter.
      oid = tid;    // remember the .id value.
    }
    //System.out.println("smallest id: "+oid+", numtuples: "+c0);
    
    // Create the result relation.
    Domain[] doms = new Domain[srcdoms.length+1];
    
    // Take off the '.id' domain.
    int j = 0;
    for(int i = 0; i < srcdoms.length; i++){
      if(srcdoms[i] != iddom)
	doms[j++] = srcdoms[i];
    }
    
    // Put the '.id' domain to the end of the domain list, since
    // Biao's compareTwoRows() routine only compares starting n
    // domains.
    doms[j++] = iddom;   // put '.id' domain to the end.
    
    //      doms[j] = domtable.getDom(".bool");
    //REVISIT: Zhongxia, I used true but check it please
    doms[j] = env.lookupDom(".bool", true);
    
    
    Object[] data = srcr.createDataArray(doms, srcr.numtuples);
    for(int i = 0; i < doms.length - 1; i++)
      data[i] = srcr.data[srcr.indexOfDomain(doms[i])];

    srcr.data = data;
    srcr.domains = doms;
    srcr.numattrs = doms.length;
    
    // temparary information pass out.
    srcr.numsortattrs = pos;
    srcr.capacity = c0;
    
    return srcr;
  }      
  /**
   * Return an un-named memory-version relation (no saving to disk).
   * Usual calling procedure :
   *          actualizer = new Actualizer(...);
   *          actualizer.actualize(virdomarray);
   *          destrel = actualizer.getRel();
   */
  public Relation getRel(){
    // Return the destination relation (might be null).
    return(destrel);
  }      
  /**
   * Return a named/disk-saved relation.
   * Usual calling procedure :
   *          actualizer = new Actualizer(...);
   *          actualizer.actualize(virdomarray);
   *          destrel = actualizer.getRel(destrelname);
   */
  public Relation getRel(String relname)
       throws InterpretError
  {
    if(destrel == null) return null;
    
    destrel.setName(relname);
    
    // Make the new relation appear in the reltable.
    //      reltable.addRel(destrel);
    env.put(destrel);
    
    destrel.save();   // REVISIT: for print relation use only.
    
    // Return the resulting relation.
    return(destrel);
  }      

   /*
    * Add new temp relation names to Vectors provided for IDList virtual domain 
    * tree specified in node, and
    * add these names to tempnames vector for future clean-up,
    * if they doesn't exist already in the Vectors.
    * This is a self-recursive method to process an entire virtual IDList
    * domain.
    * 
    * JRelixNS modified/enhanced this method:
    *   Allow for Relation as a valid
    *   identifier node where only domains where allowed previously.
    */
  private void preProcIDListDom(SimpleNode node, Vector doms, Vector rname,
				Vector rindx) throws InterpretError
  {
//    System.out.println("preIDL . Nodename " +node.name );
//    System.out.println("preIDL . type " +node.type );
//    System.out.println("preIDl . opcode " +node.opcode );
//    System.out.println("preIDL . opbits " +node.bits );
    
    switch(node.type){
    case OP_IDENTIFIER:
      String name;
      
      // REVISIT: should modify this to fit for top-level relations.
      
      //         Domain d = domtable.getDom(node.name);
      //REVISIT: Zhongxia, I used true but check it please
      Domain d = env.lookupDom(node.name, true);
      
      // START Modified Since jRelixNS.
      		 if(Relation.isSigmaJoinImplemented != true){ //Andy May 12, 2002
//      if(env.isSigmaJoinImplemented != true){
	if(d == null)   // Shouldn't happen.
	  throw new InterpretError("preProcIDListDom: domain '" +
				   node.name + "' not found.");
	
	Integer indx = new Integer(alldoms.indexOf(d));
	if(!rindx.contains(indx)){
	  name = interpreter.nextTempName();
	  doms.addElement(d.name);
	  rname.addElement(name);
	  rindx.addElement(indx);
	  
	  // Add this name to tempnames vector for future clean-up.
	  tempnames.addElement(name);
	}else
	  name = (String)rname.elementAt(rindx.indexOf(indx));
	
      } else {
	// Modified Since jRelixNS: allow for Relation as a valid
	// identifier node where only domains where allowed previously.
	if(d == null)
	  {
	    if (env.lookupRel(node.name, true) == null)//Shouldn't happen.
	      throw new InterpretError("preProcIDListDom: domain-relation '" +
				       node.name + "' not found.");
	  }
	else
	  {
	    // Add a new temp relation name if one doesn't exist already.
	    Integer indx = new Integer(alldoms.indexOf(d));
	    if(!rindx.contains(indx)){
	      name = interpreter.nextTempName();
	      doms.addElement(d.name); // Same as identifier's node.name.
	      rname.addElement(name);
	      rindx.addElement(indx);
	      
			  // Add this name to tempnames vector for future clean-up.
	      tempnames.addElement(name);
	    }else
			   name = (String)rname.elementAt(rindx.indexOf(indx));
	  }
      }
      // End Modified Since jRelixNS
      //node.name = name;
      break;
    case OP_JOINOPERATOR:
      SimpleNode left = (SimpleNode)node.jjtGetChild(0);
		 SimpleNode right = null;
		 if(node.jjtGetNumChildren() == 2)
		   right = (SimpleNode)node.jjtGetChild(1);
		 else
		   right = (SimpleNode)node.jjtGetChild(3);
		 preProcIDListDom(left, doms, rname, rindx);
		 preProcIDListDom(right, doms, rname, rindx);
		 break;
    case OP_SELECT:
    case OP_TSELECT:
    case OP_PROJECT:{
      // Get the source relation node.
      SimpleNode srcnode = (SimpleNode)node.jjtGetChild(1);
      preProcIDListDom(srcnode, doms, rname, rindx);
      break;
    }
    case OP_QSELECT:{ //Added by hongyu Sept 21,01
      SimpleNode srcnode = null;
      if(node.jjtGetNumChildren() == 2)
	srcnode = (SimpleNode)node.jjtGetChild(1);
      else srcnode = (SimpleNode)node.jjtGetChild(2);
      preProcIDListDom(srcnode, doms, rname, rindx);
      break;
    }
    // End of Hongyu addition
    case OP_RED:
    case OP_EQUIV:
    case OP_FUN://From sungsoo. Added by hongyu. Feb 2002
    case OP_PAR://From sungsoo. Added by hongyu. Feb 2002
      {
	// Get the source relation node.
	//if added by Yi Zheng July 8, 2002
	if(node.opcode != OP_REDFUNCALL)
	{
		SimpleNode srcnode = (SimpleNode)node.jjtGetChild(0);
		preProcIDListDom(srcnode, doms, rname, rindx);
	}
	else
	{
		SimpleNode srcnode = (SimpleNode)node.jjtGetChild(1);
		preProcIDListDom(srcnode, doms, rname, rindx);
	}
	break;
      }
     case OP_ARRAY: {	//Yi Zheng June 24, 2002
     	SimpleNode srcnode = (SimpleNode) node.jjtGetChild(0);
     	preProcIDListDom(srcnode, doms, rname, rindx);
     	break;
     
     }//end of addition June 24, 2002
    default:
      throw new InterpretError("preProcIDListDom: to be implemented.");
    }
  }      
  /**
   * For an IDLIST-typed virtual domain, we must generate additional
   * domain nodes to hold the intermediate relations. We put these 
   * relations to a hashtable and put the intermediate domains in the
   * nodes' info fields.
   * Note, flag == true : called by nt==true (top-level approach)
   *       flag == false : called by tuple-tuple approach, i.e. we only
   *           need the resulting domlist but don't want to take off
   *           intermediate nodes in this routine.
   */
  private Relation processIDListDom(SimpleNode node, Hashtable htable,
				    boolean flag)throws InterpretError
  {
//    System.out.println("IDL . Nodename " +node.name );
//    System.out.println("IDL . type " +node.type );
//    System.out.println("IDL . opcode " +node.opcode );
//    System.out.println("IDL . opbits " +node.bits );
    
    switch(node.type){
    case OP_IDENTIFIER:
      String rname = node.name;
      
      //Relation r = reltable.getRel("."+rname);
      //REVISIT: Zhongxia, I used true but check it please
      Relation r = env.lookupRel("."+rname, true);
      
      if(r == null){
	
	// if((r = reltable.getRel(rname)) == null)
	//REVISIT: Zhongxia, I used true but check it please
	if((r = env.lookupRel(rname, true)) == null)
	  
	  // should never happen..
	  throw new InterpretError("source relation '" +
				   rname + "' not found.");
      }
      return r;
    case OP_JOINOPERATOR:{
      Relation r1, r2;
      Domain[] doms;
      
      if(node.jjtGetNumChildren() == 2){
	// No join attributes presented.
	r1 = processIDListDom((SimpleNode)node.jjtGetChild(0), htable, flag);
	r2 = processIDListDom((SimpleNode)node.jjtGetChild(1), htable, flag);
	doms = r1.mujoinType(node.opcode, null, r2, null);
      }
      else{
	// Join attributes are given.
	r1 = processIDListDom((SimpleNode)node.jjtGetChild(0), htable, flag);
	r2 = processIDListDom((SimpleNode)node.jjtGetChild(3), htable, flag);
	Domain[] doms1, doms2;
	Vector domsv1 = new Vector();
	Vector domsv2 = new Vector();
	
	// Add .id to vectors.
	
	// domsv1.addElement(domtable.getDom(".id"));
	//REVISIT: Zhongxia, I used true but check it please
	domsv1.addElement(env.lookupDom(".id", true));
	
	// domsv2.addElement(domtable.getDom(".id"));
	//REVISIT: Zhongxia, I used true but check it please
	domsv2.addElement(env.lookupDom(".id", true));
	
	
	SimpleNode expr1, expr2;
	expr1 = (SimpleNode)node.jjtGetChild(1);
	expr2 = (SimpleNode)node.jjtGetChild(2);
	
	if(expr1.jjtGetNumChildren() != expr2.jjtGetNumChildren())
	  // Should never happen...
	  throw new InterpretError("different number of join attributes.");
	
	for(int i = 0; i < expr1.jjtGetNumChildren(); i++){
	  String name = ((SimpleNode)expr1.jjtGetChild(i)).name;
	  
	  //               Domain d = domtable.getDom(name);
	  //REVISIT: Zhongxia, I used true but check it please
	  Domain d = env.lookupDom(name, true);
	  
	  if(d != null) domsv1.addElement(d);
	  else throw new InterpretError("domain '"+d+"' not found.");
	  
	  name = ((SimpleNode)expr2.jjtGetChild(i)).name;
	  
	  // d = domtable.getDom(name);
	  //REVISIT: Zhongxia, I used true but check it please
	  d = env.lookupDom(name, true);
	  
	  if(d != null) domsv2.addElement(d);
	  else throw new InterpretError("domain '"+d+"' not found.");
	}
	doms1 = new Domain[domsv1.size()];
	doms2 = new Domain[domsv2.size()];
	domsv1.copyInto(doms1);
	domsv2.copyInto(doms2);
	
	doms = r1.mujoinType(node.opcode, doms1, r2, doms2);
      }
      
      if(node.info == null && flag){
	// We have to make a intermediate domain for this column.
	String newdomname = ""+fdomcounter++;
	
	Domain[] darray = new Domain[doms.length - 1];
	for(int i = 0; i < darray.length; i++)
	  darray[i] = doms[i+1];
	
	// Have to insert this intermediate domains into domtable, since
	// this is required by Biao's evaluateTLExpression() routine in
	// Interpreter. This is however too bad actually. ;-( Don't
	// forget to remove it after actualization. See cleanup().
	//            domtable.addDom(newdomname, IDLIST, darray);
	Domain newDom = new Domain(newdomname, IDLIST, darray, env);
	env.put(newDom);
	
	// Generate a new intermediate column in relation. Note, we
	// must use THIS domain's reference, or the domain comparison
	// routine in Relation.class won't work correctly.
	
	// Domain newdom = domtable.getDom(newdomname);
	//REVISIT: Zhongxia, I used true but check it please
	Domain newdom = env.lookupDom(newdomname, true);
	
	alldoms.addElement(newdom);
	
	// There must be a relation associated with the IDLIST domain..
	Relation newRel = new Relation("."+newdomname, doms, 0, 
				       doms.length, 0, env);
	htable.put(newdom, newRel);
	
	tempnames.addElement(newdomname);
	
	// Put the new domain in this node's 'info' field.
	node.info = newdom;
	return newRel;
      }else{
	return new Relation("", doms, 0, doms.length, 0, env);
      }
    }
    case OP_RED:
    case OP_EQUIV:
      // REVISIT: do we need to add the intermediate domain in
      // domtable here for these two cases?
      // case OP_FUN and case OP_PAR? hognyu Feb 2002
    case OP_SELECT:
      
    case OP_PROJECT:
    case OP_TSELECT:
      // REVISIT: we need to analyze the source relation part, as that
      // can be a multi-level join operations' result, i.e.
      //     let x be [A] in (S ijion T ijoin U);
      // or  let x be  where A=3 in (S ijoin T ijoin U);
      Relation srcr = null;
      
      // Get the source relation node.
      SimpleNode srcnode = null;
      if(node.type == OP_RED || node.type == OP_EQUIV)
      {
      	if(node.opcode != OP_REDFUNCALL)
			srcnode = ((SimpleNode)node.jjtGetChild(0)).jjtDuplicate();
		else
			srcnode = ((SimpleNode)node.jjtGetChild(1)).jjtDuplicate();
	  }	
      else
      {
      	if(node.opcode != OP_REDFUNCALL)
			srcnode = ((SimpleNode)node.jjtGetChild(1)).jjtDuplicate();
		else
			srcnode = ((SimpleNode)node.jjtGetChild(2)).jjtDuplicate();
		
	  }
      
      // Get the projection id list if it's OP_PROJECT or OP_TSELECT.
      Domain[] prjlist = null;
      if(node.type == OP_PROJECT || node.type == OP_TSELECT){
	SimpleNode exprlist = (SimpleNode)node.jjtGetChild(0).jjtGetChild(0);
	
	// prjlist should contain the '.id' domain.
	prjlist = new Domain[exprlist.jjtGetNumChildren() + 1];
	
	//            prjlist[0] = domtable.getDom(".id");
	//REVISIT: Zhongxia, I used true but check it please
	prjlist[0] = env.lookupDom(".id", true);
	
	for(int i = 0; i < exprlist.jjtGetNumChildren(); i++){
	  SimpleNode prjid = (SimpleNode)exprlist.jjtGetChild(i);
	  if(prjid.type != OP_IDENTIFIER)
	    throw new InterpretError("invalid projection expression.");
	  
	  
	  //               Domain d = domtable.getDom(prjid.name);
	  //REVISIT: Zhongxia, I used true but check it please
	  Domain d = env.lookupDom(prjid.name, true);
	  
	  if(d != null)
	    prjlist[i+1] = d;
	  else  // Should never happen..
	    throw new InterpretError("domain '"+ prjid.name +
				     "' in the projection list not found.");
	}
      }
      
      // If this node is not an IDENTIFIER, we must replace it with
      // a new intermediate node/domain, and actualize the new
      // domain first. i.e.
      //      Project                 Project           (tmp)
      //      /     \                  /    \                   
      //  Projector  Join        Projector  Identifier       Join  
      //    /        /  \   =>       /        \              /   \
      // ExprList  Join  U       ExprList     tmp          Join   U
      //  /  \     /  \           /   \                    /  \
      // A    B   S    T         A     B                  S    T
      if(srcnode.type == OP_IDENTIFIER || !flag){
	srcr = processIDListDom(srcnode, htable, flag);
      }else if(srcnode.type == OP_JOINOPERATOR || srcnode.type == OP_SELECT
	       || srcnode.type == OP_PROJECT || srcnode.type == OP_TSELECT){
	// we need to take off the srcnode and put it in nestdoms.
	String newdomname = ""+fdomcounter++;
	
	// Each nested domain has an hashtable(newdom, newrel)
	// Refer this part to that constructor Actualizer().
	Hashtable domrelhash = new Hashtable();
	
	// Get the idlist of this domain.
	srcnode.info = newdomname;  // temperarily put an Object here,
	// otherwise, processIDListDom() will generate a new domain
	// for this node.
	srcr = processIDListDom(srcnode, domrelhash, flag);
	Domain darray[] = new Domain[srcr.domains.length-1];
	for(int i=0; i < srcr.domains.length-1; i++)
	  darray[i] = srcr.domains[i+1];
	
	// Have to insert this intermediate domain into domtable,
	// since its required by Biao's evaluateTLExpressin()
	// routine in Interpreter. Don't forget to remove it after
	// actualization. See cleanup().
	//            domtable.addDom(newdomname, IDLIST, darray);
	Domain newDom = new Domain(newdomname, IDLIST, darray, env);
	env.put(newDom);

	// Generate a new inermediate column in relation. Note,
	// we must use THIS domain's reference. Refer to
	// processIDList().
	
	// Domain newdom = domtable.getDom(newdomname);
	//REVISIT: Zhongxia, I used true but check it please
	Domain newdom = env.lookupDom(newdomname, true);
	
	
	srcnode.info = newdom;
	
	// REVISIT: do we need the following 2 lines?
	
	// Relation rel = reltable.getRel("."+newdomname);
	//REVISIT: Zhongxia, I used true but check it please
	Relation rel = env.lookupRel("."+newdomname, true);
	
	domrelhash.put(newdom, rel);
	
	alldoms.addElement(newdom);
	tempnames.addElement(newdom.name);
	nestdoms.addElement(newdom);
	vvtree.put(newdom, srcnode);
	nesthashs.put(newdom, domrelhash);
	
	// Current node needs to stop here. i.e. we need to generate
	// a 'domain' node of 'Identifier' type.
	SimpleNode leaf = SimpleNode.jjtCreateNode("Identifier", 
						   OP_IDENTIFIER, OP_IDENTIFIER, newdomname);
	leaf.bits = alldoms.indexOf(newdom);
	
	if(node.type == OP_RED || node.type == OP_EQUIV)
	{
		if(node.opcode != OP_REDFUNCALL)
	  	node.jjtReplaceChild(0, leaf);
	  	else
	  	node.jjtReplaceChild(1, leaf);
	  	
	 }
	else
	{
		if(node.opcode != OP_REDFUNCALL)
	  	node.jjtReplaceChild(1, leaf);
	  	else
	  	node.jjtReplaceChild(2, leaf);
	 }
      }else
	throw new InterpretError("processIDListDom(): operation not implemented.");
      
      // Return the appropriate relation object.
      if(node.type == OP_PROJECT || node.type == OP_TSELECT)
			return new Relation("", prjlist, 0, prjlist.length, 0, env);
      else
			return srcr;
    default:
      throw new InterpretError("processIDListDom: to be implemented.");
    }
  }
  
  private int processRedNode(SimpleNode node, SimpleNode n, int j)
       throws InterpretError{
	 
	 // Initialize the summary field.
	 n.info = null;
	 
	 // We need to take off this node and put it in vertdom.
	 SimpleNode newnode = n.jjtDuplicate();
	 
	 if(newnode.type == OP_EQUIV){
	   // Get the child of expression list.
	   //if added by Yi Zheng July 8, 2002
	   if(newnode.opcode != OP_REDFUNCALL)
	   {
	   	SimpleNode child = (SimpleNode)newnode.jjtGetChild(1);
	   	newnode.jjtReplaceChild(1, child.jjtDuplicate());
	   }
	   else
	   {
	   	SimpleNode child = (SimpleNode)newnode.jjtGetChild(2);
	   	newnode.jjtReplaceChild(2, child.jjtDuplicate());
	   }
	 }
	 
	 // Will actualize this special domain in first pass.
	 // REVISIT: the 'type' value is not available at the
	 // moment, so we use '0'. Since no other domains will be
	 // defined on this 'newdom' ('cos it's a temparary
	 // domain), it will be OK even we use a false 'type' value.
	 String newdomname = "" + fdomcounter++;
	 Domain newdom = new Domain(newdomname, 0, 0, newnode); //note newnode is the declaration itself, not let .. be .., so it can have more than one children
	 
	 // This must be a new domain so we don't need to check if
	 // it's contained in the vertdom hashtable.
	  vertdom.put(newdom, new Integer(vertlevel));
	  
	  if(maxvertlevel < ++vertlevel)
	    maxvertlevel = vertlevel;
	  
	  // Get newtype since typeArray[] needs it.
	  //if added by Yi Zheng July 8, 2002
	  SimpleNode child;
	  if(newnode.opcode != OP_REDFUNCALL)
	  {
	  	child = ((SimpleNode)newnode.jjtGetChild(0)).jjtDuplicate();
	  	newnode.jjtReplaceChild(0, child);
	  }
	  else
	  {
	  	child = ((SimpleNode)newnode.jjtGetChild(1)).jjtDuplicate();
	  	newnode.jjtReplaceChild(1, child);
	  }
	  int newtype = buildTree(child);
	  //newtype = builtTree(newnode);
	  
	  // If it's reductions on nested domains, we have more work to
	  // do. We need to generate temperary IDLIST domain in domtable.
	  // This should also involves creating a "." relation in reltable.
	  // This is needed for tuple-by-tuple approach since we can't
	  // pass all the syntax tree to top-level evaluator without
	  // checking if there exists reduction nodes. Reduction nodes
	  // must be taken off from the tree.
	  if(newtype == IDLIST && !nt){
	    Relation r = processIDListDom(child, null, false);
	    Domain[] darray = new Domain[r.domains.length - 1];
	    for(int k = 0; k < darray.length; k++){
	      darray[k] = r.domains[k + 1];
	    }
	    
	    // Have to insert this intermediate domain into domtable,
	    // as this is required by our reduction routine.
	    // domtable.addDom(newdomname, IDLIST, darray);
	    Domain newDom = new Domain(newdomname, IDLIST, darray, env);
	    env.put(newDom);
	    
	    // Have to replace the same entry in vertdom hashtable,
	    // and replace newdom with *this* domain.
	    Integer lv = (Integer) vertdom.get(newdom); // save the level.
	    vertdom.remove(newdom);     // first remove it from vertdom.
	    
	    // newdom = domtable.getDom(newdomname);  // reset newdom.
	    //REVISIT: Zhongxia, I used true but check it please
	    newdom = env.lookupDom(newdomname, true);  // reset newdom.
	    
	    vertdom.put(newdom, lv);    // replace the old newdom.
	    
	    // Don't forget to remove this domain from domtable.
	    tempnames.addElement(newdomname);
	  }
	  
	  newdom.type = newtype;    // set the correct type value.
	  
	  // reset the vertlevel.
	  vertlevel--;
	  
	  alldoms.addElement(newdom);
	  vvtree.put(newdom, newnode);
	  
	  // Current node needs to stop here, i.e. we need to generate a 
	  // 'domain' node (of 'Identifier' type).
	  SimpleNode leaf = new SimpleNode("Identifier");
	  leaf.type = leaf.opcode = OP_IDENTIFIER;
	  
	  // Will search this domain column later.
	  leaf.name = newdom.name;  
	  leaf.bits = alldoms.indexOf(newdom);
	  node.jjtReplaceChild(j, leaf);
	  
	  return newtype;
  }      
  /**
   * Reset the value of node.bits since we are storing the position
   * of attributes here and since that positions may be changed by
   * equivalent operations' sorting.
   */
  private void resetNodeBits(SimpleNode node)
       throws InterpretError
  {
  	//System.out.println("in resetNodeBits for " + node.name);
  	//Utility.dump(node, "-->");
  	
    int i;
    if(node.type == OP_IDENTIFIER){
      for(i = 0; i < alldoms.size(); i++){
	Domain d = (Domain)alldoms.elementAt(i);
	//System.out.println("comparing with domain " + d.name);
	if(d.name.equals(node.name))
	  break;
      }
      if(i == alldoms.size()){
	// Should never happen.
	throw new InterpretError("resetNodeBits: system error.");
      }
      //System.out.println("Setting node.bits at 3419 to " + i);
      node.bits = i;
    }else{
      for(i = 0; i < node.jjtGetNumChildren(); i++)
    	if(node.opcode != OP_COMPCALL)  //added by Yi Zheng July 9, 2002 to bypass OP_REDFUNCALL
			resetNodeBits((SimpleNode)node.jjtGetChild(i));
    }
  } 
  
  /* 
   Given an expression tree and an opcode of an operator, this function 
   returns the node that contains this operator or returns null in case the operator 
   is not found in the expression tree.  - Andrey Rozenberg
 */
	private SimpleNode IsInTree(SimpleNode node, int op_type)
	{
		
		Vector tnodes = new Vector();
		tnodes.addElement(node);

		while(tnodes.size()!=0)
		{
			SimpleNode tNode = (SimpleNode)tnodes.lastElement();
			tnodes.removeElementAt(tnodes.size()-1);
			if (tNode.type==op_type)
				return tNode;
			else
			{
				int NumOfChildren=tNode.jjtGetNumChildren();
				for (int i=NumOfChildren;i>0;i--)
					tnodes.addElement(tNode.jjtGetChild(i-1));
			}	
		}
		
		return null;		
	}
     /**
   * Returns boolean value converted from a
   * boolean compatible nullary relation/IDLIST cell.
   * If relation is not boolean compatible, rel.boolCompatibleValue()
   * will throw an InterpretError.
   * @return  a boolean true if nullary relation contains true, false otherwise.
   * @since jRelixNS
   */
  private boolean actBoolRelCell(SimpleNode node, Vector vdom, Vector rname,
				 Vector rindx, String tname0, String tname1)
       throws InterpretError{
/* //	 System.out.println(" actBoolRelCell");
	 // Variables for IDLIST-typed domain and when nt == false;
	 if(!nt && (rname ==null)){
//	   System.out.println(" (!nt && (rname ==null)) ");
	   if (rname ==null) rname = new Vector();
	   if (rindx ==null) rindx = new Vector();
	   if (vdom  ==null) vdom  = new Vector();
	   if (tname0==null) tname0= interpreter.nextTempName();
	   if (tname1==null) tname1= interpreter.nextTempName();
	   
	   this.tempnames.addElement(tname0);//remember to clear this temp rel.
	   this.tempnames.addElement(tname1);//remember to clear this temp rel.
	   // Fill the vectors.
	   if (NRenv.row == 1){
//	     System.out.println(" NRenv.row == 1 ");
	     preProcIDListDom(node, vdom, rname, rindx);
	   }
	 }
	 
	 Relation rel = actRelCell(node, vdom, rname, rindx, this.relhtbl,
				   tname0, tname1);
	 return rel.boolCompatibleValue();*/
	 
	 //Andy May 12, 2002
	      if (tname0==null) 
       tname0= interpreter.nextTempName();
     this.tempnames.addElement(tname0);//remember to clear this temp rel.
     Relation rel = actRelCell(node, vdom, rname, rindx, this.relhtbl,
       tname0, tname1);
     return rel.boolCompatibleValue();

  }
 /* 
  private void preProcBoolean(SimpleNode node, Vector doms, Vector rname,
			      Vector rindx) throws InterpretError
  {
    switch(node.type){
    case OP_BIOPERATOR:
      SimpleNode left = (SimpleNode)node.jjtGetChild(0);
      SimpleNode right = (SimpleNode)node.jjtGetChild(1);
      
      switch(node.opcode){
      case OP_OR:
      case OP_AND:
	preProcBoolean(left,doms, rname, rindx); 
	preProcBoolean(right,doms, rname, rindx);
	break;
      case OP_EQ:
      case OP_NEQ:
      case OP_GT:
      case OP_LT:
      case OP_GE:
      case OP_LE:
	break;
      default:
	throw new InterpretError("preProcBoolean: OP_BIO to be implemented..");
      }
      return;
    case OP_UOPERATOR:
      preProcBoolean((SimpleNode)node.jjtGetChild(0),doms, rname, rindx);
    case OP_IDENTIFIER:
      // this node must of boolean (byte) type.
      return;
    case OP_LITERAL:
      return;
    case OP_RED:
    case OP_EQUIV://should added sungsoo's code hongyu Feb 2002
      // case OP_FUN and OP_PAR? 
      preProcBoolean((SimpleNode)node.jjtGetChild(0),doms, rname, rindx);
    case OP_IFTHENELSE:
      preProcBoolean((SimpleNode)node.jjtGetChild(0),doms, rname, rindx);
      preProcBoolean((SimpleNode)node.jjtGetChild(1),doms, rname, rindx);
      preProcBoolean((SimpleNode)node.jjtGetChild(2),doms, rname, rindx);
      return;
    case IDLIST:
      // throw new InterpretError("preProcBoolean: IDLIST node.type to be implemented..");
    default:
      if(!nt){
	preProcIDListDom(node, doms, rname, rindx);
      }else
	throw new InterpretError("preProcBoolean: nt not implemented for BOOLEAN-IDLIST node...");
    }
  }   
  */ //Andy May 12, 2002
  /* The followings are songsoo's code. Added by hongyu Feb 2002 */
  
  //Partial Functional Mapping Implementation by Sungsoo Kang
  private void actualizeParFun(Domain dom) throws InterpretError
  {
    int i, oldpos, virtype=dom.type;
    SimpleNode virtree = (SimpleNode)vvtree.get(dom);
    
    String tname0=null, tname1=null;
    Hashtable relhtbl=null;
    Vector rname = null;
    Vector rindx = null;
    Vector vdom = null;
    Relation dest = null;
    
    //Fisrt we need to sort on the by-list and order-list
    Vector bydoms = new Vector();
    Vector orderdoms = new Vector();
    Vector byorderdoms = new Vector();
    boolean updatesurrogate = true;
    //Generate orderdoms list vector
    //Generate dydoms list vector
    //if added by Yi Zheng July 9, 2002
    SimpleNode exprlist = null;
    if(virtree.opcode != OP_REDFUNCALL)
    	exprlist = (SimpleNode)virtree.jjtGetChild(2);
    else
    	exprlist = (SimpleNode)virtree.jjtGetChild(3);
    	
    for(i=0;i<exprlist.jjtGetNumChildren();i++)
      {
		SimpleNode child = (SimpleNode) exprlist.jjtGetChild(i);
		Domain d = getDom(child.name);
		if(d!=null)
		  {
		    
		    bydoms.addElement(d);
		    byorderdoms.addElement(d);
		  }
		else
		  throw new InterpretError("actualizeParFunc:domain not found.");
      }
    
    //if added by Yi Zheng July  9, 2002
    SimpleNode orderexprlist = null;
    if(virtree.opcode != OP_REDFUNCALL)
    	orderexprlist = (SimpleNode)virtree.jjtGetChild(1);
    else
     	orderexprlist = (SimpleNode)virtree.jjtGetChild(2);
   
    for(i=0;i<orderexprlist.jjtGetNumChildren();i++)
      {
		SimpleNode orderchild = (SimpleNode)orderexprlist.jjtGetChild(i);
		Domain d = getDom(orderchild.name);
		if(d!=null)
		  {
		    orderdoms.addElement(d);
		    byorderdoms.addElement(d);
		  }
		else
		  throw new InterpretError("ActualizeParFunc:order domain not found");
      }
    
    
    //Generate byroderdoms array for sorting
    
    Domain[] byorderarray = new Domain[byorderdoms.size()];
    byorderdoms.copyInto(byorderarray);
    
    Domain[] byarray = new Domain[bydoms.size()];
    bydoms.copyInto(byarray);
    
    Domain[] orderarray = new Domain[orderdoms.size()];
    orderdoms.copyInto(orderarray);
    
//    Global.println("age " + age()+": start sort");
    destrel.sort(byorderarray);
//    Global.println("age "+age()+":finish sort");
    
    for(i=byorderarray.length-1;i>=0;i--)
      {
	alldoms.removeElement(byorderarray[i]);
	alldoms.insertElementAt(byorderarray[i],0);
      }
    
    for(Enumeration e = vvtree.elements(); e.hasMoreElements();)
      {
	resetNodeBits((SimpleNode)e.nextElement());
      }
    
    
    if(virtype == IDLIST && !nt)
      {
	relhtbl = new Hashtable();
	rname = new Vector();
	rindx = new Vector();
	vdom = new Vector();
	tname0=interpreter.nextTempName();
	tname1=interpreter.nextTempName();
	tempnames.addElement(tname0);
	tempnames.addElement(tname1);
	
	if((dest = env.lookupRel("."+dom.name, true))==null)
	  throw new InterpretError("actualizseParFunc: relation ./"+dom.name+"'not found.");
	preProcIDListDom(virtree, vdom, rname, rindx);
      }
    NREnvironment NRenv = env.addNRFrame(destrel);
    //Remember the position of domain for the future use
    int pos = alldoms.indexOf(dom);
    
     //bug fix by Yi Zheng June 23, 2002
    virtree.bits = pos;
    //end of Yi's fix June 23,2002

    //Remember the position of by list domain for the future use
    int bypos[] = new int[byarray.length];
    for(i=0;i<byarray.length;i++)
      bypos[i]=alldoms.indexOf(byarray[i]);
    
    //Remember the position of order list domain for the future use
    int orderpos[]=new int[orderarray.length];
    for(i=0;i<orderarray.length;i++)
      orderpos[i] = alldoms.indexOf(orderarray[i]);
    
    //Get the initial values for sort attributes
    currow = oldpos = 0;
    
    
    //Initial value for bymemory
    Vector bymemory = new Vector();
    
    for(i=0;i<byarray.length;i++)
      {
	switch(byarray[i].type)
	  {
	case NUMERIC:
	    bymemory.addElement(((number[])destrel.data[bypos[i]])[currow]);
	    break;
	    
	  case INTEGER:
	  case SHORT:
	  case BOOLEAN:
	    Integer intval = new Integer(((int[])destrel.data[bypos[i]])[currow]);
	    bymemory.addElement(intval);
	    break;
	    
	  case LONG:
	    Long longval = new Long(((long[])destrel.data[bypos[i]])[currow]);
	    bymemory.addElement(longval);
	    break;
	    
	  case FLOAT:
	    Float floatval = new Float(((float[])destrel.data[bypos[i]])[currow]);
	    bymemory.addElement(floatval);
	    break;
	    
	  case DOUBLE:
	    Double doubleval = new Double(((double[])destrel.data[bypos[i]])[currow]);
	    bymemory.addElement(doubleval);
	    break;
	  case TEXT:
	  case STRING:
	  case EXPR:
	    bymemory.addElement(((String[])destrel.data[bypos[i]])[currow]);
	    break;
	  case IDLIST:
	    //?? hongyu Feb 2002
	    break;
	  default:
	    throw new InterpretError("actualizeParFunc(): to be implemented...");
	  }
      }//end of bymemory initialization
    
    //Initial value for ordermemory
    Vector ordermemory = new Vector();
    for(i=0;i<orderarray.length;i++)
      {
	switch(orderarray[i].type)
	  {
	case NUMERIC:
	    ordermemory.addElement(((number[])destrel.data[orderpos[i]])[currow]);
	    break;
	    
	  case INTEGER:
	  case SHORT:
	  case BOOLEAN:
	    Integer intval = new Integer(((int[])destrel.data[orderpos[i]])[currow]);
	    ordermemory.addElement(intval);
	    break;
	    
	  case LONG:
	    Long longval = new Long(((long[])destrel.data[orderpos[i]])[currow]);
	    ordermemory.addElement(longval);
	    break;
	    
	  case FLOAT:
	    Float floatval = new Float(((float[])destrel.data[orderpos[i]])[currow]);
	    ordermemory.addElement(floatval);
	    break;
	    
	  case DOUBLE:
	    Double doubleval = new Double(((double[])destrel.data[orderpos[i]])[currow]);
	    ordermemory.addElement(doubleval);
	    break;
	  case TEXT:
	  case STRING:
	  case EXPR:
	    ordermemory.addElement(((String[])destrel.data[orderpos[i]])[currow]);
	    break;
	  case IDLIST: //?? hongyu Feb 2002
	    break;
	  default:
	    throw new InterpretError("actualizeParFunc(): to be implemented...");
	  }
      }//end of ordermemory initialization
    
    
    //Actualize domain tuple by tuple
    long sur = 0;
    for(currow=0; currow < destrel.numtuples;currow++)
      {
      	//System.out.println("currow = " + currow);
	NRenv.row = currow;
	boolean breakflag = false;
	boolean breakflag1 = false;
	updatesurrogate = true;
	//Check group change
	for(i=0;i<byarray.length;i++)
	  {
	    switch(byarray[i].type)
	      {
	   	    case NUMERIC:
		if(((number)bymemory.elementAt(i)).compareTo(((number[])destrel.data[bypos[i]])[currow])!=0){
		    bymemory.setElementAt((((number[])destrel.data[bypos[i]])[currow]), i);
		    breakflag=true;
		}
		break;
	      case INTEGER:
	      case SHORT:
		if(((Integer)bymemory.elementAt(i)).intValue() !=((int[])destrel.data[bypos[i]])[currow])
		  {
		    bymemory.setElementAt(new Integer(((int[])destrel.data[bypos[i]])[currow]),i);
		    breakflag=true;
		  }
		break;
		
	      case LONG:
		if(((Long)bymemory.elementAt(i)).longValue() !=((long[])destrel.data[bypos[i]])[currow])
		  {
		    bymemory.setElementAt(new Long(((long[])destrel.data[bypos[i]])[currow]),i);
		    breakflag=true;
		  }
		break;
	      case DOUBLE:
		if(((Double)bymemory.elementAt(i)).doubleValue() !=((double[])destrel.data[bypos[i]])[currow])
		  {
		    bymemory.setElementAt(new Double(((double[])destrel.data[bypos[i]])[currow]),i);
		    breakflag=true;
		  }
		break;
	      case FLOAT:
		if(((Float)bymemory.elementAt(i)).floatValue() !=((float[])destrel.data[bypos[i]])[currow])
		  {
		    bymemory.setElementAt(new Float(((float[])destrel.data[bypos[i]])[currow]),i);
		    breakflag=true;
		  }
		break;
	      case STRING:
	      case TEXT:
	      case EXPR:
		if(((String)bymemory.elementAt(i)).compareTo(((String[])destrel.data[bypos[i]])[currow])!=0)
		  {
		    bymemory.setElementAt((((String[])destrel.data[bypos[i]])[currow]),i);
		    breakflag=true;
		  }
		break;
	      case IDLIST: //?? hongyu Feb 2002
		break;
	      default:
		throw new InterpretError("actualizeParFunc():to be implemented...");
	      }//end of switch
	    
	  }//end of group change checking
	
	//Group changed
	if(breakflag || currow==0)
	  {
//	  	if(breakflag)
	  		//System.out.println("group changed");
//	  	else
	  		//System.out.println("first row");
	    if(virtype==IDLIST)
	      {
		virtree.info = null;
		Relation r = actRelCell(virtree, vdom, rname, rindx, relhtbl, tname0, tname1);
		sur = Surrogate.next();
		updatesurrogate = true; //added by Yi Zheng July 9, 2002
			//System.out.println("virtype == IDLIST, sur = " + sur);
	      }
	    else
	      {
		AccumulateDom(virtree,ACC_RESET, virtype);
			}//added by Yi Zheng July 9, 2002
		for(i=0;i<orderarray.length;i++)
		  {
		    switch(orderarray[i].type)
		      {
					    case NUMERIC:
			ordermemory.setElementAt(((number[])destrel.data[orderpos[i]])[currow], i);
			break;

		      case INTEGER:
		      case SHORT:
		      case BOOLEAN:
			Integer intval = new Integer(((int[])destrel.data[orderpos[i]])[currow]);
			ordermemory.setElementAt(intval,i);
			break;
			
		      case LONG:
			Long longval = new Long(((long[])destrel.data[orderpos[i]])[currow]);
			ordermemory.setElementAt(longval,i);
			break;
			
		      case FLOAT:
			Float floatval = new Float(((float[])destrel.data[orderpos[i]])[currow]);
			ordermemory.setElementAt(floatval,i);
			break;
			
		      case DOUBLE:
			Double doubleval = new Double(((double[])destrel.data[orderpos[i]])[currow]);
			ordermemory.setElementAt(doubleval,i);
			break;
		      case TEXT:
		      case STRING:
		      case EXPR:
			ordermemory.setElementAt(((String[])destrel.data[orderpos[i]])[currow],i);
			break;
		      case IDLIST: //?? hongyu Feb 2002
			break;
		      default:
			throw new InterpretError("actualizeParFunc(): to be implemented...");
		      }
		  }//end of ordermemory initialization
	 //     } //commented out by Yi Zheng July 8, 2002
	  }
	else{
	  //System.out.println("group unchanged, not first row");
	  for(i=0;i<orderarray.length;i++)
	    {
	      switch(orderarray[i].type)
		{
				case NUMERIC:
		    if(((number)ordermemory.elementAt(i)).compareTo(((number[])destrel.data[orderpos[i]])[currow]) != 0){
			ordermemory.setElementAt((((number[])destrel.data[orderpos[i]])[currow]),i);
			breakflag1 = true;
		    }
		    break;

		case INTEGER:
		case SHORT:
		  if(((Integer)ordermemory.elementAt(i)).intValue() !=((int[])destrel.data[orderpos[i]])[currow])
		    {
		      ordermemory.setElementAt(new Integer(((int[])destrel.data[orderpos[i]])[currow]),i);
		      breakflag1=true;
		    }
		  break;
		  
		case LONG:
		  if(((Long)ordermemory.elementAt(i)).longValue() !=((long[])destrel.data[orderpos[i]])[currow])
		    {
		      ordermemory.setElementAt(new Long(((long[])destrel.data[orderpos[i]])[currow]),i);
		      breakflag1=true;
		    }
		  break;
		case DOUBLE:
		  if(((Double)ordermemory.elementAt(i)).doubleValue() !=((double[])destrel.data[orderpos[i]])[currow])
		    {
		      ordermemory.setElementAt(new Double(((double[])destrel.data[orderpos[i]])[currow]),i);
		      breakflag=true;
		    }
		  break;
		case FLOAT:
		  if(((Float)ordermemory.elementAt(i)).floatValue() !=((float[])destrel.data[orderpos[i]])[currow])
		    {
		      ordermemory.setElementAt(new Float(((float[])destrel.data[orderpos[i]])[currow]),i);
		      breakflag1=true;
		    }
		  break;
		case STRING:
		case TEXT:
		case EXPR:
		  if(((String)ordermemory.elementAt(i)).compareTo(((String[])destrel.data[orderpos[i]])[currow])!=0)
		    {
		      ordermemory.setElementAt((((String[])destrel.data[orderpos[i]])[currow]),i);
		      breakflag1=true;
		    }
		  break;
		case IDLIST: //?? hongyu Feb 2002
		  break;
		default:
		  throw new InterpretError("actualizeParFunc():to be implemented...");
		}//end of switch
	      
	    }//end of order change checking
	  
	  if(breakflag1)
	    {
	    	//System.out.println("order changed");
	      //actualize
	      if(virtype==IDLIST)
		{
		  Relation r = actRelCell(virtree, vdom, rname, rindx, relhtbl, tname0, tname1);
		  sur = Surrogate.next();
		  updatesurrogate = true;
		  //System.out.println("virtype == IDLIST, sur = " + sur);
		  //break;
		}
	      else
		{
		  AccumulateDom(virtree, ACC_ACCUM,virtype);
		//  sur = Surrogate.next(); //	REVISIT
		}
	      
	    }
	  else
	    {
	    	//System.out.println("order unchanged");
	      if(virtype==IDLIST)
		{
		  updatesurrogate = false;
		}
	      else
		AccumulateDom(virtree,ACC_BYPASS,virtype);
	      
	    }
	}
	//Write accumulator value in data[] array. what about the first and last value?
	switch(virtype)
	  {
	 	case NUMERIC:
	    ((number[])destrel.data[pos])[currow] = (number)virtree.info;
	    break;

	  case INTEGER:
	    ((int[])destrel.data[pos])[currow] = ((Integer)virtree.info).intValue();
	    break;
	  case SHORT:
	    ((short[])destrel.data[pos])[currow]=(short)((Integer)virtree.info).intValue();
	    break;
	  case LONG:
	    ((long[])destrel.data[pos])[currow]=((Long)virtree.info).longValue();
	    break;
	  case FLOAT:
	    ((float[])destrel.data[pos])[currow]=((Float)virtree.info).floatValue();
	    break;
	  case DOUBLE:
	    ((double[])destrel.data[pos])[currow]=((Double)virtree.info).doubleValue();
	    break;
	  case STMT:
	  case EXPR:
	  case TEXT:
	  case STRING: //added by Yi Zheng July 9, 2002
	    ((String[])destrel.data[pos])[currow] = (String)virtree.info;
	    break;
	  case IDLIST:
	    //sur = Surrogate.next();
	    if(updatesurrogate)
	      {
		Relation r = (Relation) virtree.info;
		r.load(true);  // have to load it. It seems be saved and cleared.
		
		// We must write the new surrogate for the new sub-relation.
		Domain[] darry = new Domain[r.domains.length + 1];
		
		darry[0] = env.lookupDom(".id", true);
		
		for(int k = 0; k < r.domains.length; k++)
		  darry[k+1] = r.domains[k];
		Object[] dat = Relation.createDataArray(darry, r.numtuples);
		
		for(int k = 0; k < r.domains.length; k++)
		  dat[k+1] = r.data[k];
		
		for(int k = 0; k < r.numtuples; k++)
		  ((long[])dat[0])[k] = sur;
		
		dest.append(dat, r.numtuples);
		dest.load();
	      }
	    ((long[])destrel.data[pos])[currow] =sur;
	    
	    break;
	  default:
	    throw new InterpretError("actualizeParFunc(): to be implemented...");
	  }//end switch
	
      }//end of actualizing domain tuple by tuple
      env.removeNRFrame(); //added by Yi Zheng July 9, 2002
  }
  
  private void actualizeFun(Domain dom) throws InterpretError
  {
    
    int i, oldpos, virtype=dom.type;
    SimpleNode virtree = (SimpleNode)vvtree.get(dom);
    
    String tname0=null, tname1=null;
    Hashtable relhtbl=null;
    Vector rname = null;
    Vector rindx = null;
    Vector vdom = null;
    Relation dest = null;
    
    //Fisrt we need to sort on the by-list and order-list
    
    Vector orderdoms = new Vector();
    
    //Generate orderdoms list vector
    //if added by Yi Zheng July 9, 2002
    SimpleNode orderexprlist = null;
    if(virtree.opcode != OP_REDFUNCALL)
    	orderexprlist = (SimpleNode)virtree.jjtGetChild(1);
   	else
    	orderexprlist = (SimpleNode)virtree.jjtGetChild(2);
   		
    for(i=0;i<orderexprlist.jjtGetNumChildren();i++)
      {
		SimpleNode orderchild = (SimpleNode)orderexprlist.jjtGetChild(i);
		Domain d = getDom(orderchild.name);
		if(d!=null)
		  {
		    orderdoms.addElement(d);
		    
		  }
		else
		  throw new InterpretError("ActualizeParFunc:order domain not found");
      }
    
    //Generate oderdoms array for sorting
    
    
    Domain[] orderarray = new Domain[orderdoms.size()];
    orderdoms.copyInto(orderarray);
    
//    Global.println("age " + age()+": start sort");
    destrel.sort(orderarray);
//    Global.println("age "+age()+":finish sort");
    
    for(i=orderarray.length-1;i>=0;i--)
      {
	alldoms.removeElement(orderarray[i]);
	alldoms.insertElementAt(orderarray[i],0);
      }
    
    for(Enumeration e = vvtree.elements(); e.hasMoreElements();)
      {
	resetNodeBits((SimpleNode)e.nextElement());
      }
    
    
    if(virtype == IDLIST && !nt)
      {
	relhtbl = new Hashtable();
	rname = new Vector();
	rindx = new Vector();
	vdom = new Vector();
	tname0=interpreter.nextTempName();
	tname1=interpreter.nextTempName();
	tempnames.addElement(tname0);
	tempnames.addElement(tname1);
	
	if((dest = env.lookupRel("."+dom.name, true))==null)
	  throw new InterpretError("actualizseParFunc: relation ./"+dom.name+"'not found.");
	preProcIDListDom(virtree, vdom, rname, rindx);
	
      }
    NREnvironment NRenv = env.addNRFrame(destrel);
    
    //Remember the position of domain for the future use
    int pos = alldoms.indexOf(dom);
    
    //bug fix by Yi Zheng June 23, 2002
    virtree.bits = pos;
    //end of bug fix June 23, 2002
    //Remember the position of order list domain for the future use
    int orderpos[]=new int[orderarray.length];
    for(i=0;i<orderarray.length;i++)
      orderpos[i] = alldoms.indexOf(orderarray[i]);
    
    //Get the initial values for sort attributes
    currow = oldpos = 0;
    
    
    //Initial value for ordermemory
    Vector ordermemory = new Vector();
    for(i=0;i<orderarray.length;i++)
      {
	switch(orderarray[i].type)
	  {
	case NUMERIC:
	    ordermemory.addElement(((number[])destrel.data[orderpos[i]])[currow]);
	    break;
	    
	  case INTEGER:
	  case SHORT:
	  case BOOLEAN:
	    //Integer intval=new Integer(((int[])destrel.data[bypos[i]])[currow]);
	    int a = ((int[])(destrel.data[orderpos[i]]))[currow];
	    Integer intval=new Integer( a );
	    ordermemory.addElement(intval);
	    break;
	    
	  case LONG:
	    Long longval = new Long(((long[])destrel.data[orderpos[i]])[currow]);
	    ordermemory.addElement(longval);
	    break;
	    
	  case FLOAT:
	    Float floatval = new Float(((float[])destrel.data[orderpos[i]])[currow]);
	    ordermemory.addElement(floatval);
	    break;
	    
	  case DOUBLE:
	    Double doubleval = new Double(((double[])destrel.data[orderpos[i]])[currow]);
	    ordermemory.addElement(doubleval);
	    break;
	  case TEXT:
	  case STRING:
	  case EXPR:
	    ordermemory.addElement(((String[])destrel.data[orderpos[i]])[currow]);
	    break;
	  case IDLIST: //?? hongyu Feb 2002
	    
	    break;
	  default:
	    throw new InterpretError("actualizeFunc(): to be implemented...");
	  }
      }//end of ordermemory initialization
    
    
    //Actualize domain tuple by tuple
    long sur = 0;
    for(currow=0; currow < destrel.numtuples;currow++)
      {
	NRenv.row = currow;
	boolean breakflag1=false;
	boolean updatesurrogate = true;
	//Check ORDER changes
	for(i=0;i<orderarray.length;i++)
	  {
	    switch(orderarray[i].type)
	      {
	      case INTEGER:
	      case SHORT:
		if(((Integer)ordermemory.elementAt(i)).intValue() !=((int[])destrel.data[orderpos[i]])[currow])
		  {
		    ordermemory.setElementAt(new Integer(((int[])destrel.data[orderpos[i]])[currow]),i);
		    breakflag1=true;
		  }
		break;
		
	      case LONG:
		if(((Long)ordermemory.elementAt(i)).longValue() !=((long[])destrel.data[orderpos[i]])[currow])
		  {
		    ordermemory.setElementAt(new Long(((long[])destrel.data[orderpos[i]])[currow]),i);
		    breakflag1=true;
		  }
		break;
	      case DOUBLE:
		if(((Double)ordermemory.elementAt(i)).doubleValue() !=((double[])destrel.data[orderpos[i]])[currow])
		  {
		    ordermemory.setElementAt(new Double(((double[])destrel.data[orderpos[i]])[currow]),i);
		    breakflag1=true;
		  }
		break;
	      case FLOAT:
		if(((Float)ordermemory.elementAt(i)).floatValue() !=((float[])destrel.data[orderpos[i]])[currow])
		  {
		    ordermemory.setElementAt(new Float(((float[])destrel.data[orderpos[i]])[currow]),i);
		    breakflag1=true;
		  }
		break;
		    case NUMERIC:
		  if(((number)ordermemory.elementAt(i)).compareTo(((number[])destrel.data[orderpos[i]])[currow])!=0){
		      ordermemory.setElementAt((((number[])destrel.data[orderpos[i]])[currow]),i);
		      breakflag1 = true;
		  }
		  break;

	      case STRING:
	      case TEXT:
	      case EXPR:
		if(((String)ordermemory.elementAt(i)).compareTo(((String[])destrel.data[orderpos[i]])[currow])!=0)
		  {
		    ordermemory.setElementAt((((String[])destrel.data[orderpos[i]])[currow]),i);
		    breakflag1=true;
		  }
		break;
	      case IDLIST: //?? hongyu Feb 2002
		break;
	      default:
		throw new InterpretError("actualizeParFunc():to be implemented...");
	      }//end of switch
	    
	  }//end of order change checking
	
	if(breakflag1 && currow!=0)
	  {
	    //accumulate
	    if(virtype==IDLIST)
	      {
		Relation r = actRelCell(virtree, vdom, rname, rindx, relhtbl, tname0, tname1);
		sur = Surrogate.next();
		updatesurrogate = true;
	      }
	    else
	      AccumulateDom(virtree,ACC_ACCUM,virtype);
	  }
	else if(currow == 0)
	  {
	    if(virtype==IDLIST)
	      {
		virtree.info = null;
		Relation r = actRelCell(virtree, vdom, rname, rindx, relhtbl, tname0, tname1);
		sur = Surrogate.next();
		updatesurrogate = true;
	      }
	    else
	      AccumulateDom(virtree,ACC_RESET,virtype);
	  }
	else
	  {
	    if(virtype==IDLIST)
	      {
		updatesurrogate = false;
	      }
	    else
	      AccumulateDom(virtree,ACC_BYPASS,virtype);
	  }
	switch(virtype)
	  {
	  	case NUMERIC:
	    ((number[])destrel.data[pos])[currow] = (number)virtree.info;
	    break;

	  case INTEGER:
	    ((int[])destrel.data[pos])[currow] = ((Integer)virtree.info).intValue();
	    break;
	  case SHORT:
	    ((short[])destrel.data[pos])[currow]=(short)((Integer)virtree.info).intValue();
	    break;
	  case LONG:
	    ((long[])destrel.data[pos])[currow]=((Long)virtree.info).longValue();
	    break;
	  case FLOAT:
	    ((float[])destrel.data[pos])[currow]=((Float)virtree.info).floatValue();
	    break;
	  case DOUBLE:
	    ((double[])destrel.data[pos])[currow]=((Double)virtree.info).doubleValue();
	    break;
	  case STMT:
	  case EXPR:
	  case TEXT:
	  case STRING: //added by Yi Zheng July 9, 2002
	    ((String[])destrel.data[pos])[currow] = (String)virtree.info;
	    break;
	  case IDLIST:


	    
	    if(updatesurrogate){
	      Relation r = (Relation) virtree.info;
	      //r.load();  // have to load it. It seems be saved and cleared.
	      
	      // We must write the new surrogate for the new sub-relation.
	      Domain[] darry = new Domain[r.domains.length + 1];
	      
	      darry[0] = env.lookupDom(".id", true);
	      
	      for(int k = 0; k < r.domains.length; k++)
		darry[k+1] = r.domains[k];
	      Object[] dat = Relation.createDataArray(darry, r.numtuples);
	      
	      for(int k = 0; k < r.domains.length; k++)
		dat[k+1] = r.data[k];
	      
	      for(int k = 0; k < r.numtuples; k++)
		((long[])dat[0])[k] = sur;
	      
	      dest.append(dat, r.numtuples);
	      dest.load();
	    }
	    ((long[])destrel.data[pos])[currow] =sur;
	    break;
	  default:
	    throw new InterpretError("actualizeFunc(): to be implemented...");
	  }//end switch
      }//end of actualizing domain tuple by tuple
      env.removeNRFrame(); //added by Yi Zheng July 9, 2002
  }
  
  
  //New Accumulation function for Functional and Partial Functional mapping
  //6/12/2001 Sungsoo Kang
  private void AccumulateDom(SimpleNode node, int accum_type, int virtual_type) throws InterpretError
  {
    //double actValue= actDoubleCell((SimpleNode)node.jjtGetChild(0));
    //if added by Yi Zheng July 9, 2002
    SimpleNode child_node = null;
    if(node.opcode != OP_REDFUNCALL)
    	child_node = (SimpleNode)node.jjtGetChild(0);
    else
     	child_node = (SimpleNode)node.jjtGetChild(1);
   
    switch(accum_type)
      {
      case ACC_RESET:
	switch(virtual_type)
	  {
	  case LONG:
	    Long init_longval;
//	    switch(destrel.domains[node.bits+1].type) //Yi Zheng June 23, 2002
		switch(destrel.domains[node.bits].type)
	      {
	      case LONG:
		init_longval = new Long((long)actLongCell(child_node));//new Long(((long[])destrel.data[node.bits+1])[currow]);
		break;
	      case INTEGER:
	      case SHORT:
		init_longval = new Long((long)actIntCell(child_node));//new Long((long)((short[])destrel.data[node.bits+1])[currow]);
		break;
	      case FLOAT:
	      case DOUBLE:
		init_longval = new Long((long)actDoubleCell(child_node));//new Long((long)((double[])destrel.data[node.bits+1])[currow]);
		break;
	    case NUMERIC://case added by Yiyi June 5'02
		init_longval = new Long(actNumCell(child_node).numToLong());
		break;

	      default:
		init_longval = new Long(0);
		break;
	      }
	    node.info = init_longval;
	    break;
	 	case NUMERIC:{//case added by Yiyi June 5'02
	    number init_numberval;
	    switch(destrel.domains[node.bits+1].type){
	    case LONG:
		init_numberval = number.longToNum(actLongCell(child_node));
		break;
	    case INTEGER:
	    case SHORT:
		init_numberval = number.intToNum(actIntCell(child_node));
		break;
	    case FLOAT:
	    case DOUBLE:
		init_numberval = number.doubleToNum(actDoubleCell(child_node));
		break;
	    default:
		init_numberval = new number("+000 0");
		break;
	    }
	    node.info = init_numberval;
	    break;
	}

	  case SHORT:
	  case INTEGER:
	    Integer init_intval;
//	    switch(destrel.domains[node.bits+1].type) //Yi Zheng June 23, 2002
		switch(destrel.domains[node.bits].type)
	      {
	      case LONG:
		init_intval = new Integer((int)actLongCell(child_node));//new Long(((long[])destrel.data[node.bits+1])[currow]);
		break;
	    case NUMERIC:	//case added by Yiyi June 5'02
		init_intval = new Integer(actNumCell(child_node).numToInt());
		break;

	      case INTEGER:
	      case SHORT:
		init_intval = new Integer((int)actIntCell(child_node));//new Long((long)((short[])destrel.data[node.bits+1])[currow]);
		break;
	      case FLOAT:
	      case DOUBLE:
		init_intval = new Integer((int)actDoubleCell(child_node));//new Long((long)((double[])destrel.data[node.bits+1])[currow]);
		break;
		
	      default:
		init_intval = new Integer(0);
		break;
	      }
	    node.info = init_intval;
	    break;
	  case FLOAT:
	  case DOUBLE:
	    Double init_dblval;
//	    switch(destrel.domains[node.bits+1].type) //Yi Zheng June 23, 2002
		switch(destrel.domains[node.bits].type)
	      {
	      case LONG:
		init_dblval = new Double((double)actLongCell(child_node));//new Long(((long[])destrel.data[node.bits+1])[currow]);
		break;
	      case INTEGER:
	      case SHORT:
		init_dblval = new Double((double)actIntCell(child_node));//new Long((long)((short[])destrel.data[node.bits+1])[currow]);
		break;
	    case NUMERIC: //case added by Yiyi June 5'02
		init_dblval = new Double(actNumCell(child_node).numToDouble());
		break;

	      case FLOAT:
	      case DOUBLE:
		init_dblval = new Double((double)actDoubleCell(child_node));//new Long((long)((double[])destrel.data[node.bits+1])[currow]);
		break;
	      default:
		init_dblval = new Double(0);
		break;
	      }
	    node.info = init_dblval;
	    break;
	  case STRING:
	    String init_val;
	    init_val = actStrCell(child_node);//new String(((String[])destrel.data[node.bits+1])[currow]);
	    node.info = init_val;
	    break;
	  }
	break;
	
      case ACC_BYPASS:
	break;
	
      case ACC_ACCUM:
	switch(virtual_type)
	  {
	  case STRING://case added by Yi Zheng July 9, 2002
	    String actStrValue = actStrCell(child_node);
	    switch(node.opcode)
	      {
		
		//added by Yi Zheng July 5, 2002
		//let a be fun nop of b
		case OP_NOP:
			boolean rb = new Random().nextBoolean();
			if(rb)
				node.info = new String(actStrValue);
			break;
		//end of addition July 5, 2002
		//added by Yi Zheng July 9, 2002
		case OP_REDFUNCALL:
			actStrCell(node);
			break;
		//end of addition July 9, 2002

	      default:
		//System.out.println("To be implemented");
		throw new InterpretError("actualize Fun: To be implemented");
	      }
	    break;
	  case LONG:
	    long actLongValue = actLongCell(child_node);
	    switch(node.opcode)
	      {
	      case OP_MINUS:
		node.info = new Long(((Long)node.info).longValue()-actLongValue);
		break;
	      case OP_PLUS:
		node.info = new Long(((Long)node.info).longValue()+actLongValue);
		break;
	      case OP_MUL:
		node.info = new Long(((Long)node.info).longValue()*actLongValue);
		break;
	      case OP_DIV:
		node.info = new Long(((Long)node.info).longValue()/actLongValue);
		break;
	      case OP_MIN:
		node.info = new Long(Math.min(((Long)node.info).longValue(), actLongValue));
		break;
	      case OP_MAX:
		
		node.info = new Long(Math.max(((Long)node.info).longValue(), actLongValue));
		break;
		
		//added by Yi Zheng July 5, 2002
		//let a be fun nop of b
		case OP_NOP:
			boolean rb = new Random().nextBoolean();
			if(rb)
				node.info = new Long(actLongValue);
			break;
		//end of addition July 5, 2002
		//added by Yi Zheng July 9, 2002
		case OP_REDFUNCALL:
			actLongCell(node);
			break;
		//end of addition July 9, 2002

	      default:
		//System.out.println("To be implemented");
		throw new InterpretError("actualize Fun: To be implemented");
	      }
	    break;
	  	case NUMERIC: //case added by Yiyi June 5'02
	    number actNumValue = actNumCell(child_node);
	    switch(node.opcode){
	    case OP_MINUS:
		node.info = ((number)node.info).subtract(actNumValue);
		break;
	    case OP_PLUS:
		node.info = ((number)node.info).add(actNumValue);
		break;
	    case OP_MUL:
		node.info = ((number)node.info).multiply(actNumValue);
		break;
	    case OP_DIV:
		node.info = ((number)node.info).divide(actNumValue);
		break;
		 case OP_MIN:
		node.info = number.min((number)(node.info), actNumValue);
		break;
	    case OP_MAX:
		node.info = number.max((number)(node.info), actNumValue);
		break;
	    }
	    break;
	  case SHORT:
	  case INTEGER:
	    int actIntValue= actIntCell(child_node);
	    
	    switch(node.opcode)
	      {
	      case OP_MINUS:
//		node.info = new Integer(((Integer)node.info).intValue()-actIntValue);
//modified by Yi Zheng to suit the definition of alternate sum, July 9, 2002
		node.info = new Integer(actIntValue - ((Integer)node.info).intValue());
		break;
	      case OP_PLUS:
		node.info = new Integer(((Integer)node.info).intValue()+actIntValue);
		break;
	      case OP_MUL:
		node.info = new Integer(((Integer)node.info).intValue()*actIntValue);
		break;
	      case OP_DIV:
//		node.info = new Integer(((Integer)node.info).intValue()/actIntValue);
//modified by Yi Zheng to suit the definition of alternate quotient, July 9, 2002
		node.info = new Integer(actIntValue/((Integer)node.info).intValue());
		break;
	      case OP_MIN:
		node.info = new Integer(Math.min(((Integer)node.info).intValue(), actIntValue));
		break;
	      case OP_MAX:
		node.info = new Integer(Math.max(((Integer)node.info).intValue(), actIntValue));
		break;
		//added by Yi Zheng July 5, 2002
		//let a be fun nop of b
		case OP_NOP:
		boolean rb = new Random().nextBoolean();
		if(rb)
			node.info = new Integer(actIntValue);
		break;
		
		//added by Yi Zheng July 9, 2002
		case OP_REDFUNCALL:
			actIntCell(node);
			break;
		//end of addition July 9, 2002
	      default:
		//System.out.println("To be implemented");
		throw new InterpretError("actualize Fun: To be implemented");
		//end of addition July 5, 2002
	      }
	    break;
	  case FLOAT:
	  case DOUBLE:
	    double actDblValue= actDoubleCell(child_node);
	    
	    switch(node.opcode)
	      {
	      case OP_PLUS:
		node.info = new Double(((Double)node.info).doubleValue()+actDblValue);
		break;
	      case OP_MINUS:
		node.info = new Double(((Double)node.info).doubleValue()-actDblValue);
		break;
	      case OP_MUL:
		node.info = new Double(((Double)node.info).doubleValue()*actDblValue);
		break;
	      case OP_DIV:
		node.info = new Double(((Double)node.info).doubleValue()/actDblValue);
		break;
	      case OP_MIN:
		node.info = new Double(Math.min(((Double)node.info).doubleValue(), actDblValue));
		break;
	      case OP_MAX:
		if(node.info == null)  node.info = new Double(Double.MIN_VALUE);
		node.info = new Double(Math.max(((Double)node.info).doubleValue(), actDblValue));
		break;
		//added by Yi Zheng July 5, 2002
		//let a be fun nop of b
		case OP_NOP:
		boolean rb = new Random().nextBoolean();
		if(rb)
			node.info = new Double(actDblValue);
		break;
		//added by Yi Zheng July 9, 2002
		case OP_REDFUNCALL:
			actDoubleCell(node);
			break;
		//end of addition July 9, 2002

	      default:
		//System.out.println("To be implemented");
		throw new InterpretError("actualize Fun: To be implemented");
		//end of addition July 5, 2002

	      }
	    break;
	  }
      }
  }
  
  /**
   * method for implementing function call in vertical domain algebra
   * @author Yi Zheng July 9, 2002
   */
  private void RedFunCallAccumNode(SimpleNode node, int nodeType, Object actValueObj)
 	throws InterpretError
  {
  	String tempFile1 = null; //used for name of node.info, if nodeType is IDLIST
  	String tempFile2 = null; //used for name of actValueObj, if nodeType is IDLIST
  	
  		if(node.info == null)
  		{
  			switch (nodeType)
  			{
  				case INTEGER:
  					node.info = (Integer) actValueObj;
  					break;
  				case LONG:
  					node.info = (Long) actValueObj;
  					break;
  				case BOOLEAN:
  					node.info = (Boolean) actValueObj;
  					break;
  				case DOUBLE:
  					node.info = (Double) actValueObj;
  					break;
  				case STRING:
  					node.info = (String) actValueObj;
  					break;
  				case IDLIST:
  					node.info = (Relation) actValueObj;
  					break;
  				default:
  					throw new InterpretError("unexpected node type " + nodeType + Utility.type2String(nodeType));
  			}
			
		}
		else
		{
			//step 1: find the top level computation to call
			//REVISIT: implement nested computation
			SimpleNode redfuncall = (SimpleNode) node.jjtGetChild(0);
			String compName = ((SimpleNode)redfuncall.jjtGetChild(0)).name;
			Relation comp = env.lookupRel(compName, true);
			if((comp == null) || (comp.rvc != COMPUTATION))
				throw new InterpretError(compName + " is not a legal computation");
			Computation cp = ((Computation) comp).duplicate();//get a fresh copy
			
			//step 2: construct the select condition
			//assumption: node.info is the first in parameter, actValue is the second
			//however, the exact position of the in parameters is determined by the definition
			//of the computation
			
			//step 2.1: find the compblock to run
			int blockType = 0;
			if(!redfuncall.isBitSet(0))
			{
				//throw new InterpretError("parameter list not found for computation call in vertical expression");
				//August 13, 2002
				// red cplx1() is syntactic sugar for cplx(in, in, out)
				blockType = 3;		
	
			}else
			{
			  SimpleNode callParams = (SimpleNode) redfuncall.jjtGetChild(1);
			  int numParams = callParams.jjtGetNumChildren();
			  if(numParams != cp.numattrs)
				throw new InterpretError("wrong number of call parameters");
			  else if(numParams != 3 || cp.numattrs != 3)
				throw new InterpretError("wrong number of call parameters");
			   for(int i = 0; i < numParams; i++)
			   {
				SimpleNode param = (SimpleNode) callParams.jjtGetChild(i);
				if(!param.isBitSet(0)) //input param
					blockType |= 1 << i;
			   }

		     }
	
			for(int k = 0; k < 3; k++)
			{
				if(cp.domains[k].type != nodeType)
						throw new InterpretError("type mismatch: computation doesn't work with " + Utility.type2String(nodeType));
			}
				
	//		int blockType = 0;
			CompBlock runBlock = null;
			for(int j = 0; j < cp.blocks.length; j++)
			{
				if(cp.blocks[j].blockType == blockType)
				{
					runBlock = cp.blocks[j];
					break;
				}
			}
			if(runBlock == null)
				throw new InterpretError("no runnable block found");
			
			//step 2.2: verify that runBlock can be used as AssoCommuOperator or Ordered operator, as required by node.type
			if((runBlock.blockType != 3) //(in, in, out)
				&& (runBlock.blockType != 5) //(in, out, in)
				&& (runBlock.blockType != 6) //(out, in, in)
				)
				throw new InterpretError("computation block cannot be used in reduction or functional mapping");

			switch(node.type)
			{
				case OP_RED:
				case OP_EQUIV:
					if(!runBlock.redop)
						throw new InterpretError("computation block cannot be used in reduction");
					break;
				case OP_FUN:
				case OP_PAR:
					if((!runBlock.redop) && (!runBlock.funop))
						throw new InterpretError("computation block cannot be used in functional mapping");
					break;
			}
			
			//step 2.3: find out the in parameters' name and the output domain
			String firstIn = null;
			String secondIn = null;
			int outDom = 0; //position of output domain in the domains of cp
			switch(runBlock.blockType)
			{
				case 3: //in, in, out
					firstIn = cp.domains[0].name;
					secondIn = cp.domains[1].name;
					outDom = 2;
					break;
				case 5: //in, out, in
					firstIn = cp.domains[0].name;
					secondIn = cp.domains[2].name;
					outDom = 1;
					break;
				case 6: //out, in, in
					firstIn = cp.domains[1].name;
					secondIn = cp.domains[2].name;
					outDom = 0;
					break;
			}
			
			
			//step 2.4: construct the select conditions
			SimpleNode Compare1 = new SimpleNode("Compare");
		    Compare1.set(OP_BIOPERATOR, OP_EQ);
		    SimpleNode id1 = new SimpleNode("Identifier");
		    id1.set(OP_IDENTIFIER, OP_IDENTIFIER,firstIn);
		    id1.jjtSetParent(Compare1); //link up Compare and id
		    Compare1.jjtAddChild(id1);  //link up Compare and id
		    
		    if(nodeType != IDLIST)
		    {
			    SimpleNode literal1 = new SimpleNode("Literal");
			    literal1.set(OP_LITERAL, OP_BOOLEAN + nodeType);
			    literal1.setObject(node.info);
			    literal1.jjtSetParent(Compare1);
			    Compare1.jjtAddChild(literal1);
		  	}
		  	else
		  	{
		  		SimpleNode id3 = new SimpleNode("Identifier");
		  		tempFile1 = Interpreter.nextTempName();
		  		Relation rtemp1 = ((Relation)node.info).project(((Relation)node.info).domains, tempFile1);
		  		env.put(rtemp1);
		  		id3.set(OP_IDENTIFIER, OP_IDENTIFIER, tempFile1);
		  		id3.jjtSetParent(Compare1);
		  		Compare1.jjtAddChild(id3);
		  	}
		  	
			SimpleNode Compare2 = new SimpleNode("Compare");
		    Compare2.set(OP_BIOPERATOR, OP_EQ);
		    SimpleNode id2 = new SimpleNode("Identifier");
		    id2.set(OP_IDENTIFIER, OP_IDENTIFIER,secondIn);
		    id2.jjtSetParent(Compare2); //link up Compare and id
		    Compare2.jjtAddChild(id2);  //link up Compare and id
		    
		    if(nodeType != IDLIST)
		    {
			    SimpleNode literal2 = new SimpleNode("Literal");
			    literal2.set(OP_LITERAL, OP_BOOLEAN + nodeType);
			    literal2.setObject(actValueObj);
			    literal2.jjtSetParent(Compare2);
			    Compare2.jjtAddChild(literal2);
		  	}
		    else
		    {
		  		SimpleNode id4 = new SimpleNode("Identifier");
		  		tempFile2 = Interpreter.nextTempName();
		  		Relation rtemp2 = ((Relation)actValueObj).project(((Relation)actValueObj).domains, tempFile2);
		  		env.put(rtemp2);
		  		id4.set(OP_IDENTIFIER, OP_IDENTIFIER, tempFile2);
		  		id4.jjtSetParent(Compare2);
		  		Compare2.jjtAddChild(id4);
		    	
		    }
		    
		    SimpleNode And = new SimpleNode("And");
		    And.set(OP_BIOPERATOR, OP_AND);
		    Compare1.jjtSetParent(And); //link up And and compare1
		    And.jjtAddChild(Compare1);  
		    Compare2.jjtSetParent(And); //link up And and Compare2
		    And.jjtAddChild(Compare2);
		    
			//step 3: call applySelect with And
			String fileTemp = Interpreter.nextTempName();
			Relation temp = cp.applySelect(And, fileTemp, env);
			//step 4: extract result and put in node.info
			temp.load(true);
			switch (nodeType)
			{
				case INTEGER:
					node.info = new Integer(((int[])temp.data[outDom])[0]);
					break;
				case BOOLEAN:
					byte b = ((byte[])temp.data[outDom])[0];
					if (b == 1)
						node.info = new Boolean(true);
					else
						node.info = new Boolean(false);
					break;
				case DOUBLE:
					node.info = new Double(((double[])temp.data[outDom])[0]);
					break;
				case LONG:
					node.info = new Long(((long[])temp.data[outDom])[0]);
					break;
				case STRING:
					node.info = ((String[])temp.data[outDom])[0];
					break;
				case IDLIST:
					Relation dotRel = env.lookupRel("." + cp.domains[outDom].name, true);
					if(dotRel == null)
						throw new InterpretError("relation ." + cp.domains[outDom].name + " not found");
					long surr = ((long[])temp.data[outDom])[0];
					node.info = dotRel.getRelation(surr);
					break;
			}
			temp.unload();
			//step 5: clean up	
			Interpreter.deleteFile(fileTemp);
			if(tempFile1 != null) env.removeRel(tempFile1, true);
			if(tempFile2 != null) env.removeRel(tempFile2, true);
			
		}
	}
	
	   //START Since JRelCaNu d.
   /**
    * JRelCaNu extended the methods below:
    * All methods named '*WithNull'  below have been extended from 
    * methods named '*' from above.
    */
   /**
	* Actualize the EQUIV-typed virtual domains. Note: we are now 
	* actualizing it one by one. An optimization could be made by 
	* actualize the domains with exactly the same by-list.
        * 
        * JRelCaNu modified/enhanced this method:
	*  - Changed from calling actXYZCell() to calling actXYZCellWithNull().
	*  - Use JRelix CastToSometype.value() instead of Java (SomeType) cast.
	*  - info is now a Byte object instead of Boolean to handle Dc and Dk.
	*/
   private void actualizeEquivWithNull(Domain dom)
	   throws InterpretError{
	  int i, oldpos, virtype = dom.type;
	  SimpleNode virtree = (SimpleNode)vvtree.get(dom);

	  // Variables for IDLIST-typed domain and when nt == false;
	  String tname0 = null, tname1 = null;
	  Hashtable relhtbl = null;
	  Vector rname = null;
	  Vector rindx = null;
	  Vector vdom = null;
	  Relation dest = null;

	  // First we need to sort on the by-list.
	  Vector bydoms = new Vector();
	  SimpleNode exprlist = (SimpleNode)virtree.jjtGetChild(1);
	  for(i = 0; i < exprlist.jjtGetNumChildren(); i++){
	SimpleNode child = (SimpleNode) exprlist.jjtGetChild(i);
	Domain d = getDom(child.name);
	if(d != null)
	   bydoms.addElement(d);
		 else
	   // Should never happen...
			throw new InterpretError("actualizeEquiv: domain not found.");
	  }
	  Domain[] byarray = new Domain[bydoms.size()];
	  bydoms.copyInto(byarray);

	  Global.println("age "+age()+": start sort");
	  destrel.sort(byarray);
	  Global.println("age "+age()+": finish sort");

	  // Since we did sort on the by-list, we must rearrange the 
	  // alldoms vector elements correspondingly.
	  for(i = byarray.length - 1; i >=0; i--){
	alldoms.removeElement(byarray[i]);
	alldoms.insertElementAt(byarray[i], 0);
	  }
	  
	  // Since alldoms changed, we must reset the sequence number stored
	  // in node.bits.
	  for(Enumeration e = vvtree.elements(); e.hasMoreElements();){
		 resetNodeBits((SimpleNode)e.nextElement());
	  }

	  if(virtype == IDLIST && !nt){
		 relhtbl = new Hashtable();
		 rname = new Vector();
		 rindx = new Vector();
		 vdom = new Vector();
		 tname0 = interpreter.nextTempName();
		 tname1 = interpreter.nextTempName();

		 tempnames.addElement(tname0);  // remember to clear this temp rel.
		 tempnames.addElement(tname1);  // remember to clear this temp rel.

//         if((dest = reltable.getRel("." + dom.name)) == null)
//REVISIT: Zhongxia, I used true but check it please
		 if((dest = env.lookupRel("." + dom.name, true)) == null)

			// Should never happen..
			throw new InterpretError("actualizeequiv: relation .'" +
			   dom.name + "' not found.");

		 // Fill the vectors.
		 preProcIDListDom(virtree, vdom, rname, rindx);
	  }

	  // Remember this domain's position in 'alldom'.
	  int pos = alldoms.indexOf(dom);

	  // Get the position of the domains in by-list.
	  int bypos[] = new int[byarray.length];
	  for(i = 0; i < byarray.length; i++)
		 bypos[i] = alldoms.indexOf(byarray[i]);

	  // Get the initial values for sort attributes.
	  currow = oldpos = 0;

	  //YuLing changed the following loop  in March 2001 to 
	  //fix the by STRING problem
	  Vector vals=new Vector();
	  for(i = 0; i < byarray.length; i++){


	     switch(byarray[i].type){
	       case INTEGER:
	       case SHORT:
	       case BOOLEAN:
		 Integer intval=new
		 Integer(((int[])destrel.data[bypos[i]])[currow]);
	         vals.addElement (intval);
		 break;

	       case LONG:
		 Long longval=new
		 Long(((long[])destrel.data[bypos[i]])[currow]);
	         vals.addElement (longval);
		 break;
	       case FLOAT:
	         Float floatval=new
		 Float(((float[])destrel.data[bypos[i]])[currow]);
	         vals.addElement (floatval);
		 break;
	       case DOUBLE:
	         Double doubleval=new
		 Double(((double[])destrel.data[bypos[i]])[currow]);
	         vals.addElement (doubleval);
		 break;
	       case TEXT:
	       case STRING:
	       case EXPR:
	         vals.addElement(((String[])destrel.data[bypos[i]])[currow]);
		 break;
	       case IDLIST:
	         break;
	       default:
		  throw new InterpretError("actualizeEquiv(): to be implemented..");
	    }
	 } 

	  //YuLing's modification ends here
	        

	  // Now actualize the virtual domain tuple by tuple.
	  for(; currow < destrel.numtuples; currow++){

	// If the sort values are changed, we must save the values to
	// the destrel on the fly.

		int breakflag=0;
		for (i=0;i < byarray.length; i++)
		{
		 //This loop is changed by YuLing in March, 2001
		 //to deal with the by String problem of equiv

		 switch(byarray[i].type){
		   case INTEGER:
		   case SHORT:
	             if(((Integer)(vals.elementAt(i))).intValue() !=
		          ((int[])destrel.data[bypos[i]])[currow]){
		           vals.setElementAt( new Integer(
		          ((int[])destrel.data[bypos[i]])[currow]),i);
			  breakflag=1;
			  }
		     break;
		   case LONG:
	             if(((Long)(vals.elementAt(i))).longValue() !=
		          ((long[])destrel.data[bypos[i]])[currow]){
		           vals.setElementAt( new Long(
		          ((long[])destrel.data[bypos[i]])[currow]),i);
			  breakflag=1;
			  }
		     break;
		   case DOUBLE:
	             if(((Double)(vals.elementAt(i))).doubleValue() !=
		          ((double[])destrel.data[bypos[i]])[currow]){
		           vals.setElementAt( new Double(
		          ((double[])destrel.data[bypos[i]])[currow]),i);
			  breakflag=1;
			  }
		     break;
		   case FLOAT:
	             if(((Float)(vals.elementAt(i))).floatValue() !=
		          ((float[])destrel.data[bypos[i]])[currow]){
		           vals.setElementAt( new Float(
		          ((float[])destrel.data[bypos[i]])[currow]),i);
			  breakflag=1;
			  }
		     break;
		    
		   case STRING:
		   case TEXT:
		   case EXPR:

	             if(((String)vals.elementAt(i)).compareTo(
		        ((String[])destrel.data[bypos[i]])[currow])!=0){
		     vals.setElementAt (
		         (((String[])destrel.data[bypos[i]])[currow]),i);
			  breakflag=1;
		        }
		     break;

                   case IDLIST:
		     break;
		   default:
		     throw new InterpretError("actualizeEquiv(): to be implemented..");

		 }//end of switch
		 if(breakflag==1) break;
	}//end of for loop
		 //YuLing's modification ends here

	if(i != byarray.length){
	   // sort value changed, so must write the sub-sum value to
	   // the cell.
	   for(int j = oldpos; j < currow; j++){
		  switch(virtype){
		  case INTEGER:
			 ((int[])destrel.data[pos])[j] = 
					((Integer)virtree.info).intValue();
			 break;
			   case SHORT:
				  ((short[])destrel.data[pos])[j] = 
						 (short)((Integer)virtree.info).intValue();
			 break;
			   case LONG:
				  ((long[])destrel.data[pos])[j] = 
						 ((Long)virtree.info).longValue();
			 break;
			   case FLOAT:
//System.out.println("in Actualizing 881, pos, j, relation name:" + pos + " " + j + " " + destrel.name);
				  ((float[])destrel.data[pos])[j] = 
						 (float)((Double)virtree.info).doubleValue();
			 break;
			   case DOUBLE:
				  ((double[])destrel.data[pos])[j] = 
						 ((Double)virtree.info).doubleValue();
			 break;
			   case STRING:
		  case STMT:
		  case EXPR:
		  case TEXT:
				  ((String[])destrel.data[pos])[j] = (String)virtree.info;
			 break;
			   case IDLIST:
				  break;
			   default:
				  throw new InterpretError("actualizeEquiv(): to be implemented..");
		  }
	   } 

	   // reset the sum field.
	   switch(virtype){
	   case INTEGER:
			case SHORT:
			   //virtree.info = new Integer(0);
			   virtree.info = null;//YuLing to fix equiv min
		  break;
			case LONG:
			   //virtree.info = new Long(0);
			   virtree.info=null;//YuLing to fix equiv min
		  break;
			case DOUBLE:
			case FLOAT:
			   //virtree.info = new Double(0.0);
			   virtree.info = null; //YuLing to fix equiv min
			   break;
			case STRING:
			case STMT:
			case EXPR:
			case TEXT:
			   virtree.info = "";
			   break;
			case IDLIST:
			   break;
			default:
			   throw new InterpretError("actualizeEquiv(): to be implemented..");
			}
			oldpos = currow;
	}

		 // Main loops come here...
		 switch(virtype){
		 case INTEGER:
		 case SHORT:
			actIntCellWithNull(virtree);   // the sum value is stored in the node.
			break;
		 case LONG:
			actLongCellWithNull(virtree);
			break;
		 case FLOAT:
		 case DOUBLE:
			actDoubleCellWithNull(virtree);
			break;
		 case STRING:
		 case STMT:
		 case EXPR:
		 case TEXT:
			actStrCellWithNull(virtree);
			break;
		 case IDLIST:
			Relation r = actRelCellWithNull(virtree, vdom, rname, rindx, relhtbl,
							 tname0, tname1);
			break;
		 default:
			throw new InterpretError("actualizeEquiv(): to be implemented..");
		 }
	     }

	  // Write the sub-sum value for the last group in data[] array.
	  for(int j = oldpos; j < currow; j++){
		 switch(virtype){
			case INTEGER:
			   ((int[])destrel.data[pos])[j] = 
					((Integer)virtree.info).intValue();
			   break;
			case SHORT:
			   ((short[])destrel.data[pos])[j] = 
						 (short)((Integer)virtree.info).intValue();
			   break;
			case LONG:
			   ((long[])destrel.data[pos])[j] = 
						 ((Long)virtree.info).longValue();
			   break;
			case FLOAT:
//System.out.println("in Actualizing 973, pos, j, relation name:" + pos + " " + j + " " + destrel.name);
			   ((float[])destrel.data[pos])[j] = 
						 (float)((Double)virtree.info).doubleValue();
			   break;
			case DOUBLE:
			   ((double[])destrel.data[pos])[j] = 
						 ((Double)virtree.info).doubleValue();
			   break;
			case STRING:
			case STMT:
			case EXPR:
			case TEXT:
			   ((String[])destrel.data[pos])[j] = 
						 (String)virtree.info;
			   break;
			case IDLIST:
			   break;
			default:
			   throw new InterpretError("actualizeEquiv(): to be implemented..");
		 }
	  } 
   }         

   /**
	* Actualize the virtual domains (including RED virtual domain).
        *  
        *  JRelixNS modified/enhanced this method:
        *   1) Put the relhtbl in object scope this.relhtbl, if it is null;
        *   as nullary-Relation/IDLIST and boolean are now compatible,
        *   we need to pass the relhtbl for scalars as well.
        *   2) We must set the NRenv.row for all types as scalar expressions
        *   may now contain IDLISTs.
	*   3) For empty nested relation, use Surrogate 0.
	*/
   private void actualizingWithNull(Domain[] doms, int type)
	   throws InterpretError{

	  // Intermediate arrays are used for speeding up the tuple-level
	  //  actualization.
//System.out.println("enter into actualizing");
	  int[] virtype = new int[doms.length];
	  SimpleNode[] virtree = new SimpleNode[doms.length];
	  int pos[] = new int[doms.length];

	  // Varialbes for IDLIST domains and when nt == false.
	  String[] tname0 = null, tname1 = null;
	  int idcount = 0;
	  Hashtable relhtbl = null;
	  Vector[] rname = null;
	  Vector[] rindx = null;
	  Vector[] vdom = null;
	  Relation[] dest = null;

	  // Following loop is only for IDLIST domains and when nt == false.
//System.out.println("nt:" + nt + " IDLIST " + IDLIST);
	  if(!nt){
//System.out.println("doms.length:" + doms.length);
		 for(int i = 0; i < doms.length; i++) {
//System.out.println("doms[" + i + "].type:" + doms[i].type);
			if(doms[i].type == IDLIST) idcount++;
		}

		 relhtbl = new Hashtable();
                 // START Since jRelixNS
                 // Put the relhtbl in this.relhtbl, if it is null.
                 if (Relation.isSigmaJoinImplemented == true){
                    if (this.relhtbl == null) 
                      this.relhtbl = relhtbl;
                 }
                 // END Since jRelixNS
		 rname = new Vector[idcount];
		 rindx = new Vector[idcount];
		 vdom = new Vector[idcount];
		 dest = new Relation[idcount];
		 tname0 = new String[idcount];
		 tname1 = new String[idcount];

		 idcount = 0;    // reset idcount.
	  }

	  for(int i = 0; i < doms.length; i++){
		 // Fill the virtual domain's tree array.
		 virtree[i] = (SimpleNode)vvtree.get(doms[i]);

		 // Fill the types array of virtual domains. 
		 if((virtype[i] = doms[i].type) == IDLIST && !nt){
			vdom[idcount] = new Vector();
			rname[idcount] = new Vector();
			rindx[idcount] = new Vector();
			tname0[idcount] = interpreter.nextTempName();
			tname1[idcount] = interpreter.nextTempName();

			// Add this name to tempnames vector for future clean-up.
			tempnames.addElement(tname0[idcount]);
			tempnames.addElement(tname1[idcount]);


			preProcIDListDom(virtree[i], vdom[idcount], rname[idcount],
				 rindx[idcount]);

//            dest[idcount] = reltable.getRel("."+doms[i].name);
//REVISIT: Zhongxia, I used true but check it please
			dest[idcount] = env.lookupRel("."+doms[i].name, true);

			idcount++;
		 }

		 // Fill the virtual domain's position array.
		 pos[i] = alldoms.indexOf(doms[i]);
//System.out.println("pos[i]:" + pos[i]);
	  }

// NREnvironment nrEnv = new NREnvironment(destrel, null, env);
// relhtbl.put("#env2", nrEnv);
   NREnvironment NRenv = env.addNRFrame(destrel);
   if (Relation.isSigmaJoinImplemented == true)
     this.NRenv = NRenv;
	  // Now actualize the virtual domain tuple by tuple.
	  for(currow = 0; currow < destrel.numtuples; currow++){
//System.out.println("currow:" + currow);
		 idcount = 0;
		 for(int j = 0; j < doms.length; j++){
//System.out.println("virtype[j]:" + virtype[j]);
           // START Since jRelixNS
           // Since it is NOT only the IDLIST domain tree may contain an IDLIST
           // domain with nullary-Relation/IDLIST - boolean compatiblity,
           // (e.g. IF-THEN-ELSE statements that may returning any type,
           // such as numeric or string, may still have an IDLIST
           // in the IF clause.) we must set the NRenv.row for all types. 
           if (Relation.isSigmaJoinImplemented == true)
	      NRenv.row = currow;
           // END Since jRelixNS
	   switch(virtype[j]){
	  // START Since JRelCaNu d.
	  // Changed from actXYZCell() to actXYZCellWithNull().
	   case INTEGER:
			   ((int[])destrel.data[pos[j]])[currow]
						= actIntCellWithNull(virtree[j]);
		  break;
	   case BOOLEAN:
			 ((byte[])destrel.data[pos[j]])[currow] = actBoolCellWithNull(virtree[j]);
		  break;
	  // END Since JRelCaNu d.
	   case IDLIST: {
		  // START Since jRelixNS
                  if (Relation.isSigmaJoinImplemented != true)
		     NRenv.row = currow;
                  // END Since jRelixNS

			   Relation r = actRelCellWithNull(virtree[j], vdom[idcount],
				 rname[idcount], rindx[idcount], relhtbl,
				 tname0[idcount],tname1[idcount]);
			   if(virtree[j].type != OP_RED && virtree[j].type != OP_EQUIV){
                              // BEGIN jRelixNS BUGFIX 
	                      // Don't load data if already in memeory.
	                      // IDLISTs may not have been saved to files and will
	                      // cause an error in this case if we do a load();
			      if (r.data == null) 
		              // END Since jRelixNS BUGFIX
				  r.load();   // have to load since its saved and cleared.
				  // Generate new surrogate in destrel.
		                  // START Since jRelixNS
				  // For empty nested relation, use Surrogate 0.
				  long sur;
                                  if (Relation.isSigmaJoinImplemented != true)
				    sur = Surrogate.next();
                                  else
                                  {
				    if (r.numtuples==0)
				      sur = 0;
                                    else
				      sur = Surrogate.next();
                                  }
                                  // END Since jRelixNS

				  ((long[])destrel.data[pos[j]])[currow] = sur;

				  // We must write the new surrogate for the new sub-relation.
				  // REVISIT: can be well fine-tuned!
				  Domain[] darry = new Domain[r.domains.length + 1];

//                  darry[0] = domtable.getDom(".id");
//REVISIT: Zhongxia, I used true but check it please
				  darry[0] = env.lookupDom(".id", true);

				  for(int i = 0; i < r.domains.length; i++)
					 darry[i+1] = r.domains[i];
			   
				  Object[] dat = Relation.createDataArray(darry, r.numtuples);
				  for(int i = 0; i < r.domains.length; i++)
					 dat[i+1] = r.data[i];
				  for(int i = 0; i < r.numtuples; i++)
					 ((long[])dat[0])[i] = sur;

				  dest[idcount].append(dat, r.numtuples);
			   }else{
				  // type is OP_RED or OP_EQUIV, Do nothing.
			   }
			   idcount++;
		  break;
			}
	   case SHORT:
                           //START Since JRelCaNu e.
		           // Use CastToShort.value() instead of Java (short) cast.
			   ((short[])destrel.data[pos[j]])[currow]
						 = CastToShort.value(actIntCellWithNull(virtree[j]));
						 //(short) actIntCellWithNull(virtree[j]);
                           //END Since JRelCaNu e.
		  break;
	   case LONG:
			   ((long[])destrel.data[pos[j]])[currow]
						 = actLongCellWithNull(virtree[j]);
		  break;
	   case FLOAT:
                           //START Since JRelCaNu e.
		           // Use CastToFloat.value() instead of Java (float) cast.
			   ((float[])destrel.data[pos[j]])[currow] 
						 = CastToFloat.value(actDoubleCellWithNull(virtree[j]));
						 //(float) actDoubleCellWithNull(virtree[j]);
                           //END Since JRelCaNu e.
		  break;
	   case DOUBLE:
			   ((double[])destrel.data[pos[j]])[currow]
						 = actDoubleCellWithNull(virtree[j]);
		  break;
	   case STRING:
	   case STMT:
	   case EXPR:
	   case TEXT:
			   ((String[])destrel.data[pos[j]])[currow] = actStrCellWithNull(virtree[j]);
		  break;
			default:
			   throw new InterpretError("actualizing(): to be implemented..");
	   }
	}
	  }
	  env.removeNRFrame();

	  // Check if there're OP_RED domains. They require special process.
	  idcount = 0;
	  for(int i = 0; i < doms.length; i++){
		 if(virtype[i] == IDLIST && virtree[i].type == OP_RED){
			// Generate a new surrogate.
			long sur = Surrogate.next();

			// Get the reducted relation and try to append it to
			// nested domain's data field.
			Relation r = (Relation) virtree[i].info;
			r.load();  // have to load it. It seems be saved and cleared.

			// We must write the new surrogate for the new sub-relation.
			Domain[] darry = new Domain[r.domains.length + 1];

//            darry[0] = domtable.getDom(".id");
//REVISIT: Zhongxia, I used true but check it please
			darry[0] = env.lookupDom(".id", true);

			for(int j = 0; j < r.domains.length; j++)
			   darry[j+1] = r.domains[j];
			   
			Object[] dat = Relation.createDataArray(darry, r.numtuples);
			for(int j = 0; j < r.domains.length; j++)
			   dat[j+1] = r.data[j];
			for(int j = 0; j < r.numtuples; j++)
			   ((long[])dat[0])[j] = sur;

			dest[idcount].append(dat, r.numtuples);
			dest[idcount].load();  // have to load it since append to disk.

			// Save the new surrogate into virtree[i].info for next step's
			// use. Note, we don't need the reducted relation stored in
			// virtree[i].info now.
			virtree[i].info = new Long(sur);
		 }
		 idcount++;
	  }

	  // Must go for another pass for RED operations.
	  if(type == 1){
		 for(currow = 0; currow < destrel.numtuples; currow++){
	   for(int j = 0; j < doms.length; j++){
		  switch(virtype[j]){
		  case INTEGER:
				  ((int[])destrel.data[pos[j]])[currow]
					   = ((Integer)virtree[j].info).intValue();
			 break;
		  case BOOLEAN:
		                 // START Since JRelCaNu f.
		                  // info is a Byte object.
				((byte[])destrel.data[pos[j]])[currow] = ((Byte)virtree[j].info).byteValue();
		                 // END Since JRelCaNu f.
			 break;
			   case IDLIST:
				  ((long[])destrel.data[pos[j]])[currow] = 
						 ((Long)virtree[j].info).longValue();
				  break;
		  case SHORT:
                                 //START Since JRelCaNu e.
		                 // Use CastToShort.value instead of Java (short) cast.
				  ((short[])destrel.data[pos[j]])[currow]
					   = CastToShort.value(((Integer)virtree[j].info).intValue());
                                 //END Since JRelCaNu e.
			 break;
		  case LONG:
				  ((long[])destrel.data[pos[j]])[currow]
					   = ((Long)virtree[j].info).longValue();
			 break;
		  case FLOAT:
//System.out.println("in Actualizing 2280, srcrel.numattrs, j, currow, relation name:" + srcrel.numattrs + " " + j + " " + currow + " " + destrel.name);
                                  //START Since JRelCaNu e.
		                  // Use CastToFloat.value instead of Java (float) cast.
		                  // Also info is a Double object in actDoubleCellWithNull.
				  ((float[])destrel.data[pos[j]])[currow]
					   = CastToFloat.value(((Double)virtree[j].info).doubleValue());
					   //(float)((Integer)virtree[j].info).doubleValue();
                                  //END Since JRelCaNu e.
			 break;
		  case DOUBLE:
                                  //START Since JRelCaNu e.
		                  // info is a Double object in actDoubleCellWithNull.
				  ((double[])destrel.data[pos[j]])[currow]
					   = ((Double)(virtree[j].info)).doubleValue();
					   //((Integer)virtree[j].info).doubleValue();
                                  //END Since JRelCaNu e.
			 break;
		  case STRING:
		  case STMT:
		  case EXPR:
		  case TEXT:
				  ((String[])destrel.data[srcrel.numattrs+j])[currow]
					   = (String)virtree[j].info;
			 break;
			   default:
				  throw new InterpretError("actualize(): to be implemented..");
		  }
	   }
		 }
	  }
   }      

    /**
     *  Actualize a boolean typed cell.
     *  
     *  JRelCaNu: we have to change the type from Boolean to Byte.
     */
    private byte actBoolCellWithNull(SimpleNode node) throws InterpretError {
        switch (node.type) {
        case OP_BIOPERATOR:
            SimpleNode left = (SimpleNode) node.jjtGetChild(0);
            SimpleNode right = (SimpleNode) node.jjtGetChild(1);

            switch (node.opcode) {
            case OP_OR: {
                    byte bm = actBoolCellWithNull(left);

                    if (bm == BOOL_TRUE)
                        return BOOL_TRUE;
                    byte bn = actBoolCellWithNull(right);

                    if (bn == BOOL_TRUE)
                        return BOOL_TRUE;
                    else if (bm == BOOL_DK || bn == BOOL_DK)
                        return BOOL_DK;
                    else if (bm == BOOL_DC && bn == BOOL_DC)
                        return BOOL_DC;
                    else //if (bm == BOOL_FALSE || bn  == BOOL_FALSE)
                        return BOOL_FALSE;
                }

            case OP_AND: {
                    byte bm = actBoolCellWithNull(left);

                    if (bm == BOOL_FALSE)
                        return BOOL_FALSE;
                    byte bn = actBoolCellWithNull(right);

                    if (bn == BOOL_FALSE)
                        return BOOL_FALSE;
                    else if (bm == BOOL_DK || bn == BOOL_DK)
                        return BOOL_DK;
                    else if (bm == BOOL_DC && bn == BOOL_DC)
                        return BOOL_DC;
                    else //if (bm == BOOL_TRUE || bn  == BOOL_TRUE)
                        return BOOL_TRUE;
                }

            case OP_EQ:
            case OP_NEQ:
            case OP_GT:
            case OP_LT:
            case OP_GE:
            case OP_LE:
                // We must initialize to some value to satisfy javac.
                // However, we must make sure they are assigned
                // in the switch statement for DC and DK Pre-processing.
                // We could check DK and DC within each operator, but this 
                // pre-checking saves quite a few lines of repeated code. 
                byte bm = BOOL_DK;
                byte bn = BOOL_DK;
                int im = INT_DK;
                int in = INT_DK;
                long lm = LONG_DK;
                long ln = LONG_DK;
                double dm = DOUBLE_DK;
                double dn = DOUBLE_DK;
                String sm = STRING_DK;
                String sn = STRING_DK;

                // First, look for and process any DC and DK null values.
                switch (node.bits) {
                case BOOLEAN:
                    bm = actBoolCellWithNull(left);
                    if (bm == BOOL_DC)
                        return bm;
                    bn = actBoolCellWithNull(right);
                    if (bn == BOOL_DC)
                        return BOOL_DC;
                    else if (bn == BOOL_DK || bn == BOOL_DK)
                        return BOOL_DK;
                    break;

                case SHORT:
                case INTEGER:
                    im = actIntCellWithNull(left);
                    if (im == INT_DC)
                        return BOOL_DC;
                    in = actIntCellWithNull(right);
                    if (in == INT_DC)
                        return BOOL_DC;
                    else if (im == INT_DK || in == INT_DK)
                        return BOOL_DK;
                    break;

                case LONG:
                    lm = actLongCellWithNull(left);
                    if (lm == LONG_DC)
                        return BOOL_DC;
                    ln = actLongCellWithNull(right);
                    if (ln == LONG_DC)
                        return BOOL_DC;
                    else if (lm == LONG_DK || ln == LONG_DK)
                        return BOOL_DK;
                    break;

                case FLOAT:
                case DOUBLE:
                    dm = actDoubleCellWithNull(left);
                    if (dm == DOUBLE_DC)
                        return BOOL_DC;
                    dn = actDoubleCellWithNull(right);
                    if (dn == DOUBLE_DC)
                        return BOOL_DC;
                    else if (dm == DOUBLE_DK || dn == DOUBLE_DK)
                        return BOOL_DK;
                    break;

                case STRING:
                    sm = actStrCellWithNull(left);
                    if (sm.compareTo(STRING_DC) == 0)
                        return BOOL_DC;
                    sn = actStrCellWithNull(right);
                    if (sn.compareTo(STRING_DC) == 0)
                        return BOOL_DC;
                    if (sm.compareTo(STRING_DK) == 0 || sn.compareTo(STRING_DK) == 0)
                        return BOOL_DC;
                    break;

                case DC:
                    return BOOL_DC;

                case DK:
                    return BOOL_DK;

                case NUMERIC:
                case IDLIST:
                default:
                    Utility.dump(node, "actBoolCellWithNull ERROR=>");
                    throw new InterpretError("actBoolCellWithNull: to be implemented.");
                }
                // There are no DC and DK null values when we reach here.
                switch (node.opcode) {
                case OP_EQ: {
                        switch (node.bits) {
                        case BOOLEAN:
                            return (bm == bn) ? BOOL_TRUE : BOOL_FALSE;

                        case SHORT:
                        case INTEGER:
                            return (im == in) ? BOOL_TRUE : BOOL_FALSE;

                        case LONG:
                            return (lm == ln) ? BOOL_TRUE : BOOL_FALSE;

                        case FLOAT:
                        case DOUBLE:
                            return (dm == dn) ? BOOL_TRUE : BOOL_FALSE;

                        case STRING:
                            return (sm.compareTo(sn) == 0) ? BOOL_TRUE : BOOL_FALSE;

                        case NUMERIC:
                        case IDLIST:
                        default:
                            throw new InterpretError("actBoolCellWithNull: to be implemented.");
                        }
                    }

                case OP_NEQ: {
                        switch (node.bits) {
                        case BOOLEAN:
                            return (bm != bn) ? BOOL_TRUE : BOOL_FALSE;

                        case SHORT:
                        case INTEGER:
                            return (im != in) ? BOOL_TRUE : BOOL_FALSE;

                        case LONG:
                            return (lm != ln) ? BOOL_TRUE : BOOL_FALSE;

                        case FLOAT:
                        case DOUBLE:
                            return (dm != dn) ? BOOL_TRUE : BOOL_FALSE;

                        case STRING:
                            return (sm.compareTo(sn) != 0) ? BOOL_TRUE : BOOL_FALSE;

                        case NUMERIC:
                        case IDLIST:
                        default:
                            throw new InterpretError("actBoolCellWithNull: to be implemented.");
                        }
                    }

                case OP_GT: {
                        switch (node.bits) {
                        case BOOLEAN:
                            return (bm > bn) ? BOOL_TRUE : BOOL_FALSE;

                        case SHORT:
                        case INTEGER:
                            return (im > in) ? BOOL_TRUE : BOOL_FALSE;

                        case LONG:
                            return (lm > ln) ? BOOL_TRUE : BOOL_FALSE;

                        case FLOAT:
                        case DOUBLE:
                            return (dm > dn) ? BOOL_TRUE : BOOL_FALSE;

                        case STRING:
                            return (sm.compareTo(sn) > 0) ? BOOL_TRUE : BOOL_FALSE;

                        case NUMERIC:
                        case IDLIST:
                        default:
                            throw new InterpretError("actBoolCellWithNull: to be implemented.");
                        }
                    }

                case OP_LT: {
                        switch (node.bits) {
                        case BOOLEAN:
                            return (bm < bn) ? BOOL_TRUE : BOOL_FALSE;

                        case SHORT:
                        case INTEGER:
                            return (im < in) ? BOOL_TRUE : BOOL_FALSE;

                        case LONG:
                            return (lm < ln) ? BOOL_TRUE : BOOL_FALSE;

                        case FLOAT:
                        case DOUBLE:
                            return (dm < dn) ? BOOL_TRUE : BOOL_FALSE;

                        case STRING:
                            return (sm.compareTo(sn) < 0) ? BOOL_TRUE : BOOL_FALSE;

                        case NUMERIC:
                        case IDLIST:
                        default:
                            throw new InterpretError("actBoolCellWithNull: to be implemented.");
                        }
                    }

                case OP_GE: {
                        switch (node.bits) {
                        case BOOLEAN:
                            return (bm >= bn) ? BOOL_TRUE : BOOL_FALSE;

                        case SHORT:
                        case INTEGER:
                            return (im >= in) ? BOOL_TRUE : BOOL_FALSE;

                        case LONG:
                            return (lm >= ln) ? BOOL_TRUE : BOOL_FALSE;

                        case FLOAT:
                        case DOUBLE:
                            return (dm >= dn) ? BOOL_TRUE : BOOL_FALSE;

                        case STRING:
                            return (sm.compareTo(sn) >= 0) ? BOOL_TRUE : BOOL_FALSE;

                        case NUMERIC:
                        case IDLIST:
                        default:
                            throw new InterpretError("actBoolCellWithNull: to be implemented.");
                        }
                    }

                case OP_LE: {
                        switch (node.bits) {
                        case BOOLEAN:
                            return (bm <= bn) ? BOOL_TRUE : BOOL_FALSE;

                        case SHORT:
                        case INTEGER:
                            return (im <= in) ? BOOL_TRUE : BOOL_FALSE;

                        case LONG:
                            return (lm <= ln) ? BOOL_TRUE : BOOL_FALSE;

                        case FLOAT:
                        case DOUBLE:
                            return (dm <= dn) ? BOOL_TRUE : BOOL_FALSE;

                        case STRING:
                            return (sm.compareTo(sn) <= 0) ? BOOL_TRUE : BOOL_FALSE;

                        case NUMERIC:
                        case IDLIST:
                        default:
                            throw new InterpretError("actBoolCellWithNull: to be implemented.");
                        }
                    }

                default:
                    Utility.dump(node, "switched on node.type " + node.type + " in actBoolCellWithNull->");
                    throw new InterpretError("actBoolCellWithNull: to be implemented.");
                }

            case OP_CAST:
                // Cast from type had been stored in left.bits of 
                // the the left OP_TYPE node in buildTree and
                // traverseType; so, we can use it in the switch.
                switch (left.bits) {
                case BOOLEAN:
                    return actBoolCellWithNull(right);

                case STRING:
                    return CastToBoolByte.value(actStrCellWithNull(right));

                case SHORT:
                    return CastToBoolByte.value((short)CastToShort.value(actIntCellWithNull(right)));

                case INTEGER:
                    return CastToBoolByte.value(actIntCellWithNull(right));

                case LONG:
                    return CastToBoolByte.value(actLongCellWithNull(right));

                case DOUBLE:
                    return CastToBoolByte.value(actDoubleCellWithNull(right));

                case FLOAT:
                    // Must cast to float first, otherwise, we may get
                    // a bigger number than possible for float in the
                    // result.
                    return CastToBoolByte.value(CastToFloat.value(actDoubleCellWithNull(right)));

                case IDLIST:
                    return (actBoolRelCellWithNull(right, null, null, null, null, null)) ? BOOL_TRUE : BOOL_FALSE;

                case NUMERIC:
                default:
                    Utility.dump(node, "switched on node.bits" + node.bits + " in actBoolCellWithNull->");
                    throw new InterpretError("actBoolCellWithNull: to be implemented.");
                }

            default:
                throw new InterpretError("actBoolCellWithNull: OP_BIO to be implemented..");
            }

        case OP_UOPERATOR:
            if (node.opcode == OP_UNOT) {
                byte bm = actBoolCellWithNull((SimpleNode) node.jjtGetChild(0));

                if (bm == BOOL_TRUE)
                    return BOOL_FALSE;
                else if (bm == BOOL_FALSE)
                    return BOOL_TRUE;
                else 
                    return bm; // either BOOL_DC or BOOL_DK; 
            } else {
                Utility.dump(node, "UOPERATOR node.opcode:" + node.opcode);
                throw new InterpretError("actBoolCellWithNull: UOPERATOR is not UNOT...");
            }

        case OP_IDENTIFIER:
    
            // Check if an OP_IDENTIFIER is defined as a BOOLEAN or IDLIST 
            // in the original relation and handle accordingly.
            // Specifically, have to call actBoolRelCellWithNull for IDLIST.
            if (Relation.isSigmaJoinImplemented == true) {
                switch (destrel.domains[node.bits].type) {
                case BOOLEAN:
                    return ((byte[]) destrel.data[node.bits])[currow];

                case IDLIST:
                    return (actBoolRelCellWithNull(node, null, null, null, null, null)) ? BOOL_TRUE : BOOL_FALSE;

                default:
                    throw new InterpretError("actBoolCellWithNull: OP_BIO to be implemented..");
                }
            } else {
                return ((byte[]) destrel.data[node.bits])[currow];
            }

        case OP_LITERAL:
            if (node.opcode == OP_DC) 
                return ((byte) BOOL_DC);
            else if (node.opcode == OP_DK) 
                return ((byte) BOOL_DK);
            else if (node.info instanceof Byte)
                return (((Byte) node.info).byteValue());
            else {
                throw new InterpretError("actBoolCellWithNull: Not a boolean literal.");
            }

        case OP_RED:
        case OP_EQUIV:
            byte actValue = (byte) BOOL_DK;
		 
            if (node.info == null)  node.info = new Byte(BOOL_DC);
            switch (node.opcode) {
            case OP_OR: {
                    byte bm = ((Byte) node.info).byteValue();

                    if (bm == BOOL_TRUE)
                        node.info = new Byte(BOOL_TRUE);
                    byte bn = actValue = actBoolCellWithNull((SimpleNode) node.jjtGetChild(0));

                    if (bn == BOOL_TRUE)
                        node.info = new Byte(BOOL_TRUE);
                    else if (bm == BOOL_DK || bn == BOOL_DK)
                        node.info = new Byte(BOOL_DK);
                    else if (bm == BOOL_DC && bn == BOOL_DC)
                        node.info = new Byte(BOOL_DC);
                    else //if (bm == BOOL_FALSE || bn  == BOOL_FALSE)
                        node.info = new Byte(BOOL_FALSE);
                }
                break;

            case OP_AND: {
                    byte bm = ((Byte) node.info).byteValue();

                    if (bm == BOOL_FALSE)
                        node.info = new Byte(BOOL_FALSE);
                    byte bn = actValue = actBoolCellWithNull((SimpleNode) node.jjtGetChild(0));

                    if (bn == BOOL_FALSE)
                        node.info = new Byte(BOOL_FALSE);
                    else if (bm == BOOL_DK || bn == BOOL_DK)
                        node.info = new Byte(BOOL_DK);
                    else if (bm == BOOL_DC && bn == BOOL_DC)
                        node.info = new Byte(BOOL_DC);
                    else //if (bm == BOOL_TRUE || bn  == BOOL_TRUE)
                        node.info = new Byte(BOOL_TRUE);
                }
                break;

            default:
                throw new InterpretError("actBoolCellWithNull: Red/Equiv to be implemented.");
            }
            return actValue;

        case OP_IFTHENELSE:
            // If it is DC, we just say it is false.
            // If it is DK, we throw an exception?
            // Within a boolean expression, we do not do anything except to
            // follow the logic. External caller must say DK or DC is okay or
            // not.
            byte ifVal = actBoolCellWithNull((SimpleNode) node.jjtGetChild(0));

            if (ifVal == (byte) BOOL_TRUE)
                return actBoolCellWithNull((SimpleNode) node.jjtGetChild(1));
            else if (ifVal == (byte) BOOL_FALSE)
                return actBoolCellWithNull((SimpleNode) node.jjtGetChild(2));
            else if (ifVal == (byte) BOOL_DC) {
                System.out.println("WARNING: actBoolCellWithNull: 'dc' encounterred by if-then-else.");
                System.out.println("Actualizing using the 'then' value");
                return actBoolCellWithNull((SimpleNode) node.jjtGetChild(2));
            } else //if(ifVal  == (byte)BOOL_DK)
                throw new InterpretError("actBoolCellWithNull: 'dk' encounterred by if-then-else.");

        default:
            // START Since jRelixNS
            if (Relation.isSigmaJoinImplemented == true)
                return (actBoolRelCellWithNull(node, null, null, null, null, null)) ? (byte) BOOL_TRUE : (byte) BOOL_FALSE;
            else //if (Relation.isSigmaJoinImplemented != true)
                throw new InterpretError("actBoolCellWithNull: to be implemented..");
            // END Since jRelixNS
        }
    }      

    /**
     *  Actualize a double typed cell.
     */
    private double actDoubleCellWithNull(SimpleNode node) throws InterpretError {
        double ACT_TYPED_DC = DOUBLE_DC;
        double ACT_TYPED_DK = DOUBLE_DK;

        switch (node.type) {
        case OP_IDENTIFIER:
            // Start CAST
            // We must handle the JRelix DC and DK values.
            // Thus, we invoke the JRelix CastToXYZ.value() methods
            // instead of using the Java primitive type casts.
            switch (destrel.domains[node.bits].type) {
            case DOUBLE:
                return CastToDouble.value(((double[]) destrel.data[node.bits])[currow]);

            case FLOAT:
                return CastToDouble.value(((float[]) destrel.data[node.bits])[currow]);

            case INTEGER:
                return CastToDouble.value(((int[]) destrel.data[node.bits])[currow]);

            case SHORT:
                return CastToDouble.value(((short[]) destrel.data[node.bits])[currow]);

            case LONG:
                return CastToDouble.value(((long[]) destrel.data[node.bits])[currow]);

            case NUMERIC:
                return CastToDouble.value(((number[]) destrel.data[node.bits])[currow]);

            default:
                throw new InterpretError("actDoubleCellWithNull :to be implemented.");
            }

        case OP_BIOPERATOR:
            SimpleNode left = (SimpleNode) node.jjtGetChild(0);
            SimpleNode right = (SimpleNode) node.jjtGetChild(1);
            double theBiOpVal = ACT_TYPED_DK;
            double mVal = ACT_TYPED_DK;
            double nVal = ACT_TYPED_DK;;
            if (node.opcode != OP_CAST) {
                mVal = actDoubleCellWithNull(left);
                if (mVal == ACT_TYPED_DK)
                    return ACT_TYPED_DK;
                nVal = actDoubleCellWithNull(right);
                if (nVal == ACT_TYPED_DK)
                    return ACT_TYPED_DK;
                else if (mVal == ACT_TYPED_DC && nVal == ACT_TYPED_DC)
                    return ACT_TYPED_DC;
                if (nVal == ACT_TYPED_DC)
                    return mVal;
            }

            switch (node.opcode) {
            case OP_PLUS:
                if (mVal == ACT_TYPED_DC)
                    return nVal;
                theBiOpVal = mVal + nVal;
                break;

            case OP_MINUS:
                if (mVal == ACT_TYPED_DC)
                    theBiOpVal = -nVal;
                else 
                    theBiOpVal = mVal - nVal;
                break;

            case OP_MUL:
                if (mVal == ACT_TYPED_DC)
                    return nVal;
                theBiOpVal = mVal * nVal;
                break;

            case OP_DIV:
                if (mVal == ACT_TYPED_DC)
                    theBiOpVal = 1 / nVal;
                else 
                    theBiOpVal = mVal / nVal;
                break;

            case OP_CAST:
                // Cast from type stored in left.bits of 
                // the the left OP_TYPE node in buildTree and
                // traverseType.
                switch (left.bits) {
                case BOOLEAN:
                    return CastToDouble.value(actBoolCellWithNull(right));

                case STRING:
                    return CastToDouble.value(actStrCellWithNull(right));

                case INTEGER:
                    return CastToDouble.value(actIntCellWithNull(right));

                case SHORT:
                    return CastToDouble.value((short)CastToShort.value(actIntCellWithNull(right)));

                case LONG:
                    return CastToDouble.value(actLongCellWithNull(right));

                case FLOAT:
                    // Must cast to float first, otherwise, we may get
                    // a bigger number than possible for float in the
                    // result.
                    return CastToDouble.value(CastToFloat.value(actDoubleCellWithNull(right)));

                case DOUBLE:
                    return CastToDouble.value(actDoubleCellWithNull(right));

                case NUMERIC:
                    return CastToDouble.value(actNumCellWithNull(right));

                default:
                    throw new InterpretError("actDoubleCellWithNull :to be implemented.");
                }

            default:
                throw new InterpretError("actDoubleCellWithNull :bio to be implemented.. .");
            }
            if (theBiOpVal <= ACT_TYPED_DC)
                throw new InterpretError("actDoubleCellWithNull :OP_BIOPERATOR result out of range.");
            else
                return theBiOpVal;

        case OP_UOPERATOR:
            switch (node.opcode) {
            case OP_UPLUS:
                return (actDoubleCellWithNull((SimpleNode) node.jjtGetChild(0)));

            case OP_UMINUS:
                mVal = actDoubleCellWithNull((SimpleNode) node.jjtGetChild(0));
                if (mVal == ACT_TYPED_DC || mVal == ACT_TYPED_DK)
                    return mVal;
                else
                if (-mVal <= ACT_TYPED_DC) 
                    throw new InterpretError("actDoubleCellWithNull :OP_UMINUS result out of range.");
                else 
                    return -mVal;

            default:
                throw new InterpretError("actDoubleCellWithNull :uni to be implemented.");
            }

        case OP_LITERAL:
            if (node.opcode == OP_DC)
                return (ACT_TYPED_DC);
            else if (node.opcode == OP_DK) 
                return (ACT_TYPED_DK);
            else if (node.info instanceof Double) 
                mVal = CastToDouble.value(((Double) node.info).doubleValue());
            else if (node.info instanceof Float) 
                mVal = CastToDouble.value(((Float) node.info).floatValue());
            else if (node.info instanceof Integer) 
                mVal = CastToDouble.value(((Integer) node.info).intValue());
            else if (node.info instanceof Short) 
                mVal = CastToDouble.value(((Short) node.info).shortValue());
            else if (node.info instanceof Long) 
                mVal = CastToDouble.value(((Long) node.info).longValue());
            else
                throw new InterpretError("actDoubleCellWithNull :to be implemented.");
            if (mVal == ACT_TYPED_DC || mVal == ACT_TYPED_DK)
                // The Parser/Interpreter must ensure that the values
                // of the literals are not reserved for null values.
                throw new InterpretError("actDoubleCellWithNull :literal used value reserved for Dc or Dk .");
            else
                return mVal;

        case OP_RED:
        case OP_EQUIV:
            // Initialize the actualized value to DK.
            double actValue = ACT_TYPED_DK;

            // Initialize the vertical value holder.
            if (node.info == null)  node.info = new Double(mVal = ACT_TYPED_DC);
            else mVal = ((Double) node.info).doubleValue();

            // Any binary operation with DK results in DK.
            if (mVal == ACT_TYPED_DK) 
                return ACT_TYPED_DK;

            actValue = nVal = actDoubleCellWithNull((SimpleNode) node.jjtGetChild(0));
            if (nVal == ACT_TYPED_DK) {
                node.info = new Double(ACT_TYPED_DK);
                return ACT_TYPED_DK;
            }

            // With both DC operands, DC is the result.
            if (mVal == ACT_TYPED_DC && nVal == ACT_TYPED_DC)
                return ACT_TYPED_DC;

            // With DC as 2nd operand, mVal/node.info is unchanged
            if (nVal == ACT_TYPED_DC) 
                return actValue;// which is ACT_TYPED_DC;

            theBiOpVal = mVal;
            switch (node.opcode) {
            case OP_PLUS: {
                    if (mVal == ACT_TYPED_DC) theBiOpVal = nVal;
                    else  theBiOpVal = mVal + nVal;
                    break;
                }

            case OP_MUL: {
                    if (mVal == ACT_TYPED_DC) theBiOpVal = nVal;
                    else  theBiOpVal = mVal * nVal;
                    break;
                }

            case OP_MIN: {
                    if (mVal == ACT_TYPED_DC) theBiOpVal = nVal;
                    else  theBiOpVal = Math.min(mVal, nVal);
                    break;
                }

            case OP_MAX: {
                    if (mVal == ACT_TYPED_DC) theBiOpVal = nVal;
                    else  theBiOpVal = Math.max(mVal, nVal);
                    break;
                }

            default:
                throw new InterpretError("actDoubleCellWithNull :Red/Equiv to be implemented.");
            }
            if (theBiOpVal <= ACT_TYPED_DC && mVal != ACT_TYPED_DC)
                throw new InterpretError("actDoubleCellWithNull :Red/Equiv out of range.");
            if (theBiOpVal != mVal)
                node.info = new Double(theBiOpVal);

            return actValue;

        case OP_IFTHENELSE: {
                byte ifVal = actBoolCellWithNull((SimpleNode) node.jjtGetChild(0));

                if (ifVal == (byte) BOOL_TRUE)
                    return actDoubleCellWithNull((SimpleNode) node.jjtGetChild(1));
                else if (ifVal == (byte) BOOL_FALSE)
                    return actDoubleCellWithNull((SimpleNode) node.jjtGetChild(2));
                else if (ifVal == (byte) BOOL_DC) {
                    System.out.println("WARNING: actDoubleCellWithNull :'dc' encounterred by if-then-else.");
                    System.out.println("Actualizing using the 'then' value");
                    return actDoubleCellWithNull((SimpleNode) node.jjtGetChild(2));
                } else// if BOOL_DK 
                    throw new InterpretError("actDoubleCellWithNull :'dk' encounterred by if-then-else.");
            }

        default:
            throw new InterpretError("actDoubleCellWithNull :to be implemented.");
        }
    }      

    /**
     *  Actualize an int typed cell.
     */
    private int actIntCellWithNull(SimpleNode node) throws InterpretError {
        int ACT_TYPED_DC = INT_DC;
        int ACT_TYPED_DK = INT_DK;

        switch (node.type) {
        case OP_IDENTIFIER:
            // Start CAST
            // We must handle the JRelix DC and DK values.
            // Thus, we invoke the JRelix CastToXYZ.value() methods
            // instead of using the Java primitive type casts.
            switch (destrel.domains[node.bits].type) {
            case DOUBLE:
                return CastToInt.value(((double[]) destrel.data[node.bits])[currow]);

            case FLOAT:
                return CastToInt.value(((float[]) destrel.data[node.bits])[currow]);

            case INTEGER:
                return CastToInt.value(((int[]) destrel.data[node.bits])[currow]);

            case SHORT:
                return CastToInt.value(((short[]) destrel.data[node.bits])[currow]);

            case LONG:
                return CastToInt.value(((long[]) destrel.data[node.bits])[currow]);

            case NUMERIC:
                return CastToInt.value(((number[]) destrel.data[node.bits])[currow]);

            default:
                throw new InterpretError("actIntCellWithNull to be implemented.");
            }

        case OP_BIOPERATOR:
            SimpleNode left = (SimpleNode) node.jjtGetChild(0);
            SimpleNode right = (SimpleNode) node.jjtGetChild(1);
            int theBiOpVal = ACT_TYPED_DK;
            int mVal = ACT_TYPED_DK;
            int nVal = ACT_TYPED_DK;;
            if (node.opcode != OP_CAST) {
                mVal = actIntCellWithNull(left);
                if (mVal == ACT_TYPED_DK)
                    return ACT_TYPED_DK;
                nVal = actIntCellWithNull(right);
                if (nVal == ACT_TYPED_DK)
                    return ACT_TYPED_DK;
                else if (mVal == ACT_TYPED_DC && nVal == ACT_TYPED_DC)
                    return ACT_TYPED_DC;
                if (nVal == ACT_TYPED_DC)
                    return mVal;
            }

            switch (node.opcode) {
            case OP_PLUS:
                if (mVal == ACT_TYPED_DC)
                    return nVal;
                theBiOpVal = mVal + nVal;
                break;

            case OP_MINUS:
                if (mVal == ACT_TYPED_DC)
                    theBiOpVal = -nVal;
                else 
                    theBiOpVal = mVal - nVal;
                break;

            case OP_MUL:
                if (mVal == ACT_TYPED_DC)
                    return nVal;
                theBiOpVal = mVal * nVal;
                break;

            case OP_DIV:
                if (mVal == ACT_TYPED_DC)
                    theBiOpVal = 1 / nVal;
                else 
                    theBiOpVal = mVal / nVal;
                break;

            case OP_CAST:
                // Cast from type stored in left.bits of 
                // the the left OP_TYPE node in buildTree and
                // traverseType.
                switch (left.bits) {
                case BOOLEAN:
                    return CastToInt.value(actBoolCellWithNull(right));

                case STRING:
                    return CastToInt.value(actStrCellWithNull(right));

                case INTEGER:
                    return CastToInt.value(actIntCellWithNull(right));

                case SHORT:
                    return CastToInt.value((short)CastToShort.value(actIntCellWithNull(right)));

                case LONG:
                    return CastToInt.value(actLongCellWithNull(right));

                case FLOAT:
                    // Must cast to float first, otherwise, we may get
                    // a bigger number than possible for float in the
                    // result.
                    return CastToInt.value(CastToFloat.value(actDoubleCellWithNull(right)));

                case DOUBLE:
                    return CastToInt.value(actDoubleCellWithNull(right));

                case NUMERIC:
                    return CastToInt.value(actNumCellWithNull(right));

                default:
                    throw new InterpretError("actIntCellWithNull to be implemented.");
                }

            default:
                throw new InterpretError("actIntCellWithNull bio to be implemented.. .");
            }
            if (theBiOpVal <= ACT_TYPED_DC)
                throw new InterpretError("actIntCellWithNull OP_BIOPERATOR result out of range.");
            else
                return theBiOpVal;

        case OP_UOPERATOR:
            switch (node.opcode) {
            case OP_UPLUS:
                return (actIntCellWithNull((SimpleNode) node.jjtGetChild(0)));

            case OP_UMINUS:
                mVal = actIntCellWithNull((SimpleNode) node.jjtGetChild(0));
                if (mVal == ACT_TYPED_DC || mVal == ACT_TYPED_DK)
                    return mVal;
                else
                if (-mVal <= ACT_TYPED_DC) 
                    throw new InterpretError("actIntCellWithNull OP_UMINUS result out of range.");
                else 
                    return -mVal;

            default:
                throw new InterpretError("actIntCellWithNull uni to be implemented.");
            }

        case OP_LITERAL:
            if (node.opcode == OP_DC)
                return (ACT_TYPED_DC);
            else if (node.opcode == OP_DK) 
                return (ACT_TYPED_DK);
            else if (node.info instanceof Double) 
                mVal = CastToInt.value(((Double) node.info).doubleValue());
            else if (node.info instanceof Float) 
                mVal = CastToInt.value(((Float) node.info).floatValue());
            else if (node.info instanceof Integer) 
                mVal = CastToInt.value(((Integer) node.info).intValue());
            else if (node.info instanceof Short) 
                mVal = CastToInt.value(((Short) node.info).shortValue());
            else if (node.info instanceof Long) 
                mVal = CastToInt.value(((Long) node.info).longValue());
            else
                throw new InterpretError("actIntCellWithNull to be implemented.");
            if (mVal == ACT_TYPED_DC || mVal == ACT_TYPED_DK)
                // The Parser/Interpreter must ensure that the values
                // of the literals are not reserved for null values.
                throw new InterpretError("actIntCellWithNull literal used value reserved for Dc or Dk .");
            else
                return mVal;

        case OP_RED:
        case OP_EQUIV:
            // Initialize the actualized value to DK.
            int actValue = ACT_TYPED_DK;

            // Initialize the vertical value holder.
            if (node.info == null)  node.info = new Integer(mVal = ACT_TYPED_DC);
            else mVal = ((Integer) node.info).intValue();

            // Any binary operation with DK results in DK.
            if (mVal == ACT_TYPED_DK) 
                return ACT_TYPED_DK;

            actValue = nVal = actIntCellWithNull((SimpleNode) node.jjtGetChild(0));
            if (nVal == ACT_TYPED_DK) {
                node.info = new Integer(ACT_TYPED_DK);
                return ACT_TYPED_DK;
            }

            // With both DC operands, DC is the result.
            if (mVal == ACT_TYPED_DC && nVal == ACT_TYPED_DC)
                return ACT_TYPED_DC;

            // With DC as 2nd operand, mVal/node.info is unchanged
            if (nVal == ACT_TYPED_DC) 
                return actValue;// which is ACT_TYPED_DC;

            theBiOpVal = mVal;
            switch (node.opcode) {
            case OP_PLUS: {
                    if (mVal == ACT_TYPED_DC) theBiOpVal = nVal;
                    else  theBiOpVal = mVal + nVal;
                    break;
                }

            case OP_MUL: {
                    if (mVal == ACT_TYPED_DC) theBiOpVal = nVal;
                    else  theBiOpVal = mVal * nVal;
                    break;
                }

            case OP_MIN: {
                    if (mVal == ACT_TYPED_DC) theBiOpVal = nVal;
                    else  theBiOpVal = Math.min(mVal, nVal);
                    break;
                }

            case OP_MAX: {
                    if (mVal == ACT_TYPED_DC) theBiOpVal = nVal;
                    else  theBiOpVal = Math.max(mVal, nVal);
                    break;
                }

            default:
                throw new InterpretError("actIntCellWithNull Red/Equiv to be implemented.");
            }
            if (theBiOpVal <= ACT_TYPED_DC && mVal != ACT_TYPED_DC)
                throw new InterpretError("actIntCellWithNull Red/Equiv out of range.");
            if (theBiOpVal != mVal)
                node.info = new Integer(theBiOpVal);
            return actValue;

        case OP_IFTHENELSE: {
                byte ifVal = actBoolCellWithNull((SimpleNode) node.jjtGetChild(0));

                if (ifVal == (byte) BOOL_TRUE)
                    return actIntCellWithNull((SimpleNode) node.jjtGetChild(1));
                else if (ifVal == (byte) BOOL_FALSE)
                    return actIntCellWithNull((SimpleNode) node.jjtGetChild(2));
                else if (ifVal == (byte) BOOL_DC) {
                    System.out.println("WARNING: actIntCellWithNull 'dc' encounterred by if-then-else.");
                    System.out.println("Actualizing using the 'then' value");
                    return actIntCellWithNull((SimpleNode) node.jjtGetChild(2));
                } else// if BOOL_DK 
                    throw new InterpretError("actIntCellWithNull 'dk' encounterred by if-then-else.");
            }

        default:
            throw new InterpretError("actIntCellWithNull to be implemented.");
        }
    }      

    /**
     *  Actualize a long typed cell.
     */
    private long actLongCellWithNull(SimpleNode node) throws InterpretError {
        long ACT_TYPED_DC = LONG_DC;
        long ACT_TYPED_DK = LONG_DK;

        switch (node.type) {
        case OP_IDENTIFIER:
            // Start CAST
            // We must handle the JRelix DC and DK values.
            // Thus, we invoke the JRelix CastToXYZ.value() methods
            // instead of using the Java primitive type casts.
            switch (destrel.domains[node.bits].type) {
            case DOUBLE:
                return CastToLong.value(((double[]) destrel.data[node.bits])[currow]);

            case FLOAT:
                return CastToLong.value(((float[]) destrel.data[node.bits])[currow]);

            case INTEGER:
                return CastToLong.value(((int[]) destrel.data[node.bits])[currow]);

            case SHORT:
                return CastToLong.value(((short[]) destrel.data[node.bits])[currow]);

            case LONG:
                return CastToLong.value(((long[]) destrel.data[node.bits])[currow]);

            case NUMERIC:
                return CastToLong.value(((number[]) destrel.data[node.bits])[currow]);

            default:
                throw new InterpretError("actLongCellWithNull :to be implemented.");
            }

        case OP_BIOPERATOR:
            SimpleNode left = (SimpleNode) node.jjtGetChild(0);
            SimpleNode right = (SimpleNode) node.jjtGetChild(1);
            long theBiOpVal = ACT_TYPED_DK;
            long mVal = ACT_TYPED_DK;
            long nVal = ACT_TYPED_DK;;
            if (node.opcode != OP_CAST) {
                mVal = actLongCellWithNull(left);
                if (mVal == ACT_TYPED_DK)
                    return ACT_TYPED_DK;
                nVal = actLongCellWithNull(right);
                if (nVal == ACT_TYPED_DK)
                    return ACT_TYPED_DK;
                else if (mVal == ACT_TYPED_DC && nVal == ACT_TYPED_DC)
                    return ACT_TYPED_DC;
                if (nVal == ACT_TYPED_DC)
                    return mVal;
            }

            switch (node.opcode) {
            case OP_PLUS:
                if (mVal == ACT_TYPED_DC)
                    return nVal;
                theBiOpVal = mVal + nVal;
                break;

            case OP_MINUS:
                if (mVal == ACT_TYPED_DC)
                    theBiOpVal = -nVal;
                else 
                    theBiOpVal = mVal - nVal;
                break;

            case OP_MUL:
                if (mVal == ACT_TYPED_DC)
                    return nVal;
                theBiOpVal = mVal * nVal;
                break;

            case OP_DIV:
                if (mVal == ACT_TYPED_DC)
                    theBiOpVal = 1 / nVal;
                else 
                    theBiOpVal = mVal / nVal;
                break;

            case OP_CAST:
                // Cast from type stored in left.bits of 
                // the the left OP_TYPE node in buildTree and
                // traverseType.
                switch (left.bits) {
                case BOOLEAN:
                    return CastToLong.value(actBoolCellWithNull(right));

                case STRING:
                    return CastToLong.value(actStrCellWithNull(right));

                case INTEGER:
                    return CastToLong.value(actIntCellWithNull(right));

                case SHORT:
                    return CastToLong.value((short)CastToShort.value(actIntCellWithNull(right)));

                case LONG:
                    return CastToLong.value(actLongCellWithNull(right));

                case FLOAT:
                    // Must cast to float first, otherwise, we may get
                    // a bigger number than possible for float in the
                    // result.
                    return CastToLong.value(CastToFloat.value(actDoubleCellWithNull(right)));

                case DOUBLE:
                    return CastToLong.value(actDoubleCellWithNull(right));

                case NUMERIC:
                    return CastToLong.value(actNumCellWithNull(right));

                default:
                    throw new InterpretError("actLongCellWithNull :to be implemented.");
                }

            default:
                throw new InterpretError("actLongCellWithNull :bio to be implemented.. .");
            }
            if (theBiOpVal <= ACT_TYPED_DC)
                throw new InterpretError("actLongCellWithNull :OP_BIOPERATOR result out of range.");
            else
                return theBiOpVal;

        case OP_UOPERATOR:
            switch (node.opcode) {
            case OP_UPLUS:
                return (actLongCellWithNull((SimpleNode) node.jjtGetChild(0)));

            case OP_UMINUS:
                mVal = actLongCellWithNull((SimpleNode) node.jjtGetChild(0));
                if (mVal == ACT_TYPED_DC || mVal == ACT_TYPED_DK)
                    return mVal;
                else
                if (-mVal <= ACT_TYPED_DC) 
                    throw new InterpretError("actLongCellWithNull :OP_UMINUS result out of range.");
                else 
                    return -mVal;

            default:
                throw new InterpretError("actLongCellWithNull :uni to be implemented.");
            }

        case OP_LITERAL:
            if (node.opcode == OP_DC)
                return (ACT_TYPED_DC);
            else if (node.opcode == OP_DK) 
                return (ACT_TYPED_DK);
            else if (node.info instanceof Double) 
                mVal = CastToLong.value(((Double) node.info).doubleValue());
            else if (node.info instanceof Float) 
                mVal = CastToLong.value(((Float) node.info).floatValue());
            else if (node.info instanceof Integer) 
                mVal = CastToLong.value(((Integer) node.info).intValue());
            else if (node.info instanceof Short) 
                mVal = CastToLong.value(((Short) node.info).shortValue());
            else if (node.info instanceof Long) 
                mVal = CastToLong.value(((Long) node.info).longValue());
            else
                throw new InterpretError("actLongCellWithNull :to be implemented.");
            if (mVal == ACT_TYPED_DC || mVal == ACT_TYPED_DK)
                // The Parser/Interpreter must ensure that the values
                // of the literals are not reserved for null values.
                throw new InterpretError("actLongCellWithNull :literal used value reserved for Dc or Dk .");
            else
                return mVal;

        case OP_RED:
        case OP_EQUIV:
            // Initialize the actualized value to DK.
            long actValue = ACT_TYPED_DK;

            // Initialize the vertical value holder.
            if (node.info == null)  node.info = new Long(mVal = ACT_TYPED_DC);
            else mVal = ((Long) node.info).longValue();

            // Any binary operation with DK results in DK.
            if (mVal == ACT_TYPED_DK) 
                return ACT_TYPED_DK;

            actValue = nVal = actLongCellWithNull((SimpleNode) node.jjtGetChild(0));
            if (nVal == ACT_TYPED_DK) {
                node.info = new Long(ACT_TYPED_DK);
                return ACT_TYPED_DK;
            }

            // With both DC operands, DC is the result.
            if (mVal == ACT_TYPED_DC && nVal == ACT_TYPED_DC)
                return ACT_TYPED_DC;

            // With DC as 2nd operand, mVal/node.info is unchanged
            if (nVal == ACT_TYPED_DC) 
                return actValue;// which is ACT_TYPED_DC;

            theBiOpVal = mVal;
            switch (node.opcode) {
            case OP_PLUS: {
                    if (mVal == ACT_TYPED_DC) theBiOpVal = nVal;
                    else  theBiOpVal = mVal + nVal;
                    break;
                }

            case OP_MUL: {
                    if (mVal == ACT_TYPED_DC) theBiOpVal = nVal;
                    else  theBiOpVal = mVal * nVal;
                    break;
                }

            case OP_MIN: {
                    if (mVal == ACT_TYPED_DC) theBiOpVal = nVal;
                    else  theBiOpVal = Math.min(mVal, nVal);
                    break;
                }

            case OP_MAX: {
                    if (mVal == ACT_TYPED_DC) theBiOpVal = nVal;
                    else  theBiOpVal = Math.max(mVal, nVal);
                    break;
                }

            default:
                throw new InterpretError("actLongCellWithNull :Red/Equiv to be implemented.");
            }
            if (theBiOpVal <= ACT_TYPED_DC && mVal != ACT_TYPED_DC)
                throw new InterpretError("actLongCellWithNull :Red/Equiv out of range.");
            if (theBiOpVal != mVal)
                node.info = new Long(theBiOpVal);
            return actValue;

        case OP_IFTHENELSE: {
                byte ifVal = actBoolCellWithNull((SimpleNode) node.jjtGetChild(0));

                if (ifVal == (byte) BOOL_TRUE)
                    return actLongCellWithNull((SimpleNode) node.jjtGetChild(1));
                else if (ifVal == (byte) BOOL_FALSE)
                    return actLongCellWithNull((SimpleNode) node.jjtGetChild(2));
                else if (ifVal == (byte) BOOL_DC) {
                    System.out.println("WARNING: actLongCellWithNull :'dc' encounterred by if-then-else.");
                    System.out.println("Actualizing using the 'then' value");
                    return actLongCellWithNull((SimpleNode) node.jjtGetChild(2));
                } else// if BOOL_DK 
                    throw new InterpretError("actLongCellWithNull :'dk' encounterred by if-then-else.");
            }

        default:
            throw new InterpretError("actLongCellWithNull :to be implemented.");
        }
    }      

    /**
     * Actualize an IDLIST typed cell.
     */
    private Relation actRelCellWithNull(SimpleNode node, Vector dnames, Vector rname,
        Vector rindx, Hashtable relhtbl, String tname0, String tname1)
        throws InterpretError {
        Relation r = null, subr;

        switch (node.type) {
        case OP_RED:
        case OP_EQUIV:
            SimpleNode srcnode = (SimpleNode) node.jjtGetChild(0);

            r = actRelCellWithNull(srcnode, dnames, rname, rindx, relhtbl, tname1, null);

            if (node.info == null)   node.info = r;
            else {
                node.info = ((Relation) node.info).mujoin(node.opcode, null,
                            r, null, tname0);
                ((Relation) node.info).print();
                ((Relation) node.info).load();
            }

            break;

        default:

            /*         for(int i = 0; i < rname.size(); i++){
             int index = ((Integer)rindx.elementAt(i)).intValue();
             long surg = ((long[])destrel.data[index])[currow];
             subr = Relation.getRelation((String)dnames.elementAt(i), surg);
             subr.name = (String)rname.elementAt(i);

             relhtbl.put(subr.name, subr);
             }
             */
            // Note, Biao has changed the getRelation() method so that we don't
            // need to save above 'subr', and evaluateTLExpression() will not
            // try to load relations from disk if there're data in memory.
            // However, evaluateTLExpression() and Relation.mujoin() methods
            // still saves their result to disk.
            r = interpreter.evaluateTLExpression(node, tname0, this.relhtbl, env);
            break;
        }
        //relhtbl.clear();
        return r;
    }         

    /**
     * Returns boolean value converted from a
     * boolean compatible nullary relation/IDLIST cell.
     * If relation is not boolean compatible, rel.boolCompatibleValue()
     * will throw an InterpretError.
     * @return  a boolean true if nullary relation contains true, false otherwise.
     * @since jRelixNS
     */
    private boolean actBoolRelCellWithNull(SimpleNode node, Vector vdom, Vector rname,
        Vector rindx, String tname0, String tname1)
        throws InterpretError {
        if (tname0 == null) 
            tname0 = interpreter.nextTempName();
        this.tempnames.addElement(tname0);//remember to clear this temp rel.
        Relation rel = actRelCellWithNull(node, vdom, rname, rindx, this.relhtbl,
                tname0, tname1);

        return rel.boolCompatibleValue();
    }

    /**
     *  Actualize a String typed cell.
     */
    private String actStrCellWithNull(SimpleNode node) throws InterpretError {

        switch (node.type) {
        case OP_IDENTIFIER:
            return ((String[]) destrel.data[node.bits])[currow];

        case OP_BIOPERATOR:
            SimpleNode left = (SimpleNode) node.jjtGetChild(0);
            SimpleNode right = (SimpleNode) node.jjtGetChild(1);

            switch (node.opcode) {
            case OP_CAT:
                String sm = actStrCellWithNull(left);

                if (sm.compareTo(STRING_DK) == 0)
                    return STRING_DK;
                String sn = actStrCellWithNull(right);

                if (sn.compareTo(STRING_DK) == 0)
                    return STRING_DK;
                if (sm.compareTo(STRING_DC) == 0)
                    return sn;
                if (sn.compareTo(STRING_DC) == 0)
                    return sm;
                String sBiOpVal = new String(sm + sn);;
                if (sBiOpVal.compareTo(STRING_DC) == 0 || sBiOpVal.compareTo(STRING_DK) == 0)
                    // Should we do this checking for OPs in other types?
                    throw new InterpretError("actStrCellWithNull: concat to Dc or Dk.");
                else
                    return sBiOpVal;

            case OP_CAST:
                // Cast from type stored in left.bits of 
                // the the left OP_TYPE node in buildTree and
                // traverseType.
                switch (left.bits) {
                case BOOLEAN:
                    return CastToStrn.value(actBoolCellWithNull(right));

                case SHORT:
                    // Must cast to short first, otherwise, we may get
                    // a bigger number than possible for short in the
                    // string result.
                    return  CastToStrn.value((short)CastToShort.value(actIntCellWithNull(right)));

                case INTEGER:
                    return  CastToStrn.value(actIntCellWithNull(right));

                case LONG:
                    return  CastToStrn.value(actLongCellWithNull(right));

                case FLOAT:
                    // Must cast to float first, otherwise, we may get
                    // a bigger number than possible for float in the
                    // string result.
                    return CastToStrn.value(CastToFloat.value(actDoubleCellWithNull(right)));

                case DOUBLE:
                    return  CastToStrn.value(actDoubleCellWithNull(right));

                case STRING:
                    return  CastToStrn.value(actStrCellWithNull(right));

                case NUMERIC:
                    return  CastToStrn.value(actNumCellWithNull(right));

                default:
                    throw new InterpretError("actStrCellWithNull: to be implemented.");
                }

            default:
                throw new InterpretError("actStrCellWithNull: bio to be implemented.");
            }

        case OP_UOPERATOR:
            switch (node.opcode) {
            default:
                throw new InterpretError("actStrCellWithNull: uni to be implemented.");
            }

        case OP_LITERAL:
            if (node.opcode == OP_DC)
                return (STRING_DC);
            else if (node.opcode == OP_DK) 
                return (STRING_DK);
            else if (node.info instanceof String)
                return ((String) node.info); 
            else
                throw new InterpretError("actStrCellWithNull: lit to be implemented.");

        case OP_RED:
        case OP_EQUIV:
            throw new InterpretError("actStrCellWithNull: red to be implemented.");

        case OP_IFTHENELSE: {
                byte ifVal = actBoolCellWithNull((SimpleNode) node.jjtGetChild(0));

                if (ifVal == (byte) BOOL_TRUE)
                    return actStrCellWithNull((SimpleNode) node.jjtGetChild(1));
                else if (ifVal == (byte) BOOL_FALSE)
                    return actStrCellWithNull((SimpleNode) node.jjtGetChild(2));
                else if (ifVal == (byte) BOOL_DC) {
                    System.out.println("WARNING: actStrCellWithNull: 'dc' encounterred by if-then-else.");
                    System.out.println("Actualizing using the 'then' value");
                    return actStrCellWithNull((SimpleNode) node.jjtGetChild(2));
                } else// if BOOL_DK 
                    throw new InterpretError("actStrCellWithNull: 'dk' encounterred by if-then-else.");
            }

        default:
            throw new InterpretError("actStrCellWithNull: to be implemented.");
        }
    }      

    /**
     *  Actualize a Numeric-typed cell.
     */
 
 /*
    private number actNumCell(SimpleNode node) throws InterpretError {

        switch (node.type) {
        case OP_IDENTIFIER:
            return ((number[]) destrel.data[node.bits])[currow];

        case OP_BIOPERATOR:
            SimpleNode left = (SimpleNode) node.jjtGetChild(0);
            SimpleNode right = (SimpleNode) node.jjtGetChild(1);

            switch (node.opcode) {
            case OP_PLUS:
                throw new InterpretError("actNumCell: OP_PLUS to be implemented.");

            case OP_MINUS:
                throw new InterpretError("actNumCell: OP_MINUS to be implemented.");

            case OP_MUL:
                throw new InterpretError("actNumCell: OP_MUL to be implemented.");

            case OP_DIV:
                throw new InterpretError("actNumCell: OP_DIV to be implemented.");

            case OP_MIN:
                throw new InterpretError("actNumCell: OP_MIN to be implemented.");

            case OP_MAX:
                throw new InterpretError("actNumCell: OP_MAX to be implemented.");

            case OP_CAST:
                // Cast from type stored in left.bits of 
                // the the left OP_TYPE node in buildTree and
                // traverseType.
                switch (left.bits) {
                case BOOLEAN:
                    return CastToNum.value(CastToBoolByte.value(actBoolCell(right)));

                case SHORT:
                    // Must cast to short first, otherwise, we may get
                    // a bigger number than possible for short in the
                    // result.
                    return  CastToNum.value((short)CastToShort.value(actIntCell(right)));

                case INTEGER:
                    return  CastToNum.value(actIntCell(right));

                case LONG:
                    return  CastToNum.value(actLongCell(right));

                case FLOAT:
                    // Must cast to float first, otherwise, we may get
                    // a bigger number than possible for float in the
                    // result.
                    return CastToNum.value(CastToFloat.value(actDoubleCell(right)));

                case DOUBLE:
                    return  CastToNum.value(actDoubleCell(right));

                case STRING:
                    return  CastToNum.value(actStrCell(right));

                case NUMERIC:
                    return  CastToNum.value(actNumCell(right));
 //YYY               
                //added by Yi Zheng Sept 28, 2002
               // case DC:
               // 	return NUMERIC_DC;
               // case DK:
                //	return NUMERIC_DK;
                
                //end of addition Sept 28, 2002

                default:
                    throw new InterpretError("actNumCell: to be implemented.");
                }

            default:
                throw new InterpretError("actNumCell: bio to be implemented.");
            }

        case OP_UOPERATOR:
            switch (node.opcode) {
            case OP_UPLUS:
            case OP_UMINUS:
            default:
                throw new InterpretError("actNumCell: uni to be implemented.");
            }

        case OP_LITERAL:
            throw new InterpretError("actNumCell: lit to be implemented.");

        case OP_RED:
        case OP_EQUIV:
            switch (node.opcode) {
            case OP_PLUS:
            case OP_MUL:
            case OP_MIN:
            case OP_MAX:
            default:
                throw new InterpretError("actNumCell: red to be implemented.");
            }

        case OP_IFTHENELSE: {
                byte ifVal = actBoolCell((SimpleNode) node.jjtGetChild(0)) ? BOOL_TRUE : BOOL_FALSE;

                if (ifVal == (byte) BOOL_TRUE)
                    return actNumCell((SimpleNode) node.jjtGetChild(1));
                else if (ifVal == (byte) BOOL_FALSE)
                    return actNumCell((SimpleNode) node.jjtGetChild(2));
                //else if( ifVal == (byte)BOOL_DC) {
                //System.out.println("WARNING: actNumCell: 'dc' encounterred by if-then-else.");
                //System.out.println("Actualizing using the 'then' value");
                //return actNumCell((SimpleNode)node.jjtGetChild(2));
                //}
                //else// if BOOL_DK 
                //throw new InterpretError("actNumCell: 'dk' encounterred by if-then-else.");
            }

        default:
            throw new InterpretError("actNumCell: to be implemented.");
        }
    }      
*/
    /**
     *  Actualize a Numeric-typed cell.
     */
    private number actNumCellWithNull(SimpleNode node) throws InterpretError {

        switch (node.type) {
        case OP_IDENTIFIER:
            return ((number[]) destrel.data[node.bits])[currow];

        case OP_BIOPERATOR:
            SimpleNode left = (SimpleNode) node.jjtGetChild(0);
            SimpleNode right = (SimpleNode) node.jjtGetChild(1);

            switch (node.opcode) {
            case OP_PLUS:
                throw new InterpretError("actNumCellWithNull: OP_PLUS to be implemented.");

            case OP_MINUS:
                throw new InterpretError("actNumCellWithNull: OP_MINUS to be implemented.");

            case OP_MUL:
                throw new InterpretError("actNumCellWithNull: OP_MUL to be implemented.");

            case OP_DIV:
                throw new InterpretError("actNumCellWithNull: OP_DIV to be implemented.");

            case OP_MIN:
                throw new InterpretError("actNumCellWithNull: OP_MIN to be implemented.");

            case OP_MAX:
                throw new InterpretError("actNumCellWithNull: OP_MAX to be implemented.");

            case OP_CAST:
                // Cast from type stored in left.bits of 
                // the the left OP_TYPE node in buildTree and
                // traverseType.
                switch (left.bits) {
                case BOOLEAN:
                    return CastToNum.value(actBoolCellWithNull(right));

                case SHORT:
                    // Must cast to short first, otherwise, we may get
                    // a bigger number than possible for short in the
                    // result.
                    return  CastToNum.value((short)CastToShort.value(actIntCellWithNull(right)));

                case INTEGER:
                    return  CastToNum.value(actIntCellWithNull(right));

                case LONG:
                    return  CastToNum.value(actLongCellWithNull(right));

                case FLOAT:
                    // Must cast to float first, otherwise, we may get
                    // a bigger number than possible for float in the
                    // result.
                    return CastToNum.value(CastToFloat.value(actDoubleCellWithNull(right)));

                case DOUBLE:
                    return  CastToNum.value(actDoubleCellWithNull(right));

                case STRING:
                    return  CastToNum.value(actStrCellWithNull(right));

                case NUMERIC:
                    return  CastToNum.value(actNumCellWithNull(right));

                default:
                    throw new InterpretError("actNumCellWithNull: to be implemented.");
                }

            default:
                throw new InterpretError("actNumCellWithNull: bio to be implemented.");
            }

        case OP_UOPERATOR:
            switch (node.opcode) {
            case OP_UPLUS:
            case OP_UMINUS:
            default:
                throw new InterpretError("actNumCellWithNull: uni to be implemented.");
            }

        case OP_LITERAL:
            throw new InterpretError("actNumCellWithNull: lit to be implemented.");

        case OP_RED:
        case OP_EQUIV:
            switch (node.opcode) {
            case OP_PLUS:
            case OP_MUL:
            case OP_MIN:
            case OP_MAX:
            default:
                throw new InterpretError("actNumCellWithNull: red to be implemented.");
            }

        case OP_IFTHENELSE: {
                byte ifVal = actBoolCellWithNull((SimpleNode) node.jjtGetChild(0));

                if (ifVal == (byte) BOOL_TRUE)
                    return actNumCellWithNull((SimpleNode) node.jjtGetChild(1));
                else if (ifVal == (byte) BOOL_FALSE)
                    return actNumCellWithNull((SimpleNode) node.jjtGetChild(2));
                else if (ifVal == (byte) BOOL_DC) {
                    System.out.println("WARNING: actNumCellWithNull: 'dc' encounterred by if-then-else.");
                    System.out.println("Actualizing using the 'then' value");
                    return actNumCellWithNull((SimpleNode) node.jjtGetChild(2));
                } else// if BOOL_DK 
                    throw new InterpretError("actNumCellWithNull: 'dk' encounterred by if-then-else.");
            }

        default:
            throw new InterpretError("actNumCellWithNull: to be implemented.");
        }
    }      
    //End Since JRelCaNu d.
}






















